/*
 * arch/ppc/5xxx_io/ac97.c
 *
 * Driver for the MPC5200 PSC2 in AC97 mode
 *
 * Author: Reinhard Meyer <info@emk-elektronik.de>
 *
 * 2006 (c) EMK Elektronik GmbH.  This file is licensed under the terms
 * of the GNU General Public License version 2.  This program is licensed
 * "as is" without any warranty of any kind, whether express or implied.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <linux/miscdevice.h>
#include <linux/proc_fs.h>

#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/poll.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/delay.h>
#include <linux/wrapper.h>	/* for mem_map_reserve */

#include <linux/init.h>
#include <asm/mpc5xxx.h>
#include <bestcomm_api.h>
#include <task_api/bestcomm_cntrl.h>
#include <mgt5200/mgt5200.h>
#include <mgt5200/sdma.h>
#include <linux/soundcard.h>
#include <linux/ac97_codec.h>

#define AC97_MINOR		240
#define	DRIVER_NAME		"ac97"
#define	DRIVER_TITLE		"MPC5200 AC97 driver EXPERIMENTAL!"
#define	AUTHOR			"Reinhard Meyer <info@emk-elektronik.de>"
struct ac97regparm {
	unsigned int regnum;
	unsigned int regval;
};

#define AC97_READ_REGISTER		_IOR('r', 1, struct ac97regparm)
#define AC97_WRITE_REGISTER		_IOW('r', 2, struct ac97regparm)

/* note: framerate is fixed at 48kHz */
#define	FRAME_WORDS		13	/* 13 words per frame */
#define	FRAME_BYTES		(13*4)	/* 52 bytes per frame */
#define	FRAMES_PER_BUFFER	48	/* 48 frames per Buffer = 1ms of time per buffer */

#define DMABUF_SIZE		(FRAME_BYTES*FRAMES_PER_BUFFER)
#define DMABUF_NUM		16	/* that gives us up to 16ms of IRQ slack */

#define	WORDS_PER_FRAME		1	/* 16 bits each for L + R */
#define	WORDS_PER_BUFFER	(FRAMES_PER_BUFFER*WORDS_PER_FRAME)
#define	BUFFERS_PER_RING	100
#define WORDS_PER_RING		(WORDS_PER_BUFFER*BUFFERS_PER_RING)

/* local forwards */
static void psc_init (void);
static void psc_enable (void);
static void psc_reset (void);
static int read_proc (char *, char **, off_t, int, int *, void *);

static struct mpc5xxx_psc *psc = (struct mpc5xxx_psc *) MPC5xxx_PSC2;

static long driver_open_count;

static spinlock_t lock;

static const char *codec_name = "generic AC97 codec";

static int modcnt;

/*
 * AC97 register structure with OSS assignments
 */
struct ac97reg {
	unsigned short readval;	/* read value */
	unsigned short writeval;	/* write value */
	unsigned int flags;	/* flags as follows: */
#define	REG_READ	0x01	/* readable register */
#define	REG_WRITE	0x02	/* writeable register */
#define	REG_READ1	0x10	/* read register once */
#define	REG_WRITE1	0x20	/* write register once */
#define	REG_READn	0x40	/* read perpetually */
#define	REG_READOK	0x80	/* successful read */
	int oss_channel;
	int ac_minval;
	int ac_maxval;
	int ac_mask;
	int channels;
	int oss_savedval;
};

static struct ac97reg ac97_regs[AC97_REG_CNT] = {
	/* output volume (after adder) registers */
	[AC97_MASTER_VOL_STEREO / 2] =
		{0, 0, 0, SOUND_MIXER_VOLUME, 0x1f, 0x00, 0x9f1f, 2},
	[AC97_MASTER_VOL_MONO / 2] =
		{0, 0, 0, SOUND_MIXER_SPEAKER, 0x1f, 0x00, 0x801f, 1},
	[AC97_HEADPHONE_VOL / 2] =
		{0, 0, 0, SOUND_MIXER_PHONEOUT, 0x1f, 0x00, 0x9f1f, 2},
	/* input volume (before adder) registers */
	[AC97_MIC_VOL / 2] =
		{0, 0, 0, SOUND_MIXER_MIC, 0x1f, 0x00, 0x801f, 1},
	[AC97_LINEIN_VOL / 2] =
		{0, 0, 0, SOUND_MIXER_LINE, 0x1f, 0x00, 0x9f1f, 2},
	[AC97_CD_VOL / 2] = {0, 0, 0, SOUND_MIXER_CD, 0x1f, 0x00, 0x9f1f, 2},
	[AC97_VIDEO_VOL / 2] =
		{0, 0, 0, SOUND_MIXER_VIDEO, 0x1f, 0x00, 0x9f1f, 2},
	[AC97_AUX_VOL / 2] =
		{0, 0, 0, SOUND_MIXER_LINE1, 0x1f, 0x00, 0x9f1f, 2},
	[AC97_PCMOUT_VOL / 2] =
		{0, 0, 0, SOUND_MIXER_PCM, 0x1f, 0x00, 0x9f1f, 2},
	[AC97_VIDEO_VOL / 2] =
		{0, 0, 0, SOUND_MIXER_VIDEO, 0x1f, 0x00, 0x9f1f, 2},
	/* recording registers */
	[AC97_RECORD_GAIN / 2] =
		{0, 0, 0, SOUND_MIXER_RECLEV, 0x0f, 0x00, 0x8f0f, 2},
};
static unsigned next_reg = 0;

/*
 * register access
 */
static void ac97_reg_init (short regnum, unsigned short writeval,
			   unsigned int flags)
{
	regnum /= 2;
	if (regnum < 0 || regnum >= AC97_REG_CNT)
		return;
	ac97_regs[regnum].writeval = writeval;
	ac97_regs[regnum].flags = flags;
}

/*
 * find AC97 register for OSS channel
 */
struct ac97reg *find_ac97_reg (int oss_channel)
{
	int i;

	for (i = 0; i < AC97_REG_CNT; i++) {
		if (ac97_regs[i].channels > 0
		    && ac97_regs[i].oss_channel == oss_channel) {
			return ac97_regs + i;
		}
	}
	return NULL;
}

/*
 * ring buffer structure:
 * tail==head --> empty
 */
struct ring {
	unsigned int *start;	/* points to first word */
	unsigned int *end;	/* points to first word behind */
	unsigned int *head;	/* points to place newest data will be placed */
	unsigned int *tail;	/* points to oldest data */
	unsigned int used;	/* used space */
};

/*
 * receive variables
 */
static DECLARE_WAIT_QUEUE_HEAD (rx_wait);
static int rx_newdata = 0;
static struct ring rx_ring = { NULL, NULL, NULL, NULL };
static TaskId rx_task;
static BDIdx rx_bdi;
static void *rx_buf_vaddr = NULL;
static dma_addr_t rx_buf_paddr = 0;

#define rx_buf_p(i)	((void *)(rx_buf_paddr + (i) * DMABUF_SIZE))
#define rx_buf_v(p)	((unsigned int *)((dma_addr_t)rx_buf_vaddr + ((p) - rx_buf_paddr)))

/*
 * transmit variables
 */
static DECLARE_WAIT_QUEUE_HEAD (tx_wait);
static struct ring tx_ring = { NULL, NULL, NULL, NULL };
static TaskId tx_task;
static BDIdx tx_bdi;
static void *tx_buf_vaddr = NULL;
static dma_addr_t tx_buf_paddr = 0;

#define tx_buf_p(i)	((void *)(tx_buf_paddr + (i) * DMABUF_SIZE))
#define tx_buf_v(p)	((unsigned int *)((dma_addr_t)tx_buf_vaddr + ((p) - tx_buf_paddr)))

/*
 * statistical variables
 */
static unsigned int rx_ring_dataloss;
static unsigned int rx_ring_interrupts;
static unsigned int rx_ring_framemiss;
static unsigned int tx_ring_dataloss;
static unsigned int tx_ring_interrupts;
static unsigned int rx_notx_irq_counter;
static unsigned int tx_irq_hangs;

/*
 * called by rx_irq to copy one BDs data to the ring
 */
static void rx_save_data (unsigned int *buf)
{
	int i;

	/* calculate free space in ring */
	if (WORDS_PER_RING - rx_ring.used >= WORDS_PER_BUFFER) {
		/* enough space, copy data */
		for (i = 0; i < FRAMES_PER_BUFFER; i++) {
			/* check here for valid frame alignment */
			if ((buf[0] & 0x80000800) != 0x80000800) {
				rx_ring_framemiss++;
			} else {
				/* just copy the timeslots we need - if they are there */
				if ((buf[0] & 0x18000000) == 0x18000000) {
					*rx_ring.head++ = (buf[3] & 0xffff0000)	/* L channel */
						|(buf[4] >> 16);	/* R channel */
					if (rx_ring.head == rx_ring.end)
						rx_ring.head = rx_ring.start;
					rx_ring.used++;
				}
				/* read register value there? */
				if ((buf[0] & 0x60000000) == 0x60000000) {
					unsigned short reg =
						(buf[1] >> 25) & 0x3f;
					unsigned short val = buf[2] >> 16;

					ac97_regs[reg].readval = val;
					ac97_regs[reg].flags &= ~REG_READ1;
					ac97_regs[reg].flags |= REG_READOK;
				}
			}
			buf += FRAME_WORDS;
		}
	} else {
		/* not enough space, remember error */
		rx_ring_dataloss++;
		/* search register answers here */
		for (i = 0; i < FRAMES_PER_BUFFER; i++) {
			/* check here for valid frame alignment */
			if ((buf[0] & 0x80000800) != 0x80000800) {
				rx_ring_framemiss++;
			} else {
				/* read register value there? */
				if ((buf[0] & 0x60000000) == 0x60000000) {
					unsigned short reg =
						(buf[1] >> 25) & 0x3f;
					unsigned short val = buf[2] >> 16;

					ac97_regs[reg].readval = val;
					ac97_regs[reg].flags &= ~REG_READ1;
					ac97_regs[reg].flags |= REG_READOK;
				}
			}
			buf += FRAME_WORDS;
		}
	}
	rx_newdata = 1;
}

/*
 * optionally add a register access
 */
static void tx_add_reg (unsigned int *buf)
{
	if (ac97_regs[next_reg].flags & (REG_READ1 | REG_WRITE1 | REG_READn)) {
		buf[0] |= 0xe0000000;
		if (ac97_regs[next_reg].flags & REG_WRITE1) {
			buf[1] = next_reg << 25;
			buf[2] = ac97_regs[next_reg].writeval << 16;
			ac97_regs[next_reg].flags &= ~REG_WRITE1;
		} else {
			buf[1] = (next_reg << 25) | 0x80000000;
			buf[2] = 0;
		}
	}
	next_reg = (next_reg + 1) % AC97_REG_CNT;
}

/*
 * called by tx_irq to copy one BDs data from the ring
 */
static void tx_get_data (unsigned int *buf)
{
	int i;

	/* calculate available data in ring */
	if (tx_ring.used >= WORDS_PER_BUFFER) {
		/* enough data, copy data */
		for (i = 0; i < FRAMES_PER_BUFFER; i++) {
			/* just copy the timeslots we need */
			buf[0] = 0x98000800;
			buf[3] = *tx_ring.tail & 0xffff0000;	/* L channel */
			buf[4] = *tx_ring.tail++ << 16;	/* R channel */
			if (tx_ring.tail == tx_ring.end)
				tx_ring.tail = tx_ring.start;
			tx_ring.used--;
			if (i == 0)
				tx_add_reg (buf);
			buf += FRAME_WORDS;
		}
	} else {
		/* not enough data, remember error */
		tx_ring_dataloss++;
		for (i = 0; i < FRAMES_PER_BUFFER; i++) {
			/* just copy the timeslots we need */
			buf[0] = 0;
			buf[1] = 0;
			buf[2] = 0;
			buf[3] = 0;
			buf[4] = 0;
			if (i == 0)
				tx_add_reg (buf);
			buf += FRAME_WORDS;
		}
	}
}

/*
 * Transmit IRQ when one BD has been sent
 */
static void ac97_tx_irq (int irq, void *dev_id, struct pt_regs *regs)
{
	TaskBD1_t *bd;

	TaskIntClear (tx_task);

	tx_ring_interrupts++;
	bd = (TaskBD1_t *) TaskGetBD (tx_task, tx_bdi);

	/* fill all empty BDs */
	while (!(bd->Status & SDMA_BD_MASK_READY)) {
		void *p = (void *) bd->DataPtr[0];

		/* fill with new data */
		tx_get_data (tx_buf_v (p));
		/* re-assign the BD to the task */
		tx_bdi = TaskBDRelease (tx_task);
		TaskBDAssign (tx_task, p, NULL, DMABUF_SIZE, 0);
		bd = (TaskBD1_t *) TaskGetBD (tx_task, tx_bdi);
	}
	wake_up (&tx_wait);
	rx_notx_irq_counter = 0;
}

/*
 * Receive IRQ when one BD has been filled
 */
static void ac97_rx_irq (int irq, void *dev_id, struct pt_regs *regs)
{
	TaskBD1_t *bd;

	TaskIntClear (rx_task);

	rx_ring_interrupts++;
	bd = (TaskBD1_t *) TaskGetBD (rx_task, rx_bdi);

	/* process ALL filled BDs */
	while (!(bd->Status & SDMA_BD_MASK_READY)) {
		void *p = (void *) bd->DataPtr[0];

		/* save recieved data */
		rx_save_data (rx_buf_v (p));
		/* re-assign the BD to the task */
		rx_bdi = TaskBDRelease (rx_task);
		TaskBDAssign (rx_task, p, NULL, DMABUF_SIZE, 0);
		bd = (TaskBD1_t *) TaskGetBD (rx_task, rx_bdi);
	}
	wake_up (&rx_wait);
	rx_notx_irq_counter++;
	if (rx_notx_irq_counter > 100) {
		tx_irq_hangs++;
		rx_notx_irq_counter = 0;
	}
}

/*
 * USER read data
 */
static ssize_t ac97_read (struct file *file, char *buf,
			  size_t count, loff_t * ppos)
{
	ssize_t ret = 0;

	/* sanity checks: */
	/* count % 4 must be 0 and count must be positive */
	if (count % 4 || count <= 0)
		return -EINVAL;
	/* seeking not allowed */
	if (ppos != &file->f_pos)
		return -ESPIPE;

	/* loop until request satisfied */
	while (count > 0) {
		int avail;
		unsigned long flags;

		spin_lock_irqsave (&lock, flags);
		avail = rx_ring.used * 4;
		/*  head <= tail --> data wraps in ring, cut copy at end of ring */
		if (avail > 0 && rx_ring.head <= rx_ring.tail)
			avail = (rx_ring.end - rx_ring.tail) * 4;
		spin_unlock_irqrestore (&lock, flags);

		if (avail > count)
			avail = count;	/* limit to users request */
		/* "avail" NOW is the smaller of the amount that will be read, */
		/* and the amount that is requested in this read (or partial). */
		if (avail <= 0) {
			if (file->f_flags & O_NONBLOCK)
				return ret ? ret : -EAGAIN;
			interruptible_sleep_on (&rx_wait);
			if (signal_pending (current))
				return ret ? ret : -ERESTARTSYS;
			continue;
		}
		/* there are bytes in the buffer to read. */
		copy_to_user (buf, rx_ring.tail, avail);

		spin_lock_irqsave (&lock, flags);
		rx_ring.tail += avail / 4;
		if (rx_ring.tail == rx_ring.end)
			rx_ring.tail = rx_ring.start;
		rx_ring.used -= avail / 4;
		spin_unlock_irqrestore (&lock, flags);

		count -= avail;
		buf += avail;
		ret += avail;
	}
	return ret;
}

/*
 * USER write data
 */
static ssize_t ac97_write (struct file *file, const char *buf, size_t count,
			   loff_t * ppos)
{
	ssize_t ret = 0;

	/* sanity checks: */
	/* count % 4 must be 0 and count must be positive */
	if (count % 4 || count <= 0)
		return -EINVAL;
	/* seeking not allowed */
	if (ppos != &file->f_pos)
		return -ESPIPE;

	/* loop until request satisfied */
	while (count > 0) {
		int avail;
		unsigned long flags;

		spin_lock_irqsave (&lock, flags);
		avail = (WORDS_PER_RING - tx_ring.used) * 4;
		/*  head >= tail --> free space wraps in ring, cut copy at end of ring */
		if (avail > 0 && tx_ring.head >= tx_ring.tail)
			avail = (tx_ring.end - tx_ring.head) * 4;
		spin_unlock_irqrestore (&lock, flags);

		if (avail > count)
			avail = count;	/* limit to users request */
		/* "avail" NOW is the smaller of the amount that will be written, */
		/* and the amount that is requested in this write (or partial). */
		if (avail <= 0) {
			if (file->f_flags & O_NONBLOCK)
				return ret ? ret : -EAGAIN;
			interruptible_sleep_on (&tx_wait);
			if (signal_pending (current))
				return ret ? ret : -ERESTARTSYS;
			continue;
		}
		/* there are bytes in the buffer to write. */
		copy_from_user (tx_ring.head, buf, avail);

		spin_lock_irqsave (&lock, flags);
		tx_ring.head += avail / 4;
		if (tx_ring.head == tx_ring.end)
			tx_ring.head = tx_ring.start;
		tx_ring.used += avail / 4;
		spin_unlock_irqrestore (&lock, flags);

		count -= avail;
		buf += avail;
		ret += avail;
	}
	return ret;
}

/*
 * USER open
 */
static int ac97_open (struct inode *inode, struct file *file)
{
	int i;
	unsigned long flags;

	spin_lock_irqsave (&lock, flags);
	if (driver_open_count == 0) {
		psc_reset ();
		rx_ring.head = rx_ring.tail = rx_ring.start;
		rx_ring.used = 0;
		rx_ring_dataloss = 0;
		rx_ring_interrupts = 0;
		rx_ring_framemiss = 0;
		rx_bdi = 0;
		rx_notx_irq_counter = 0;
		tx_ring.head = tx_ring.tail = tx_ring.start;
		tx_ring.used = 0;
		tx_ring_dataloss = 0;
		tx_ring_interrupts = 0;
		tx_bdi = 0;
		tx_irq_hangs = 0;
		TaskBDReset (rx_task);
		TaskBDReset (tx_task);
		/* assigne the BDs */
		for (i = 0; i < DMABUF_NUM; i++) {
			TaskBDAssign (rx_task, rx_buf_p (i), NULL,
				      DMABUF_SIZE, 0);
			TaskBDAssign (tx_task, tx_buf_p (i), NULL,
				      DMABUF_SIZE, 0);
		}
		/* start the DMA tasks */
		TaskStart (rx_task, TASK_AUTOSTART_ENABLE, rx_task,
			   TASK_INTERRUPT_ENABLE);
		TaskStart (tx_task, TASK_AUTOSTART_ENABLE, tx_task,
			   TASK_INTERRUPT_ENABLE);
		/* start the PSC, causing DMA */
		psc_enable ();
	}
	driver_open_count++;

	spin_unlock_irqrestore (&lock, flags);

	return 0;
}

/*
 * USER close
 */
static int ac97_release (struct inode *inode, struct file *file)
{
	unsigned long flags;

	spin_lock_irqsave (&lock, flags);
	driver_open_count--;
	if (driver_open_count == 0) {
		while (tx_ring.used >= WORDS_PER_BUFFER) {
			interruptible_sleep_on (&tx_wait);
		}
		psc_reset ();
		TaskStop (rx_task);
		TaskStop (tx_task);
		TaskBDReset (rx_task);
		TaskBDReset (tx_task);
		psc_reset ();
	}
	spin_unlock_irqrestore (&lock, flags);
	return 0;
}

static unsigned int ac97_poll (struct file *file, poll_table * wait)
{
	int ret = 0;

	poll_wait (file, &rx_wait, wait);

	if (rx_newdata)
		ret = (POLLIN | POLLRDNORM);

	return ret;
}

/*
 * convert OSS value to AC97 setting
 */
static unsigned short oss2ac (struct ac97reg *r, int oss_val)
{
	int oss_left, oss_right;
	unsigned short val_left, val_right;

	oss_left = oss_val & 0xff;
	oss_right = (oss_val >> 8) & 0xff;
	if (oss_left == 0 && oss_right == 0)
		return r->ac_mask;
	if (oss_left > 100)
		oss_left = 100;
	val_left =
		r->ac_minval +
		((r->ac_maxval - r->ac_minval) * oss_left) / 100;
	if (oss_right > 100)
		oss_right = 100;
	val_right =
		r->ac_minval +
		((r->ac_maxval - r->ac_minval) * oss_right) / 100;
	if (r->channels == 2)
		return (val_left << 8) | val_right;
	else
		return (val_left);
}

/*
 * convert AC97 setting to OSS value
 */
static unsigned short ac2oss (struct ac97reg *r, unsigned short ac_val)
{
	unsigned short ac_left, ac_right;
	int val_left, val_right;

	if (ac_val & 0x8000)
		return 0;
	ac_left = ((ac_val & r->ac_mask) >> 8) & 0x7f;
	ac_right = (ac_val & r->ac_mask) & 0xff;
	val_left =
		((ac_left - r->ac_minval) * 100) / (r->ac_maxval -
						    r->ac_minval);
	val_right =
		((ac_right - r->ac_minval) * 100) / (r->ac_maxval -
						     r->ac_minval);
	if (r->channels == 2)
		return (val_right << 8) | val_left;
	else
		return (val_right << 8) | val_right;
}

/*
 * ioctl - the wildcard interface to a driver
 */
static int ac97_ioctl (struct inode *inode, struct file *file,
		       unsigned int cmd, unsigned long arg)
{
	unsigned long flags;
	struct ac97regparm parm;
	int i, val;
	audio_buf_info abi;
	mixer_info mi;
	_old_mixer_info omi;
	struct ac97reg *r;

	switch (cmd) {

		/* private interface function to read a AC97 register */
	case AC97_READ_REGISTER:
		if (copy_from_user
		    (&parm, (struct ac97regparm *) arg, sizeof parm))
			return -EFAULT;
		parm.regnum /= 2;
		if (parm.regnum < 0 || parm.regnum >= AC97_REG_CNT)
			return -EINVAL;
		if (ac97_regs[parm.regnum].flags & REG_READ) {
			if (ac97_regs[parm.regnum].flags & REG_READOK) {
				parm.regval = ac97_regs[parm.regnum].readval;
				ac97_regs[parm.regnum].flags &= ~REG_READOK;
				ac97_regs[parm.regnum].flags |= REG_READ1;
				if (copy_to_user
				    ((struct ac97regparm *) arg, &parm,
				     sizeof parm))
					return -EFAULT;
				return 0;
			}
			ac97_regs[parm.regnum].flags |= REG_READ1;
			return -EAGAIN;
		}
		return -EINVAL;

		/* private interface function to write a AC97 register */
	case AC97_WRITE_REGISTER:
		if (copy_from_user
		    (&parm, (struct ac97regparm *) arg, sizeof parm))
			return -EFAULT;
		parm.regnum /= 2;
		if (parm.regnum < 0 || parm.regnum >= AC97_REG_CNT)
			return -EINVAL;
		if (ac97_regs[parm.regnum].flags & REG_WRITE) {
			ac97_regs[parm.regnum].writeval = parm.regval;
			ac97_regs[parm.regnum].flags |= REG_WRITE1;
			return 0;
		}
		return -EINVAL;

		/*
		 * OSS get version
		 */
	case OSS_GETVERSION:
		/* return actual value */
		i = SOUND_VERSION;
		if (copy_to_user ((int *) arg, &i, sizeof i))
			return -EFAULT;
		return 0;

		/*
		 * OSS defined functions for DSP follow here
		 */

		/* Reset DSP immediately */
	case SNDCTL_DSP_RESET:
		/* currently we need to do nothing */
		return 0;

		/* Wait for output to complete and then stop */
	case SNDCTL_DSP_SYNC:
		/* currently we need to do nothing */
		return 0;

		/* Set Sampling Rate */
	case SNDCTL_DSP_SPEED:
		/* read requested value */
		if (copy_from_user (&i, (int *) arg, sizeof i))
			return -EFAULT;
		/* ignore!!! request and return actual value */
		i = 48000;
		if (copy_to_user ((int *) arg, &i, sizeof i))
			return -EFAULT;
		return 0;

		/* Get Fragment Size */
	case SNDCTL_DSP_GETBLKSIZE:
		/* return actual value */
		i = WORDS_PER_BUFFER * 4;
		if (copy_to_user ((int *) arg, &i, sizeof i))
			return -EFAULT;
		return 0;

		/* Set Data Format */
	case SNDCTL_DSP_SETFMT:
		/* read requested value */
		if (copy_from_user (&i, (int *) arg, sizeof i))
			return -EFAULT;
		/* ignore!!! request and return actual format */
		i = AFMT_U16_BE;
		if (copy_to_user ((int *) arg, &i, sizeof i))
			return -EFAULT;
		return 0;

		/* Set Number of Channels */
	case SNDCTL_DSP_CHANNELS:
		/* read requested value */
		if (copy_from_user (&i, (int *) arg, sizeof i))
			return -EFAULT;
		/* ignore!!! request and return actual format */
		i = 2;
		if (copy_to_user ((int *) arg, &i, sizeof i))
			return -EFAULT;
		return 0;

		/* Output might pause a while announcement */
	case SNDCTL_DSP_POST:
		/* currently we need to do nothing */
		return 0;

		/* Set Fragment */
	case SNDCTL_DSP_SETFRAGMENT:
		/* read requested value */
		if (copy_from_user (&i, (int *) arg, sizeof i))
			return -EFAULT;
		/* ignore!!! request and return actual format */
		i = WORDS_PER_BUFFER * 4;
		if (copy_to_user ((int *) arg, &i, sizeof i))
			return -EFAULT;
		return 0;

		/* Get available Data Formats */
	case SNDCTL_DSP_GETFMTS:
		/* return actual format available */
		i = AFMT_U16_BE;
		if (copy_to_user ((int *) arg, &i, sizeof i))
			return -EFAULT;
		return 0;

		/* Get Output Buffer Space */
	case SNDCTL_DSP_GETOSPACE:
		spin_lock_irqsave (&lock, flags);
		abi.fragstotal = BUFFERS_PER_RING;
		abi.fragsize = WORDS_PER_BUFFER * 4;
		abi.bytes = (WORDS_PER_RING - tx_ring.used) * 4;
		abi.fragments = abi.bytes / abi.fragsize;
		spin_unlock_irqrestore (&lock, flags);
		if (copy_to_user ((audio_buf_info *) arg, &abi, sizeof abi))
			return -EFAULT;
		return 0;

		/* Get Input Buffer Space */
	case SNDCTL_DSP_GETISPACE:
		spin_lock_irqsave (&lock, flags);
		abi.fragstotal = BUFFERS_PER_RING;
		abi.fragsize = WORDS_PER_BUFFER * 4;
		abi.bytes = rx_ring.used * 4;
		abi.fragments = abi.bytes / abi.fragsize;
		spin_unlock_irqrestore (&lock, flags);
		if (copy_to_user ((audio_buf_info *) arg, &abi, sizeof abi))
			return -EFAULT;
		return 0;

		/* Get available Capabilities */
	case SNDCTL_DSP_GETCAPS:
		i = DSP_CAP_MULTI | DSP_CAP_BATCH | DSP_CAP_DUPLEX;
		if (copy_to_user ((int *) arg, &i, sizeof i))
			return -EFAULT;
		return 0;

		/*
		 * OSS defined functions for MIXER follow here
		 */

		/* Get Mixer Info (new style) */
	case SOUND_MIXER_INFO:
		strncpy (mi.id, codec_name, sizeof mi.id);
		strncpy (mi.name, codec_name, sizeof mi.name);
		mi.modify_counter = modcnt;
		if (copy_to_user ((mixer_info *) arg, &mi, sizeof mi))
			return -EFAULT;
		return 0;

		/* Get Mixer Info (old style) */
	case SOUND_OLD_MIXER_INFO:
		strncpy (omi.id, codec_name, sizeof omi.id);
		strncpy (omi.name, codec_name, sizeof omi.name);
		if (copy_to_user ((_old_mixer_info *) arg, &omi, sizeof omi))
			return -EFAULT;
		return 0;

		/* most of the mixer ioctl must be handled here: */
	default:
		/* check for mixer ioctl */
		if (_IOC_TYPE (cmd) != 'M'
		    || _SIOC_SIZE (cmd) != sizeof (int)) {
			printk (KERN_INFO "AC97: ioctl %08x unknown\n", cmd);
			return -EINVAL;
		}
		/* here could be a mixer ioctl */
		if (_SIOC_DIR (cmd) == _SIOC_READ) {
			/* read operations */
			switch (_IOC_NR (cmd)) {
			case SOUND_MIXER_RECSRC:
				/* read current recording source */
				val = 0;
				break;

			case SOUND_MIXER_DEVMASK:
				/* read supported mixers */
				val = 0;
				for (i = 0; i < AC97_REG_CNT; i++) {
					if (ac97_regs[i].channels >= 1)
						val |= (1 << ac97_regs[i].oss_channel);
				}
				break;

			case SOUND_MIXER_RECMASK:
				/* read supported recording sources */
				val = 0;
				break;

			case SOUND_MIXER_STEREODEVS:
				/* read channels supporting stereo */
				val = 0;
				for (i = 0; i < AC97_REG_CNT; i++) {
					if (ac97_regs[i].channels >= 2)
						val |= (1 << ac97_regs[i].oss_channel);
				}
				break;

			case SOUND_MIXER_CAPS:
				/* read capabilities */
				val = SOUND_CAP_EXCL_INPUT;
				break;

			default:
				/* read a specific mixer */
				i = _IOC_NR (cmd);
				r = find_ac97_reg (i);
				if (r == NULL)
					return -EINVAL;
				r->oss_savedval = ac2oss (r, r->readval);
				val = r->oss_savedval;
				break;
			}
			return put_user (val, (int *) arg);
		}

		if (_SIOC_DIR (cmd) == (_SIOC_WRITE | _SIOC_READ)) {
			modcnt++;
			if (get_user (val, (int *) arg))
				return -EFAULT;

			switch (_IOC_NR (cmd)) {
			case SOUND_MIXER_RECSRC:
				/* Arg contains a bit for each recording source */
				return -EINVAL;

			default:
				/* write a specific mixer */
				i = _IOC_NR (cmd);
				r = find_ac97_reg (i);
				if (r == NULL)
					return -EINVAL;
				r->oss_savedval = val;
				r->writeval = oss2ac (r, r->oss_savedval);
				r->flags |= REG_WRITE1;
				return 0;
			}
		}
		return -EINVAL;

	}
}

/*
 * initialize PSC2 and Port Configuration for AC97
 */
static void psc_init (void)
{
	struct mpc5xxx_gpio *gpio = (struct mpc5xxx_gpio *) MPC5xxx_GPIO;

	/* stop any activity that might be pending */
	psc->command = (MPC5xxx_PSC_TX_DISABLE | MPC5xxx_PSC_RX_DISABLE);

	/* PSC2 in AC97 */
	gpio->port_config = (gpio->port_config & ~0x00000070) | 0x00000020;

	/* reset PSC again */
	psc_reset ();
	psc->mode = 0;
	/* set alarm threshold levels */
	psc->rfalarm = 256;
	psc->tfalarm = 256;
	/* set granularity */
	psc->rfcntl = 4;
	psc->tfcntl = 4;
	/* disable PSC interrupts to CPU */
	psc->mpc5xxx_psc_imr = 0;
	/* set PSC for AC97 mode */
	psc->sicr = 0x03000000;
	/* reset PSC again */
	psc_reset ();
	/* reset the codec */
	psc->op1 = 0x02;
	udelay (10);
	psc->op0 = 0x02;
}

/*
 * enable PSC RX and TX
 */
static void psc_enable (void)
{
	psc->command = MPC5xxx_PSC_RX_ENABLE;
	psc->command = MPC5xxx_PSC_TX_ENABLE;
}

/*
 * hard stop PSC RX and TX, clear any pending Error Bits
 */
static void psc_reset (void)
{
	psc->command = MPC5xxx_PSC_RST_RX;
	psc->command = MPC5xxx_PSC_RST_TX;
	psc->command = MPC5xxx_PSC_SEL_MODE_REG_1;
	psc->command = MPC5xxx_PSC_RST_ERR_STAT;
}

/*
 * file operations for /dev/ac97
 */
static struct file_operations ac97_fops = {
      owner:THIS_MODULE,
      read:ac97_read,
      write:ac97_write,
      open:ac97_open,
      release:ac97_release,
      poll:ac97_poll,
      ioctl:ac97_ioctl,
};

/*
 * entry for /dev/ac97: MAJOR:misc, MINOR:as defined
 */
static struct miscdevice ac97_miscdev = {
	AC97_MINOR,
	DRIVER_NAME,
	&ac97_fops
};

/*
 * initialize this driver. called once during kernel boot or when module is loaded
 * (do we really want to support this as a module????
 */
static int __init init_driver (void)
{
	int ret = -1;
	TaskSetupParamSet_t ac97_setup_rx;
	TaskSetupParamSet_t ac97_setup_tx;

	spin_lock_init (&lock);

	if (WORDS_PER_RING % WORDS_PER_BUFFER) {
		printk (KERN_ERR DRIVER_NAME
			": invalid ring buffer size %d - must be multiple of %d\n",
			WORDS_PER_RING, WORDS_PER_BUFFER);
		goto out;
	}

	rx_ring.start = kmalloc (WORDS_PER_RING * 4, GFP_KERNEL);
	tx_ring.start = kmalloc (WORDS_PER_RING * 4, GFP_KERNEL);
	if (rx_ring.start == NULL || tx_ring.start == NULL) {
		printk (KERN_ERR DRIVER_NAME
			": can't allocate ring buffer\n");
		goto out;
	}
	rx_ring.head = rx_ring.tail = rx_ring.start;
	rx_ring.end = rx_ring.start + WORDS_PER_RING;
	rx_ring.used = 0;
	tx_ring.head = tx_ring.tail = tx_ring.start;
	tx_ring.end = tx_ring.start + WORDS_PER_RING;
	tx_ring.used = 0;

	if ((rx_buf_vaddr = pci_alloc_consistent (NULL,
						  DMABUF_SIZE *
						  DMABUF_NUM,
						  &rx_buf_paddr)) == NULL) {
		printk (KERN_ERR DRIVER_NAME ": can't allocate rx buffers\n");
		goto out;
	}

	if ((tx_buf_vaddr = pci_alloc_consistent (NULL,
						  DMABUF_SIZE *
						  DMABUF_NUM,
						  &tx_buf_paddr)) == NULL) {
		printk (KERN_ERR DRIVER_NAME ": can't allocate tx buffers\n");
		goto out;
	}

	ret = misc_register (&ac97_miscdev);
	if (ret) {
		printk (KERN_ERR DRIVER_NAME
			": can't misc_register on minor=%d\n", AC97_MINOR);
		goto out;
	}

	psc_init ();

	memset (&ac97_setup_rx, 0, sizeof (ac97_setup_rx));
	ac97_setup_rx.Initiator = (MPC5200Initiator_t) INITIATOR_PSC2_RX;
	ac97_setup_rx.NumBD = DMABUF_NUM;
	ac97_setup_rx.Size.MaxBuf = DMABUF_SIZE;
	ac97_setup_rx.StartAddrSrc = (volatile u32) &(psc->rfdata);
	ac97_setup_rx.IncrDst = 4;
	ac97_setup_rx.IncrSrc = 0;
	ac97_setup_rx.SzDst = 4;
	rx_task = TaskSetup (TASK_GEN_RX_BD, &ac97_setup_rx);

	memset (&ac97_setup_tx, 0, sizeof (ac97_setup_tx));
	ac97_setup_tx.Initiator = (MPC5200Initiator_t) INITIATOR_PSC2_TX;
	ac97_setup_tx.NumBD = DMABUF_NUM;
	ac97_setup_tx.Size.MaxBuf = DMABUF_SIZE;
	ac97_setup_tx.StartAddrDst = (volatile u32) &(psc->tfdata);
	ac97_setup_tx.IncrSrc = 4;
	ac97_setup_tx.IncrDst = 0;
	ac97_setup_tx.SzSrc = 4;
	tx_task = TaskSetup (TASK_GEN_TX_BD, &ac97_setup_tx);

	printk (KERN_INFO DRIVER_TITLE " - " AUTHOR
		"\n  PSC2 16 Bit BE Stereo (%d BDs x %d bytes); ring size %d words.\n",
		DMABUF_NUM, DMABUF_SIZE, WORDS_PER_RING);

	create_proc_read_entry (DRIVER_NAME, 0, 0, read_proc, NULL);

	/* kill any PSC operation */
	psc_reset ();

	rx_ring.head = rx_ring.tail = rx_ring.start;
	rx_ring.used = 0;
	rx_ring_dataloss = 0;
	rx_ring_interrupts = 0;
	rx_ring_framemiss = 0;
	rx_bdi = 0;
	rx_notx_irq_counter = 0;

	tx_ring.head = tx_ring.tail = tx_ring.start;
	tx_ring.used = 0;
	tx_ring_dataloss = 0;
	tx_ring_interrupts = 0;
	tx_bdi = 0;
	tx_irq_hangs = 0;

	/* set the AC97 registers to set and query */
	next_reg = 0;
	ac97_reg_init (AC97_RESET, 0, REG_READ | REG_READn);
	ac97_reg_init (AC97_MASTER_VOL_STEREO, 0,
		       REG_READ | REG_WRITE | REG_WRITE1 | REG_READn);
	ac97_reg_init (AC97_HEADPHONE_VOL, 0,
		       REG_READ | REG_WRITE | REG_WRITE1 | REG_READn);
	ac97_reg_init (AC97_MASTER_VOL_MONO, 0,
		       REG_READ | REG_WRITE | REG_WRITE1 | REG_READn);
	ac97_reg_init (AC97_MASTER_TONE, 0, REG_READ | REG_WRITE | REG_READn);
	ac97_reg_init (AC97_PCBEEP_VOL, 0, REG_READ | REG_WRITE | REG_READn);
	ac97_reg_init (AC97_PHONE_VOL, 0, REG_READ | REG_WRITE | REG_READn);
	ac97_reg_init (AC97_MIC_VOL, 0,
		       REG_READ | REG_WRITE | REG_WRITE1 | REG_READn);
	ac97_reg_init (AC97_LINEIN_VOL, 0, REG_READ | REG_WRITE | REG_READn);
	ac97_reg_init (AC97_CD_VOL, 0, REG_READ | REG_WRITE | REG_READn);
	ac97_reg_init (AC97_VIDEO_VOL, 0, REG_READ | REG_WRITE | REG_READn);
	ac97_reg_init (AC97_AUX_VOL, 0, REG_READ | REG_WRITE | REG_READn);
	ac97_reg_init (AC97_PCMOUT_VOL, 0,
		       REG_READ | REG_WRITE | REG_WRITE1 | REG_READn);
	ac97_reg_init (AC97_RECORD_SELECT, AC97_RECMUX_MIC,
		       REG_READ | REG_WRITE | REG_WRITE1 | REG_READn);
	ac97_reg_init (AC97_RECORD_GAIN, 0x0f0f,
		       REG_READ | REG_WRITE | REG_WRITE1 | REG_READn);
	ac97_reg_init (AC97_RECORD_GAIN_MIC, 0,
		       REG_READ | REG_WRITE | REG_READn);
	ac97_reg_init (AC97_GENERAL_PURPOSE, AC97_GP_MS,
		       REG_READ | REG_WRITE | REG_WRITE1 | REG_READn);
	ac97_reg_init (AC97_3D_CONTROL, 0, REG_READ | REG_WRITE | REG_READn);
	ac97_reg_init (AC97_MODEM_RATE, 0, REG_READ | REG_WRITE | REG_READn);
	ac97_reg_init (AC97_POWER_CONTROL, 0,
		       REG_READ | REG_WRITE | REG_READn);

	/* register IRQ functions */
	if (request_irq (MPC5xxx_SDMA_IRQ_BASE + rx_task, ac97_rx_irq,
			 SA_INTERRUPT, DRIVER_NAME " rx dma", NULL)) {
		printk (KERN_ERR DRIVER_NAME
			": SDMA rx irq allocation failed\n");
		return -EINVAL;
	}
	if (request_irq (MPC5xxx_SDMA_IRQ_BASE + tx_task, ac97_tx_irq,
			 SA_INTERRUPT, DRIVER_NAME " tx dma", NULL)) {
		printk (KERN_ERR DRIVER_NAME
			": SDMA tx irq allocation failed\n");
		return -EINVAL;
	}

      out:
	if (ret) {
		if (rx_buf_vaddr) {
			pci_free_consistent (NULL,
					     DMABUF_SIZE * DMABUF_NUM,
					     rx_buf_vaddr, rx_buf_paddr);
			rx_buf_vaddr = NULL;
		}
		if (rx_ring.start) {
			kfree (rx_ring.start);
			rx_ring.start = rx_ring.end = rx_ring.head =
				rx_ring.tail = NULL;
		}
		if (tx_ring.start) {
			kfree (tx_ring.start);
			tx_ring.start = tx_ring.end = tx_ring.head =
				tx_ring.tail = NULL;
		}
		remove_proc_entry (DRIVER_NAME, NULL);
	}

	return ret;

}

/*
 * de-initialize this driver. called during reboot or module unload
 */
static void __exit deinit_driver (void)
{
	psc_reset ();
	TaskStop (rx_task);
	TaskStop (tx_task);
	TaskBDReset (rx_task);
	TaskBDReset (tx_task);
	free_irq (MPC5xxx_SDMA_IRQ_BASE + rx_task, NULL);
	free_irq (MPC5xxx_SDMA_IRQ_BASE + tx_task, NULL);

	remove_proc_entry (DRIVER_NAME, NULL);

	pci_free_consistent (NULL, DMABUF_SIZE * DMABUF_NUM,
			     rx_buf_vaddr, rx_buf_paddr);
	kfree (rx_ring.start);
	kfree (tx_ring.start);

	misc_deregister (&ac97_miscdev);
}

/*
 * Info exported via "DRIVER_NAME".
 */
static int proc_output (char *buf)
{
	char *p = buf;
	int i;

	p += sprintf (p,
		      "rx head, tail, used: %5d, %5d, %5d\n"
		      "tx head, tail, used: %5d, %5d, %5d\n"
		      "rx taskid: %2d IRQs: %10d overflows:  %10d framingmissed: %10d\n"
		      "tx taskid: %2d IRQs: %10d underflows: %10d hangs:         %10d\n",
		      rx_ring.head - rx_ring.start,
		      rx_ring.tail - rx_ring.start, rx_ring.used,
		      tx_ring.head - tx_ring.start,
		      tx_ring.tail - tx_ring.start, tx_ring.used, rx_task,
		      rx_ring_interrupts, rx_ring_dataloss, rx_ring_framemiss,
		      tx_task, tx_ring_interrupts, tx_ring_dataloss,
		      tx_irq_hangs);
	p += sprintf (p,
		      "PSC status:%04x isr:%04x ip:%02x sicr:%08x\n"
		      " RF stat:%04x cntl:%04x alarm:%3u num:%3u rptr:%3u wptr:%3u\n"
		      " TF stat:%04x cntl:%04x alarm:%3u num:%3u rptr:%3u wptr:%3u\n",
		      psc->mpc5xxx_psc_status, psc->mpc5xxx_psc_isr, psc->ip,
		      psc->sicr, psc->rfstat, psc->rfcntl, psc->rfalarm,
		      psc->rfnum, psc->rfrptr, psc->rfwptr, psc->tfstat,
		      psc->tfcntl, psc->tfalarm, psc->tfnum, psc->tfrptr,
		      psc->tfwptr);
	for (i = 0; i < AC97_REG_CNT; i++)
		if (ac97_regs[i].flags & (REG_READ | REG_WRITE))
			p += sprintf (p,
				      "%02x: f=%02x r=%04x w=%04x oss=%2d min=%02x max=%02x mute=%04x c=%d saved=%3d\n",
				      i * 2, ac97_regs[i].flags,
				      ac97_regs[i].readval,
				      ac97_regs[i].writeval,
				      ac97_regs[i].oss_channel,
				      ac97_regs[i].ac_minval,
				      ac97_regs[i].ac_maxval,
				      ac97_regs[i].ac_mask,
				      ac97_regs[i].channels,
				      ac97_regs[i].oss_savedval);
	return p - buf;
}

/*
 * function to give status and info output via "DRIVER_NAME"
 */
static int read_proc (char *page, char **start, off_t off, int count,
		      int *eof, void *data)
{
	int len = proc_output (page);

	if (len <= off + count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;
	return len;
}

/*
 * module entry points
 */
module_init (init_driver);
module_exit (deinit_driver);

/*
 * declarations
 */
MODULE_DESCRIPTION (DRIVER_TITLE "\n");
MODULE_LICENSE ("GPL");
MODULE_AUTHOR (AUTHOR);
EXPORT_NO_SYMBOLS;
