/*
 * Routines for MPC5xxx GPIO Simple Interrupts handling.
 * Only PSC_3_8 pin (that is GPIO_SINT_2 interrupt) is supported for now.
 *
 * 2006 (c) Wolfgang Denk, DENX Software Engineering, <wd@denx.de>
 *
 * This file is licensed under the term of  the GNU General Public License
 * version 2. The program licensed "as is" without any warranty of any
 * kind, whether express or implied.
*/

#include <linux/irq.h>
#include <asm/mpc5xxx.h>
#include <asm/bitops.h>

/*
 * Defines and prototypes for an external usage
 */
/* GPIO_SINT2 (PSC3_8 pin)  */
#define	GSI2			(1 << 2)
/* Interrupt on any transition */
#define	ANY_EDGE_INT_T		(0)
/* Interrupt on rising edge */
#define	RISING_EDGE_INT_T	(1)
/* Interrupt on falling edge */
#define	FALLING_EDGE_INT_T	(2)
/* Interrupt on pulse (any two transitions)  */
#define	PULSE_INT_T		(3)

int request_gpio_irq (int num, int itype, void (*handler)(void));
int free_gpio_irq (int num);

/*
 * Defines and variables for a local usage
 */
/* GPIO_STD Interrupt Source for MPC5xxx Interrupt Controller */
#define	MPC5xxx_GPIO_STD_IRQ	(MPC5xxx_MAIN_IRQ_BASE + 7)
/* GPIO Simple Interrupt Master Enable */
#define	MPC5xxx_GSI_ME_BIT	(1 << 4)

/* Already-initialized flag */
static int gpio_std_inited = 0;
/*
 * For Shared Interrupt handling we need unique identifier, let it
 * be the pointer to our static variable
 */
static void *gpio_dev_id = (void *)(&gpio_std_inited);
/* pointer to GSI2 user handler */
static void (*uhandler)(void) = NULL;

/*
 *	@brief gpio_std_handler: main handler, that is called on
 *		MPC5xxx_GPIO_STD_IRQ and check gpio GSI2 interrupt occurance
 *	@param irq: MPC5xxx_GPIO_STD_IRQ, not used
 *	@param dev_id: gpio_dev_id, not used
 *	@param regs: not used
 */
static void gpio_std_handler (int irq, void *dev_id, struct pt_regs *regs)
{
	volatile struct mpc5xxx_gpio *pgpio =
					(struct mpc5xxx_gpio *)MPC5xxx_GPIO;

	/* Check GPIO Simple Interrupt Status Register */
	if (pgpio->sint_istat & GSI2) {
		/*
		 * GSI2 Interrupt has occured, call user handler. Really,
		 * if there's no handler, something goes wrong
		 */
		if (uhandler)
			(*uhandler)();
		/*
		 * Clear only GSI2 Status read-write-clear bit, '='
		 * instead oft '|=' to leave other bits in sint_istat
		 * unchanged
		 */
		pgpio->sint_istat = GSI2;
	}
}

/*
 *	@brief init_gpio_simple_interrupt: function initialize GPIO pin
 *		as a simple interrupt
 *	@param num: gpio simple interrupt number
 *	@param itpe: ANY_EDGE_INT_T, RISING_EDGE_INT_T, FALLING_EDGE_INT_T
 *		or PULSE_INT_T
 */
static void init_gpio_simple_interrupt (int num, int itype)
{
	volatile struct mpc5xxx_gpio *pgpio =
					(struct mpc5xxx_gpio *)MPC5xxx_GPIO;

	/*
	 * Clear GPIO Interrupt Master Enable before beginning to program
	 * GSI2 interrupt
	 */
	pgpio->gpio_control &= ~MPC5xxx_GSI_ME_BIT;

	/* Configure GPIO as input */
	pgpio->sint_ddr &= ~num;

	/* Configure GPIO Interrupt Type */
	pgpio->sint_itype |= (unsigned short)(itype << (__ilog2(num) << 1));

	/* Enable GSI2 Interrupt */
	pgpio->sint_gpioe |= num;
	pgpio->sint_inten |= num;

	/* Set  GPIO Interrupt Master Enable */
	pgpio->gpio_control |= MPC5xxx_GSI_ME_BIT;
}

/*
 *	@brief free_gpio_simple_interrupt: hardware deactivation of gpio
 *		simple interrupt
 *	@param num: gpio simple interrupt number
 */
static void free_gpio_simple_interrupt (int num)
{
	volatile struct mpc5xxx_gpio *pgpio =
					(struct mpc5xxx_gpio *)MPC5xxx_GPIO;

	/*
	 * Clear GPIO Interrupt Master Enable before beginning to program
	 * gpio interrupt
	 */
	pgpio->gpio_control &= ~MPC5xxx_GSI_ME_BIT;

	/* Configure GPIO as input */
	pgpio->sint_ddr &= ~num;

	/* Disable GSI2 Interrupt */
        pgpio->sint_gpioe &= ~num;

	/* Set  GPIO Interrupt Master Enable */
	pgpio->gpio_control |= MPC5xxx_GSI_ME_BIT;
}

/*
 *	@brief init_gpio_irq: function request GPIO_Std (main source 7)
 *		interrupt and call GSI2 gpio simple interrupt initialization
 *		routine
 *	@param num: not used, reserved for future
 *	@param itype: ANY_EDGE_INT_T, RISING_EDGE_INT_T, FALLING_EDGE_INT_T
 *		or PULSE_INT_T
 *	@param handler: pointer to function that takes no parameters. It will
 *		 be called on each GSI2 interrupt.
 *	@return 0 : Success
 *		!0: Error code
 */
int request_gpio_irq (int num, int itype, void (*handler)(void))
{
	/* Check parameters */
	if (!handler || (itype > PULSE_INT_T))
		return -EINVAL;

	/*
	 * For now, working only with PSC_3_8 pin, that is GPIO_SINT_2 simple
	 * interrupt
	 */
	num = GSI2;

	/* Get Main Interrupt Source 7 (GPIO_std) */
	if (!gpio_std_inited) {
		if (request_irq(MPC5xxx_GPIO_STD_IRQ, gpio_std_handler,
					SA_SHIRQ, "psc3_8_irq", gpio_dev_id)) {
			printk ("%s: can't get assigned gpio_std_irq %d!\n",
					__FUNCTION__, MPC5xxx_GPIO_STD_IRQ);
			return -EBUSY;
		}
		gpio_std_inited++;
	} else {
		return -EBUSY; /* resource is inited and used already */
	}

	/* Function to call on interrupt */
        uhandler = handler;

	/* Perform hardware initialization */
	init_gpio_simple_interrupt (num, itype);

	return 0;
}

/*
 *	@brief free_gpio_irq: free GPIO interrupt previously got by
 *		request_gpio_irq() call
 *	@param num: not used, reserved for future
 */
int free_gpio_irq (int num)
{
	/*
	 * For now, working only with PSC_3_8 pin, that is GPIO_SINT_2 simple
	 * interrupt
	 */
	num = GSI2;

	if (gpio_std_inited) {
		/* hardware deactiavtion */
		free_gpio_simple_interrupt (num);
		gpio_std_inited = 0;
	} else {
		/* GPIO_STD_IRQ was not requested before freeing */
		return -EINVAL;
	}
	/*
	 * For now only one GPIO interrupt may be used, so free GPIO_Std
	 * Interrupt Source always
	 */
	free_irq (MPC5xxx_GPIO_STD_IRQ, gpio_dev_id);

	return 0;
}
