/*
 * Authors: Mark A. Greer <mgreer@mvista.com>, Stefan Roese <sr@denx.de>
 *
 * Adapted from ev64260.c by Stefan Roese (sr@denx.de)
 *
 * Copyright 2001-2003 MontaVista Software Inc.
 * Copyright 2005 DENX Software Engineering (wd@denx.de)
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

#include <linux/config.h>
#include <linux/delay.h>
#include <linux/pci.h>
#include <linux/ide.h>
#include <linux/irq.h>
#include <linux/fs.h>
#include <linux/seq_file.h>
#include <linux/serial.h>

#include <asm/bootinfo.h>
#include <asm/machdep.h>
#include <asm/gt64260.h>
#include <asm/ppcboot.h>
#include <asm/todc.h>
#include <asm/time.h>

#include <platforms/p3g4.h>


#define BOARD_VENDOR	"Prodrive"
#define BOARD_MACHINE	"P3G4"


ulong	p3g4_mem_size = 0;
bd_t	ppcboot_bd;
int	ppcboot_bd_valid = 0;

unsigned char __res[sizeof(bd_t)];

extern void gen550_progress(char *, unsigned short);
extern void gen550_init(int, struct serial_struct *);


static const unsigned int cpu_7xx[16] = { /* 7xx & 74xx (but not 745x) */
	18, 15, 14, 2, 4, 13, 5, 9, 6, 11, 8, 10, 16, 12, 7, 0
};
static const unsigned int cpu_745x[2][16] = { /* PLL_EXT 0 & 1 */
	{ 1, 15, 14,  2,  4, 13,  5,  9,  6, 11,  8, 10, 16, 12,  7,  0 },
	{ 0, 30,  0,  2,  0, 26,  0, 18,  0, 22, 20, 24, 28, 32,  0,  0 }
};


static int
p3g4_get_bus_speed(void)
{
	int	speed;

	if (ppcboot_bd_valid) {
		speed = ppcboot_bd.bi_busfreq;
	} else {
		speed = 100000000; /* XXX Only 100MHz is stable */
	}

	return speed;
}

static int
p3g4_get_cpu_speed(void)
{
	unsigned long	pvr, hid1, pll_ext;

	pvr = PVR_VER(mfspr(PVR));

	if (pvr != PVR_VER(PVR_7450)) {
		hid1 = mfspr(HID1) >> 28;
		return p3g4_get_bus_speed() * cpu_7xx[hid1]/2;
	} else {
		hid1 = (mfspr(HID1) & 0x0001e000) >> 13;
		pll_ext = 0; /* No way to read; must get from schematic */
		return p3g4_get_bus_speed() * cpu_745x[pll_ext][hid1]/2;
	}
}

unsigned long __init
p3g4_find_end_of_memory(void)
{
	if(!ppcboot_bd_valid) {
		return gt64260_get_mem_size();
	}
	return ppcboot_bd.bi_memsize;
}

static void __init
p3g4_pci_fixups(void)
{
}


/*
 * Marvell/Galileo P3G4 Board PCI interrupt routing.
 */
static int __init
p3g4_map_irq(struct pci_dev *dev, unsigned char idsel, unsigned char pin)
{
	struct pci_controller	*hose = pci_bus_to_hose(dev->bus->number);

	if (hose->index == 0) {
#ifndef CONFIG_P3G4_PRPMC_SLOT_B
		static char pci_irq_table[][4] =
			/*
			 *	PCI IDSEL/INTPIN->INTLINE
			 * 	   A   B   C   D
			 */
			{
				{P3G4_PCI_INTA, P3G4_PCI_INTB, 0, 0}, /* IDSEL 16 - Peripheral/Oxford */
				{P3G4_PCI_INTA, P3G4_PCI_INTB, P3G4_PCI_INTC, P3G4_PCI_INTD}, /* IDSEL 17 - Universe */
				{P3G4_PCI_INTB, P3G4_PCI_INTC, P3G4_PCI_INTD, P3G4_PCI_INTA}, /* IDSEL 18 */
				{P3G4_PCI_INTB, P3G4_PCI_INTC, P3G4_PCI_INTD, P3G4_PCI_INTA}, /* IDSEL 19 */
			};

		const long min_idsel = 16, max_idsel = 19, irqs_per_slot = 4;
#else /* CONFIG_P3G4_PRPMC_SLOT_B */
		static char pci_irq_table[][4] =
			/*
			 *	PCI IDSEL/INTPIN->INTLINE
			 * 	   A   B   C   D
			 */
			{
				{P3G4_PCI_INTD, P3G4_PCI_INTA, P3G4_PCI_INTB, P3G4_PCI_INTC}, /* device 6 (pmc) */
				{0, 0, 0, 0}, /* device 7, not available */
				{P3G4_PCI_INTB, P3G4_PCI_INTC, P3G4_PCI_INTD, P3G4_PCI_INTA}, /* device 8 */
				{P3G4_PCI_INTC, P3G4_PCI_INTD, P3G4_PCI_INTA, P3G4_PCI_INTB}, /* device 9 */
				{P3G4_PCI_INTD, P3G4_PCI_INTA, P3G4_PCI_INTB, P3G4_PCI_INTC}, /* device 10 */
			};

		const long min_idsel = 6, max_idsel = 10, irqs_per_slot = 4;
#endif /* CONFIG_P3G4_PRPMC_SLOT_B */

		return PCI_IRQ_TABLE_LOOKUP;
	}

	return 0;
}

static void __init
p3g4_setup_peripherals(void)
{
	/*
	 * Set up windows to SRAM, RTC/TODC and DUART on device module
	 * (CS 0, 1 & 2)
	 */
	gt64260_cpu_boot_set_window(P3G4_EMB_FLASH_BASE, P3G4_EMB_FLASH_SIZE);
	gt64260_cpu_cs_set_window(0, P3G4_EXT_FLASH_BASE, P3G4_EXT_FLASH_SIZE);

	gt_clr_bits(GT64260_CPU_CONFIG, ((1<<28) | (1<<29)));
	gt_set_bits(GT64260_CPU_CONFIG, (1<<27));

	gt_set_bits(GT64260_PCI_0_SLAVE_PCI_DECODE_CNTL, ((1<<0) | (1<<3)));
	gt_set_bits(GT64260_PCI_1_SLAVE_PCI_DECODE_CNTL, ((1<<0) | (1<<3)));

        /*
         * Enabling of PCI internal-vs-external arbitration
         * is a platform- and errata-dependent decision.
         */
        if( gt64260_revision == GT64260 )  {
                /* FEr#35 */
                gt_clr_bits(GT64260_PCI_0_ARBITER_CNTL, (1<<31));
                gt_clr_bits(GT64260_PCI_1_ARBITER_CNTL, (1<<31));
        } else if( gt64260_revision == GT64260A )  {
                gt_set_bits(GT64260_PCI_0_ARBITER_CNTL, (1<<31));
                gt_set_bits(GT64260_PCI_1_ARBITER_CNTL, (1<<31));
        }

        gt_set_bits(GT64260_CPU_MASTER_CNTL, (1<<9)); /* Only 1 cpu */

	/*
	 * The P3G4 uses several Multi-Purpose Pins (MPP) on the 64260
	 * bridge as interrupt inputs (via the General Purpose Ports (GPP)
	 * register).  Need to route the MPP inputs to the GPP and set the
	 * polarity correctly.
	 *
	 * In MPP Control 0 Register
	 *   MPP 0 -> GPP 0 (PCI0 INTA) bits 0-3 -> 0
	 *   MPP 1 -> GPP 1 (PCI0 INTB) bits 4-7 -> 0
	 *   MPP 2 -> GPP 2 (PCI0 INTC) bits 8-11 -> 0
	 *   MPP 3 -> GPP 3 (PCI0 INTD) bits 12-15 -> 0
	 */
	gt_clr_bits(GT64260_MPP_CNTL_0, (0xf<<(GPP0<<2)) | (0xf<<(GPP1<<2)) | (0xf<<(GPP2<<2)) | (0xf<<(GPP3<<2)));

	/* External interrupts are inputs */
	gt_clr_bits(GT64260_GPP_IO_CNTL, GPP_EXTERNAL_INTERRUPTS);
	/* External interrupts are active low */
	gt_set_bits(GT64260_GPP_LEVEL_CNTL, GPP_EXTERNAL_INTERRUPTS);

	/* Clear any pending interrupts for these inputs and enable them. */
	gt_write(GT64260_GPP_INTR_CAUSE, ~GPP_EXTERNAL_INTERRUPTS);
	gt_set_bits(GT64260_GPP_INTR_MASK, GPP_EXTERNAL_INTERRUPTS);

	/*
	 * Set MPSC Multiplex RMII
	 * NOTE: ethernet driver modifies bit 0 and 1
	 */
	gt_write(GT64260_MPP_SERIAL_PORTS_MULTIPLEX, 0x00001102);
	return;
}


static void __init
p3g4_setup_bridge(void)
{
        gt64260_bridge_info_t           info;
	int				i;
	u32				val;
	u8				save_exclude;

        GT64260_BRIDGE_INFO_DEFAULT(&info, p3g4_find_end_of_memory());

	/*
	 * Assume that GT64260_CPU_SCS_DECODE_WINDOWS,
	 * GT64260_CPU_SNOOP_WINDOWS, GT64260_PCI_SCS_WINDOWS, and
	 * GT64260_PCI_SNOOP_WINDOWS are all the equal.
	 * Also assume that GT64260_CPU_PROT_WINDOWS >=
	 * GT64260_CPU_SCS_DECODE_WINDOWS and
	 * GT64260_PCI_ACC_CNTL_WINDOWS >= GT64260_PCI_SCS_WINDOWS.
	 */
	for (i=0; i<GT64260_CPU_SCS_DECODE_WINDOWS; i++) {
		info.cpu_prot_options[i] = 0;
		info.cpu_snoop_options[i] = GT64260_CPU_SNOOP_WB;
		info.pci_0_acc_cntl_options[i] =
			/* Breaks PCI (especially slot 4)
			   GT64260_PCI_ACC_CNTL_PREFETCHEN |
			*/
			GT64260_PCI_ACC_CNTL_DREADEN |
			GT64260_PCI_ACC_CNTL_RDPREFETCH |
			GT64260_PCI_ACC_CNTL_RDLINEPREFETCH |
			GT64260_PCI_ACC_CNTL_RDMULPREFETCH |
			GT64260_PCI_ACC_CNTL_SWAP_NONE |
			GT64260_PCI_ACC_CNTL_MBURST_4_WORDS;
		info.pci_0_snoop_options[i] = GT64260_PCI_SNOOP_WB;
		info.pci_1_acc_cntl_options[i] =
			/* Breaks PCI (especially slot 4)
			   GT64260_PCI_ACC_CNTL_PREFETCHEN |
			*/
			GT64260_PCI_ACC_CNTL_DREADEN |
			GT64260_PCI_ACC_CNTL_RDPREFETCH |
			GT64260_PCI_ACC_CNTL_RDLINEPREFETCH |
			GT64260_PCI_ACC_CNTL_RDMULPREFETCH |
			GT64260_PCI_ACC_CNTL_SWAP_NONE |
			GT64260_PCI_ACC_CNTL_MBURST_4_WORDS;
		info.pci_1_snoop_options[i] = GT64260_PCI_SNOOP_WB;
	}

	info.pci_0_latency = 0x8;
	info.pci_1_latency = 0x8;

	info.hose_a = pcibios_alloc_controller();
	if (!info.hose_a) {
		if (ppc_md.progress)
			ppc_md.progress("p3g4_setup_bridge: hose_a failed", 0);
		printk("p3g4_setup_bridge: can't set up first hose\n");
		return;
	}

	info.hose_b = pcibios_alloc_controller();
	if (!info.hose_b) {
		if (ppc_md.progress)
			ppc_md.progress("p3g4_setup_bridge: hose_b failed", 0);
		printk("p3g4_setup_bridge: can't set up second hose\n");
		return;
	}

	info.phys_base_addr = gt64260_base;

	/* Map in the bridge's registers */
	gt64260_base = (u32)ioremap(gt64260_base, GT64260_INTERNAL_SPACE_SIZE);

        /* Lookup PCI host bridges */
        if (gt64260_find_bridges(&info, p3g4_map_irq)) {
                printk("Bridge initialization failed.\n");
		iounmap((void *)gt64260_base);
        }

	/*
	 * Dave Wilhardt found that bit 4 in the PCI Command registers must
	 * be set if you are using cache coherency.
	 *
	 * Note: he also said that bit 4 must be on in all PCI devices but
	 *       that has not been implemented yet.
	 */
	save_exclude = gt64260_pci_exclude_bridge;
	gt64260_pci_exclude_bridge = FALSE;

	early_read_config_dword(info.hose_a,
				info.hose_a->first_busno,
				PCI_DEVFN(0,0),
				PCI_COMMAND,
				&val);
	val |= 0x10;
	early_write_config_dword(info.hose_a,
				 info.hose_a->first_busno,
				 PCI_DEVFN(0,0),
				 PCI_COMMAND,
				 val);

	early_read_config_dword(info.hose_b,
				info.hose_b->first_busno,
				PCI_DEVFN(0,0),
				PCI_COMMAND,
				&val);
	val |= 0x10;
	early_write_config_dword(info.hose_b,
				 info.hose_b->first_busno,
				 PCI_DEVFN(0,0),
				 PCI_COMMAND,
				 val);

	gt64260_pci_exclude_bridge = save_exclude;
	return;
}

static void __init
p3g4_early_serial_map(void)
{
#ifdef	CONFIG_KGDB
	static char	first_time = 1;


#if defined(CONFIG_KGDB_TTYS0)
#define KGDB_PORT 0
#elif defined(CONFIG_KGDB_TTYS1)
#define KGDB_PORT 1
#else
#error "Invalid kgdb_tty port"
#endif

	if (first_time) {
		gt_early_mpsc_init(KGDB_PORT, B9600|CS8|CREAD|HUPCL|CLOCAL);
		first_time = 0;
	}

	return;
#endif
}

static void __init
p3g4_setup_arch(void)
{
	if ( ppc_md.progress )
		ppc_md.progress("p3g4_setup_arch: enter", 0);

#ifdef CONFIG_BLK_DEV_INITRD
	if (initrd_start)
		ROOT_DEV = MKDEV(RAMDISK_MAJOR, 0);
	else
#endif
#ifdef	CONFIG_ROOT_NFS
		ROOT_DEV = to_kdev_t(0x00FF);	/* /dev/nfs pseudo device */
#else
	ROOT_DEV = to_kdev_t(0x0802);	/* /dev/sda2 SCSI disk */
#endif

	if ( ppc_md.progress )
		ppc_md.progress("p3g4_setup_arch: Enabling L2 cache", 0);

	/* Enable L2 and L3 caches (if 745x) */
	_set_L2CR(_get_L2CR() | L2CR_L2E);
	_set_L3CR(_get_L3CR() | L3CR_L3E);

	if ( ppc_md.progress )
		ppc_md.progress("p3g4_setup_arch: Initializing bridge", 0);

	p3g4_setup_bridge();		/* set up PCI bridge(s) */
	p3g4_setup_peripherals();	/* set up chip selects/GPP/MPP etc */

	if ( ppc_md.progress )
		ppc_md.progress("p3g4_setup_arch: bridge init complete", 0);

#ifdef	CONFIG_DUMMY_CONSOLE
	conswitchp = &dummy_con;
#endif
#if defined(CONFIG_SERIAL) || defined(CONFIG_GT64260_CONSOLE)
	p3g4_early_serial_map();
#endif

	printk("Prodrive P3G4 port (C) 2005 DENX Software Engineering (wd@denx.de)\n");

	if ( ppc_md.progress )
		ppc_md.progress("p3g4_setup_arch: exit", 0);

	return;
}

static void
p3g4_reset_board(void *addr)
{
	__cli();

	/* disable and invalidate the L2 cache */
	_set_L2CR(0);
	_set_L2CR(0x200000);

	/* flush and disable L1 I/D cache */
	__asm__ __volatile__
		("mfspr   3,1008\n\t"
		 "ori	5,5,0xcc00\n\t"
		 "ori	4,3,0xc00\n\t"
		 "andc	5,3,5\n\t"
		 "sync\n\t"
		 "mtspr	1008,4\n\t"
		 "isync\n\t"
		 "sync\n\t"
		 "mtspr	1008,5\n\t"
		 "isync\n\t"
		 "sync\n\t");

	/* unmap any other random cs's that might overlap with bootcs */
	gt64260_cpu_cs_set_window(0,0,0);
	gt64260_cpu_cs_set_window(1,0,0);
	gt64260_cpu_cs_set_window(2,0,0);
	gt64260_cpu_cs_set_window(3,0,0);

	/* map bootrom back in to gt @ reset defaults */
	gt64260_cpu_boot_set_window(0xff800000, 8*1024*1024);

	/* move gt reg base back to default, setup default pci0 swapping
	 * config... */
	gt_write(GT64260_INTERNAL_SPACE_DECODE,
		 (1<<24) | GT64260_INTERNAL_SPACE_DEFAULT_ADDR>>20);

	/* NOTE: FROM NOW ON no more GT_REGS accesses.. 0x1 is not mapped
	 * via BAT or MMU, and MSR IR/DR is ON */
#if 0
	/* BROKEN... IR/DR is still on !!  won't work!! */
	/* Set exception prefix high - to the firmware */
	_nmask_and_or_msr(0, MSR_IP);

	out_8((u_char *)P3G4_BOARD_MODRST_REG, 0x01);
#else
	/* SRR0 has system reset vector, SRR1 has default MSR value */
	/* rfi restores MSR from SRR1 and sets the PC to the SRR0 value */
	/* NOTE: assumes reset vector is at 0xfff00100 */
	__asm__ __volatile__
		("mtspr   26, %0\n\t"
		 "li      4,(1<<6)\n\t"
		 "mtspr   27,4\n\t"
		 "rfi\n\t"
		 :: "r" (addr):"r4");
#endif
	return;
}

static void
p3g4_restart(char *cmd)
{
	volatile ulong	i = 10000000;

	p3g4_reset_board((void *)0xfff00100);

	while (i-- > 0);
	panic("restart failed\n");
}

static void
p3g4_halt(void)
{
	__cli();
	while (1);
	/* NOTREACHED */
}

static void
p3g4_power_off(void)
{
	p3g4_halt();
	/* NOTREACHED */
}

static int
p3g4_show_cpuinfo(struct seq_file *m)
{
	uint pvid;

	pvid = mfspr(PVR);
	seq_printf(m, "vendor\t\t: " BOARD_VENDOR "\n");
	seq_printf(m, "machine\t\t: " BOARD_MACHINE "\n");
	seq_printf(m, "cpu MHz\t\t: %d\n", p3g4_get_cpu_speed()/1000/1000);
	seq_printf(m, "bus MHz\t\t: %d\n", p3g4_get_bus_speed()/1000/1000);

	return 0;
}

/* DS1501 RTC has too much variation to use RTC for calibration */
static void __init
p3g4_calibrate_decr(void)
{
	ulong freq;

	freq = p3g4_get_bus_speed()/4;

	printk("time_init: decrementer frequency = %lu.%.6lu MHz\n",
	       freq/1000000, freq%1000000);

	tb_ticks_per_jiffy = freq / HZ;
	tb_to_us = mulhwu_scale_factor(freq, 1000000);

	return;
}

static void parse_ppcbootinfo(unsigned long r3,
			      unsigned long r4, unsigned long r5,
			      unsigned long r6, unsigned long r7)
{
	bd_t *bd=NULL;
	char *cmdline_start=NULL;
	int cmdline_len=0;

	if(r3) {
		if((r3 & 0xf0000000) == 0) r3 += KERNELBASE;
		if((r3 & 0xf0000000) == KERNELBASE) {
			bd=(void *)r3;

			/* hack for ppcboot loaders that report freqs in Mhz */
			if(bd->bi_intfreq<1000000) bd->bi_intfreq*=1000000;
			if(bd->bi_busfreq<1000000) bd->bi_busfreq*=1000000;

			memcpy(&ppcboot_bd,bd,sizeof(ppcboot_bd));
			memcpy(&__res[0], bd, sizeof(ppcboot_bd));
			ppcboot_bd_valid=1;
		}
	}

#ifdef CONFIG_BLK_DEV_INITRD
	if(r4 && r5 && r5>r4) {
		if((r4 & 0xf0000000) == 0) r4 += KERNELBASE;
		if((r5 & 0xf0000000) == 0) r5 += KERNELBASE;
		if((r4 & 0xf0000000) == KERNELBASE) {
			initrd_start=r4;
			initrd_end=r5;
			initrd_below_start_ok = 1;
		}
	}
#endif /* CONFIG_BLK_DEV_INITRD */

	if(r6 && r7 && r7>r6) {
		if((r6 & 0xf0000000) == 0) r6 += KERNELBASE;
		if((r7 & 0xf0000000) == 0) r7 += KERNELBASE;
		if((r6 & 0xf0000000) == KERNELBASE) {
			cmdline_start=(void *)r6;
			cmdline_len=(r7-r6);
			strncpy(cmd_line,cmdline_start,cmdline_len);
		}
	}

	if(ppcboot_bd_valid) {
		printk("found bd_t @%p\n", bd);
		printk("memstart=%08lx\n", bd->bi_memstart);
		printk("memsize=%08lx\n", bd->bi_memsize);
		printk("enetaddr=%02x%02x%02x%02x%02x%02x\n",
		       bd->bi_enetaddr[0],
		       bd->bi_enetaddr[1],
		       bd->bi_enetaddr[2],
		       bd->bi_enetaddr[3],
		       bd->bi_enetaddr[4],
		       bd->bi_enetaddr[5]
			);
		printk("intfreq=%ld\n", bd->bi_intfreq);
		printk("busfreq=%ld\n", bd->bi_busfreq);
		printk("baudrate=%ld\n", bd->bi_baudrate);
	}

#ifdef CONFIG_BLK_DEV_INITRD
	if(initrd_start) {
		printk("found initrd @%lx-%lx\n", initrd_start, initrd_end);
	}
#endif /* CONFIG_BLK_DEV_INITRD */

	if(cmdline_start && cmdline_len) {
		printk("found cmdline: '%s'\n", cmd_line);
	}
}

#if defined(CONFIG_SERIAL_TEXT_DEBUG) || defined(CONFIG_KGDB)
static void __init
p3g4_map_io(void)
{
	io_block_mapping(0xfb000000, 0xfb000000, 0x01000000, _PAGE_IO);
}
#endif

void __init
platform_init(unsigned long r3, unsigned long r4, unsigned long r5,
	      unsigned long r6, unsigned long r7)
{
#ifdef CONFIG_BLK_DEV_INITRD
	initrd_start=initrd_end=0;
	initrd_below_start_ok=0;
#endif /* CONFIG_BLK_DEV_INITRD */

	ppcboot_bd_valid=0;
	memset(&ppcboot_bd,0,sizeof(ppcboot_bd));

	parse_ppcbootinfo(r3, r4, r5, r6, r7);

	isa_mem_base = 0;
	isa_io_base = GT64260_PCI_0_IO_START_PROC;
	pci_dram_offset = GT64260_PCI_0_MEM_START_PROC;

	loops_per_jiffy = p3g4_get_cpu_speed() / HZ;

	ppc_md.setup_arch = p3g4_setup_arch;
	ppc_md.show_cpuinfo = p3g4_show_cpuinfo;
	ppc_md.irq_cannonicalize = NULL;
	ppc_md.init_IRQ = gt64260_init_irq;
	ppc_md.get_irq = gt64260_get_irq;

	ppc_md.pcibios_fixup = p3g4_pci_fixups;

	ppc_md.restart = p3g4_restart;
	ppc_md.power_off = p3g4_power_off;
	ppc_md.halt = p3g4_halt;

	ppc_md.find_end_of_memory = p3g4_find_end_of_memory;

	ppc_md.init = NULL;

	ppc_md.time_init = NULL;
	ppc_md.set_rtc_time = NULL;
	ppc_md.get_rtc_time = NULL;

	ppc_md.calibrate_decr = p3g4_calibrate_decr;

#ifdef	CONFIG_GT64260_NEW_BASE
	gt64260_base = CONFIG_GT64260_NEW_REG_BASE;
#else
	gt64260_base = CONFIG_GT64260_ORIG_REG_BASE;
#endif

#ifdef	CONFIG_SERIAL
#if defined(CONFIG_SERIAL_TEXT_DEBUG)
	ppc_md.setup_io_mappings = p3g4_map_io;
	ppc_md.progress = gen550_progress;
#endif /* CONFIG_SERIAL_TEXT_DEBUG */
#if defined(CONFIG_KGDB)
	ppc_md.setup_io_mappings = p3g4_map_io;
	ppc_md.early_serial_map = p3g4_early_serial_map;
#endif /* CONFIG_KGDB */
#elif defined(CONFIG_GT64260_CONSOLE)
#ifdef	CONFIG_SERIAL_TEXT_DEBUG
	ppc_md.setup_io_mappings = p3g4_map_io;
	ppc_md.progress = gt64260_mpsc_progress;
#endif	/* CONFIG_SERIAL_TEXT_DEBUG */
#ifdef	CONFIG_KGDB
	ppc_md.setup_io_mappings = p3g4_map_io;
	ppc_md.early_serial_map = p3g4_early_serial_map;
#endif	/* CONFIG_KGDB */

#endif /* CONFIG_SERIAL */

	return;
}
