/*
 * arch/ppc/platforms/sorcery_cpld_pic.c
 *
 * Functions for interrupt multiplexor in Sorcery CPLD.
 *
 * This file is licensed under the terms of the GNU General Public License
 * version 2.  This program is licensed "as is" without any warranty of
 * any kind, whether express or implied.
 */

#include <linux/init.h>
#include <linux/irq.h>

#include <asm/io.h>
#include <platforms/sorcery.h>

static void
sorcery_cpld_ic_enable(unsigned int irq)
{
	u16 val;
	cpld_regs_t *cpld_regs = (cpld_regs_t *)SORCERY_CPLD_REG_BASE;

	if ((irq < SORCERY_CPLD_INT_BASE + SORCERY_CPLD_MASK0_INT_NUM)
	    && (irq >= SORCERY_CPLD_INT_BASE)) {
		val = in_be16(&(cpld_regs->mask0));
		val |= 1 << (irq - SORCERY_CPLD_INT_BASE);
		val |= SORCERY_CPLD_MASK0_MIE;
		out_be16(&(cpld_regs->mask0), val);
	}
	if ((irq < SORCERY_CPLD_INT_BASE + SORCERY_CPLD_MASK0_INT_NUM + 
	    SORCERY_CPLD_MASK1_INT_NUM)
	    && (irq >= SORCERY_CPLD_INT_BASE + SORCERY_CPLD_MASK0_INT_NUM)) {
		val = in_be16(&(cpld_regs->mask1));
		val |= 1 << (irq - SORCERY_CPLD_INT_BASE - 
			SORCERY_CPLD_MASK0_INT_NUM);
		val |= SORCERY_CPLD_MASK1_MIE;
		out_be16(&(cpld_regs->mask1), val);
	}
}

static void
sorcery_cpld_ic_disable(unsigned int irq)
{
	u16 val;
	cpld_regs_t *cpld_regs = (cpld_regs_t *)SORCERY_CPLD_REG_BASE;

	if ((irq < SORCERY_CPLD_INT_BASE + SORCERY_CPLD_MASK0_INT_NUM)
	    && (irq >= SORCERY_CPLD_INT_BASE)) {
		val = in_be16(&(cpld_regs->mask0));
		val &= ~(1 << (irq - SORCERY_CPLD_INT_BASE));
		out_be16(&(cpld_regs->mask0), val);
	}
	if ((irq < SORCERY_CPLD_INT_BASE + SORCERY_CPLD_MASK0_INT_NUM + 
	    SORCERY_CPLD_MASK1_INT_NUM)
	    && (irq >= SORCERY_CPLD_INT_BASE + SORCERY_CPLD_MASK0_INT_NUM)) {
		val = in_be16(&(cpld_regs->mask1));
		val &= ~(1 << (irq - SORCERY_CPLD_INT_BASE - 
			SORCERY_CPLD_MASK0_INT_NUM));
		out_be16(&(cpld_regs->mask1), val);
	}
}

static void
sorcery_cpld_ic_disable_and_ack(unsigned int irq)
{
	u16 val, val1;
	cpld_regs_t *cpld_regs = (cpld_regs_t *)SORCERY_CPLD_REG_BASE;

	if ((irq < SORCERY_CPLD_INT_BASE + SORCERY_CPLD_MASK0_INT_NUM)
	    && (irq >= SORCERY_CPLD_INT_BASE)) {

		/* disable MIE */
		val = in_be16(&(cpld_regs->mask0));
		val1 = val;
		val &= ~(SORCERY_CPLD_MASK0_MIE);
		out_be16(&(cpld_regs->mask0), val);
		
		/* disable req irq */
		sorcery_cpld_ic_disable(irq);
		
		/* ack & enable MIE */
		val1 |= 1 << (irq - SORCERY_CPLD_INT_BASE);
		val1 |= SORCERY_CPLD_MASK0_MIE;
		out_be16(&(cpld_regs->status_clear0), val1);
	}
	if ((irq < SORCERY_CPLD_INT_BASE + SORCERY_CPLD_MASK0_INT_NUM + 
	    SORCERY_CPLD_MASK1_INT_NUM)
	    && (irq >= SORCERY_CPLD_INT_BASE + SORCERY_CPLD_MASK0_INT_NUM)) {

		/* disable MIE */
		val = in_be16(&(cpld_regs->mask1));
		val1 = val;
		val &= ~(SORCERY_CPLD_MASK1_MIE);
		out_be16(&(cpld_regs->mask1), val);
		
		/* disable req irq */
		sorcery_cpld_ic_disable(irq);

		/* ack req & enable MIE */
		val1 |= 1 << (irq - SORCERY_CPLD_INT_BASE - 
			SORCERY_CPLD_MASK0_INT_NUM);
		val1 |= SORCERY_CPLD_MASK1_MIE;
		out_be16(&(cpld_regs->status_clear1), val1);
	}
}

static void
sorcery_cpld_ic_end(unsigned int irq)
{
	if (!(irq_desc[irq].status & (IRQ_DISABLED | IRQ_INPROGRESS))) {
		sorcery_cpld_ic_enable(irq);
	}
}

static struct hw_interrupt_type sorcery_cpld_ic = {
	.typename	= "Sorcery CPLD PIC",
	.startup	= NULL,
	.shutdown	= NULL,
	.enable		= sorcery_cpld_ic_enable,
	.disable	= sorcery_cpld_ic_disable,
	.ack		= sorcery_cpld_ic_disable_and_ack,
	.end		= sorcery_cpld_ic_end,
	.set_affinity	= 0
};

int
sorcery_get_irq(struct pt_regs *regs)
{
	int irq;
	cpld_regs_t *cpld_regs = (cpld_regs_t *)SORCERY_CPLD_REG_BASE;
	unsigned long enabled_irqs_asserted;

	irq = mpc8220_get_irq(regs);

	if (irq == SORCERY_CPLD_INT0){
		enabled_irqs_asserted = in_be16(&cpld_regs->mask0) 
			& in_be16(&cpld_regs->status_clear0);
		if(enabled_irqs_asserted == 0)
			BUG();
		else
			return ffs(enabled_irqs_asserted) + 
				SORCERY_CPLD_INT_BASE - 1;
	}
	if(irq == SORCERY_CPLD_INT1){
		enabled_irqs_asserted = in_be16(&cpld_regs->mask1)
			& in_be16(&cpld_regs->status_clear1);
		if(enabled_irqs_asserted == 0)
			BUG();
		else
			return ffs(enabled_irqs_asserted) + 
				SORCERY_CPLD_INT_BASE + 
				SORCERY_CPLD_MASK0_INT_NUM - 1;
	}
	return irq;
}

void __init
sorcery_init_irq(void)
{
	int i;
	u16 val;
	cpld_regs_t *cpld_regs = (cpld_regs_t *)SORCERY_CPLD_REG_BASE;
	struct mpc8220_intr *intr = (struct mpc8220_intr *)MPC8220_INTCTL;

	/* initialize MPC8220 interrupt controller */
	mpc8220_init_irq();
	
	/* reset and disable all CPLD interrupts */
	out_be16(&cpld_regs->mask0, 0);
	out_be16(&cpld_regs->mask1, 0);
	
	out_be16(&cpld_regs->status_clear0, 0xffff);
	out_be16(&cpld_regs->status_clear1, 0xffff);

	/* assign sorcery ic */
	for (i = 0; i < SORCERY_CPLD_MASK0_INT_NUM + 
	    SORCERY_CPLD_MASK1_INT_NUM; i++) {
		irq_desc[SORCERY_CPLD_INT_BASE + i].handler = &sorcery_cpld_ic;
		irq_desc[SORCERY_CPLD_INT_BASE + i].status &= ~IRQ_LEVEL;
	}
	
	/* Make IRQ0, IRQ1 falling edge sensitive */
	val = in_be32(&intr->ctrl);
	out_be32(&intr->ctrl, val | 0x00f00000);

	/* enable SORCERY_CPLD_INT0, SORCERY_CPLD_INT1 and make it visible 
	   in /proc */
	request_irq(SORCERY_CPLD_INT0, no_action, 0, "Sorcery CPLD PIC", 0);
	request_irq(SORCERY_CPLD_INT1, no_action, 0, "Sorcery CPLD PIC", 0);

}
