/*
 * arch/ppc/platforms/sorcery_pci.c
 *
 * Board setup routines for the Xerox Sorcery-C MPC8220 board.
 *
 * Author: MontaVista Software, Inc.
 *         source@mvista.com
 *
 * Based on code from Motorola Inc.
 *
 * Copyright 2003 Motorola Inc.
 *
 * 2004 (c) MontaVista SoftOCware, Inc. This file is licensed under the terms
 * of the GNU General Public License version 2. This program is licensed
 * "as is" without any warranty of any kind, whether express or implied.
 */

#include <linux/module.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/string.h>
#include <linux/init.h>
#include <linux/bootmem.h>

#include <asm/io.h>
#include <asm/prom.h>
#include <asm/pci-bridge.h>
#include <asm/machdep.h>
#include <asm/mpc8220.h>

#include <platforms/sorcery.h>
#include <platforms/sorcery_pci.h>

/* Forward declarations */
static int sorcery_exclude_device(unsigned char bus, unsigned char devfn);
static int sorcery_map_irq(struct pci_dev *dev, unsigned char idsel,
                        unsigned char pin);

/*****************************************************************************/
#define cfg_read(val, addr, type, op)		*val = op((type)(addr)); 
#define cfg_write(val, addr, type, op)		op((type *)(addr), (val)); 

#define PCI_OP(rw, size, type, op, mask)				\
int sorcery_pci_##rw##_config_##size(struct pci_dev *dev, int offset,	\
					type val)			\
{									\
	struct pci_controller *hose = dev->sysdata;			\
	u8 cfg_type = 0;						\
	u32 addr = 0;							\
									\
	if (ppc_md.pci_exclude_device)					\
		if (ppc_md.pci_exclude_device(dev->bus->number,		\
		    dev->devfn))					\
			return PCIBIOS_DEVICE_NOT_FOUND;		\
									\
	if (hose->set_cfg_type)						\
		if (dev->bus->number != hose->first_busno)		\
			cfg_type = 1;					\
									\
	addr = (((offset & 0xfc) | cfg_type) | (dev->devfn << 8)	\
		| ((dev->bus->number - hose->bus_offset) << 16)		\
		| 0x80000000);						\
									\
	out_be32(hose->cfg_addr, addr);					\
	__asm__ __volatile__("sync");					\
	cfg_##rw(val, hose->cfg_data + (offset & mask), type, op);	\
									\
	out_be32(hose->cfg_addr, addr & 0x7fffffff);			\
	__asm__ __volatile__("sync");					\
									\
	return PCIBIOS_SUCCESSFUL;					\
}

PCI_OP(read, byte, u8 *, in_8, 3)
PCI_OP(read, word, u16 *, in_le16, 2)
PCI_OP(write, byte, u8, out_8, 3)
PCI_OP(write, word, u16, out_le16, 2)
PCI_OP(write, dword, u32, out_le32, 0)

/* for 8220 errata 1028 - PCI type 1 Config space reads of dwords don't work
 *  so this code does 2 word accesses and combines results.
 */
int sorcery_pci_read_config_dword(struct pci_dev *dev, int offset, u32 *val)
{
	struct pci_controller *hose = dev->sysdata;
	u32 addr;
	u32 tmpv;
	u32 mask = 2;    /* word access */
	u8 cfg_type = 0;

	if (ppc_md.pci_exclude_device)
		if (ppc_md.pci_exclude_device(dev->bus->number, dev->devfn))
			return PCIBIOS_DEVICE_NOT_FOUND;
	if (hose->set_cfg_type)
		if (dev->bus->number != hose->first_busno)
			cfg_type = 1;

	/* Read lower 16 bits */
	addr = (((offset & 0xfc) | cfg_type) | (dev->devfn << 8)
		| ((dev->bus->number - hose->bus_offset) << 16) | 0x80000000);
	out_be32(hose->cfg_addr, addr);
	__asm__ __volatile__("sync");
	*val = (u32) in_le16((u16 *) (hose->cfg_data + (offset & mask)));
	out_be32(hose->cfg_addr, addr & 0x7fffffff);
	__asm__ __volatile__("sync");

	/* Read upper 16 bits */
	offset += 2;
	addr = (((offset & 0xfc) | cfg_type) | (dev->devfn << 8)
	    | ((dev->bus->number - hose->bus_offset) << 16) | 0x80000000);
	out_be32(hose->cfg_addr, addr);
	__asm__ __volatile__("sync");
	tmpv = (u32) in_le16((u16 *) (hose->cfg_data + (offset & mask)));
	out_be32(hose->cfg_addr, addr & 0x7fffffff);
	__asm__ __volatile__("sync");

	/* combine results into dword value */
	*val = (tmpv << 16) | *val;

	return PCIBIOS_SUCCESSFUL;
}

/*****************************************************************************/
struct pci_ops sorcery_pci_ops =
{
	.read_byte	= sorcery_pci_read_config_byte,
	.read_word	= sorcery_pci_read_config_word,
	.read_dword	= sorcery_pci_read_config_dword,
	.write_byte	= sorcery_pci_write_config_byte,
	.write_word	= sorcery_pci_write_config_word,
	.write_dword	= sorcery_pci_write_config_dword
};


/*****************************************************************************/
static void __init setup_sorcery_pci(struct pci_controller *hose, u32 cfg_addr, 
				u32 cfg_data)
{
	hose->ops = &sorcery_pci_ops;
	hose->cfg_addr = (volatile u32 *) ioremap(cfg_addr, 4);
	hose->cfg_data = (volatile unsigned char *) ioremap(cfg_data, 4);
}

/*****************************************************************************/
static void sorcery_pci_init(void)
{
	struct mpc8220_xipci *xipci;
	u32 win0, win1, win2;
	xipci = (struct mpc8220_xipci *) SORCERY_PCI_REG_MMAP;
	win0 = (u32) SORCERY_PCI_SPACE_MEM;
	win1 = (u32) SORCERY_PCI_SPACE_IO;
	win2 = (u32) SORCERY_PCI_SPACE_CFG;

	/* Assert PCI reset */
	out_be32 (&xipci->glb_stat_ctl, PCI_GLB_STAT_CTRL_PR);
	
	/* Disable prefetching but read-multiples will still prefetch */
	out_be32 (&xipci->target_ctrl, 0x00000000);

	printk (KERN_DEBUG "XLB_CONFIG = %08x\n", 
		in_be32((volatile u32 *)MPC_XLB_CONFIG));
	/* Set the XLB arbiter master priority enable register (MBAR + 0x264) */
	out_be32((volatile u32 *)MPC_XLB_PRIEN, 0x000000ff);
	printk(KERN_DEBUG "Sorcery XLB: Arbiter Master Priority Enable Register = %x\n", 
		in_be32((volatile u32 *)MPC_XLB_PRIEN));

	/* Set all XLB masters to the same priority (MBAR + 0x268) */
	/* PCI is required to be at least the same priority as the CPU */
	out_be32((volatile unsigned *)MPC_XLB_PRI, 0x00000000);
	printk(KERN_DEBUG "Sorcery XLB: Arbiter Master Priority Register = %x\n",
		in_be32((volatile unsigned *)MPC_XLB_PRI));

	/* Initiator windows */
	out_be32 (&xipci->init_win0,  (win0 >> 16) | win0 | 0x003f0000);   /* 1GB */
	out_be32 (&xipci->init_win1, ((win1 >> 16) | win1 ));
	out_be32 (&xipci->init_win2, ((win2 >> 16) | win2 ));

	out_be32 (&xipci->init_win_cfg,
		PCI_INIT_WIN_CFG_WIN0_CTRL_EN |
		PCI_INIT_WIN_CFG_WIN1_CTRL_EN | PCI_INIT_WIN_CFG_WIN1_CTRL_IO |
		PCI_INIT_WIN_CFG_WIN2_CTRL_EN | PCI_INIT_WIN_CFG_WIN2_CTRL_IO);

	out_be32 (&xipci->init_ctrl, 0x00000000);

	/* Enable bus master and mem access */
	out_be32 (&xipci->stat_cmd_reg, PCI_STAT_CMD_B | PCI_STAT_CMD_M);

	/* Cache line size and master latency */
	out_be32 (&xipci->bist_htyp_lat_cshl, (0xf8 << PCI_CFG1_LT_SHIFT));
	
	out_be32 (&xipci->base0, PCI_BASE_ADDR_REG0); /* 256MB - MBAR space */
	out_be32 (&xipci->base1, PCI_BASE_ADDR_REG1); /* 1GB - SDRAM space */

	out_be32 (&xipci->target_bar0,
		PCI_TARGET_BASE_ADDR_REG0 | PCI_TARGET_BASE_ADDR_EN);
	out_be32 (&xipci->target_bar1,
		PCI_TARGET_BASE_ADDR_REG1 | PCI_TARGET_BASE_ADDR_EN);

	/* Deassert reset bit */
	out_be32 (&xipci->glb_stat_ctl, 0x00000000);

	/* Set PCIGNT1, PCIREQ1, PCIREQ0/PCIGNTIN, PCIGNT0/PCIREQOUT, PCIREQ2, PCIGNT2 */
	out_be32((volatile u32 *)MPC8220_PCFG3,
		(in_be32((volatile u32 *)MPC8220_PCFG3) & 0xFC3FCE7F));
	out_be32((volatile u32 *)MPC8220_PCFG3,
		(in_be32((volatile u32 *)MPC8220_PCFG3) | 0x01400180));

}

/*****************************************************************************/
void sorcery_find_bridges(void)
{
	struct pci_controller *hose;

	pci_dram_offset = 0;

	/* Allocate memory for pci controller, aka hose */
	hose = pcibios_alloc_controller();
	if (!hose) {
		printk(KERN_WARNING "PCI: System does not support PCI\n");
		return;
	}

	/* Initialize PCI Windows and Registers */
	sorcery_pci_init();

	/* Check if we get the hose right */
	setup_sorcery_pci(hose, SORCERY_PCI_CFG_ADDRESS, SORCERY_PCI_CFG_DATA);
	printk(KERN_DEBUG "Setup Sorcery PCI Controller:\n");
	printk(KERN_DEBUG "Hose = 0x%08x, hose->cfg_addr=%p,	\
		hose->cfg_data=%p\n", (int)hose, hose->cfg_addr, 
		hose->cfg_data);

	/* Host sorcery pci controller's device number is 16 */
	hose->first_busno = 0x00;
	hose->last_busno = 0xff;
	hose->pci_mem_offset = SORCERY_PCI_OFF_MEM_PHY;

	/* Setup bridge memory/IO ranges & resources */
	hose->io_space.start	= SORCERY_PCI_IO_LOWER;
	hose->io_space.end	= SORCERY_PCI_IO_UPPER;
	hose->mem_space.start	= SORCERY_PCI_MEM_LOWER;
	hose->mem_space.end	= SORCERY_PCI_MEM_UPPER;
	hose->io_base_phys	= SORCERY_PCI_IO_PHY_START;

	/* Both get the same 81m */
	printk(KERN_DEBUG "Hose IO Base Physical = 0x%08x\n", 
		(int)hose->io_base_phys);
	hose->io_base_virt = ioremap(hose->io_base_phys, 0x10000);
	printk(KERN_DEBUG "Hose IO Base Virtual = 0x%08x\n", 
		(int)hose->io_base_virt);

	isa_io_base = (unsigned long)hose->io_base_virt;

	hose->io_resource.start		= SORCERY_PCI_IO_LOWER;
	hose->io_resource.end		= SORCERY_PCI_IO_UPPER;
	hose->io_resource.flags		= IORESOURCE_IO;
	hose->io_resource.name		= "PCI I/O";
	hose->mem_resources[0].start	= SORCERY_PCI_MEM_LOWER;
	hose->mem_resources[0].end	= SORCERY_PCI_MEM_UPPER;
	hose->mem_resources[0].flags	= IORESOURCE_MEM;
	hose->mem_resources[0].name	= "PCI Memory";

	/* Auto bus scan only finds one bus */
	ppc_md.pci_exclude_device	= sorcery_exclude_device;
	hose->last_busno		= pciauto_bus_scan(hose, 
						hose->first_busno);

	ppc_md.pcibios_fixup		= NULL;
	ppc_md.pcibios_fixup_resources	= NULL;
	ppc_md.pci_swizzle		= common_swizzle;
	ppc_md.pci_map_irq		= sorcery_map_irq;
}

/*****************************************************************************/
static int sorcery_exclude_device(unsigned char bus, unsigned char devfn)
{
	/* We prevent us from seeing ourselves to avoid having
	 * the kernel try to remap our BAR #1 and mess up bus
	 * master from external PCI devices
	 */
	if ((bus == 0) && (PCI_SLOT(devfn) == SORCERY_HOST_BRIDGE_IDSEL))
		return PCIBIOS_DEVICE_NOT_FOUND;
	else
		return PCIBIOS_SUCCESSFUL;
}

/*****************************************************************************/
static int sorcery_map_irq(struct pci_dev *dev, unsigned char idsel, 
			unsigned char pin)
{
	/*
	 *	PCI IDSEL/INTPIN->INTLINE
	 * 	   A | B | C | D
	 */
	static char pci_irq_table[][4] =
	{
		{ SORCERY_MERC_INT,    SORCERY_MERC_INT,    SORCERY_MERC_INT,    
			SORCERY_MERC_INT    },  /* IDSEL 17 - PCI slot 1 */
		{ SORCERY_AQUAR_INT,   SORCERY_AQUAR_INT,   SORCERY_AQUAR_INT,   
			SORCERY_AQUAR_INT   },  /* IDSEL 18 - PCI slot 2 */
		{ SORCERY_SATA_INT,    SORCERY_SATA_INT,    SORCERY_SATA_INT,    
			SORCERY_SATA_INT    },  /* IDSEL 19 - PCI slot 3 */
		{ SORCERY_RISER_B_INT, SORCERY_RISER_A_INT, SORCERY_RISER_C_INT,
			SORCERY_RISER_D_INT },  /* IDSEL 20 - PCI slot 4 */
	};

	const long min_idsel = 17, max_idsel = 20, irqs_per_slot = 4;

	return PCI_IRQ_TABLE_LOOKUP;
}

