/*
 * (C) Copyright 2006
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 */


/*
 * This file implements power failure handling for the V38B board. The
 * interface uses /proc entires and allows to:
 * - enable/disable the backup battery power
 * - check the backup battery power status
 * - check the current power failure status
 * - check the time (epoch) of the most recent power failure
 * - use select(2) to check for updates of the power failuer time 
 */

#include <linux/ctype.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/poll.h>
#include <linux/proc_fs.h>
#include <linux/spinlock.h>
#include <linux/types.h>
#include <linux/time.h>
#include <linux/wait.h>

#include <asm/mpc5xxx.h>
#include <asm-ppc/poll.h>


#define PF_DEBUG
#undef PF_DEBUG

#ifdef PF_DEBUG
#define DEBUGF(x, args...) printk(x, ## args)
#else
#define DEBUGF(x, args...)
#endif


#define MODULE_VERSION	"1.0"
#define MODULE_NAME	"v38b_proc_marel"


#define IRDA_TX		0x10000000UL
#define PSC6_2		IRDA_TX		/* PSC6_2 is called IrDA_TX */
#define GPIO_WKUP_7	0x80000000UL


/* names we create in /proc */
static const char *proc_dir_name = "marel";
static const char *proc_entry_board_status_name = "board-status";
static const char *proc_entry_battery_enable_name = "battery-enable";

/* strings returned whan reading from our entires in /proc */
static const char *pf_time_str = "PowerFailureTime=";
static const char *pf_status_str = "PowerFailureStatus=";
static const char *battery_enable_str = "BatteryEnable=";

/* /proc entry */
static struct proc_dir_entry *proc_dir;


DECLARE_WAIT_QUEUE_HEAD(pf_queue);	/* wait queue */

/* two variables in critical section */
static time_t pf_time = 0;	/* time of the most recent power failure */
static time_t pf_time_read = 0;	/* time of the most recently read failure */

spinlock_t pf_times_lock = SPIN_LOCK_UNLOCKED;

/* GPIO registers */
struct mpc5xxx_gpio *gpio = (struct mpc5xxx_gpio *)MPC5xxx_GPIO;



/*
 * Power-failure interrupt handler.
 */
void pf_irq(int irq, void *dev_id, struct pt_regs *regs)
{
	struct timeval tv;

	DEBUGF("pf_irq: power-fail int fired!\n");

	do_gettimeofday(&tv);

	/*
	 * Update the time of the most recent power failure. Note that time_t
	 * uses one machine word and reading/writing it is atomic on PPC. Care
	 * must be taken if this is to be ported to other architectures, as
	 * the atomicity must be preserved.
	 */
	pf_time = tv.tv_sec;

	wake_up_interruptible(&pf_queue);
}


/*
 * Set up the interrupt for the power-fail event (IRQ0, falling edge active).
 */
static int __init intr_setup(void)
{
	int r;

	r = request_irq(MPC5xxx_IRQ0, pf_irq, SA_INTERRUPT,
			"power-fail", NULL);
	if (r < 0) {
		printk("%s: could not allocate IRQ %d\n", __FUNCTION__,
		       MPC5xxx_IRQ0);
		return -EINVAL;
	}

	return 0;
}


/*
 * Updates the read time, and outputs the time of last failure, and current
 * status of power failure. Reading of pf_time written by the interrupt
 * handler must be atomic - see note in the handler. If the atomicity can't
 * be guaranteed, then the interrupts must be disabled in the critical section
 * (e.g., spin_lock_irqsave()/spin_unlock_irqrestore() must be used instead of
 * spin_lock()/spin_uinlock()).
 */
static int read_board_status(char *page, char **start, off_t off, int count,
                         int *eof, void *data)
{
	int len;
	time_t tmp_time;

	/* update the time of the most recent read of the power failure time */
	spin_lock(&pf_times_lock);
	pf_time_read = pf_time;
	tmp_time = pf_time_read;
	spin_unlock(&pf_times_lock);

	len = sprintf(page + off, "%s%ld\n\%s%u\n",
		      pf_time_str, tmp_time,
		      pf_status_str,
		      !(((*(volatile u32 *)MPC5xxx_WU_GPIO_DATAIN) & GPIO_WKUP_7)
			== GPIO_WKUP_7));

	if (len <= off + count) {
		*eof = 1;
	}

	*start = page + off;
	len -= off;
	if (len > count) {
		len = count;
	}

	if (len < 0) {
		len = 0;
	}

	return len;
}


/*
 * Checks if we have a power failure time update that has not beed read.
 * Reports the state via select() system call, which reports "exception"
 * condition on the filedes associated with the /proc entry when there is
 * power failure time update. Reading of pf_time written by the interrupt
 * handler must be atomic - see note in the handler. If the atomicity can't be
 * guaranteed, then the interrupts must be disabled in the critical section
 * (e.g., spin_lock_irqsave()/spin_unlock_irqrestore() must be used instead of
 * spin_lock()/spin_uinlock()).
 */
static unsigned int poll_board_status(struct file *f,
				      struct poll_table_struct *pt)
{
	unsigned int poll_mask;

	poll_wait(f, &pf_queue, pt);
	spin_lock(&pf_times_lock);
	poll_mask = (pf_time > pf_time_read) ? POLLPRI : 0x0;
	spin_unlock(&pf_times_lock);

	return poll_mask;
}


/*
 * Used to report the state of the battery backup pin.
 */
static int read_battery_enable(char *page, char **start, off_t off, int count,
			       int *eof, void *data)
{
	int len;

	len = sprintf(page + off, "%s%u\n", battery_enable_str,
		      (gpio->simple_dvo & PSC6_2) == PSC6_2);

	if (len <= off + count) {
		*eof = 1;
	}

	*start = page + off;
	len -= off;
	if (len > count) {
		len = count;
	}

	if (len < 0) {
		len = 0;
	}

	return len;
}


/*
 * Enable ('1') or disable ('0') backup battery.
 */
static int write_battery_enable(struct file *file, const char *buffer,
				unsigned long count, void *data)
{
	int r;
	char val;

	r = copy_from_user(&val, buffer, sizeof(val));
	if (r) {
		printk(KERN_WARNING "%s(): copy_from_user() failed\n",
		       __FUNCTION__);
		return r;
	}

	switch (val) {
	case '0':
		/* drive PSC6_2 pin low */
		gpio->simple_dvo &= ~PSC6_2;
		break;
	case '1':
		/* drive PSC6_2 pin high */
		gpio->simple_dvo |= PSC6_2;
		break;
	default:
		DEBUGF("write_battery_enable: unknown value 0x%x\n", val);
		/* dolce far niente */
		break;
	}

	return count;
}


/*
 * Initalisation: create /proc entries, configure pins, setup irq.
 */
static int __init init_proc_marel(void)
{
	int error;
	static struct proc_dir_entry *entry;

	proc_dir = proc_mkdir(proc_dir_name, NULL);
	if (proc_dir == NULL) {
		printk("%s: unable to create /proc/%s\n", __FUNCTION__,
			proc_dir_name);
		return -EINVAL;
	}


	/* create entry for the board status */
	entry = create_proc_entry(proc_entry_board_status_name,
				  S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH,
				  proc_dir);
	if (entry == NULL) {
		printk("%s: Unable to create proc entry '%s'\n", __FUNCTION__,
		proc_entry_board_status_name);
		return -EINVAL;
	}
	entry->owner = THIS_MODULE;
	entry->read_proc = read_board_status;
	entry->proc_fops->poll = poll_board_status;

	/* create entry for battery enable */
	entry = create_proc_entry(proc_entry_battery_enable_name,
				  S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH,
				  proc_dir);
	if (entry == NULL) {
		printk("%s: Unable to create proc entry '%s'\n", __FUNCTION__,
		proc_entry_battery_enable_name);
		return -EINVAL;
	}
	entry->owner = THIS_MODULE;
	entry->read_proc = read_battery_enable;
	entry->write_proc = write_battery_enable;


	/* configure PSC6_2 pin (BKUP_POWER_ON) as GPIO output */
	gpio->port_config &= ~0x00700000;
	gpio->simple_gpioe |= PSC6_2;
	gpio->simple_ddr |= PSC6_2;


	/* configure GPIO_WKUP_7 pin as GPIO input */
	*(volatile u32 *) MPC5xxx_WU_GPIO_ENABLE |= GPIO_WKUP_7;


	/* allocate interrupt for power-fail event */
	error = intr_setup();
	if (error)
		return error;

	return 0;
}


/*
 * Clean up after ourselves.
 */
static void __exit exit_proc_marel(void)
{
	free_irq(MPC5xxx_IRQ0, NULL);
	remove_proc_entry(proc_entry_board_status_name, proc_dir);
	remove_proc_entry(proc_entry_battery_enable_name, proc_dir);
	remove_proc_entry(proc_dir_name, NULL);
}


module_init(init_proc_marel);
module_exit(exit_proc_marel);

MODULE_AUTHOR("Wolfgang Denk <wd@denx.de>");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Power failure handling for the V38B board");

EXPORT_NO_SYMBOLS;
