/***********************************************************************
 *
 * (C) Copyright 2005
 * DENX Software Engineering
 * Heiko Schocher, hs@denx.de
 * All rights reserved.
 *
 * This driver provides access to the FIFO IDT72V225
 * on a 3DMSM Evaluationboard with a MPC5200
 *
 *
 ***********************************************************************/

/*
 * Standard in kernel modules
 */
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/smp_lock.h>
#include <linux/version.h>
#include <linux/string.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <asm/types.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/time.h>
#include <asm/mpc5xxx.h>

#include <linux/proc_fs.h>
#include <linux/vmalloc.h>
#include <linux/config.h>

#ifndef KERNEL_VERSION
# define KERNEL_VERSION(a,b,c)	(((a) << 16) + ((b) << 8) + (c))
#endif

#if (!defined(LINUX_VERSION_CODE) || LINUX_VERSION_CODE < KERNEL_VERSION(2,2,0))
#error "You need to use at least Linux kernel version 2.2.0"
#endif

/***************************************************************************
 * Global stuff
 ***************************************************************************
 */
/* #define	DEBUG	1*/

/*
 * Deal with CONFIG_MODVERSIONS
 */
#if CONFIG_MODVERSIONS==1
/* # define MODVERSIONS */
# include <linux/modversions.h>
#endif

/*
 * For character devices
 */
#include <linux/fs.h>		/* character device definitions	*/
#include <linux/wrapper.h>	/* wrapper for compatibility with future versions */

#define	CFG_FIFO_PADDR	0xe4000000		/* we make this later in a .h File */
#define CFG_FIFO_IRQ	MPC5xxx_IRQ1

static 	int dmsm_fifo_pbase = CFG_FIFO_PADDR;
static	int dmsm_fifo_base;

#ifdef MODULE
MODULE_PARM(dmsm_fifo_pbase, "i");
__setup("dmsm_fifo_pbase=", dmsm_fifo_pbase);
#endif

/*
 * Device Declarations
 */

/*
 * The name for our device, as it will appear in /proc/devices
 */
#define DEVICE_NAME	"3dmsmfifo"
#define	DMSM_VERSION	"$Revision: 1.0 $ hs@denx.de"
#define DMSM_MAJOR	240

#define DMSM_IRQ                CFG_FIFO_IRQ
#define DMSM_BLOCK_POINTS	CONFIG_FIFO_BLOCK_SIZE
#define DMSM_BLOCK_SIZE         (DMSM_BLOCK_POINTS * 2)
#define DMSM_NEW_BLOCKS         CONFIG_FIFO_NUMBER_OF_BLOCKS

extern int errno;

/***************************************************************************
 * Local stuff
 ***************************************************************************
 */

#ifdef	DEBUG
# define debugk(fmt,args...)	printk(fmt ,##args)
# define DEBUG_OPTARG(arg)	arg,
#else
# define debugk(fmt,args...)
# define DEBUG_OPTARG(arg)
#endif

/*
 * Prototypes
 */
static int		dmsm_open (struct inode *, struct file *);
static int		dmsm_release (struct inode *, struct file *);
static ssize_t 		dmsm_read (struct file *, char *, size_t,
				loff_t *);
int init_module(void);
void cleanup_module(void);

static int dmsm_init_fifo (void);
static int dmsm_free_fifo (void);

/* avoid multiple access */
static char dmsm_status = 0;
static unsigned long dmsm_port;

typedef struct _dmsm_block
{
	char	data[DMSM_BLOCK_SIZE];
	int	pos;
	struct _dmsm_block	*next;
}DMSM_BLOCK;

static int dmsm_new_blocks (unsigned int);
static int dmsm_release_memory (void);
static int dmsm_free_block (void);
static DMSM_BLOCK *dmsm_get_block (void);
static void dmsm_interrupt(int, void *, struct pt_regs *);

static 	int	dmsm_size = 0;
static 	int	dmsm_max = 0;
static 	int	dmsm_buffer_full = 0;
static 	int	dmsm_inuse = 0;
static	u32	dmsm_maxtime = 0;
static	u32	dmsm_lasttime = 0;
static	int	dmsm_IsFull = 0;
static	int	dmsm_IsFree = 0;
static	DMSM_BLOCK	*dmsm_used = NULL;
static	DMSM_BLOCK	*dmsm_used_last = NULL;
static	DMSM_BLOCK	*dmsm_free = NULL;
static	DMSM_BLOCK	*dmsm_free_last = NULL;
static	DMSM_BLOCK	dmsm_temp;

static	wait_queue_head_t	wq;

/***************************************************************************
 * Driver interface
 ***************************************************************************
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)

static struct file_operations dmsm_fops = {
	NULL,			/* lseek	*/
	dmsm_read,		/* read		*/
	NULL,			/* write	*/
	NULL,			/* readdir	*/
	NULL,			/* select	*/
	NULL,			/* ioctl	*/
	NULL,			/* mmap		*/
	dmsm_open,		/* open		*/
	NULL,			/* flush	*/
	dmsm_release		/* close	*/
};

#else	/* Linux kernel >= 2.3.x */

static struct file_operations dmsm_fops = {
	owner:			THIS_MODULE,
	open:			dmsm_open,
	release:		dmsm_release,
	read:			dmsm_read,
};
#endif

#define DMSM_USE_PROC

#ifdef DMSM_USE_PROC /* don't waste space if unused */
/*
 * The proc filesystem: function to read and entry
 */

static int dmsm_read_procmem (char *buf, char **start, off_t offset,
                   int count, int *eof, void *data)
{
	int	len = 0;

	len += sprintf (buf + len, "\n");
	len += sprintf (buf + len, "BlockSize: %d ", CONFIG_FIFO_BLOCK_SIZE);
	len += sprintf (buf + len, "Size: %d ", dmsm_size);
	len += sprintf (buf + len, "used: %d ", dmsm_inuse);
	len += sprintf (buf + len, "maxU: %d ", dmsm_max);
	len += sprintf (buf + len, "Full: %d ", dmsm_buffer_full);
	len += sprintf (buf + len, "lastTime: %d ", dmsm_lasttime);
	len += sprintf (buf + len, "maxTime: %d ", dmsm_maxtime);
	len += sprintf (buf + len, "\n");
	return len;
}

#ifdef USE_PROC_REGISTER

static int dmsm_get_info (char *buf, char **start, off_t offset,
                int len, int unused)
{
    int eof = 0;
    return dmsm_read_procmem(buf, start, offset, len, &eof, NULL);
}


struct proc_dir_entry dmsm_proc_entry = {
        0,                 /* low_ino: the inode -- dynamic */
        9, "3dmsmfifo",     /* len of name and name */
        S_IFREG | S_IRUGO, /* mode */
        1, 0, 0,           /* nlinks, owner, group */
        0, NULL,           /* size - unused; operations -- use default */
        dmsm_get_info,   /* function used to read data */
        /* nothing more */
    };


static inline void create_proc_read_entry (const char *name, mode_t mode,
                struct proc_dir_entry *base, void *read_func, void *data)
{
    proc_register_dynamic (&proc_root, &dmsm_proc_entry);
}

static inline void remove_proc_entry (char *name, void *parent)
{
    proc_unregister (&proc_root, dmsm_proc_entry.low_ino);
}

#endif /* USE_PROC_REGISTER */
#endif /* DMSM_USE_PROC */

/***************************************************************************
 *
 * Initialize the driver - Register the character device
 */
static int __init dmsm_init (void)
{
	printk (KERN_INFO "3DMSM FIFO DRIVER  " DMSM_VERSION "\n");
	if (register_chrdev (DMSM_MAJOR, DEVICE_NAME, & dmsm_fops))
	{
		printk (KERN_ERR "%s_drv: register_chrdev failed\n", DEVICE_NAME);
		return -EIO;
	}
	init_waitqueue_head (&wq);

#ifdef DMSM_USE_PROC /* only when available */
    create_proc_read_entry(DEVICE_NAME, 0, NULL, dmsm_read_procmem, NULL);
#endif
	return 0;
}

/***************************************************************************
 *
 * called whenever a process attempts to open the device
 * the device can only opened by one process
 */
static int dmsm_open (struct inode *inode, struct file *file)
{
	lock_kernel ();

	if (dmsm_status) {
		unlock_kernel ();
		return -EBUSY;
	}
	dmsm_status = 1;

	if (dmsm_new_blocks (DMSM_NEW_BLOCKS)) {
                printk (KERN_ERR "%s: not enough memory!\n", DEVICE_NAME);
		dmsm_release_memory ();
		unlock_kernel ();
		return -ENOMEM;
	}

	/* enable Memory Controller */
	dmsm_init_fifo ();

        if (request_irq (DMSM_IRQ, dmsm_interrupt, SA_INTERRUPT, DEVICE_NAME, (void *)&dmsm_port)) {
                printk (KERN_ERR "%s: cannot register IRQ %d\n", DEVICE_NAME, DMSM_IRQ);
		dmsm_release_memory ();
		dmsm_free_fifo ();
		unlock_kernel ();
                return -EIO;
        }

	disable_irq (DMSM_IRQ);
        /* clear the FIFO, to synchronize it */
	memcpy (&dmsm_temp, (char *)dmsm_fifo_base, DMSM_BLOCK_SIZE);
	enable_irq (DMSM_IRQ);

	unlock_kernel ();
	MOD_INC_USE_COUNT;

	debugk ("%s open success\n", DEVICE_NAME);
	dmsm_max = 0;
	dmsm_buffer_full = 0;
	dmsm_IsFull = 0;
	dmsm_IsFree = 0;
	return 0;
}



/***************************************************************************
 *
 * Called when a process closes the device.
 */
static int dmsm_release (struct inode *inode, struct file *file)
{
	lock_kernel ();
	dmsm_status = 0;

	free_irq (DMSM_IRQ, (void *) &dmsm_port);
	dmsm_free_fifo ();

	dmsm_release_memory ();
	unlock_kernel ();

	MOD_DEC_USE_COUNT;

	debugk ("%s close success\n", DEVICE_NAME);
	return 0;
}

/***************************************************************************
 *
 * Called when a process reads from the device.
 */
#define DMSM_BL_SIZE	(DMSM_BLOCK_SIZE - dmsm_used->pos)

static ssize_t dmsm_read (struct file *filep, char *buf, size_t count,
				loff_t *f_pos)
{
	int	len = 0;

	debugk ("%s read count: %d Pos:%d inuse:%d\n", DEVICE_NAME, count, (int)*f_pos, dmsm_inuse);
	if ((dmsm_IsFull) && (dmsm_IsFree == 0)) {
		/* There was an overflow -> post this as an error */
		dmsm_IsFull = 0;
		dmsm_IsFree = 1;
		errno = ENOMEM;
		return -1;
	}
	if (dmsm_inuse == 0) {
		if (filep->f_flags & O_NONBLOCK) {
			return -EAGAIN;
		} else {
			interruptible_sleep_on (&wq);
		}
	}
	while (len < count) {
		debugk ("%s read len: %d count: %d used: %p\n", DEVICE_NAME, len, count, dmsm_used);
		if (dmsm_used == NULL) {return len;}
		debugk ("%s read BL_SIZE: %d\n", DEVICE_NAME, DMSM_BL_SIZE);
		if ((DMSM_BL_SIZE + len) < count) {
			copy_to_user (buf, &(dmsm_used->data[dmsm_used->pos]), DMSM_BL_SIZE);
			len += DMSM_BL_SIZE;
			dmsm_free_block ();
		} else {
			int size = count - len;
			debugk ("%s read size: %d\n", DEVICE_NAME, size);
			copy_to_user (buf, &(dmsm_used->data[dmsm_used->pos]), size);
			len += size;
			dmsm_used->pos += size;
			if (dmsm_used->pos >= DMSM_BLOCK_SIZE) {
				dmsm_free_block();
			}
		}
	}
	return len;
}


/***************************************************************************
 * Managing the Buffer
 ***************************************************************************
 */
static int dmsm_new_blocks (unsigned int count)
{
	DMSM_BLOCK	*new;
	unsigned int	i;
	
	for (i = 0; i < count; i++) {
		new = (DMSM_BLOCK *)kmalloc (sizeof(DMSM_BLOCK), GFP_KERNEL);
		if (!new) {
			debugk ("%s no memory for block %d\n", DEVICE_NAME, dmsm_size);
			return -ENOMEM;
		}
		debugk ("%s allocate block %d\n", DEVICE_NAME, dmsm_size);
		new->next = NULL;
		new->pos = 0;
		dmsm_size ++;
		if (dmsm_free == NULL) {
			dmsm_free = new;
		} else {
			dmsm_free_last->next = new;
		}
		dmsm_free_last = new;
	}
	debugk ("%s allocate %d blocks OK\n", DEVICE_NAME, count);
	return 0;
}

static int dmsm_release_memory (void)
{
	DMSM_BLOCK	*tmp;

	dmsm_size = 0;
	dmsm_max  = 0;
	dmsm_inuse = 0;
	dmsm_buffer_full = 0;
	while (dmsm_used != NULL) {
		tmp = dmsm_used;
		kfree (tmp);
		dmsm_used = dmsm_used->next;
		debugk ("%s release one used block\n", DEVICE_NAME);
	}
	dmsm_used_last = NULL;
	while (dmsm_free != NULL) {
		tmp = dmsm_free;
		kfree (tmp);
		debugk ("%s release one free block\n", DEVICE_NAME);
		dmsm_free = dmsm_free->next;
	}
	dmsm_free_last = NULL;
	return 0;
}

static int dmsm_free_block (void)
{
	DMSM_BLOCK	*tmp;

	if (dmsm_used == NULL) {
		return -1;
	}
	/* we must disable the interrupt */
	disable_irq (DMSM_IRQ);
	tmp = dmsm_used;
	dmsm_used = dmsm_used->next;
	dmsm_inuse --;
	if (dmsm_free == NULL) {
		dmsm_free = tmp;
	} else {
		dmsm_free_last->next = tmp;
	}
	tmp->next = NULL;
	dmsm_free_last = tmp;
	if (dmsm_inuse < (DMSM_NEW_BLOCKS / 2)) {
		dmsm_IsFree = 0;
		dmsm_IsFull = 0;
	}
	/* now not forget to enable the interrupt */
	enable_irq (DMSM_IRQ);
	return 0;
}

static inline DMSM_BLOCK *dmsm_get_block (void)
{
	DMSM_BLOCK	*tmp = NULL;

	if (dmsm_free == NULL) {
		dmsm_buffer_full ++;
	} else {
		tmp = dmsm_free;
		dmsm_free = dmsm_free->next;
		if (dmsm_used == NULL) {
			dmsm_used = tmp;
		} else {
			dmsm_used_last->next = tmp;
		}
		tmp->next = NULL;
		tmp->pos = 0;
		dmsm_used_last = tmp;
		dmsm_inuse ++;
		if (dmsm_inuse > dmsm_max) dmsm_max = dmsm_inuse;
	}
	return tmp;
}

/***************************************************************************
 * Hardware related stuff
 ***************************************************************************
 */
#if	(CONFIG_FIFO_LP_CS == 3)
#define MPC5xxx_CSSTART           (MPC5xxx_MBAR + 0x001c)
#define MPC5xxx_CSEND             (MPC5xxx_MBAR + 0x0020)
#define MPC5xxx_CSCONFIG          (MPC5xxx_MBAR + 0x030c)
#endif
#if	(CONFIG_FIFO_LP_CS == 4)
#define MPC5xxx_CSSTART           (MPC5xxx_MBAR + 0x0024)
#define MPC5xxx_CSEND             (MPC5xxx_MBAR + 0x0028)
#define MPC5xxx_CSCONFIG          (MPC5xxx_MBAR + 0x0310)
#endif
#if	(CONFIG_FIFO_LP_CS == 5)
#define MPC5xxx_CSSTART           (MPC5xxx_MBAR + 0x002c)
#define MPC5xxx_CSEND             (MPC5xxx_MBAR + 0x0030)
#define MPC5xxx_CSCONFIG          (MPC5xxx_MBAR + 0x0314)
#endif

static int dmsm_init_fifo (void)

{
        struct mpc5xxx_intr *intr = (struct mpc5xxx_intr *)MPC5xxx_INTR;
	u32	adren;

	/* enable Memory Controller */
	out_be32 ((u32 *)MPC5xxx_CSSTART, dmsm_fifo_pbase >> 16);
	out_be32 ((u32 *)MPC5xxx_CSEND, (dmsm_fifo_pbase >> 16) + 1);
	out_be32 ((u32 *)MPC5xxx_CSCONFIG, 0x00001501);
	adren = in_be32 ((u32 *)MPC5xxx_ADREN);
	adren |= (0x00010000 << CONFIG_FIFO_LP_CS);
	out_be32 ((u32 *)MPC5xxx_ADREN, adren);

        /* map physical address */
        dmsm_fifo_base = (unsigned long)ioremap(dmsm_fifo_pbase, 0x10000);
        if(!dmsm_fifo_base) {
                printk("%s: ioremap FIFO failed @%x\n", DEVICE_NAME, dmsm_fifo_pbase);
                return -EIO;
        }

	/* IRQ edge sensitive falling edge */
	adren = in_be32(&intr->ctrl);
	adren |= 0x00200000;
	adren &= ~0x00100000;
        out_be32(&intr->ctrl, adren);
        /* Clear Interrupt */
        adren = in_be32(&intr->ctrl);
        adren |= 0x04000000;
        out_be32(&intr->ctrl, adren);
	return 0;
}

static int dmsm_free_fifo (void)
{
	u32	adren;

	iounmap ((void *)dmsm_fifo_base);
	/* disable Memory Controller */
	adren = in_be32 ((u32 *)MPC5xxx_ADREN);
	adren &= ~(0x00010000 << CONFIG_FIFO_LP_CS);
	out_be32 ((u32 *)MPC5xxx_ADREN, adren);

	return 0;
}

#define rdtscl(dest) \
  __asm__ __volatile__("mftbl %0; nop" : "=r" (dest))

static void dmsm_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	DMSM_BLOCK	*new;
	u16		*data;
	u32		starttime, stoptime;

	rdtscl(starttime);
	new = dmsm_get_block ();
	if (new) {
		data = (u16 *) new;
	} else {
		data = (u16 *) &dmsm_temp;
		dmsm_IsFull = 1;
	}
	/* First read is an error -> ignore it !!! */
	*data = readw (dmsm_fifo_base);
	/* Now read the Fifo */
	memcpy (data, (char *)dmsm_fifo_base, DMSM_BLOCK_SIZE);
	wake_up (&wq);
	rdtscl(stoptime);
	dmsm_lasttime = stoptime - starttime;
	if (dmsm_maxtime < dmsm_lasttime) {
		dmsm_maxtime = dmsm_lasttime;
	}
}

/***************************************************************************
 * Module Declarations
 ***************************************************************************
 */
module_init(dmsm_init);

#ifdef MODULE
/*
 * Cleanup - unregister the appropriate file from /proc
 */
static void dmsm_cleanup (void)
{
	if (unregister_chrdev (DMSM_MAJOR, DEVICE_NAME))
	{
		printk (KERN_ERR "%s: failed to unregister character device\n", DEVICE_NAME);
	}

#ifdef DMSM_USE_PROC
    remove_proc_entry("dmsm_inf", 0);
#endif

	debugk ("%s Driver de-registered\n", DEVICE_NAME);
}

module_exit(dmsm_cleanup);
#endif	/* MODULE */

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Heiko Schocher <hs@denx.de>");
MODULE_DESCRIPTION("DMSM driver");
