/*
 * Driver for the PDSP 1880 Display
 * (C) Copyright 2006 Heiko Schocher, DENX <hs@denx.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/version.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/poll.h>
#include <linux/delay.h>
#include <linux/display.h>

#include <linux/proc_fs.h>

#include <linux/miscdevice.h>
#include <asm/io.h>

#define DISPLAY_VERSION "1.0"
#define DISPLAY_NAME "display"

#define DISPLAY_MINOR	251

/* proc interface for driver */
#define DISPLAY_PROC	"driver/display"

#define DISPLAY_LEN	8

#define CONFIG_DISPLAY_PHYS_BASE	0x80600000

#define CTRL_BRIGHTNESS_MASK	0x07
#define CTRL_BLINK		0x10
#define CTRL_SELF_OK		0x20
#define CTRL_SELF_START		0x40
#define CTRL_CLEAR		0x80

volatile char *display_buf = NULL;
volatile char *display_ctrl;

typedef struct {
	char	flash_ram[32];
	char	udc_addr[8];
	char	udc_ram[8];
	char	control[8];
	char	ch_ram[8];
} DISPLAY;

static DISPLAY		*disp_base = NULL;



/* Froward declarations */

/* Driver's banner */
static char display_banner[] __initdata = KERN_INFO DISPLAY_NAME ": "
						DISPLAY_VERSION "\n";

static void display_clear (void)
{
	/* Clear display */
	*display_ctrl	= CTRL_CLEAR;
	mdelay(110 * 3);
}

static int display_set_brightness (char value)
{
	char tmp = *display_ctrl;

	tmp &= ~CTRL_BRIGHTNESS_MASK;
	tmp |= (value & CTRL_BRIGHTNESS_MASK);
	*display_ctrl = tmp;

	return 0;
}

/*
 * Device file operations
 */
static ssize_t display_write(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
	int	p = (int )*ppos;

	if (p >= DISPLAY_LEN) return 0;
	if (count > DISPLAY_LEN - p) count = DISPLAY_LEN - p;

	if (copy_from_user((void *) &display_buf[p], buf, count))
		return -EFAULT;

	*ppos += count;

	return count;
}

static ssize_t display_read(struct file *file, char *buf, size_t count, loff_t *ppos)
{
	unsigned long p = *ppos;

	/* on the UC101 board we can only write to the display :-( */
	if (p >= DISPLAY_LEN) return 0;
	if (count > DISPLAY_LEN - p) count = DISPLAY_LEN - p;
	if (copy_to_user(buf, (void *) &display_buf[p], count))
		return -EFAULT;

	*ppos += count;

	return count;

}

static unsigned int display_poll(struct file *file, struct poll_table_struct *ptbl)
{
	/* It is always possible to read from the display */
	return POLLIN | POLLRDNORM;
}

static loff_t display_lseek(struct file *file, loff_t offset, int orig)
{

	switch (orig) {
		case 0:
			file->f_pos = offset;
			return file->f_pos;
		case 1:
			file->f_pos += offset;
			return file->f_pos;
		case 2:
			file->f_pos = 
				DISPLAY_LEN - offset - 1;
			return file->f_pos;
		default:
			return -EINVAL;
	}
}


static int display_ioctl (struct inode * inode, struct file * file,
			 uint cmd, unsigned long arg)
{
	int err;
	unsigned int param;

	if (_IOC_TYPE(cmd) != DISPLAY_MAGIC) return -ENOTTY;

	if (_IOC_DIR(cmd) & _IOC_WRITE) {
		if ((err = get_user(param, (unsigned int *)arg)) != 0)
			return err;
	}
	switch (cmd) {
		case DISPLAY_SET_BRIGHTNESS:
			display_set_brightness (param);
			break;
		case DISPLAY_SET_CLEAR:
			display_clear ();
			break;
		default:
			return -ENOTTY;
	}
	return 0;
}

#ifdef MODULE
static int display_open(struct inode *inode, struct file *file)
{
	MOD_INC_USE_COUNT;
	return 0;
}

static int display_close(struct inode *inode, struct file *file)
{
	MOD_DEC_USE_COUNT;
	return 0;
}
#endif

static struct file_operations display_fops = {
	.owner = THIS_MODULE,
	.read	= display_read,
	.write	= display_write,
	.poll	= display_poll,
	.llseek = display_lseek,
	.ioctl	= display_ioctl,
#ifdef MODULE
	.open	= display_open,
	.release = display_close
#endif
};

static struct miscdevice display_miscdev = {
	DISPLAY_MINOR,
	DISPLAY_NAME,
	&display_fops
};

/*
 * procfs operations
 */
#ifdef CONFIG_PROC_FS

static int display_proc_read(char *page, char **start, off_t off,
					int count, int *eof, void *data)
{
	char *p = page;
	char	buf[DISPLAY_LEN + 1];
	int	i;

	/* on the UC101 board we can only write to the display :-( */
	memcpy (buf, (char *)display_buf, DISPLAY_LEN);
	for (i = 0; i < DISPLAY_LEN; i++) {
		p += sprintf (p, "%x ", buf[i]);
	}
	p += sprintf (p, "\n");
	return p - page;
}

static int display_proc_write(struct file *file, const char *buffer,
	unsigned long count, void *data)
{
	unsigned char	*buf = (unsigned char *)display_buf;

	if ( count > DISPLAY_LEN) count = DISPLAY_LEN;
	if ( copy_from_user(buf, buffer, count) ) {
		return -EFAULT;
	}
	return count;
}
#endif

/*
 * Initialization/cleanup
 */
static int __devinit display_init(void)
{
	printk(display_banner);

#ifdef CONFIG_PROC_FS
	struct proc_dir_entry * proc;
#endif

	if (misc_register(&display_miscdev) < 0) {
		printk(KERN_ERR "DISPLAY: can't misc_register on minor=%d\n",
			DISPLAY_MINOR);
		goto error;
	}

#ifdef CONFIG_PROC_FS
	if (!(proc = create_proc_entry(DISPLAY_PROC, S_IFREG|S_IRUGO, NULL))) {
		printk(KERN_ERR "failed to create /proc/" DISPLAY_PROC "\n");
		goto error_unreg;
	}

	proc->read_proc	 = display_proc_read;
	proc->write_proc = display_proc_write;
#endif

	/* ioremap the address */
	disp_base = (DISPLAY *)ioremap (CONFIG_DISPLAY_PHYS_BASE, 0x1000);
	if (disp_base == NULL) goto error_ioremap;

	display_ctrl	= (char *)disp_base + 0x30;
	display_buf	= (char *)disp_base + 0x38;

	return 0;

error_ioremap:
#ifdef CONFIG_PROC_FS
	remove_proc_entry(DISPLAY_PROC, NULL);
#endif
error_unreg:
	misc_deregister(&display_miscdev);
error:
	return -ENODEV;
}

static void __devexit display_cleanup(void)
{
	display_buf = NULL;
	display_ctrl = NULL;

	if (disp_base != NULL) iounmap (disp_base);

#ifdef CONFIG_PROC_FS
	remove_proc_entry(DISPLAY_PROC, NULL);
#endif

	misc_deregister(&display_miscdev);
}

module_init(display_init)
module_exit(display_cleanup)

MODULE_AUTHOR("Heiko Schocher <hs@denx.de>");
MODULE_DESCRIPTION("PDSP 1880 Display driver");
MODULE_LICENSE("GPL");
EXPORT_NO_SYMBOLS;
