/***********************************************************************
 *
 * (C) Copyright 2005
 * DENX Software Engineering
 * Stefan Roese, sr@denx.de
 * All rights reserved.
 *
 * This driver provides access to the Dallas DS1337
 * which is used for instance in TQM85xx systems.
 *
 * It uses the i2c-simple driver to access the device over the I2C bus.
 *
 ***********************************************************************/

/*
 * Standard in kernel modules
 */
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/module.h>
#include <linux/miscdevice.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/rtc.h>
#include <linux/smp_lock.h>
#include <linux/version.h>
#include <linux/i2c.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/init.h>
#include <asm/types.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/time.h>
#include <asm/machdep.h>		/* ppc_md */

#ifndef KERNEL_VERSION
# define KERNEL_VERSION(a,b,c)	(((a) << 16) + ((b) << 8) + (c))
#endif

#if (!defined(LINUX_VERSION_CODE) || LINUX_VERSION_CODE < KERNEL_VERSION(2,2,0))
#error "You need to use at least Linux kernel version 2.2.0"
#endif

/***************************************************************************
 * Global stuff
 ***************************************************************************
 */
 /*#define	DEBUG	1	//	*/

/*
 * Deal with CONFIG_MODVERSIONS
 */
#if CONFIG_MODVERSIONS==1
/* # define MODVERSIONS */
# include <linux/modversions.h>
#endif

/*
 * For character devices
 */
#include <linux/fs.h>		/* character device definitions	*/
#include <linux/wrapper.h>	/* wrapper for compatibility with future versions */


/*
 * Device Declarations
 */

/*
 * The name for our device, as it will appear in /proc/devices
 */
#define DEVICE_NAME	"rtc"
#define	RTC_VERSION	"$Revision: 1.0 $ sr@denx.de"

/***************************************************************************
 * Local stuff
 ***************************************************************************
 */

#ifdef	DEBUG
# define debugk(fmt,args...)	printk(fmt ,##args)
# define DEBUG_OPTARG(arg)	arg,
#else
# define debugk(fmt,args...)
# define DEBUG_OPTARG(arg)
#endif

static struct i2c_driver ds1337_driver;

static unsigned short ignore[] = { I2C_CLIENT_END };
static unsigned short normal_addr[] = { 0x68, I2C_CLIENT_END };

static struct i2c_client_address_data addr_data = {
	normal_i2c:		normal_addr,
	normal_i2c_range:	ignore,
	probe:			ignore,
	probe_range:		ignore,
	ignore:			ignore,
	ignore_range:		ignore,
	force:			ignore,
};

/*
 * Prototypes
 */
static int		ds1337_rtc_open (struct inode *, struct file *);
static int		ds1337_rtc_release (struct inode *, struct file *);
static int		ds1337_rtc_ioctl (struct inode *, struct file *,
				 unsigned int, unsigned long);
static unsigned int	bcd2bin (unsigned int);
static unsigned int	bin2bcd (unsigned int);
static unsigned char	rtc_rd (unsigned char);
static void		rtc_wr (unsigned char, unsigned char);
static unsigned long	ds1337_get_rtc_time(void);
static int		ds1337_set_rtc_time(unsigned long now);
static void		ds1337_reset_rtc(void);
static void		ds1337_set_system_time(void);

int init_module(void);
void cleanup_module(void);

/* avoid multiple access */
static char rtc_status = 0;

/* save/restore old machdep pointers */
int		(*save_set_rtc_time)(unsigned long);
unsigned long	(*save_get_rtc_time)(void);

/***************************************************************************
 * I2C stuff
 ***************************************************************************
 */
#define DAT(x) ((unsigned int)(x->data))

static struct i2c_client *clnt;

static int
ds1337_attach(struct i2c_adapter *adap, int addr, unsigned short flags,
	       int kind)
{
	int ret;

	clnt = kmalloc(sizeof(*clnt), GFP_KERNEL);
	if (!clnt)
		return -ENOMEM;

	strcpy(clnt->name, "DS1337");
	clnt->id	= ds1337_driver.id;
	clnt->flags	= 0;
	clnt->addr	= addr;
	clnt->adapter	= adap;
	clnt->driver	= &ds1337_driver;
	clnt->data	= NULL;

	ret = i2c_attach_client(clnt);

	if (! ret) {
		ds1337_set_system_time();
	}

	return ret;
}

static int
ds1337_probe(struct i2c_adapter *adap)
{
	return i2c_probe(adap, &addr_data, ds1337_attach);
}

static int
ds1337_detach(struct i2c_client *client)
{
	i2c_detach_client(client);
	return 0;
}

static struct i2c_driver ds1337_driver = {
	name:		"DS1337",
	id:		I2C_DRIVERID_DS1337,
	flags:		I2C_DF_NOTIFY,
	attach_adapter:	ds1337_probe,
	detach_client:	ds1337_detach,
	command:	NULL
};

/***************************************************************************
 * Driver interface
 ***************************************************************************
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)

static struct file_operations ds1337_rtc_fops = {
	NULL,			/* lseek	*/
	NULL,			/* read		*/
	NULL,			/* write	*/
	NULL,			/* readdir	*/
	NULL,			/* select	*/
	ds1337_rtc_ioctl,	/* ioctl	*/
	NULL,			/* mmap		*/
	ds1337_rtc_open,		/* open		*/
	NULL,			/* flush	*/
	ds1337_rtc_release	/* close	*/
};

#else	/* Linux kernel >= 2.3.x */

static struct file_operations ds1337_rtc_fops = {
	owner:			THIS_MODULE,
	open:			ds1337_rtc_open,
	release:		ds1337_rtc_release,
	ioctl:			ds1337_rtc_ioctl,
};
#endif

/*
 * We sponge a minor off of the misc major. No need slurping
 * up another valuable major dev number for this.
 */
static struct miscdevice rtc_dev = {
	RTC_MINOR,
	DEVICE_NAME,
	&ds1337_rtc_fops,
};


/***************************************************************************
 *
 * Initialize the driver - Register the character device
 */
static int __init ds1337_rtc_init (void)
{
	int error;
	printk (KERN_INFO "DS1337 Real-Time Clock Driver " RTC_VERSION "\n");

	error = misc_register(&rtc_dev);
	if (error) {
		printk(KERN_ERR "rtc: unable to get misc minor for DS1337\n");
		return error;
	}
	i2c_add_driver(&ds1337_driver);

	return 0;
}

static void __init ds1337_set_system_time (void)
{
	unsigned long now, flags;
	extern time_t last_rtc_update;
	extern rwlock_t xtime_lock;

	/* Switching kernel RTC pointers */
	debugk ("RTC switching kernel pointers\n");

	save_set_rtc_time   = ppc_md.set_rtc_time;
	save_get_rtc_time   = ppc_md.get_rtc_time;

	ppc_md.set_rtc_time = ds1337_set_rtc_time;
	ppc_md.get_rtc_time = ds1337_get_rtc_time;

	/*
	 * Set system time
	 * Code copied from arch/ppc/kernel/time.c
	 */
	write_lock_irqsave(&xtime_lock, flags);

	now = ds1337_get_rtc_time();

	debugk ("Set System Time from RTC Time: %lu\n", now);

	xtime.tv_usec = 0;
	xtime.tv_sec  = now;

	last_rtc_update = now - 658;

	time_adjust   = 0;		/* stop active adjtime() */
	time_status  |= STA_UNSYNC;
	time_state    = TIME_ERROR;
	time_maxerror = NTP_PHASE_LIMIT;
	time_esterror = NTP_PHASE_LIMIT;

	write_unlock_irqrestore(&xtime_lock, flags);

	debugk ("RTC installed, minor %d\n", RTC_MINOR);

	/*
	 * Reset any error conditions, alarms, etc.
	 */
	ds1337_reset_rtc ();
}

/***************************************************************************
 *
 * called whenever a process attempts to open the device
 */
static int ds1337_rtc_open (struct inode *inode, struct file *file)
{
	lock_kernel();
	if (rtc_status) {
		unlock_kernel();
		return -EBUSY;
	}

	rtc_status = 1;
	unlock_kernel();

	MOD_INC_USE_COUNT;

	return 0;
}



/***************************************************************************
 *
 * Called when a process closes the device.
 */
static int ds1337_rtc_release (struct inode *inode, struct file *file)
{
	lock_kernel();
	rtc_status = 0;
	unlock_kernel();

	MOD_DEC_USE_COUNT;

	return 0;
}


/***************************************************************************
 *
 * ioctl entry point:
 */
static int ds1337_rtc_ioctl (struct inode *inode, struct file *file,
	unsigned int cmd, unsigned long arg)
{
	struct rtc_time tm;
	int leap_yr;
	static const unsigned char days_in_month[] =
		{ 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

	switch (cmd) {
	case RTC_RD_TIME:	/* Read the time/date from RTC	*/

		to_tm (ds1337_get_rtc_time(), &tm);

		/* convert from rtc_time to struct tm conventions */
		tm.tm_year -= 1900;
		tm.tm_mon  -= 1;

		debugk ("RTC_RD_TIME: %4d-%02d-%02d-%02d:%02d:%02d\n",
			tm.tm_year+1900, tm.tm_mon+1, tm.tm_mday,
			tm.tm_hour, tm.tm_min, tm.tm_sec);

		return (copy_to_user((void *)arg, &tm, sizeof(tm))) ? -EFAULT : 0;

	case RTC_SET_TIME:	/* Set the time/date in the RTC	*/

		if (!capable(CAP_SYS_TIME))
			return -EACCES;

		if (copy_from_user(&tm, (struct rtc_time *)arg, sizeof(tm)))
			return -EFAULT;

		/* convert from struct tm to rtc_time conventions */
		tm.tm_year += 1900;
		tm.tm_mon  += 1;

		/* simple but OK for all dates in time_t range */
		leap_yr = ((tm.tm_year % 4) == 0);

		if ((tm.tm_year < 1970) ||
		    (tm.tm_mon  > 12)	||
		    (tm.tm_mday == 0)	||
		    (tm.tm_mday >
			(days_in_month[tm.tm_mon] + ((tm.tm_mon==2)&&leap_yr)))	||
		    (tm.tm_hour >= 24)	||
		    (tm.tm_min  >= 60)	||
		    (tm.tm_sec  >= 60)	) {

			return (-EINVAL);
		}

		debugk ("RTC_SET_TIME: %4d-%02d-%02d-%02d:%02d:%02d\n",
			tm.tm_year, tm.tm_mon, tm.tm_mday,
			tm.tm_hour, tm.tm_min, tm.tm_sec);

		ds1337_set_rtc_time (mktime(tm.tm_year, tm.tm_mon, tm.tm_mday,
					   tm.tm_hour, tm.tm_min, tm.tm_sec ) );

		return (0);

	default:
		return (-EINVAL);
	}
}


/***************************************************************************
 * Hardware related stuff
 ***************************************************************************
 */

/*
 * RTC register addresses
 */
#define RTC_SEC_REG_ADDR	0x0
#define RTC_MIN_REG_ADDR	0x1
#define RTC_HR_REG_ADDR		0x2
#define RTC_DAY_REG_ADDR	0x3
#define RTC_DATE_REG_ADDR	0x4
#define RTC_MON_REG_ADDR	0x5
#define RTC_YR_REG_ADDR		0x6
#define RTC_CTL_REG_ADDR	0x0e
#define RTC_STAT_REG_ADDR	0x0f

/*
 * RTC control register bits
 */
#define RTC_CTL_BIT_A1IE	0x1	/* Alarm 1 interrupt enable     */
#define RTC_CTL_BIT_A2IE	0x2	/* Alarm 2 interrupt enable     */
#define RTC_CTL_BIT_INTCN	0x4	/* Interrupt control            */
#define RTC_CTL_BIT_RS1		0x8	/* Rate select 1                */
#define RTC_CTL_BIT_RS2		0x10	/* Rate select 2                */
#define RTC_CTL_BIT_DOSC	0x80	/* Disable Oscillator           */

/*
 * RTC status register bits
 */
#define RTC_STAT_BIT_A1F	0x1	/* Alarm 1 flag                 */
#define RTC_STAT_BIT_A2F	0x2	/* Alarm 2 flag                 */
#define RTC_STAT_BIT_OSF	0x80	/* Oscillator stop flag         */

/***************************************************************************
 *
 * get RTC time:
 */
static unsigned long ds1337_get_rtc_time(void)
{
	u8 sec, min, hour, mday, wday, mon_cent, year, control, status;
	int yr;

	control = rtc_rd(RTC_CTL_REG_ADDR);
	status = rtc_rd(RTC_STAT_REG_ADDR);
	sec = rtc_rd(RTC_SEC_REG_ADDR);
	min = rtc_rd(RTC_MIN_REG_ADDR);
	hour = rtc_rd(RTC_HR_REG_ADDR);
	wday = rtc_rd(RTC_DAY_REG_ADDR);
	mday = rtc_rd(RTC_DATE_REG_ADDR);
	mon_cent = rtc_rd(RTC_MON_REG_ADDR);
	year = rtc_rd(RTC_YR_REG_ADDR);

	debugk("Get RTC year: %02x mon/cent: %02x mday: %02x wday: %02x "
	       "hr: %02x min: %02x sec: %02x control: %02x status: %02x\n",
	       year, mon_cent, mday, wday, hour, min, sec, control, status);

	if (status & RTC_STAT_BIT_OSF) {
		printk("### Warning: RTC oscillator has stopped\n");
		/* clear the OSF flag */
		rtc_wr(RTC_STAT_REG_ADDR, rtc_rd(RTC_STAT_REG_ADDR) & ~RTC_STAT_BIT_OSF);
	}

	yr = bcd2bin(year) + ((mon_cent & 0x80) ? 2000 : 1900);

	debugk ("get rtc [dec] year=%d mon=%d day=%d hour=%d min=%d sec=%d\n",
		yr,
		bcd2bin(mon_cent & 0x1F),
		bcd2bin(mday & 0x3F),
		bcd2bin(hour & 0x3F),
		bcd2bin(min  & 0x7F),
		bcd2bin(sec  & 0x7F) );

	return (mktime (yr,
			bcd2bin(mon_cent & 0x1F),
			bcd2bin(mday & 0x3F),
			bcd2bin(hour & 0x3F),
			bcd2bin(min  & 0x7F),
			bcd2bin(sec  & 0x7F)
		) );
}

/***************************************************************************
 *
 * set RTC time:
 */
static int ds1337_set_rtc_time(unsigned long now)
{
	struct rtc_time tm;
	unsigned char century, year, mon, wday, mday, hour, min, sec;

	to_tm (now, &tm);

	debugk ("Set RTC [dec] year=%d mon=%d day=%d hour=%d min=%d sec=%d\n",
		tm.tm_year, tm.tm_mon, tm.tm_mday,
		tm.tm_hour, tm.tm_min, tm.tm_sec);

	century = (tm.tm_year >= 2000) ? 0x80 : 0;
	year = bin2bcd (tm.tm_year % 100);
	mon  = bin2bcd (tm.tm_mon) | century;
	wday = bin2bcd (tm.tm_wday + 1);
	mday = bin2bcd (tm.tm_mday);
	hour = bin2bcd (tm.tm_hour);
	min  = bin2bcd (tm.tm_min);
	sec  = bin2bcd (tm.tm_sec);

	debugk ("Set RTC [bcd] year=%X mon=%X day=%X "
		"hour=%X min=%X sec=%X wday=%X\n",
		year, mon, mday, hour, min, sec, wday);

	rtc_wr(RTC_YR_REG_ADDR, year);
	rtc_wr(RTC_MON_REG_ADDR, mon);
	rtc_wr(RTC_DAY_REG_ADDR, wday);
	rtc_wr(RTC_DATE_REG_ADDR, mday);
	rtc_wr(RTC_HR_REG_ADDR, hour);
	rtc_wr(RTC_MIN_REG_ADDR, min);
	rtc_wr(RTC_SEC_REG_ADDR, sec);

	return (0);
}

/*
 * Reset the RTC.  We also enable the oscillator output on the
 * SQW/INTB* pin and program it for 32,768 Hz output. Note that
 * according to the datasheet, turning on the square wave output
 * increases the current drain on the backup battery from about
 * 600 nA to 2uA.
 */
static void ds1337_reset_rtc(void)
{
	rtc_wr(RTC_CTL_REG_ADDR, RTC_CTL_BIT_RS1 | RTC_CTL_BIT_RS2);
}

/***************************************************************************
 * Module Declarations
 ***************************************************************************
 */
module_init(ds1337_rtc_init);

#ifdef MODULE
/*
 * Cleanup - unregister the appropriate file from /proc
 */
static void ds1337_rtc_cleanup (void)
{
	ppc_md.set_rtc_time = save_set_rtc_time;
	ppc_md.get_rtc_time = save_get_rtc_time;

	misc_deregister(&rtc_dev);

	debugk ("DS1337 Real-Time Clock Driver de-registered\n");
}

module_exit(ds1337_rtc_cleanup);
#endif	/* MODULE */


/***************************************************************************
 *
 * Utility functions
 */

static unsigned int bcd2bin (unsigned int n)
{
	return ((((n >> 4) & 0x0F) * 10) + (n & 0x0F));
}

static unsigned int bin2bcd (unsigned int n)
{
	return (((n / 10) << 4) | (n % 10));
}

static void rtc_wr (unsigned char reg, unsigned char val)
{
	unsigned char buf[2];

	buf[0] = reg;
	buf[1] = val;

	if (i2c_master_send(clnt, (char *)buf, 2) != 2)	{
		printk("%s(): i2c transfer error\n", __FUNCTION__);
	}
}

static unsigned char rtc_rd  (unsigned char reg)
{
	unsigned char buf[1], addr[1];
	struct i2c_msg msgs[2] = {
		{ clnt->addr, 0,        1, addr },
		{ clnt->addr, I2C_M_RD, 1, buf  }
	};

	buf[0] = 0;
	addr[0] = reg;

	if (i2c_transfer(clnt->adapter, msgs, 2) != 2) {
		printk("%s(): i2c transfer error\n", __FUNCTION__);
		return -1;
	}

	return buf[0];
}
