/*
 * Drivers for the EMK Elektronik
 * MPC5200 based Human Machine Interface HMI5200
 *
 * includes drivers for all specialized front panel
 * devices:
 *
 * LCD 240x128 with Toshiba T6963C controller mapped at CS3
 * Rotary Encoder with Push-Button wired to GPIO
 * ChipCard Interface I2C based. For this only CardDetect
 * and CardPower are controlled by this driver. Access of the
 * I2C Bus is done with the generic drivers.
 *
 * 04.08.2006 Reinhard Meyer
 */

#include <linux/module.h>

#include <linux/types.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/tty.h>
#include <linux/timer.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/wait.h>
#include <linux/string.h>
#include <linux/delay.h>
#include <linux/miscdevice.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/init.h>
#include <linux/hmi5200_panel.h>
#include <linux/proc_fs.h>

#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/mpc5xxx.h>

#if defined (CONFIG_FONT_8x8)
  /* actually we do not expect that frambuffer stuff is linked into */
  /* a kernel for the HMI5200, but if it is, we use the font here */
#include <video/font.h>
#else
  /* I DO AGREE this is not the way its supposed to be done, but how */
  /* else do we get the font here without duplicating the source? */
#include "../video/font_8x8.c"
  /* end of hack... */
#endif

#if defined (CONFIG_FB)
  /* we hope the BW version is there to be linked */
extern unsigned char linux_logo_bw[];
#else
  /* we must include the logo here */
#define INCLUDE_LINUX_LOGO_DATA
#define __HAVE_ARCH_LINUX_LOGO
#undef __HAVE_ARCH_LINUX_LOGOBW
#define __HAVE_ARCH_LINUX_LOGO16
#include <linux/linux_logo.h>
#endif

#define	HMI5200_MINOR	156
#define	DRIVER_NAME	"panel"
#define	DRIVER_TITLE	"HMI5200 LCD driver"
#define	AUTHOR		"Reinhard Meyer <info@emk-elektronik.de>"
#define	BUFLEN 30

#define	T6963CBASE	0xf0020000
#define	T6963CCS	3

/* T6963C commands */
#define	LCD_SETXY	0x21
#define	LCD_SETOFFSET	0x22
#define	LCD_SETADDR	0x24
#define	LCD_SETTXTHOME	0x40
#define	LCD_SETTXTAREA	0x41
#define	LCD_SETGFXHOME	0x42
#define	LCD_SETGFXAREA	0x43
#define	LCD_SETORMODE	0x80
#define	LCD_SETEXORMODE	0x81
#define	LCD_SETANDMODE	0x83
#define	LCD_SETTXTAMODE	0x84
#define	LCD_SETINTCG	0x80
#define	LCD_SETEXTCG	0x88
#define	LCD_DISPOFF	0x90
#define	LCD_CURSORON	0x92
#define	LCD_CURSORBLINK	0x93
#define	LCD_TXTON	0x94
#define	LCD_GFXON	0x98
#define	LCD_CURSOR(x)	(0xa0+(x)-1)
#define	LCD_AUTOWRITE	0xb0
#define	LCD_AUTOREAD	0xb1
#define	LCD_AUTOOFF	0xb2
#define	LCD_WRITEINCR	0xc0
#define	LCD_READINCR	0xc1
#define	LCD_WRITEDECR	0xc2
#define	LCD_READDECR	0xc3
#define	LCD_WRITE	0xc4
#define	LCD_READ	0xc5
#define	LCD_PEEK	0xe0
#define	LCD_COPY	0xe8

/* since GPT2,3,4 are used for rotary encoder inputs with input capture, we
 * can use some of their output related registers for variable storage:
 */
/* PWM width - 16 Bits each */
#define x_pos (*((volatile unsigned short *)(MPC5xxx_GPT+0x28)))
#define y_pos (*((volatile unsigned short *)(MPC5xxx_GPT+0x38)))
#define mode  (*((volatile unsigned short *)(MPC5xxx_GPT+0x48)))

/* local variables */
static int is_open = 0;
static LCD_ORGANISATION lcdorg = {
	LCD_TEXTBASE,
	LCD_GFXBASE,
	LCD_CGENBASE,
	LCD_LINELENGTH,
	LCD_SCREENLENGTH
};
static int timer2_irq_counter;
static int timer3_irq_counter;
static int timer4_irq_counter;

/* LCD access functions */
static void _writeLCD (int rs, int d)
{
	*((volatile u8 *) (T6963CBASE + rs)) = d;
}

static int _readLCD (int rs)
{
	return *((volatile u8 *) (T6963CBASE + rs));
}

static void _cmdLCD (int cmd)
{
	while ((_readLCD (1) & 0x03) != 0x03);
	_writeLCD (1, cmd);
}

static void _cmdLCD1 (int cmd, int data)
{
	while ((_readLCD (1) & 0x03) != 0x03);
	_writeLCD (0, data);
	while ((_readLCD (1) & 0x03) != 0x03);
	_writeLCD (1, cmd);
}

static void _cmdLCD2 (int cmd, int data)
{
	while ((_readLCD (1) & 0x03) != 0x03);
	_writeLCD (0, data);
	while ((_readLCD (1) & 0x03) != 0x03);
	_writeLCD (0, data >> 8);
	while ((_readLCD (1) & 0x03) != 0x03);
	_writeLCD (1, cmd);
}

static void _adtLCD (int data)
{
	while ((_readLCD (1) & 0x08) != 0x08);
	_writeLCD (0, data);
}

static int _adrLCD (void)
{
	while ((_readLCD (1) & 0x04) != 0x04);
	return _readLCD (0);
}

static void _initLCD (void)
{
	int i;
	u8 *p = font_vga_8x8.data;

	/* general setup */
	_cmdLCD2 (LCD_SETTXTHOME, lcdorg.textbase);
	_cmdLCD2 (LCD_SETGFXHOME, lcdorg.gfxbase);
	_cmdLCD2 (LCD_SETTXTAREA, lcdorg.width);
	_cmdLCD2 (LCD_SETGFXAREA, lcdorg.width);
	_cmdLCD (LCD_SETEXORMODE | LCD_SETEXTCG);
	_cmdLCD2 (LCD_SETOFFSET, lcdorg.cgenbase >> 11);
	_cmdLCD (LCD_TXTON | LCD_GFXON);
	/* load IBM font */
	_cmdLCD2 (LCD_SETADDR, lcdorg.cgenbase);
	_cmdLCD (LCD_AUTOWRITE);
	for (i = 0; i < 2048; i++)
		_adtLCD (*p++);
	_cmdLCD (LCD_AUTOOFF);
	x_pos = y_pos = 0;
}

static void _filltextLCD (u8 val)
{
	int i;

	_cmdLCD2 (LCD_SETADDR, lcdorg.textbase);
	_cmdLCD (LCD_AUTOWRITE);
	for (i = 0; i < lcdorg.height * lcdorg.width; i++)
		_adtLCD (val);
	_cmdLCD (LCD_AUTOOFF);
}

static void _fillgfxLCD (u8 val)
{
	int i;

	_cmdLCD2 (LCD_SETADDR, lcdorg.gfxbase);
	_cmdLCD (LCD_AUTOWRITE);
	for (i = 0; i < lcdorg.height * lcdorg.width * 8; i++)
		_adtLCD (val);
	_cmdLCD (LCD_AUTOOFF);
}

static void _plotxyLCD (int x, int y, int on)
{
	_cmdLCD2 (LCD_SETADDR, lcdorg.gfxbase + y * lcdorg.width + (x >> 3));	/* address pointer set */
	_cmdLCD ((on ? 0xf8 : 0xf0) + (~x & 7));
}

static void _markLCD (int l, int t, int r, int b, int on, int solid)
{
#define	LL	-1
#define	RR	+7
#define	TT	-1
#define	BB	+7
	int x, y;

	if (solid) {
		for (x = l * 8 LL; x <= r * 8 RR; x++)
			for (y = t * 8 TT; y <= b * 8 BB; y++)
				_plotxyLCD (x, y, on);
	} else {
		for (x = l * 8 LL; x <= r * 8 RR; x++) {
			_plotxyLCD (x, t * 8 TT, on);
			_plotxyLCD (x, b * 8 BB, on);
		}
		for (y = t * 8 TT; y <= b * 8 BB; y++) {
			_plotxyLCD (l * 8 LL, y, on);
			_plotxyLCD (r * 8 RR, y, on);
		}
	}
}

#ifdef CONFIG_PPC_5xxx_PSC_CONSOLE
static void _gotoxyLCD (int x, int y)
{
	_cmdLCD2 (LCD_SETADDR, lcdorg.textbase + (y) * lcdorg.width + x);
}

static void _scrollupLCD (void)
{
	char buf[LCD_LINELENGTH];
	int i, j;

	for (i = 1; i < lcdorg.height; i++) {
		/* copy line up */
		_gotoxyLCD (0, i);
		_cmdLCD (LCD_AUTOREAD);
		for (j = 0; j < lcdorg.width; j++)
			buf[j] = _adrLCD ();
		_cmdLCD (LCD_AUTOOFF);
		_gotoxyLCD (0, i - 1);
		_cmdLCD (LCD_AUTOWRITE);
		for (j = 0; j < lcdorg.width; j++)
			_adtLCD (buf[j]);
		_cmdLCD (LCD_AUTOOFF);
	}
	/* clear last line */
	_gotoxyLCD (0, lcdorg.height - 1);
	_cmdLCD (LCD_AUTOWRITE);
	for (j = 0; j < lcdorg.width; j++)
		_adtLCD (0);
	_cmdLCD (LCD_AUTOOFF);
}

void panel_putchar (unsigned char c)
{
	if (mode == 0)
		return;
	if (c == 13)
		x_pos = 0;
	else if (c == 10) {
	      linefeed:
		y_pos++;
		if (y_pos >= lcdorg.height) {
			_scrollupLCD ();
			y_pos--;
		}
	} else if (c == 8) {
		if (x_pos > 0)
			x_pos--;
		else if (y_pos > 0) {
			y_pos--;
			x_pos = lcdorg.width - 1;
		}
	} else if (c == 9) {
		x_pos = (x_pos + 7) & ~8;
		goto ckeol;
	} else if (c >= 0x20) {
		_gotoxyLCD (x_pos, y_pos);
		if (mode == 2)
			_cmdLCD1 (LCD_WRITEINCR, c);
		else
			_cmdLCD1 (LCD_WRITEINCR, c - 0x20);
		x_pos++;
	      ckeol:
		if (x_pos >= lcdorg.width) {
			x_pos = 0;
			goto linefeed;
		}
	}
}
#endif

/* put initial boot logo */
static void Splash (void)
{
	int x, y;
	u8 *p = linux_logo_bw;

	for (y = 0; y < 80; y++) {
		_cmdLCD2 (LCD_SETADDR, lcdorg.gfxbase + y * lcdorg.width);
		_cmdLCD (LCD_AUTOWRITE);
		for (x = 0; x < 10; x++)
			_adtLCD (~*p++);
		_cmdLCD (LCD_AUTOOFF);
	}
	y = 10;
	p = (u8 *) "Booting";
	_cmdLCD2 (LCD_SETADDR, lcdorg.textbase + y * lcdorg.width);
	_cmdLCD (LCD_AUTOWRITE);
	while (*p)
		_adtLCD (*p++);
	_cmdLCD (LCD_AUTOOFF);
	y++;
	p = (u8 *) "Operating";
	_cmdLCD2 (LCD_SETADDR, lcdorg.textbase + y * lcdorg.width);
	_cmdLCD (LCD_AUTOWRITE);
	while (*p)
		_adtLCD (*p++);
	_cmdLCD (LCD_AUTOOFF);
	y++;
	p = (u8 *) "System";
	_cmdLCD2 (LCD_SETADDR, lcdorg.textbase + y * lcdorg.width);
	_cmdLCD (LCD_AUTOWRITE);
	while (*p)
		_adtLCD (*p++);
	_cmdLCD (LCD_AUTOOFF);
}

/* incremental encoder table
 * encoder states 0-1-3-2-0
 * previous state in bits 2,3 of index, new input in bits 0,1 of index
 * result is new state in bits 2,3 and bit 0 = increment, bit 1 = decrement */
static u8 encoder_table[16] =
	{ 0, 5, 10, 12, 2, 4, 8, 13, 1, 4, 8, 14, 0, 6, 9, 12 };
static u8 encoder_state = 0;
static ENCODER encoder = { 0, -1000000, 1000000, 0 };

static void handle_switch (void)
{
	int i, q, t;

	q = *((volatile u32 *) MPC5xxx_GPT2_STATUS) & 0x00000100;
	i = *((volatile u32 *) MPC5xxx_GPT3_STATUS) & 0x00000100;
	t = *((volatile u32 *) MPC5xxx_GPT4_STATUS) & 0x00000100;

	/* analyze pulse encoder inputs */
	encoder_state &= ~3;
	if (i)
		encoder_state++;
	if (q)
		encoder_state += 2;
	encoder_state = encoder_table[encoder_state];
	if ((encoder_state & 1) && encoder.value < encoder.hilim)
		encoder.value++;
	if ((encoder_state & 2) && encoder.value > encoder.lolim)
		encoder.value--;
	encoder.state = t ? 1 : 0;
}

/* timer2 interrupt */
static void timer2_irq (int irq, void *dev_id, struct pt_regs *regs)
{
	handle_switch ();
	*((volatile u32 *) MPC5xxx_GPT2_STATUS) = 0xf;
	timer2_irq_counter++;
}

/* timer3 interrupt */
static void timer3_irq (int irq, void *dev_id, struct pt_regs *regs)
{
	handle_switch ();
	*((volatile u32 *) MPC5xxx_GPT3_STATUS) = 0xf;
	timer3_irq_counter++;
}

/* timer4 interrupt */
static void timer4_irq (int irq, void *dev_id, struct pt_regs *regs)
{
	handle_switch ();
	*((volatile u32 *) MPC5xxx_GPT4_STATUS) = 0xf;
	timer4_irq_counter++;
}

/* open the device. note: only one open is allowed */
static int panel_open (struct inode *ino, struct file *filep)
{
	if (is_open)
		return -EBUSY;
	MOD_INC_USE_COUNT;
	is_open = 1;
	return 0;
}

/* close the device */
static int panel_release (struct inode *ino, struct file *filep)
{
	if (!is_open)
		return -ENODEV;
	MOD_DEC_USE_COUNT;
	is_open = 0;
	return 0;
}

/* read bytes from display memory */
static ssize_t panel_read (struct file *file,	/* file handle */
			   char *buf,	/* buffer in user space */
			   size_t count,	/* number of bytes to read */
			   loff_t * ppos)
{				/* position to start at */
	char buffer[BUFLEN], *p;
	ssize_t handled = 0;
	unsigned long cnt, cnt2;

	/* end of memory reached? */
	if (*ppos >= LCD_MEMSIZE)
		return 0;

	/* set memory pointer to current position */
	_cmdLCD2 (LCD_SETADDR, *ppos);

	/* read until all data read or end of memory reached */
	_cmdLCD (LCD_AUTOREAD);
	while (count > 0 && *ppos < LCD_MEMSIZE) {
		cnt = count;
		if (cnt > BUFLEN)
			cnt = BUFLEN;
		if (cnt + *ppos > LCD_MEMSIZE)
			cnt = LCD_MEMSIZE - *ppos;
		p = buffer;
		cnt2 = cnt;
		while (cnt2--)
			*p++ = _adrLCD ();
		if (copy_to_user (buf, buffer, cnt))
			return -EFAULT;
		buf += cnt;
		count -= cnt;
		(*ppos) += cnt;
		handled += cnt;
	}
	_cmdLCD (LCD_AUTOOFF);
	return handled ? handled : -ENOSPC;
}

/* write bytes to display memory */
static ssize_t panel_write (struct file *file,	/* file handle */
			    const char *buf,	/* buffer in user space */
			    size_t count,	/* number of bytes to write */
			    loff_t * ppos)
{				/* position to start at */
	char buffer[BUFLEN], *p;
	ssize_t handled = 0;
	unsigned long cnt, cnt2;

	/* end of memory reached? */
	if (*ppos >= LCD_MEMSIZE)
		return -ENOSPC;

	/* set memory pointer to current position */
	_cmdLCD2 (LCD_SETADDR, *ppos);

	/* special case of writing one char must be dealt extra */
	if (count == 1) {
		if (copy_from_user (buffer, buf, 1))
			return -EFAULT;
		_cmdLCD1 (LCD_WRITEINCR, buffer[0]);
		(*ppos)++;
		return 1;
	}

	/* write until all data written or end of memory reached */
	_cmdLCD (LCD_AUTOWRITE);
	while (count > 0 && *ppos < LCD_MEMSIZE) {
		cnt = count;
		if (cnt > BUFLEN)
			cnt = BUFLEN;
		if (cnt + *ppos > LCD_MEMSIZE)
			cnt = LCD_MEMSIZE - *ppos;
		if (copy_from_user (buffer, buf, cnt))
			return -EFAULT;
		p = buffer;
		cnt2 = cnt;
		while (cnt2--)
			_adtLCD (*p++);
		buf += cnt;
		count -= cnt;
		(*ppos) += cnt;
		handled += cnt;
	}
	_cmdLCD (LCD_AUTOOFF);
	return handled ? handled : -ENOSPC;
}

static loff_t panel_lseek (struct file *file, loff_t offset, int orig)
{
	switch (orig) {
	case 0:
		file->f_pos = offset;
		return file->f_pos;
	case 1:
		file->f_pos += offset;
		return file->f_pos;
	case 2:
		file->f_pos = LCD_MEMSIZE - offset;
		return file->f_pos;
	default:
		return -EINVAL;
	}
}

static int panel_ioctl (struct inode *inode,
			struct file *file,
			unsigned int cmd, unsigned long arg)
{
	LCD_TEXTMARKER tm;
	LCD_POINT p;
	int i;

	switch (cmd) {
	case PANEL_SET_ORG:
		if (copy_from_user
		    (&lcdorg, (LCD_ORGANISATION *) arg, sizeof lcdorg))
			return -EFAULT;
		_initLCD ();
		_filltextLCD (0);
		_fillgfxLCD (0);
		break;

	case PANEL_GET_ORG:
		if (copy_to_user
		    ((LCD_ORGANISATION *) arg, &lcdorg, sizeof lcdorg))
			return -EFAULT;
		break;

	case PANEL_FILL_TXT:
		_filltextLCD (arg);
		break;

	case PANEL_FILL_GFX:
		_fillgfxLCD (arg);
		break;

	case PANEL_MARK_TXT:
		if (copy_from_user (&tm, (LCD_TEXTMARKER *) arg, sizeof tm))
			return -EFAULT;
		_markLCD (tm.left, tm.top, tm.right, tm.bottom, tm.on,
			  tm.solid);
		break;

	case PANEL_PLOT:
		if (copy_from_user (&p, (LCD_POINT *) arg, sizeof p))
			return -EFAULT;
		_plotxyLCD (p.x, p.y, p.on);
		break;

	case PANEL_GET_ENCODER:
		if (copy_to_user ((ENCODER *) arg, &encoder, sizeof encoder))
			return -EFAULT;
		break;

	case PANEL_SET_ENCODER:
		if (copy_from_user
		    (&encoder, (ENCODER *) arg, sizeof encoder))
			return -EFAULT;
		break;

	case PANEL_GET_CARD:
		i = 0;
		if (!(*((volatile u8 *) MPC5xxx_WU_GPIO_DATAIN) & 0x02))
			i |= 1;
		if (copy_to_user ((int *) arg, &i, sizeof i))
			return -EFAULT;
		break;

	case PANEL_SET_CARD:
		if (copy_from_user (&i, (int *) arg, sizeof i))
			return -EFAULT;
		if (i)
			*((volatile u8 *) MPC5xxx_WU_GPIO_DATAOUT) |= 1;
		else
			*((volatile u8 *) MPC5xxx_WU_GPIO_DATAOUT) &= ~1;
		break;

	default:
		return -EINVAL;
	}
	return 0;
}

static int proc_output (char *buf)
{
	char *p = buf;
	int i, j, c;

#if 0
	p += sprintf (p,
		      "organisation: textbase:%04x gfxbase:%04x cgenbase:%04x\n",
		      lcdorg.textbase, lcdorg.gfxbase, lcdorg.cgenbase);
	p += sprintf (p,
		      "              width:%u height:%u\n",
		      lcdorg.width, lcdorg.height);
	p += sprintf (p,
		      "              IRQs: %u %u %u Encoder:%d %d %d %d\n",
		      timer2_irq_counter, timer3_irq_counter,
		      timer4_irq_counter,
		      encoder.value, encoder.lolim, encoder.hilim,
		      encoder.state);
	p += sprintf (p, "current text display:\n");
#endif
	*p++ = '+';
	for (i = 0; i < lcdorg.width; i++)
		*p++ = '-';
	*p++ = '+';
	*p++ = '\n';
	_cmdLCD2 (LCD_SETADDR, lcdorg.textbase);
	_cmdLCD (LCD_AUTOREAD);
	for (j = 0; j < lcdorg.height; j++) {
		*p++ = '|';
		for (i = 0; i < lcdorg.width; i++) {
			c = _adrLCD () & 0xff;
			if (c < 0x20)
				c = ' ';
			*p++ = c;
		}
		*p++ = '|';
		*p++ = '\n';
	}
	_cmdLCD (LCD_AUTOOFF);
	*p++ = '+';
	for (i = 0; i < lcdorg.width; i++)
		*p++ = '-';
	*p++ = '+';
	*p++ = '\n';
	return p - buf;
}

static int read_proc (char *page, char **start, off_t off, int count,
		      int *eof, void *data)
{
	int len = proc_output (page);

	if (len <= off + count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;
	return len;
}

static struct file_operations panel_operations = {
      llseek:panel_lseek,
      ioctl:panel_ioctl,
      read:panel_read,
      write:panel_write,
      open:panel_open,
      release:panel_release,
};

static struct miscdevice panel_miscdev = {
	HMI5200_MINOR, DRIVER_NAME, &panel_operations
};

static int __init hmi5200_panel_init (void)
{
	printk (KERN_INFO DRIVER_TITLE " - " AUTHOR "\n");

	if (misc_register (&panel_miscdev) < 0) {
		return -EBUSY;
	}

	create_proc_read_entry ("panel", 0, 0, read_proc, NULL);

	/* register IRQ functions */
	if (request_irq (MPC5xxx_TIMER2_IRQ, timer2_irq,
			 SA_INTERRUPT, DRIVER_NAME " Q (timer2)", NULL)) {
		printk (KERN_ERR DRIVER_NAME
			": timer2 irq allocation failed\n");
		return -EINVAL;
	}
	if (request_irq (MPC5xxx_TIMER3_IRQ, timer3_irq,
			 SA_INTERRUPT, DRIVER_NAME " I (timer3)", NULL)) {
		printk (KERN_ERR DRIVER_NAME
			": timer3 irq allocation failed\n");
		return -EINVAL;
	}
	if (request_irq (MPC5xxx_TIMER4_IRQ, timer4_irq,
			 SA_INTERRUPT, DRIVER_NAME " P (timer4)", NULL)) {
		printk (KERN_ERR DRIVER_NAME
			": timer4 irq allocation failed\n");
		return -EINVAL;
	}

	/* initialize dedicated ports */
	/* TIMER2,3,4 as input capture with IRQ for rotary encoder and button */
	*((volatile u32 *) MPC5xxx_GPT2_ENABLE) = 0x00000101;
	*((volatile u32 *) MPC5xxx_GPT3_ENABLE) = 0x00000101;
	*((volatile u32 *) MPC5xxx_GPT4_ENABLE) = 0x00000101;
#if 0				/* don't touch those here */
	/* PSC1_4 as output to connect chip card lines to i2c bus */
	/* PSC2_4 as input with IRQ to detect chip card inserted (low active) */
	*((volatile u8 *) MPC5xxx_WU_GPIO_ENABLE) |= 1 | 2;
	*((volatile u8 *) MPC5xxx_WU_GPIO_DIR) |= 1;
	*((volatile u8 *) MPC5xxx_WU_GPIO_DATAOUT) &= ~1;
#endif
	/* initialize T6963C */
	_initLCD ();
	_filltextLCD (0);
	_fillgfxLCD (0);
	if (mode == 0)
		Splash ();
	else
		mode = 2;
	return 0;
}

static void __exit hmi5200_panel_exit (void)
{
	/* TIMER2,3,4 as simple input for rotary encoder and button */
	*((volatile u32 *) MPC5xxx_GPT2_ENABLE) = 0x00000004;
	*((volatile u32 *) MPC5xxx_GPT3_ENABLE) = 0x00000004;
	*((volatile u32 *) MPC5xxx_GPT4_ENABLE) = 0x00000004;

	remove_proc_entry (DRIVER_NAME, NULL);
	misc_deregister (&panel_miscdev);
	free_irq (MPC5xxx_GPIO_WKUP_IRQ, NULL);
	free_irq (MPC5xxx_TIMER4_IRQ, NULL);
	free_irq (MPC5xxx_TIMER3_IRQ, NULL);
	free_irq (MPC5xxx_TIMER2_IRQ, NULL);
}

module_init (hmi5200_panel_init);
module_exit (hmi5200_panel_exit);

MODULE_LICENSE ("GPL");
MODULE_AUTHOR (AUTHOR);
MODULE_DESCRIPTION (DRIVER_TITLE);
EXPORT_NO_SYMBOLS;
