/*
 * Generic I2C EEPROM driver.
 * Copyright (C) 2006 EMK Elektronik GmbH, Reinhard Meyer, info@emk-elektronik.de
 *
 * This driver is useable to read and write data from and to EEPROMs attached
 * at the I2C Bus.
 *
 * Since its impossible to safely probe for EEPROMs and find out their address
 * width and size, all have to be defined statically.
 * Also we support removeable EEPROMs like ChipCards. On the other hand
 * this allows us to easyly subpartition the EEPROMs.
 *
 * We support two modes for a partition:
 *  - raw bytes for read and write, access to the full size of the partition
 *  - actual length of file stored in the first byte(s) of the partition
 *
 * implementation restrictions:
 *  - currently gets hooked to the first adapter only
 */
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/vmalloc.h>
#include <linux/errno.h>
#include <linux/proc_fs.h>
#include <linux/i2c.h>

#include <asm/uaccess.h>
#include <asm/io.h>
#ifdef CONFIG_HMI5200
#include <asm/mpc5xxx.h>
#endif

/* MAJOR device number */
#define EEPROM_MAJOR	33
/* number of MINOR devices supported */
#define	NUM_MINOR	64

#define	DRIVER_NAME	"eeprom"
#define	DRIVER_TITLE	"I2C EEPROM driver"
#define	AUTHOR		"Reinhard Meyer <info@emk-elektronik.de>"

typedef struct ee_part EE_PART;
struct ee_part {
	/* compiled in constants */
	void *bus;		/* some sort of assigning the bus */
	int unit;		/* I2C address */
	int addr_len;		/* address length in unit 0,1,2 */
	int pagesize;		/* pagesize for write */
	int start_addr;		/* start offset of partition */
	int maxlen;		/* size of partition */
	int raw;		/* has no length info stored */
	int (*probe) (EE_PART *, int);
	char *name;		/* name of partition (for display purposes) */
	/* variables */
	int open_count;		/* count this partition is opened */
	int curlen;		/* current length */
};

#ifdef CONFIG_HMI5200
static int card_probe (EE_PART *, int);
#endif

static EE_PART ee_part[NUM_MINOR] = {
#ifdef CONFIG_TOP5200
	/* fixed partitions in the TOP5200 module EEPROM itself */
	/* the whole chip At26C64 */
	[0] = {0, 0x57, 2, 32, 0x0000, 0x2000, 1, NULL, "at24c64 on TOP5200"},
#ifdef CONFIG_LMS5200
	[8] = {0, 0x57, 2, 32, 0x0000, 0x1800, 1, NULL, "u-boot env"},
#else
	[1] = {0, 0x57, 2, 32, 0x0000, 0x0400, 0, NULL, "file1"},
	[2] = {0, 0x57, 2, 32, 0x0400, 0x0400, 0, NULL, "file2"},
	[3] = {0, 0x57, 2, 32, 0x0800, 0x0400, 0, NULL, "file3"},
	[4] = {0, 0x57, 2, 32, 0x0c00, 0x0400, 0, NULL, "file4"},
	[8] = {0, 0x57, 2, 32, 0x1000, 0x0800, 1, NULL, "u-boot env"},
#endif
	[9] = {0, 0x57, 2, 32, 0x1800, 0x0200, 0, NULL, "factory"},
	[10] = {0, 0x57, 2, 32, 0x1a00, 0x0600, 0, NULL, "setup"},
#endif
#ifdef CONFIG_HMI5200
	[16] = {0, 0x50, 2, 32, 0x0000, 0x10000, 1, card_probe,
		"ChipCard on HMI5200"},
	[17] = {0, 0x50, 2, 32, 0x0100, 0x0400, 0, card_probe, "file1"},
	[18] = {0, 0x50, 2, 32, 0x0500, 0x0400, 0, card_probe, "file2"},
	[19] = {0, 0x50, 2, 32, 0x0900, 0x0400, 0, card_probe, "file3"},
	[20] = {0, 0x50, 2, 32, 0x0d00, 0x0400, 0, card_probe, "file4"},
	[21] = {0, 0x50, 2, 32, 0x1100, 0x0400, 0, card_probe, "file5"},
	[22] = {0, 0x50, 2, 32, 0x1500, 0x0400, 0, card_probe, "file6"},
	[23] = {0, 0x50, 2, 32, 0x1900, 0x0400, 0, card_probe, "file7"},
	[24] = {0, 0x50, 2, 32, 0x1d00, 0x0300, 0, card_probe, "file8"},
	[31] = {0, 0x50, 2, 32, 0x0000, 0x0100, 0, card_probe, ".root"},
#endif
};

static struct i2c_client *this_client = NULL;
static struct i2c_driver eeprom_driver;

#ifdef CONFIG_HMI5200
static int card_present (void)
{
	return (*((volatile u8 *) MPC5xxx_WU_GPIO_DATAIN) & 0x02) ? 0 : 1;
}

static void card_connect (int on)
{
	if (on)
		*((volatile u8 *) MPC5xxx_WU_GPIO_DATAOUT) |= 1;
	else
		*((volatile u8 *) MPC5xxx_WU_GPIO_DATAOUT) &= ~1;
}

static int last_state = -1;
static int open_count = 0;

/* check for ChipCard inserted, control I2C connection to it */
/* somehow we should mutex the bus before doing so, how ? */
static int card_probe (EE_PART * ee, int updown)
{
	int card_state = card_present ();

	open_count += updown;
	if (open_count < 1)
		card_state = 0;
	if (card_state != last_state) {
		down (&this_client->adapter->lock);
		udelay (1);
		card_connect (card_state);
		udelay (1);
		up (&this_client->adapter->lock);
		last_state = card_state;
	}
	return card_state;
}
#endif

static int eeprom_get (struct i2c_client *client,
		       EE_PART * ee, int offset, char *buf, int len)
{
	int ret;
	char addrbuf[2] = { offset >> 8, offset };
	struct i2c_msg random_addr_read[2] = {
		{
		 /* "Set Current Address" */
		 ee->unit, client->flags, ee->addr_len, (u8 *) addrbuf},
		{
		 /* "Sequential Read" */
		 ee->unit, client->flags | I2C_M_RD, len, (u8 *) buf}
	};

	if ((ret = i2c_transfer (client->adapter, random_addr_read, 2)) != 2) {
		printk (KERN_ERR DRIVER_NAME
			"_get: i2c_transfer failed, ret=%d\n", ret);
		ret = -ENXIO;
	}

	return ret;
}

static int eeprom_put (struct i2c_client *client,
		       EE_PART * ee, int offset, char *buf, int len)
{
	int ret;
	char *local_buf;
	char addrbuf[2] = { offset >> 8, offset };
	struct i2c_msg page_write = {
		ee->unit, client->flags, len + ee->addr_len, NULL
	};

	local_buf = (char *) kmalloc (len + ee->addr_len, GFP_KERNEL);
	if (local_buf == NULL) {
		printk (KERN_ERR DRIVER_NAME "_put: buffer alloc failed\n");
		return -ENOMEM;
	}

	memcpy (local_buf, addrbuf, ee->addr_len);
	memcpy (local_buf + ee->addr_len, buf, len);
	page_write.buf = (u8 *) local_buf;

	if ((ret = i2c_transfer (client->adapter, &page_write, 1)) != 1) {
		printk (KERN_ERR DRIVER_NAME
			"_put: i2c_transfer failed, ret=%d\n", ret);
		ret = -ENXIO;
	}
	udelay (5000);		/* let the page be written */

	kfree (local_buf);
	return ret;
}

static int eeprom_put_pages (struct i2c_client *client,
			     EE_PART * ee, int offset, char *buf, int len)
{
	while (len > 0) {
		int ret;
		int next_page = (offset + ee->pagesize) & (-ee->pagesize);
		int rest_len = next_page - offset;

		if (rest_len > len)
			rest_len = len;
		ret = eeprom_put (client, ee, offset, buf, rest_len);
		if (ret < 0)
			return ret;
		offset += rest_len;
		buf += rest_len;
		len -= rest_len;
	}
	return 1;
}

static ssize_t eeprom_read (struct file *file,	/* file handle */
			    char *buf,	/* buffer in user space */
			    size_t count,	/* number of bytes to read */
			    loff_t * ppos)
{				/* position to start at */
	EE_PART *ee = file->private_data;
	unsigned long p = *ppos;
	char *local_buf;
	int ret;

	if (ee->probe && (*ee->probe) (ee, 0) == 0)
		return -ENXIO;

	if (p >= ee->curlen)
		return 0;

	if (count > ee->curlen - p)
		count = ee->curlen - p;

	p += ee->start_addr;
	if (!ee->raw)
		p += ee->addr_len;

	local_buf = (char *) kmalloc (count, GFP_KERNEL);
	if (local_buf == NULL) {
		printk (KERN_ERR DRIVER_NAME "_read: buffer alloc failed\n");
		return -ENOMEM;
	}

	ret = eeprom_get (this_client, ee, p, local_buf, count);
	if (ret < 0) {
		kfree (local_buf);
		return ret;
	}
	ret = copy_to_user (buf, local_buf, count);
	kfree (local_buf);
	if (ret)
		return -EFAULT;

	*ppos += count;
	return count;
}

static ssize_t eeprom_write (struct file *file,	/* file handle */
			     const char *buf,	/* buffer in user space */
			     size_t count,	/* number of bytes to write */
			     loff_t * ppos)
{				/* position to start at */
	EE_PART *ee = file->private_data;
	unsigned long p = *ppos;
	char *local_buf;
	int ret;

	if (ee->probe && (*ee->probe) (ee, 0) == 0)
		return -ENXIO;

	if (ee->raw) {
		if (p >= ee->maxlen)
			return -ENOSPC;
		if (count > ee->maxlen - p)
			count = ee->maxlen - p;
		if (p + count > ee->curlen)
			ee->curlen = p + count;
		p += ee->start_addr;
	} else {
		if (file->f_flags & O_APPEND)
			p = ee->curlen;
		if (p >= ee->maxlen - ee->addr_len)
			return -ENOSPC;
		if (count > ee->maxlen - ee->addr_len - p)
			count = ee->maxlen - ee->addr_len - p;
		if (p + count > ee->curlen)
			ee->curlen = p + count;
		p += ee->start_addr + ee->addr_len;
	}

	local_buf = (char *) kmalloc (count, GFP_KERNEL);
	if (local_buf == NULL) {
		printk (KERN_ERR DRIVER_NAME "_write: buffer alloc failed\n");
		return -ENOMEM;
	}

	if (copy_from_user (local_buf, buf, count)) {
		kfree (local_buf);
		return -EFAULT;
	}

	/* writing to EEPROM isn't as trivial as reading. EEPROMs have a page
	 * organisation. A write cannot cross a page boundary
	 */
	ret = eeprom_put_pages (this_client, ee, p, local_buf, count);
	kfree (local_buf);
	if (ret < 0)
		return ret;

	*ppos += count;
	return count;
}

static loff_t eeprom_lseek (struct file *file, loff_t offset, int orig)
{
	EE_PART *ee = file->private_data;

	if (ee->probe && (*ee->probe) (ee, 0) == 0)
		return -ENXIO;

	switch (orig) {
	case 0:
		file->f_pos = offset;
		return file->f_pos;
	case 1:
		file->f_pos += offset;
		return file->f_pos;
	case 2:
		file->f_pos = ee->curlen - offset;
		return file->f_pos;
	default:
		return -EINVAL;
	}
}

static int eeprom_open (struct inode *inode, struct file *file)
{
	EE_PART *ee;
	int minor = MINOR (inode->i_rdev);

	if (minor < 0 || minor >= NUM_MINOR || ee_part[minor].unit == 0)
		return -ENXIO;
	ee = ee_part + minor;
	if (ee->open_count > 0)
		return -EBUSY;
	if (ee->probe && (*ee->probe) (ee, 1) == 0) {
		(*ee->probe) (ee, -1);
		return -ENXIO;
	}
	ee->open_count++;
	file->private_data = ee;
	MOD_INC_USE_COUNT;
	/* determine current length */
	if (ee->raw)
		ee->curlen = ee->maxlen;	/* raw partitions have no length stored */
	else if (file->f_flags & O_TRUNC)
		ee->curlen = 0;
	else {
		/* now we must read the length info */
		unsigned char len_buffer[2];
		int ret =
			eeprom_get (this_client, ee, ee->start_addr,
				    (char *) len_buffer, ee->addr_len);
		if (ret < 0) {
			ee->curlen = 0;
			ee->open_count--;
			file->private_data = NULL;
			MOD_DEC_USE_COUNT;
			return ret;
		}
		ee->curlen = len_buffer[0];
		if (ee->addr_len > 1)
			ee->curlen += (len_buffer[1] << 8);
		if (ee->curlen > ee->maxlen - ee->addr_len)
			ee->curlen = 0;
	}
	return 0;
}

static int eeprom_release (struct inode *inode, struct file *file)
{
	EE_PART *ee;
	int minor = MINOR (inode->i_rdev);

	if (minor < 0 || minor >= NUM_MINOR || ee_part[minor].unit == 0)
		return -ENXIO;
	ee = ee_part + minor;
	if (ee->open_count < 1)
		return -EIO;

	/* on non-raw, open for write files, write the current length */
	if (!ee->raw && (file->f_flags & (O_WRONLY | O_RDWR))) {
		char lenbuf[2] = { ee->curlen, ee->curlen >> 8 };
		/* no point error checking here */
		eeprom_put (this_client, ee, ee->start_addr, lenbuf,
			    ee->addr_len);
	}
	ee->open_count--;
	file->private_data = NULL;
	MOD_DEC_USE_COUNT;
	if (ee->probe)
		(*ee->probe) (ee, -1);
	return 0;
}

static int proc_output (char *buf)
{
	char *p = buf;
	int i;

	for (i = 0; i < NUM_MINOR; i++) {
		EE_PART *ee = ee_part + i;

		if (ee->unit)
			p += sprintf (p,
				      "%2d: %08x %02x %d %2d %04x %5d(%5d) %d %d %s\n",
				      i,
				      (int) ee->bus,
				      ee->unit,
				      ee->addr_len,
				      ee->pagesize,
				      ee->start_addr,
				      ee->curlen,
				      ee->maxlen,
				      ee->raw, ee->open_count, ee->name);
	}
#ifdef CONFIG_HMI5200
	p += sprintf (p, "card_state=%d open_count=%d\n",
		      last_state, open_count);
#endif
	return p - buf;
}

static int read_proc (char *page, char **start, off_t off, int count,
		      int *eof, void *data)
{
	int len = proc_output (page);

	if (len <= off + count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;
	return len;
}

static int eeprom_probe (struct i2c_adapter *adap)
{
	if (this_client != NULL)
		return -EBUSY;
	this_client = kmalloc (sizeof (*this_client), GFP_KERNEL);
	if (this_client == NULL) {
		return -ENOMEM;
	}
	strcpy (this_client->name, eeprom_driver.name);
	this_client->id = eeprom_driver.id;
	this_client->flags = 0;
	this_client->addr = -1;
	this_client->adapter = adap;
	this_client->driver = &eeprom_driver;
	this_client->data = NULL;
	/* we do not really probe here */
	printk (KERN_INFO DRIVER_NAME "_drv: hooked to %s id:0x%x\n",
		adap->name, adap->id);
	return i2c_attach_client (this_client);
}

static int eeprom_detach (struct i2c_client *client)
{
	i2c_detach_client (client);
	if (this_client != NULL) {
		kfree (this_client);
		this_client = NULL;
	}
	return 0;
}

static int eeprom_command (struct i2c_client *client,
			   unsigned int cmd, void *arg)
{
	return -EINVAL;
}

static struct file_operations eeprom_fops = {
      llseek:eeprom_lseek,
      read:eeprom_read,
      write:eeprom_write,
      open:eeprom_open,
      release:eeprom_release,
};

static struct i2c_driver eeprom_driver = {
	.name = DRIVER_NAME,
	.id = I2C_DRIVERID_EEPROM,
	.flags = I2C_DF_NOTIFY,
	.attach_adapter = eeprom_probe,
	.detach_client = eeprom_detach,
	.command = eeprom_command
};

static int __init eeprom_init (void)
{
	int ret;

	ret = i2c_add_driver (&eeprom_driver);
	if (ret) {
		printk (KERN_ERR DRIVER_NAME
			"_drv: register I2C driver failed, ret = %d\n", ret);
		return -EIO;
	}
	ret = register_chrdev (EEPROM_MAJOR, DRIVER_NAME, &eeprom_fops);
	if (ret) {
		printk (KERN_ERR DRIVER_NAME
			"_drv: register_chrdev failed, ret = %d\n", ret);
		ret = i2c_del_driver (&eeprom_driver);
		if (ret) {
			printk (KERN_ERR
				"Unregister I2C driver failed, ret = %d\n",
				ret);
		}
		return -EIO;
	}
	create_proc_read_entry (DRIVER_NAME, 0, 0, read_proc, NULL);
#ifdef CONFIG_HMI5200
	/* PSC1_4 as output to connect chip card lines to i2c bus */
	/* PSC2_4 as input with IRQ to detect chip card inserted (low active) */
	*((volatile u8 *) MPC5xxx_WU_GPIO_ENABLE) |= 1 | 2;
	*((volatile u8 *) MPC5xxx_WU_GPIO_DIR) |= 1;
	*((volatile u8 *) MPC5xxx_WU_GPIO_DATAOUT) &= ~1;
#endif
	printk (KERN_INFO DRIVER_TITLE " - " AUTHOR "\n");
	return 0;
}


static void eeprom_exit (void)
{
	remove_proc_entry (DRIVER_NAME, NULL);
	if (unregister_chrdev (EEPROM_MAJOR, DRIVER_NAME)) {
		printk (KERN_ERR DRIVER_NAME
			"_drv: failed to unregister character device\n");
	}
	i2c_del_driver (&eeprom_driver);
}

module_init (eeprom_init);
module_exit (eeprom_exit);

MODULE_LICENSE ("GPL");
MODULE_AUTHOR (AUTHOR);
MODULE_DESCRIPTION (DRIVER_TITLE);
EXPORT_NO_SYMBOLS;
