/*
 * (C) Copyright 2005
 * DENX Software Engineering
 * Wolfgang Denk, wd@denx.de
 * All rights reserved.
 *
 * Driver for M48T58 TIMEKEEPER RTC & RAM chip on Prodrive P3G4 & SCCB Board
 *
 * Based on linux/drivers/char/m41t11.c
 *
 *  Copyright (C) 2000 MontaVista Software, Inc.
 *
 *
 *  Copyright (C) 2000 Russell King
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/module.h>
#include <linux/miscdevice.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/rtc.h>
#include <linux/smp_lock.h>
#include <linux/version.h>
#include <linux/i2c.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <asm/types.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/time.h>
#include <asm/machdep.h>		/* ppc_md */

#include "m48t58.h"


#ifndef KERNEL_VERSION
# define KERNEL_VERSION(a,b,c)	(((a) << 16) + ((b) << 8) + (c))
#endif

#if (!defined(LINUX_VERSION_CODE) || LINUX_VERSION_CODE < KERNEL_VERSION(2,2,0))
#error "You need to use at least Linux kernel version 2.2.0"
#endif


/***************************************************************************
 * Global stuff
 ***************************************************************************
 */
/* #define	DEBUG	1	//	*/

/*
 * Deal with CONFIG_MODVERSIONS
 */
#if CONFIG_MODVERSIONS==1
/* # define MODVERSIONS */
# include <linux/modversions.h>
#endif

/*
 * For character devices
 */
#include <linux/fs.h>		/* character device definitions	*/
#include <linux/wrapper.h>	/* wrapper for compatibility with future versions */


/*
 * Device Declarations
 */

/*
 * The name for our device, as it will appear in /proc/devices
 */
#define DEVICE_NAME	"rtc"
#define	RTC_VERSION	"$Revision: 1.0 $ wd@denx.de"

/***************************************************************************
 * Local stuff
 ***************************************************************************
 */

#ifdef	DEBUG
# define debugk(fmt,args...)	printk(fmt ,##args)
# define DEBUG_OPTARG(arg)	arg,
#else
# define debugk(fmt,args...)
# define DEBUG_OPTARG(arg)
#endif


/* avoid multiple access */
static char rtc_status = 0;

/* save/restore old machdep pointers */
int		(*save_set_rtc_time)(unsigned long);
unsigned long	(*save_get_rtc_time)(void);

static unsigned int vbar1;


/***************************************************************************
 *
 * Utility functions
 */

static unsigned int bcd2bin (unsigned int n)
{
	return ((((n >> 4) & 0x0F) * 10) + (n & 0x0F));
}


static unsigned int bin2bcd (unsigned int n)
{
	return (((n / 10) << 4) | (n % 10));
}


/***************************************************************************
 *
 * get RTC time:
 */
static unsigned long m48t58_get_rtc_time(void)
{
	unsigned char year, mon, wday_cent, mday, hour, min, sec;
	int yr;
	struct rtc_registers *rtc;

	/*
	 * Point page register to rtc regs
	 */
	*(volatile unsigned int *)(vbar1 + 4) = 0x1f000000;
	rtc = (struct rtc_registers *)(vbar1 + 0x7e0);

	rtc->control |= 0x40; /* set read bit */
	sec = rtc->secs;
	min = rtc->mins;
	hour = rtc->hours;
	mday = rtc->mday;
	wday_cent = rtc->wday;
	mon = rtc->mon;
	year = rtc->year;
	rtc->control &= ~0x40; /* clear read bit */

	debugk ("Get RTC year: %02x mon: %02x mday: %02x wday/cent: %02x "
		"hr: %02x min: %02x sec: %02x\n",
		year, mon, mday, wday_cent, hour, min, sec );

	if (mday & 0x40) {
		printk (KERN_CRIT "M48T58 RTC Low Voltage - date/time not reliable\n");
	}

	yr = bcd2bin(year) + ((wday_cent & 0x10) ? 2000 : 1900);

	debugk("get rtc [dec] year=%d mon=%d day=%d hour=%d min=%d sec=%d\n",
	       yr,
	       bcd2bin(mon  & 0x1F),
	       bcd2bin(mday & 0x3F),
	       bcd2bin(hour & 0x3F),
	       bcd2bin(min  & 0x7F),
	       bcd2bin(sec  & 0x7F));

	return(mktime(yr,
		      bcd2bin(mon  & 0x1F),
		      bcd2bin(mday & 0x3F),
		      bcd2bin(hour & 0x3F),
		      bcd2bin(min  & 0x7F),
		      bcd2bin(sec  & 0x7F)
		       ));
}


/***************************************************************************
 *
 * set RTC time:
 */
static int m48t58_set_rtc_time(unsigned long now)
{
	struct rtc_time tm;
	unsigned char century, year, mon, wday, mday, hour, min, sec;
	struct rtc_registers *rtc;

	to_tm(now, &tm);

	debugk("Set RTC [dec] year=%d mon=%d day=%d hour=%d min=%d sec=%d\n",
	       tm.tm_year, tm.tm_mon, tm.tm_mday,
	       tm.tm_hour, tm.tm_min, tm.tm_sec);

	century = (tm.tm_year >= 2000) ? 0x30 : 0;
	year = bin2bcd(tm.tm_year % 100);
	mon  = bin2bcd(tm.tm_mon);
	wday = bin2bcd(tm.tm_wday) | century;
	mday = bin2bcd(tm.tm_mday);
	hour = bin2bcd(tm.tm_hour);
	min  = bin2bcd(tm.tm_min);
	sec  = bin2bcd(tm.tm_sec);

	debugk("Set RTC [bcd] year=%X mon=%X day=%X "
	       "hour=%X min=%X sec=%X wday=%X\n",
	       year, mon, mday, hour, min, sec, wday);

	/*
	 * Point page register to rtc regs
	 */
	*(volatile unsigned int *)(vbar1 + 4) = 0x1f000000;
	rtc = (struct rtc_registers *)(vbar1 + 0x7e0);

	rtc->control |= 0x80; /* set write bit */
	rtc->year = year;
	rtc->mon = mon;
	rtc->wday = wday;
	rtc->mday = mday;
	rtc->hours = hour;
	rtc->mins = min;
	rtc->secs = sec;
	rtc->control &= ~0x80; /* clear write bit */

	return (0);
}


static void __init m48t58_set_system_time (void)
{
	unsigned long now, flags;
	extern time_t last_rtc_update;
	extern rwlock_t xtime_lock;
	struct rtc_registers *rtc;

	/* Switching kernel RTC pointers */
	debugk ("RTC switching kernel pointers\n");

	save_set_rtc_time   = ppc_md.set_rtc_time;
	save_get_rtc_time   = ppc_md.get_rtc_time;

	ppc_md.set_rtc_time = m48t58_set_rtc_time;
	ppc_md.get_rtc_time = m48t58_get_rtc_time;

	/*
	 * Set system time
	 * Code copied from arch/ppc/kernel/time.c
	 */
	write_lock_irqsave(&xtime_lock, flags);

	now = m48t58_get_rtc_time();

	debugk ("Set System Time from RTC Time: %lu\n", now);

	xtime.tv_usec = 0;
	xtime.tv_sec  = now;

	last_rtc_update = now - 658;

	time_adjust   = 0;		/* stop active adjtime() */
	time_status  |= STA_UNSYNC;
	time_state    = TIME_ERROR;
	time_maxerror = NTP_PHASE_LIMIT;
	time_esterror = NTP_PHASE_LIMIT;

	write_unlock_irqrestore(&xtime_lock, flags);

	debugk ("RTC installed, minor %d\n", RTC_MINOR);

	/*
	 * Point page register to rtc regs
	 */
	*(volatile unsigned int *)(vbar1 + 4) = 0x1f000000;
	rtc = (struct rtc_registers *)(vbar1 + 0x7e0);

	/*
	 * Check for low voltage
	 */
	if (rtc->mday & 0x40) {
		printk (KERN_CRIT "PCF8563 RTC Low Voltage - date/time not reliable\n");
	}
}


int rtc_open(struct inode *minode, struct file *mfile)
{
	lock_kernel();
	if (rtc_status) {
		unlock_kernel();
		return -EBUSY;
	}

	rtc_status = 1;
	unlock_kernel();

	MOD_INC_USE_COUNT;

	return 0;
}


int rtc_release(struct inode *minode, struct file *mfile)
{
	lock_kernel();
	rtc_status = 0;
	unlock_kernel();

	MOD_DEC_USE_COUNT;

	return 0;
}


static int rtc_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
		     unsigned long arg)
{
	struct rtc_time tm;
	int leap_yr;
	static const unsigned char days_in_month[] =
		{ 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

	switch (cmd) {
	case RTC_RD_TIME:	/* Read the time/date from RTC	*/

		to_tm(m48t58_get_rtc_time(), &tm);

		/* convert from rtc_time to struct tm conventions */
		tm.tm_year -= 1900;
		tm.tm_mon  -= 1;

		debugk("RTC_RD_TIME: %4d-%02d-%02d-%02d:%02d:%02d\n",
		       tm.tm_year+1900, tm.tm_mon+1, tm.tm_mday,
		       tm.tm_hour, tm.tm_min, tm.tm_sec);

		return (copy_to_user((void *)arg, &tm, sizeof(tm))) ? -EFAULT : 0;

	case RTC_SET_TIME:	/* Set the time/date in the RTC	*/

		if (!capable(CAP_SYS_TIME))
			return -EACCES;

		if (copy_from_user(&tm, (struct rtc_time *)arg, sizeof(tm)))
			return -EFAULT;

		/* convert from struct tm to rtc_time conventions */
		tm.tm_year += 1900;
		tm.tm_mon  += 1;

		/* simple but OK for all dates in time_t range */
		leap_yr = ((tm.tm_year % 4) == 0);

		if ((tm.tm_year < 1970) ||
		    (tm.tm_mon  > 12)	||
		    (tm.tm_mday == 0)	||
		    (tm.tm_mday >
		     (days_in_month[tm.tm_mon] + ((tm.tm_mon==2)&&leap_yr)))	||
		    (tm.tm_hour >= 24)	||
		    (tm.tm_min  >= 60)	||
		    (tm.tm_sec  >= 60)	) {

			return (-EINVAL);
		}

		debugk("RTC_SET_TIME: %4d-%02d-%02d-%02d:%02d:%02d\n",
		       tm.tm_year, tm.tm_mon, tm.tm_mday,
		       tm.tm_hour, tm.tm_min, tm.tm_sec);

		m48t58_set_rtc_time(mktime(tm.tm_year, tm.tm_mon, tm.tm_mday,
					   tm.tm_hour, tm.tm_min, tm.tm_sec));

		return (0);

	default:
		return (-EINVAL);
	}
}


static struct file_operations rtc_fops = {
	owner:		THIS_MODULE,
	ioctl:		rtc_ioctl,
	open:		rtc_open,
	release:	rtc_release,
};


static struct miscdevice m48t58rtc_miscdev = {
	RTC_MINOR,
	DEVICE_NAME,
	&rtc_fops
};


static int __init m48t58_init(void)
{
	int err;
	struct pci_dev	*dev;
	unsigned int bar1;

	if ((dev = pci_find_device(PCI_VENDOR_ID_OX16PCI954,
				   PCI_DEVICE_ID_OX16PCI954,
				   NULL)) == NULL) {
		printk("P3G4: No OX16PCI954 found, no RTC available!\n");
		return -ENODEV;
	}

	printk (KERN_INFO "M48T58 Real-Time Clock Driver " RTC_VERSION "\n");
	err = misc_register(&m48t58rtc_miscdev);
	if (err) {
		printk("M48T58: Register misc driver failed, errno is %d\n" ,err);
		return err;
	}

	pci_read_config_dword(dev, PCI_BASE_ADDRESS_1, &bar1);
	vbar1 = (unsigned int)ioremap(bar1, 0x1000);

	/*
	 * Set system time
	 */
	m48t58_set_system_time();

	return 0;
}

module_init(m48t58_init);


#ifdef MODULE
/*
 * Cleanup - unregister the appropriate file from /proc
 */
static void m48t58_cleanup (void)
{
	ppc_md.set_rtc_time = save_set_rtc_time;
	ppc_md.get_rtc_time = save_get_rtc_time;

	misc_deregister(&rtc_dev);

	debugk ("M48T58: Real-Time Clock Driver de-registered\n");
}

module_exit(m48t58_cleanup);
#endif	/* MODULE */
