/*
 * Driver for MCC200 GPIO.
 * (C) Copyright 2006 Dmitry Bazhenov <atrey@emcraft.com>, EC Emcraft
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/version.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/poll.h>

#include <linux/proc_fs.h>

#include <linux/miscdevice.h>
#include <asm/mpc5xxx.h>

#define MCC200_GPIO_VERSION "1.0"
#define MCC200_GPIO_NAME "mcc200_gpio"

#define MCC200_GPIO_MINOR	250

/* Number of GPIOs serviced by this driver */
#define	MCC200_GPIOS_SNUM	6

/* Used GPIO pins. 'buffer' in comments is the user's buffer in 
 mcc200_proc_read/write() calls */
#define	WKUP1		1	/* buffer[0] PSC2_4 pin */
#define	GPT7		7	/* buffer[1] TIMER7 pin */
#define	GPT6		6	/* buffer[2] TIMER6 pin */
#define	WKUP2		2	/* buffer[3] PSC3_9 pin */
#define	WKUP6		6	/* buffer[4] GPIO_WKUP_6 pin */
#define	GPIO_PSC3_5	13	/* buffer[5] PSC3_7 pin */

/* Timer GPIO configurations */
#define GPT_OUT_0	0x00000027	/* Simple GPIO output driving 0 */
#define	GPT_OUT_1	0x00000037	/* Simple GPIO output driving 1 */
#define	GPT_DISABLE	0x00000000	/* GPT pin disabled */

/* proc interface for driver */
#define MCC200_GPIO_PROC	"driver/mcc200_gpio"

/* Froward declarations */
static inline void	mcc200_wkup_enable(int pin, int data);
static inline void	mcc200_wkup_disable(int pin);
static inline void	mcc200_wkup_set(int pin, int data);
static inline int	mcc200_wkup_get(int pin);
static inline void	mcc200_gpt_disable(int pin);
static inline void	mcc200_gpt_set(int pin, int data);
static inline int	mcc200_gpt_get(int pin);
static inline void	mcc200_simp_enable(int pin, int data);
static inline void	mcc200_simp_disable(int n);
static inline void	mcc200_simp_set(int pin, int data);
static inline int	mcc200_simp_get(int pin);

/* Driver's banner */
static char mcc200_gpio_banner[] __initdata = KERN_INFO MCC200_GPIO_NAME ": "
						MCC200_GPIO_VERSION "\n";

/* Configure and enable WakeUp GPIO as a Simple GPIO Output */
static inline void mcc200_wkup_enable(int n, int data)
{
	/* Set initial value */
	*(volatile unsigned long *)MPC5xxx_WU_GPIO_DATAOUT |= data << (n + 24);
	/* Set WakeUp GPIO as an output */
	*(volatile unsigned long *)MPC5xxx_WU_GPIO_DIR |= 1 << (n + 24);
	/* Enable WakeUp GPIO pin */
	*(volatile unsigned long *)MPC5xxx_WU_GPIO_ENABLE |= 1 << (n + 24);
}

/* Disable and reset WakeUp GPIO */
static inline void mcc200_wkup_disable(int n)
{
	/* Disable WakeUp GPIO pin */
	*(volatile unsigned long *)MPC5xxx_WU_GPIO_ENABLE &= ~(1 << (n + 24));
	/* Reset value */
	*(volatile unsigned long *)MPC5xxx_WU_GPIO_DATAOUT &= ~(1 << (n + 24));
	/* Reset WakeUp GPIO direction to default */
	*(volatile unsigned long *)MPC5xxx_WU_GPIO_DIR &= ~(1 << (n + 24));
}

/* Set WakeUp GPIO output value */
static inline void mcc200_wkup_set(int n, int data)
{
	volatile unsigned long *out = (unsigned long *)MPC5xxx_WU_GPIO_DATAOUT;
	*out = data ? *out | (1 << (n + 24)) : *out & ~(1 << (n + 24));
}

/* Get latched WakeUp GPIO output value */
static inline int mcc200_wkup_get(int n)
{
	unsigned long data = *(volatile unsigned long *)MPC5xxx_WU_GPIO_DATAOUT;
	return (data >> (n + 24)) & 1;
}

/* Disable and reset Timer GPIO */
static inline void mcc200_gpt_disable(int n)
{
	volatile struct mpc5xxx_gpt *gpt =
		(volatile struct mpc5xxx_gpt *)MPC5xxx_GPT + n;
	gpt->emsr = GPT_DISABLE;
}

/* Configure Timer GPIO as a Simple GPIO output and set output value */
static inline void mcc200_gpt_set(int n, int data)
{
	volatile struct mpc5xxx_gpt *gpt =
		(volatile struct mpc5xxx_gpt *)MPC5xxx_GPT + n;
	gpt->emsr = data ? GPT_OUT_1 : GPT_OUT_0;
}

/* Get latched Timer GPIO output value */
static inline int mcc200_gpt_get(int n)
{
	volatile struct mpc5xxx_gpt *gpt =
		(volatile struct mpc5xxx_gpt *)MPC5xxx_GPT + n;
	return (gpt->emsr & 0x10) >> 4;
}

/* Enable Simple GPIO */
static inline void mcc200_simp_enable (int n, int data)
{
	volatile struct mpc5xxx_gpio *pgpio = 
		(struct mpc5xxx_gpio *)MPC5xxx_GPIO;
	/* Set initial value */
	pgpio->simple_dvo |= data << n;
	/* Set GPIO direction as an output */
	pgpio->simple_ddr |= 1 << n;
	/* Enable Simple GPIO pin */
	pgpio->simple_gpioe |= 1 << n;
}

/* Disable and reset Simple GPIO */
static inline void mcc200_simp_disable(int n)
{
	volatile struct mpc5xxx_gpio *pgpio = 
		(struct mpc5xxx_gpio *)MPC5xxx_GPIO;
	/* Disable Simple GPIO pin */
	pgpio->simple_gpioe &= ~(1 << n);
	/* Reset GPIO value */
	pgpio->simple_dvo &= ~(1 << n);
	/* Reset GPIO direction to default */
	pgpio->simple_ddr &= ~(1 << n);
}

/* Set Simple GPIO output value */
static inline void mcc200_simp_set(int n, int data)
{
	volatile struct mpc5xxx_gpio *pgpio =
		(struct mpc5xxx_gpio *)MPC5xxx_GPIO;
	pgpio->simple_dvo = data ? pgpio->simple_dvo | (1 << n) : 
					pgpio->simple_dvo & ~(1 << n);
}

/* Get latched Simple GPIO output value */
static inline int mcc200_simp_get(int n)
{
	volatile struct mpc5xxx_gpio *pgpio =
		(struct mpc5xxx_gpio *)MPC5xxx_GPIO;
	return ((pgpio->simple_dvo) >> (n)) & 1;
}

/*
 * Device file operations
 */
static ssize_t mcc200_gpio_write(struct file *file, const char *buf, size_t bufsz, loff_t *off)
{
	unsigned char state;

	if (get_user(state, buf))
		return -EFAULT;

	mcc200_wkup_set(WKUP1, state & 1);
	mcc200_gpt_set(GPT7, state & 2);
	mcc200_gpt_set(GPT6, state & 4);
	mcc200_wkup_set(WKUP2, state & 8);
	mcc200_wkup_set(WKUP6, state & 0x10);
	mcc200_simp_set(GPIO_PSC3_5, state & 0x20);

	return sizeof (state);
}

static ssize_t mcc200_gpio_read(struct file *file, char *buf, size_t bufsz, loff_t *off)
{
	unsigned char state;

	state = (mcc200_simp_get(GPIO_PSC3_5) << 5) | 
		(mcc200_wkup_get(WKUP6) << 4) |  
		(mcc200_wkup_get(WKUP2) << 3) | (mcc200_gpt_get(GPT6) << 2) | 
		(mcc200_gpt_get(GPT7) << 1) | mcc200_wkup_get(WKUP1);

	if (put_user(state, buf))
		return -EFAULT;

	return sizeof (state);
}

static unsigned int mcc200_gpio_poll(struct file *file, struct poll_table_struct *ptbl)
{
	/* It is always possible to read pin states */
	return POLLIN | POLLRDNORM;
}

#ifdef MODULE
static int mcc200_gpio_open(struct inode *inode, struct file *file)
{
	MOD_INC_USE_COUNT;
	return 0;
}

static int mcc200_gpio_close(struct inode *inode, struct file *file)
{
	MOD_DEC_USE_COUNT;
	return 0;
}
#endif

static struct file_operations mcc200_gpio_fops = {
	.owner = THIS_MODULE,
	.read	= mcc200_gpio_read,
	.write	= mcc200_gpio_write,
	.poll	= mcc200_gpio_poll,
#ifdef MODULE
	.open	= mcc200_gpio_open,
	.release = mcc200_gpio_close
#endif
};

static struct miscdevice mcc200_gpio_miscdev = {
	MCC200_GPIO_MINOR,
	MCC200_GPIO_NAME,
	&mcc200_gpio_fops
};

/*
 * procfs operations
 */
#ifdef CONFIG_PROC_FS

static int mcc200_gpio_proc_read(char *page, char **start, off_t off,
					int count, int *eof, void *data)
{
	char *p = page;

	p += sprintf(p, "%d%d%d%d%d%d\n",
				 mcc200_wkup_get(WKUP1),
				 mcc200_gpt_get(GPT7),
				 mcc200_gpt_get(GPT6),
				 mcc200_wkup_get(WKUP2),
				 mcc200_wkup_get(WKUP6),
				 mcc200_simp_get(GPIO_PSC3_5));
	return p - page;
}

static int mcc200_gpio_proc_write(struct file *file, const char *buffer,
	unsigned long count, void *data)
{
	int i;
	unsigned char vals[MCC200_GPIOS_SNUM];

	/* If the buffer is lesser then needed */
	if (count < MCC200_GPIOS_SNUM)
		return -EINVAL;

	/* If the buffer is valid */
	if ( copy_from_user(vals, buffer, MCC200_GPIOS_SNUM) ) {
		return -EFAULT;
	}

	/* Convert from ASCII to binary 0 or 1 */
	for (i=0; i<MCC200_GPIOS_SNUM; i++)
	{
		vals[i] -= '0';
		/* Only ASCII '0' and '1' are allowed */
		if (vals[i] > 1)
			return -EINVAL;
	}

	i = 0;
	mcc200_wkup_set(WKUP1, vals[i++]);
	mcc200_gpt_set(GPT7, vals[i++]);
	mcc200_gpt_set(GPT6, vals[i++]);
	mcc200_wkup_set(WKUP2, vals[i++]);
	mcc200_wkup_set(WKUP6, vals[i++]);
	mcc200_simp_set(GPIO_PSC3_5, vals[i++]);

	return count;
}
#endif

/*
 * Initialization/cleanup
 */
static int __devinit mcc200_gpio_init(void)
{
	printk(mcc200_gpio_banner);

#ifdef CONFIG_PROC_FS
	struct proc_dir_entry * proc;
#endif

	if (misc_register(&mcc200_gpio_miscdev) < 0) {
		printk(KERN_ERR "i2s: can't misc_register on minor=%d\n",
			MCC200_GPIO_MINOR);
		goto error;
	}

#ifdef CONFIG_PROC_FS
	if (!(proc = create_proc_entry(MCC200_GPIO_PROC, S_IFREG|S_IRUGO, NULL))) {
		printk(KERN_ERR "failed to create /proc/" MCC200_GPIO_PROC "\n");
		goto error_unreg;
	}

	proc->read_proc	 = mcc200_gpio_proc_read;
	proc->write_proc = mcc200_gpio_proc_write;
#endif

	/* Enable GPIO pins */
	mcc200_wkup_enable(WKUP1, 0);
	mcc200_gpt_set(GPT6, 1);
	mcc200_gpt_set(GPT7, 0);
	mcc200_wkup_enable(WKUP2, 0);
	mcc200_wkup_enable(WKUP6, 0);
	mcc200_simp_enable(GPIO_PSC3_5, 0);

	return 0;

error_unreg:
	misc_deregister(&mcc200_gpio_miscdev);
error:
	return -ENODEV;
}

static void __devexit mcc200_gpio_cleanup(void)
{
	/* Disable GPIO pins */
	mcc200_wkup_disable(WKUP1);
	mcc200_gpt_disable(GPT6);
	mcc200_gpt_disable(GPT7);
	mcc200_wkup_disable(WKUP2);
	mcc200_wkup_disable(WKUP6);
	mcc200_simp_disable(GPIO_PSC3_5);

#ifdef CONFIG_PROC_FS
	remove_proc_entry(MCC200_GPIO_PROC, NULL);
#endif

	misc_deregister(&mcc200_gpio_miscdev);
}

module_init(mcc200_gpio_init)
module_exit(mcc200_gpio_cleanup)

MODULE_AUTHOR("Dmitry Bazhenov <atrey@emcraft.com>");
MODULE_DESCRIPTION("MCC200 GPIO Driver");
MODULE_LICENSE("GPL");
EXPORT_NO_SYMBOLS;
