/***********************************************************************
 *
 * (C) Copyright 2006
 * DENX Software Engineering
 * Heiko Schocher, hs@denx.de
 * All rights reserved.
 *
 * This driver provides access to the PLD EPM 3064A
 * on a SM-MaCo4-Board with a MPC5200
 *
 ***********************************************************************/

/*
 * Standard in kernel modules
 */
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/smp_lock.h>
#include <linux/version.h>
#include <linux/string.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <asm/types.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/time.h>
#include <asm/mpc5xxx.h>

#include <linux/proc_fs.h>
#include <linux/vmalloc.h>
#include <linux/config.h>
#include <linux/pld_drv.h>

#ifndef KERNEL_VERSION
# define KERNEL_VERSION(a,b,c)	(((a) << 16) + ((b) << 8) + (c))
#endif

#if (!defined(LINUX_VERSION_CODE) || LINUX_VERSION_CODE < KERNEL_VERSION(2,2,0))
#error "You need to use at least Linux kernel version 2.2.0"
#endif

/***************************************************************************
 * Global stuff
 ***************************************************************************
 */
/*#define	DEBUG	1*/

/*
 * Deal with CONFIG_MODVERSIONS
 */
#if CONFIG_MODVERSIONS==1
/* # define MODVERSIONS */
# include <linux/modversions.h>
#endif

/*
 * For character devices
 */
#include <linux/fs.h>		/* character device definitions	*/
#include <linux/wrapper.h>	/* wrapper for compatibility with future versions */

static 	int pld_pbase = CFG_PLD_PADDR;
static	int pld_base;
static 	int lv_pbase = CFG_LV541_PADDR;
static	int lv_base;

#ifdef MODULE
MODULE_PARM(pld_pbase, "i");
__setup("pld_pbase=", pld_pbase);
MODULE_PARM(lv_pbase, "i");
__setup("lv_pbase=", lv_pbase);
#endif

static int pld_init_hardware (void);

/*
 * Device Declarations
 */

/*
 * The name for our device, as it will appear in /proc/devices
 */
#define DEVICE_NAME	"pld"
#define	PLD_VERSION	"$Revision: 1.0 $ hs@denx.de"
#define PLD_MAJOR	245

static struct proc_dir_entry *pld_proc_root = NULL;

/***************************************************************************
 * Local stuff
 **************************************************************************/


#ifdef	DEBUG
# define debugk(fmt,args...)	printk(fmt ,##args)
# define DEBUG_OPTARG(arg)	arg,
#else
# define debugk(fmt,args...)
# define DEBUG_OPTARG(arg)
#endif

/*
 * Prototypes
 */
static int		pld_open (struct inode *, struct file *);
static int		pld_release (struct inode *, struct file *);
static ssize_t 		pld_read (struct file *, char *, size_t,
				loff_t *);
static ssize_t 		pld_write (struct file *, const char *, size_t,
				loff_t *);
int init_module(void);
void pld_cleanup (void);

/***************************************************************************
 * Driver interface
 ***************************************************************************
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)

static struct file_operations pld_fops = {
	NULL,			/* lseek	*/
	pld_read,		/* read		*/
	pld_write,		/* write	*/
	NULL,			/* readdir	*/
	NULL,			/* select	*/
	NULL,			/* ioctl	*/
	NULL,			/* mmap		*/
	pld_open,		/* open		*/
	NULL,			/* flush	*/
	pld_release		/* close	*/
};

#else	/* Linux kernel >= 2.3.x */

static struct file_operations pld_fops = {
	owner:			THIS_MODULE,
	open:			pld_open,
	release:		pld_release,
	read:			pld_read,
	write:			pld_write,
};
#endif

#ifdef CONFIG_PROC_FS /* don't waste space if unused */
/*
 * PLD control via /proc/pld
 */
static int pld_cmd (unsigned int cmd, unsigned int off, unsigned int *param)
{
	char	data = (char) *param;
	debugk("cmd=%d off=%d param=%d data: %x\n", cmd, off, *param, data);

	switch (cmd) {
	case PLD_SET_F_VERS:
	case PLD_SET_L_VERS:
	case PLD_SET_DIP:
	case PLD_SET_ID:
		/* Read only */
		return -EFAULT;
		break;
	case PLD_SET_SYSTEM:
	case PLD_SET_LED1:
	case PLD_SET_LED2:
		writeb (data, pld_base + off);
		break;

	case PLD_SET_LV541:
		writeb (data, lv_base);
		break;

	case PLD_GET_F_VERS:
	case PLD_GET_L_VERS:
	case PLD_GET_DIP:
	case PLD_GET_ID:
	case PLD_GET_SYSTEM:
	case PLD_GET_LED1:
	case PLD_GET_LED2:
		*param = readb (pld_base + off);
		break;

	case PLD_GET_LV541:
		*param = readb (lv_base);
		break;

	default:
		return -ENOTTY;
		break;
	}
	return 0;
}

typedef struct {
	char *name;
	unsigned int set_cmd;
	unsigned int get_cmd;
} pld_cmds_t;

#define PLD_CMD_NUM 8
static pld_cmds_t pld_cmds[PLD_CMD_NUM] = {
	{ "f_vers", PLD_SET_F_VERS, PLD_GET_F_VERS },
	{ "l_vers", PLD_SET_L_VERS, PLD_GET_L_VERS },
	{ "dip", PLD_SET_DIP, PLD_GET_DIP },
	{ "id", PLD_SET_ID, PLD_GET_ID },
	{ "system", PLD_SET_SYSTEM, PLD_GET_SYSTEM },
	{ "LED1", PLD_SET_LED1, PLD_GET_LED1 },
	{ "LED2", PLD_SET_LED2, PLD_GET_LED2 },
	{ "LV541", PLD_SET_LV541, PLD_GET_LV541 }
};

static int
pld_write_proc (struct file *file, const char *buffer,
		   unsigned long count, void *data)
{
	char tmp[16];
	int len, cmd = pld_cmds[(int)data].set_cmd;
	unsigned int val;

	if (!count)
		return (count);

	len = sizeof(tmp) - 1;
	if (count < len)
		len = count;

	copy_from_user (tmp, buffer, len);

	tmp[len] = '\0';

	debugk("w %s=%s\n", pld_cmds[(int)data].name, tmp);

	val = simple_strtoul ((const char *)&tmp, NULL, 0);

	pld_cmd(cmd, (int)data, &val);
	return (len);
}

static int
pld_read_proc (char *page, char **start, off_t off,
		  int count, int *eof, void *data)
{
	int cmd = pld_cmds[(int)data].get_cmd;
	unsigned int val;
	int len;

	pld_cmd(cmd, (int)data, &val);
	debugk("r %s=%d %x\n", pld_cmds[(int)data].name, val, val);
	len = sprintf (page, "%x\n", val);
	*eof = 1;
	return len;
}


#endif /* CONFIG_PROC_FS */

/***************************************************************************
 *
 * Initialize the driver - Register the character device
 */
static int __init pld_init (void)
{
	int	i;

	printk (KERN_INFO "PLD DRIVER  " PLD_VERSION "\n");
	if (register_chrdev (PLD_MAJOR, DEVICE_NAME, & pld_fops))
	{
		printk (KERN_ERR "%s_drv: register_chrdev failed\n", DEVICE_NAME);
		return -EIO;
	}

	/* enable Memory Controller */
	pld_init_hardware ();

#ifdef CONFIG_PROC_FS
	/*
	 * Setup /proc/pld for PLD control
	 */
	if ((pld_proc_root = proc_mkdir(DEVICE_NAME, NULL)) == NULL) {
		printk (DEVICE_NAME ": proc_mkdir PLD failed\n");
		pld_cleanup ();
		return -ENXIO;
	}

	for (i = 0; i < PLD_CMD_NUM; i++) {
		struct proc_dir_entry *e;
		if ((e = create_proc_entry(pld_cmds[i].name, S_IRUGO | S_IWUSR,
					   pld_proc_root)) == NULL) {
			printk (DEVICE_NAME ": create_proc_entry %s failed\n",
				pld_cmds[i].name);
			pld_cleanup ();
		}
		e->write_proc = pld_write_proc;
		e->read_proc  = pld_read_proc;
		e->data = (void *)i;
	}
#endif

	return 0;
}

/***************************************************************************
 *
 * called whenever a process attempts to open the device
 * the device can only opened by one process
 */
static int pld_open (struct inode *inode, struct file *file)
{
	int	minor;

	lock_kernel ();

	minor = MINOR (inode->i_rdev);
	file->private_data = (void *)minor;
	unlock_kernel ();
	MOD_INC_USE_COUNT;

	debugk ("%s open success\n", DEVICE_NAME);
	return 0;
}



/***************************************************************************
 *
 * Called when a process closes the device.
 */
static int pld_release (struct inode *inode, struct file *file)
{
	lock_kernel ();

	unlock_kernel ();

	MOD_DEC_USE_COUNT;

	debugk ("%s close success\n", DEVICE_NAME);
	return 0;
}

/***************************************************************************
 *
 * Called when a process reads from the device.
 */
static ssize_t pld_read (struct file *filep, char *buf, size_t count,
				loff_t *f_pos)
{
	int	minor = (int) filep->private_data;
	char	data;

	debugk ("%s read count: %d Pos:%d M:%d\n", DEVICE_NAME, count, (int)*f_pos, minor);
	if ( minor == CFG_MINOR_LV541 ) {
		data = readb (lv_base);
	} else {
		if (minor >= 4) {
			return -EFAULT;
		} else {
			data = readb (pld_base + minor);
		}
	}
	if (copy_to_user(buf, &data, 1)) {
		return -EFAULT;
	}
	return 1;
}


/***************************************************************************
 *
 * Called when a process writes to the device.
 */
static ssize_t pld_write (struct file *filep, const char *buf, size_t count,
				loff_t *f_pos)
{
	int	minor = (int) filep->private_data;
	char	data;

	debugk ("%s write count: %d Pos:%d M:%d\n", DEVICE_NAME, count, (int)*f_pos, minor);
	if (copy_from_user(&data, buf, 1)) {
		return -EFAULT;
	}
	debugk ("%s write data: %x\n", DEVICE_NAME, data);
	if (( minor == CFG_MINOR_LV541 ) || (minor < 4)){
		return -EROFS;
	} else {
		writeb (data, pld_base + minor);
	}
	return 1;
}

/***************************************************************************
 * Hardware related stuff
 ***************************************************************************
 */

static int pld_init_hardware (void)
{
	u32	adren;

	/* enable Memory Controller for PLD */
	out_be32 ((u32 *)(MPC5xxx_MBAR + 0x001c), pld_pbase >> 16);
	out_be32 ((u32 *)(MPC5xxx_MBAR + 0x0020), (pld_pbase >> 16) + 1);
	out_be32 ((u32 *)(MPC5xxx_MBAR + 0x030c), 0x00001000);
	adren = in_be32 ((u32 *)MPC5xxx_ADREN);
	adren |= (0x00010000 << CFG_PLD_LP_CS);
	out_be32 ((u32 *)MPC5xxx_ADREN, adren);

	/* map physical address */
	pld_base = (unsigned long)ioremap_nocache (pld_pbase, 0x10000);
	if(!pld_base) {
		printk("%s: ioremap PLD failed @%x\n", DEVICE_NAME, pld_pbase);
		return -EIO;
	}

	adren = in_be32 ((u32 *)(MPC5xxx_MBAR + 0x0328));

	/* enable Memory Controller for LV541 */
	out_be32 ((u32 *)(MPC5xxx_MBAR + 0x0024), lv_pbase >> 16);
	out_be32 ((u32 *)(MPC5xxx_MBAR + 0x0028), (lv_pbase >> 16) + 1);
	out_be32 ((u32 *)(MPC5xxx_MBAR + 0x0310), 0x00001000);
	adren = in_be32 ((u32 *)MPC5xxx_ADREN);
	adren |= (0x00010000 << CFG_LV541_LP_CS);
	out_be32 ((u32 *)MPC5xxx_ADREN, adren);

	/* map physical address */
	lv_base = (unsigned long)ioremap_nocache (lv_pbase, 0x10000);
	if(!lv_base) {
		printk("%s: ioremap LV541 failed @%x\n", DEVICE_NAME, lv_pbase);
		return -EIO;
	}
	return 0;
}

/***************************************************************************
 * Module Declarations
 ***************************************************************************
 */
module_init(pld_init);

static int pld_exit_hardware (void)
{
	u32	adren;

	iounmap ((void *)pld_pbase);
	iounmap ((void *)lv_pbase);
	/* disable Memory Controller */
	adren = in_be32 ((u32 *)MPC5xxx_ADREN);
	adren &= ~(0x00010000 << CFG_PLD_LP_CS);
	adren &= ~(0x00010000 << CFG_LV541_LP_CS);
	out_be32 ((u32 *)MPC5xxx_ADREN, adren);
	return 0;
}

/*
 * Cleanup - unregister the appropriate file from /proc
 */
void pld_cleanup (void)
{
	pld_exit_hardware ();
	if (unregister_chrdev (PLD_MAJOR, DEVICE_NAME))
	{
		printk (KERN_ERR "%s: failed to unregister character device\n", DEVICE_NAME);
	}

#ifdef CONFIG_PROC_FS
	/*
	 * Remove PROC files in /proc/buffer
	 */
	if (pld_proc_root != NULL) {
		int i;
		for (i = 0; i < PLD_CMD_NUM; i++)
			remove_proc_entry (pld_cmds[i].name, pld_proc_root);
		remove_proc_entry ("pld", NULL);
	}
#endif


	debugk ("%s Driver de-registered\n", DEVICE_NAME);
}

module_exit(pld_cleanup);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Heiko Schocher <hs@denx.de>");
MODULE_DESCRIPTION("PLD driver");
