/*
 * Generic ST M48Txx driver.
 *
 * should be easyly adapted to other devices in the M48T-series that are
 * memory mapped and have a similiar layout.
 *
 * currently supports:
 * M48T08: 8 registers at end of chip, no way to store century data
 * M48T37: 16 registers at end of chip, including a full century register
 *
 * you MUST define the following symbols in your board specific file:
 * CFG_SRAM_PADDR
 * CFG_SRAM_SIZE
 * CFG_RTC_PADDR
 * CFG_RTC_SIZE
 *
 * Copyright (C) 2006 EMK Elektronik GmbH, info@emk-elektronik.de
 */
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/vmalloc.h>
#include <linux/errno.h>
#include <linux/rtc.h>
#include <linux/miscdevice.h>
#include <linux/proc_fs.h>

#include <asm/uaccess.h>
#include <asm/time.h>
#include <asm/io.h>

#ifdef CONFIG_TOP5200
#include <platforms/top5200.h>
#endif

#define	DRIVER_NAME		"rtc"
#define	DRIVER_TITLE		"M48Txx RTC driver with NVRAM support"
#define	AUTHOR			"Reinhard Meyer <info@emk-elektronik.de>"

/* the RTC registers */
#if defined (CONFIG_M48T08_RTC)
#define	M48_CONTROL	0
#define	M48_SEC		1
#define	M48_MIN		2
#define	M48_HOUR	3
#define	M48_DAY		4
#define	M48_DATE	5
#define	M48_MONTH	6
#define	M48_YEAR	7
#endif
#if defined (CONFIG_M48T37_RTC)
#define	M48_FLAGS	0
#define	M48_CENTURY	1
#define	M48_AL_SEC	2
#define	M48_AL_MIN	3
#define	M48_AL_HOUR	4
#define	M48_AL_DATE	5
#define	M48_IRQS	6
#define	M48_WDT		7
#define	M48_CONTROL	8
#define	M48_SEC		9
#define	M48_MIN		10
#define	M48_HOUR	11
#define	M48_DAY		12
#define	M48_DATE	13
#define	M48_MONTH	14
#define	M48_YEAR	15
#endif

static unsigned char reg_read (unsigned char reg)
{
	unsigned char val;

	val = *(volatile unsigned char *) (CFG_RTC_PADDR + reg);
	return val;
}

static void reg_write (unsigned char reg, unsigned char val)
{
	*(volatile unsigned char *) (CFG_RTC_PADDR + reg) = val;
}

static unsigned bcd2bin (unsigned char n)
{
	return ((((n >> 4) & 0x0F) * 10) + (n & 0x0F));
}

static unsigned char bin2bcd (unsigned int n)
{
	return (((n / 10) << 4) | (n % 10));
}

static void rtc_get (struct rtc_time *tmp)
{
	unsigned char sec, min, hour, day, date, month, year, century, flags;
	unsigned char ccr;	/* Clock control register */

	/* Lock RTC for read using clock control register */
	ccr = reg_read (M48_CONTROL) & 0x3f;
	reg_write (M48_CONTROL, ccr | 0x40);

	sec = reg_read (M48_SEC);
	min = reg_read (M48_MIN);
	hour = reg_read (M48_HOUR);
	day = reg_read (M48_DAY);
	date = reg_read (M48_DATE);
	month = reg_read (M48_MONTH);
	year = reg_read (M48_YEAR);
#if defined (CONFIG_M48T37_RTC)
	century = reg_read (M48_CENTURY);
	flags = reg_read (M48_FLAGS);
#else
	century = 0x20;
#endif

	/* UNLock RTC */
	reg_write (M48_CONTROL, ccr);

#if defined (CONFIG_M48T37_RTC)
	if (flags & 0x10)
		printk (KERN_INFO
			"m48t37 RTC clock reports battery low!!!\n");
#endif

	tmp->tm_sec = bcd2bin (sec & 0x7F);
	tmp->tm_min = bcd2bin (min & 0x7F);
	tmp->tm_hour = bcd2bin (hour & 0x3F);
	tmp->tm_mday = bcd2bin (date & 0x3F);
	tmp->tm_mon = bcd2bin (month & 0x1F);
	tmp->tm_year = bcd2bin (year) + bcd2bin (century) * 100;
	tmp->tm_wday = bcd2bin (day & 0x07);
	tmp->tm_yday = -1;
	tmp->tm_isdst = -1;
	tmp->tm_mon--;
	tmp->tm_year -= 1900;
}

static void rtc_set (struct rtc_time *tmp)
{
	unsigned char ccr;	/* Clock control register */

	tmp->tm_mon++;
	tmp->tm_year += 1900;

	/* Lock RTC for write using clock control register */
	ccr = reg_read (M48_CONTROL) & 0x3f;
	reg_write (M48_CONTROL, ccr | 0x80);

	reg_write (M48_YEAR, bin2bcd (tmp->tm_year % 100));
#if defined (CONFIG_M48T37_RTC)
	reg_write (M48_CENTURY, bin2bcd (tmp->tm_year / 100));
#endif
	reg_write (M48_MONTH, bin2bcd (tmp->tm_mon));
	reg_write (M48_DATE, bin2bcd (tmp->tm_mday));
	reg_write (M48_DAY, bin2bcd (tmp->tm_wday) & 0x07);	/* prevent FT bit */
	reg_write (M48_HOUR, bin2bcd (tmp->tm_hour));
	reg_write (M48_MIN, bin2bcd (tmp->tm_min));
	reg_write (M48_SEC, bin2bcd (tmp->tm_sec) & 0x7f);	/* prevent ST bit */

	/* UNLock RTC */
	reg_write (M48_CONTROL, ccr);
}

static void rtc_reset (void)
{
	unsigned char val;

	/* Clear all clock control registers */
	reg_write (M48_CONTROL, 0x80);	/* No Read Lock or calibration */

	/* Clear stop bit */
	val = reg_read (M48_SEC);
	val &= 0x7f;
	reg_write (M48_SEC, val);

	/* disable frequency test */
	val = reg_read (M48_DAY);
	val &= 0x07;
	reg_write (M48_DAY, val);

#if defined (CONFIG_M48T37_RTC)
	/* disable watchdog */
	reg_write (M48_WDT, 0);

	/* disable alarms */
	reg_write (M48_AL_SEC, 0);
	reg_write (M48_AL_MIN, 0);
	reg_write (M48_AL_HOUR, 0);
	reg_write (M48_AL_DATE, 0);

	/* disable interrupts */
	reg_write (M48_IRQS, 0);
#endif

	/* Clear write lock */
	reg_write (M48_CONTROL, 0);
}

/* open the device */
static int rtc_open (struct inode *ino, struct file *filep)
{
	MOD_INC_USE_COUNT;
	return 0;
}

/* close the device */
static int rtc_release (struct inode *ino, struct file *filep)
{
	MOD_DEC_USE_COUNT;
	return 0;
}

/* read bytes from memory */
static ssize_t rtc_read (struct file *file,	/* file handle */
			 char *buf,	/* buffer in user space */
			 size_t count,	/* number of bytes to read */
			 loff_t * ppos)
{				/* position to start at */
	unsigned long p = *ppos;

	if (p >= CFG_SRAM_SIZE)
		return 0;

	if (count > CFG_SRAM_SIZE - p)
		count = CFG_SRAM_SIZE - p;

	p += CFG_SRAM_PADDR;

	if (copy_to_user (buf, (void *) p, count))
		return -EFAULT;

	*ppos += count;

	return count;
}

/* write bytes to memory */
static ssize_t rtc_write (struct file *file,	/* file handle */
			  const char *buf,	/* buffer in user space */
			  size_t count,	/* number of bytes to write */
			  loff_t * ppos)
{				/* position to start at */
	unsigned long p = *ppos;

	if (p >= CFG_SRAM_SIZE)
		return -ENOSPC;

	if (count > CFG_SRAM_SIZE - p)
		count = CFG_SRAM_SIZE - p;

	p += CFG_SRAM_PADDR;

	if (copy_from_user ((void *) p, buf, count))
		return -EFAULT;

	*ppos += count;

	return count;
}

static loff_t rtc_lseek (struct file *file, loff_t offset, int orig)
{
	switch (orig) {
	case 0:
		file->f_pos = offset;
		return file->f_pos;
	case 1:
		file->f_pos += offset;
		return file->f_pos;
	case 2:
		file->f_pos = CFG_SRAM_SIZE - offset;
		return file->f_pos;
	default:
		return -EINVAL;
	}
}

static int rtc_ioctl (struct inode *inode,
		      struct file *file, unsigned int cmd, unsigned long arg)
{
	struct rtc_time rtc_tm;
	int ret;

	switch (cmd) {
	case RTC_RD_TIME:
		memset (&rtc_tm, 0, sizeof rtc_tm);
		rtc_get (&rtc_tm);
		ret = copy_to_user ((struct rtc_time *) arg, &rtc_tm,
				    sizeof rtc_tm);
		if (ret)
			return -EFAULT;
		return 0;

	case RTC_SET_TIME:
		ret = copy_from_user (&rtc_tm, (struct rtc_time *) arg,
				      sizeof rtc_tm);
		if (ret)
			return -EFAULT;
		rtc_set (&rtc_tm);
		return 0;

	default:
		return -EINVAL;
	}
}

static int rtc_mmap (struct file *file, struct vm_area_struct *vma)
{
	unsigned long offset = vma->vm_pgoff << PAGE_SHIFT;
	unsigned long size = vma->vm_end - vma->vm_start;

	if (offset + size > CFG_SRAM_SIZE)
		return -EINVAL;

	offset += CFG_SRAM_PADDR;

	if (remap_page_range (vma->vm_start, offset, size, vma->vm_page_prot))
		return -EAGAIN;

	return 0;
}

static int proc_output (char *buf)
{
	char *p = buf;
	unsigned char sec, min, hour, day, date, month, year, century, flags;
	unsigned char ccr;	/* Clock control register */

	/* Lock RTC for read using clock control register */
	ccr = reg_read (M48_CONTROL) & 0x3f;
	reg_write (M48_CONTROL, ccr | 0x40);

	sec = reg_read (M48_SEC);
	min = reg_read (M48_MIN);
	hour = reg_read (M48_HOUR);
	day = reg_read (M48_DAY);
	date = reg_read (M48_DATE);
	month = reg_read (M48_MONTH);
	year = reg_read (M48_YEAR);
#if defined (CONFIG_M48T37_RTC)
	century = reg_read (M48_CENTURY);
	flags = reg_read (M48_FLAGS);
#else
	century = 0x20;
#endif

	/* UNLock RTC */
	reg_write (M48_CONTROL, ccr);

#if defined (CONFIG_M48T08_RTC)
	p += sprintf (p, DRIVER_NAME " chip: M48T08\n");
#endif
#if defined (CONFIG_M48T37_RTC)
	p += sprintf (p, DRIVER_NAME " chip: M48T37\n");
	if (flags & 0x10)
		p += sprintf (p, "clock reports battery low!!!\n");
#endif
	p += sprintf (p, "%02x:%02x:%02x %02x.%02x.%02x%02x\n",
		      hour, min, sec, date, month, century, year);

	return p - buf;
}

static int read_proc (char *page, char **start, off_t off, int count,
		      int *eof, void *data)
{
	int len = proc_output (page);

	if (len <= off + count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;
	return len;
}

static struct file_operations rtc_operations = {
      llseek:rtc_lseek,
      ioctl:rtc_ioctl,
      mmap:rtc_mmap,
      read:rtc_read,
      write:rtc_write,
      open:rtc_open,
      release:rtc_release,
};

static struct miscdevice rtc_miscdev = {
	RTC_MINOR, DRIVER_NAME, &rtc_operations
};

static int __init rtc_init (void)
{
	printk (KERN_INFO DRIVER_TITLE " - " AUTHOR "\n  chip: "
#if defined (CONFIG_M48T08_RTC)
		"M48T08"
#endif
#if defined (CONFIG_M48T37_RTC)
		"M48T37"
#endif
		"\n");
	if (misc_register (&rtc_miscdev) < 0) {
		return -EBUSY;
	}
	create_proc_read_entry (DRIVER_NAME, 0, 0, read_proc, NULL);
	rtc_reset ();
	return 0;
}

static void __exit rtc_exit (void)
{
	rtc_reset ();
	remove_proc_entry (DRIVER_NAME, NULL);
	misc_deregister (&rtc_miscdev);
}

module_init (rtc_init);
module_exit (rtc_exit);

MODULE_LICENSE ("GPL");
MODULE_AUTHOR (AUTHOR);
MODULE_DESCRIPTION (DRIVER_TITLE);
EXPORT_NO_SYMBOLS;
