/*
 * Special driver for Sennheiser to access additional UARTs through a PLD
 * attached to a Microsys CR825.
 *
 * Code was liberally borrowed from arch/ppc/8260_io/uart.c.
 *
 * Note that there is no way to use this as a module because we cannot return
 * memory allocated using m8260_cpm_dpalloc() and m8260_cpm_hostalloc().
 *
 * Copyright (C) 2006 Gary Jennejohn <garyj@denx.de>
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/skbuff.h>
#include <linux/poll.h>
#include <asm/uaccess.h>
#include <asm/immap_8260.h>
#include <asm/mpc8260.h>
#include <asm/cpm_8260.h>
#include <asm/irq.h>

/* this is a character device */
#define SENNPLD_MAJ 120 /* local/experimental use */

/* prototypes */
static int
sennpld_open(struct inode *inode, struct file *filp);
static int
sennpld_close(struct inode *inode, struct file *filp);
static ssize_t
sennpld_read(struct file *filp, char *buf, size_t count, loff_t *f_pos);
static ssize_t
sennpld_write(struct file *filp, const char *buf, size_t count, loff_t *f_pos);
static unsigned int
sennpld_poll(struct file *, struct poll_table_struct *);
static int
sennpld_ioctl(struct inode *, struct file *, unsigned int, unsigned long);

/* used to wait for completed reads and poll() */
static wait_queue_head_t sennpld_wait;
static struct sk_buff_head sennpld_skbq;

/*
 * CPM Buffer Descriptor pointers.
*/
static cbd_t *rx_bd_base;
static cbd_t *rx_cur;
static cbd_t *tx_bd_base;
static cbd_t *tx_cur;

struct file_operations sennpld_fops = {
	owner: THIS_MODULE,
	read: sennpld_read,
	write: sennpld_write,
	poll: sennpld_poll,
	ioctl: sennpld_ioctl,
	open: sennpld_open,
	release: sennpld_close,
};

/*
 * The number of buffer descriptors and their sizes.
 * The customer thinks that 1kB should be enough, but
 * allocate 2 kB for RX.
*/
#define RX_NUM_FIFO	32
#define RX_BUF_SIZE	64
#define TX_NUM_FIFO	64
#define TX_BUF_SIZE	32

#define SENNPLD_VER "0.7h"

#ifndef MIN
#define MIN(a,b)	((a) < (b) ? (a) : (b))
#endif

/* not defined in cpm_8260.h */
#define BD_SC_CT	((ushort)0x0001)	/* lost CTS */
#define BD_SC_UN	((ushort)0x0002)	/* TX UNderrun */
#define BD_SC_CR	((ushort)0x0004)	/* CRC */
#define BD_SC_NO	((ushort)0x0010)	/* non-octet */
#define BD_SC_DE	((ushort)0x0080)	/* DPLL error */

#undef SENNPLD_DEBUG
#undef SENNPLD_DEBUGW
#undef SENNPLD_DEBUGR
#define SENNPLD_INTDEBUG

#define SENNPLD_TRIG
#ifdef SENNPLD_TRIG
#define SENN_SET_PD7 _IO('S', 2)
#endif

/* this will put debug output into a memory buffer without driver slowups */
#undef SENN_BUFFERED_DEBUG
#ifdef SENN_BUFFERED_DEBUG
/* an IOCTL to grab the debugging output */
#define SENN_GET_DEBUG _IOR('S', 1, char *)
#define SENN_BUF_SIZE 2048
static char senn_debug[SENN_BUF_SIZE];
static char *senn_dptr = senn_debug;
#define SENNDEBUG(f, args...) \
	{ senn_dptr += sprintf(senn_dptr, f, ##args); \
	if ((senn_dptr - senn_debug) >= (SENN_BUF_SIZE - 2)) \
		senn_dptr = senn_debug; }
#else
#define SENNDEBUG(f, args...) printk(f, ##args)
#endif
/* this allows one to easily turn off debug by pre-pending an X */
#define XSENNDEBUG(f, args...)

/* logically empty the rx buffer */
#define SENN_CLR_RXB _IO('S', 3)
/* report how many chars were lost in receive_chars() */
#define SENN_GET_LOST _IOR('S', 4, unsigned int *)
/* the counter for the above ioctl */
static unsigned int lostchars;

static int already_open = 0;

static unsigned char *rxbuf;
static unsigned char *rxhead;
static unsigned char *rxtail;
static unsigned char *rxend;
#define RXBUF_SIZE 5000

/* define this to put the RX and TX clocks both on CLK4! */
#undef ONLY_CLK4

static int
sennpld_open(struct inode *inode, struct file *filp)
{
	XSENNDEBUG("open\n");
/* allow a process to also open the driver for the SENN_GET_DEBUG IOCTL */
#ifndef SENN_BUFFERED_DEBUG
	if (already_open != 0)
		return -EBUSY;
#endif

	/* EXCL open */
	already_open++;

#ifdef SENN_BUFFERED_DEBUG
	memset(senn_debug, 0, SENN_BUF_SIZE);
	senn_dptr = senn_debug;
#endif

	return 0;
}

static int
sennpld_close(struct inode *inode, struct file *filp)
{
	unsigned long flags;

	local_irq_save(flags);
	XSENNDEBUG("close\n");
	/* flush the sk_buff write queue */
	rxhead = rxtail = rxbuf;
	if (!skb_queue_empty(&sennpld_skbq))
		skb_queue_purge(&sennpld_skbq);

	already_open = 0;
	local_irq_restore(flags);
	/* always succeeds */
	return 0;
}

#if 0 /* too complicated! */
static void
sennpld_trywrite(int whence)
{
	int	c, free, idx;
	uint addr;
unsigned char *taddr;
	volatile cbd_t *bdp;
	volatile cbd_t *abdp[TX_NUM_FIFO];
	struct sk_buff *skb;
	unsigned long flags;
#ifdef SENNPLD_TRIG
	volatile	iop8260_t	*io;
#endif
	static unsigned int txcnt = 0;
volatile	scc_trans_t	*stp;

	local_irq_save(flags);

	XSENNDEBUG("whence %d\n", whence);
	if (skb_queue_empty(&sennpld_skbq)) {
		XSENNDEBUG("WQ1 empty\n");
		local_irq_restore(flags);
		return;
	}

#ifdef SENNPLD_TRIG
	io = &immr->im_ioport;
#endif
	bdp = tx_cur;
	if (bdp->cbd_sc & BD_SC_READY) {
#if 0
		volatile cbd_t *scr;
		for (scr = tx_bd_base;; scr++) {
			if (!(scr->cbd_sc & BD_SC_READY))
				printk("B_S_R but %d NOT!\n", scr - tx_bd_base);
			if (scr->cbd_sc & BD_SC_WRAP)
				break;
		}
#endif
		SENNDEBUG("B_S_R! sent %d\n", txcnt);
		local_irq_restore(flags);
		return;
	}

	memset(abdp, 0, TX_NUM_FIFO * sizeof(cbd_t *));
	skb = NULL;
	idx = 0;

	free = TX_BUF_SIZE;
	bdp->cbd_datlen = 0;
	addr = bdp->cbd_bufaddr;

	/* use up as many skb's and bdp's as possible */
	while (1) {
		c = 0;

		XSENNDEBUG("QL %d\n", skb_queue_len(&sennpld_skbq));
		/* the queue is empty and there's no pending data */
		if (skb == NULL && skb_queue_empty(&sennpld_skbq)) {
		XSENNDEBUG("WQ2 empty\n");
			if (abdp[idx] != NULL) {
				/*
				 * Get next BD.
				*/
				if (bdp->cbd_sc & BD_SC_WRAP)
					bdp = tx_bd_base;
				else
					bdp++;
				tx_cur = (cbd_t *)bdp;
				idx++;
			}
			break;
		}

		if (skb == NULL)
			skb = skb_dequeue(&sennpld_skbq);

		c = MIN(skb->len, free);

		memcpy(__va(addr), skb->data, c);
		abdp[idx] = bdp;
		free -= c;
		bdp->cbd_datlen += c;
		addr += c;

		XSENNDEBUG("bdp %p skb %p len %d free %d idx %d\n", bdp, skb, skb->len, free, idx);

		/* either free the skb because it's empty or bump its pointers */
		if (c == skb->len) {
			dev_kfree_skb(skb);
			skb = NULL;
		} else {
printk("SPLIT skb!\n");
			skb->len -= c;
			skb->data += c;
		}

		/* this bdp is full */
		if (free == 0) {
			/*
			 * Get next BD.
			*/
			if (bdp->cbd_sc & BD_SC_WRAP)
				bdp = tx_bd_base;
			else
				bdp++;
			tx_cur = (cbd_t *)bdp;
			idx++;
			if (bdp->cbd_sc & BD_SC_READY) {
				if (skb != NULL) {
printk("SAVE skb!\n");
					/* put the sk_buff back at the head */
					skb_queue_head(&sennpld_skbq, skb);
				}
				break;
			}
			free = TX_BUF_SIZE;
			bdp->cbd_datlen = 0;
			addr = bdp->cbd_bufaddr;
		}
	}

	XSENNDEBUG("IDX %d tx_cur %p abdp[0] %p\n", idx, tx_cur, abdp[0]);
	/* only do these steps if we put data into one or more bdp's */
	if (abdp[0] != NULL) {
		for (c = 0; c < idx; c++) {
if (abdp[c]->cbd_datlen == 0)
printk("acbp[%d:%d] in abdp but len 0!\n", c, abdp[c] - tx_bd_base);
#if 0 /* XXXX */
stp = (scc_trans_t *)&immr->im_dprambase[PROFF_SCC1];
printk("Toff %d Soff %d\n", abdp[c] - (cbd_t *)tx_bd_base, (cbd_t *)stp->st_genscc.scc_tbptr - (cbd_t *)stp->st_genscc.scc_tbase);
bdp = tx_bd_base;
for (free = 0; free < TX_NUM_FIFO; free++) {
printk("BD_SC[%d] %#x ", free, bdp->cbd_sc);
bdp++;
}
printk("\n");
#endif
			abdp[c]->cbd_sc &= ~BD_SC_LAST;
			abdp[c]->cbd_sc |= BD_SC_READY;
			txcnt++;
#ifdef SENNPLD_TRIG
			io->iop_pdata |= 0x200;
			io->iop_pdata &= ~0x200;
#endif
			/* the last BDP */
			if (abdp[c + 1] == NULL) {
				abdp[c]->cbd_sc |= BD_SC_LAST;
#if 0 /* XXXX */
taddr = (unsigned char *)__va(abdp[c]->cbd_bufaddr);
for (free = 0; free < abdp[c]->cbd_datlen; free++)
printk("L[%d] %#x ", free, taddr[free]);
printk("\n");
#endif
				break;
			}
		}
	}
	local_irq_restore(flags);
}
#else
static void
sennpld_trywrite(int whence)
{
	int	c, count;
	volatile cbd_t *bdp;
	volatile cbd_t *lbdp;
	struct sk_buff *skb;
	unsigned char *buf;
	unsigned long flags;
#ifdef SENNPLD_TRIG
	volatile iop8260_t *io;
#endif

	local_irq_save(flags);
	if (skb_queue_empty(&sennpld_skbq)) {
		local_irq_restore(flags);
		return;
	}

#ifdef SENNPLD_TRIG
	io = &immr->im_ioport;
#endif
	bdp = tx_cur;
	/* only do one sk_buff per call */
	skb = skb_dequeue(&sennpld_skbq);
	count = skb->len;
	buf = skb->data;

	while (1) {
		lbdp = NULL;
		c = MIN(count, TX_BUF_SIZE);

		if (c <= 0)
			break;

		if (bdp->cbd_sc & BD_SC_READY) {
			printk("%s: BD_SC_READY c %d\n", __FUNCTION__, c);
			break;
		}

		memcpy(__va(bdp->cbd_bufaddr), buf, c);

		bdp->cbd_datlen = c;
		/* bdp->cbd_sc &= ~BD_SC_LAST; */
		bdp->cbd_sc |= BD_SC_READY|BD_SC_LAST;
#ifdef SENNPLD_TRIG
		io->iop_pdata |= 0x200;
		io->iop_pdata &= ~0x200;
#endif

		buf += c;
		count -= c;
		lbdp = bdp;

		/* Get next BD.
		*/
		if (bdp->cbd_sc & BD_SC_WRAP)
			bdp = tx_bd_base;
		else
			bdp++;
		tx_cur = (cbd_t *)bdp;
	}
	/* lbdp->cbd_sc |= BD_SC_LAST; */
	dev_kfree_skb(skb);
	local_irq_restore(flags);
}
#endif

static ssize_t
sennpld_write(struct file *filp, const char *buf, size_t count, loff_t *f_pos)
{
	struct sk_buff *skb;
#ifdef SENNPLD_DEBUGW
	static int cnt = 0;
#endif

	if (buf == NULL)
		return -EINVAL;

	if (count == 0)
		return 0;

	/*
	 * The data are always a multiple of 16 bits! This assumes
	 * 8 bits per byte!
	 */
	if (count < 2 || count % 2 != 0)
		return -EINVAL;

	/* allocate an sk_buff for the data */
	skb = dev_alloc_skb(count);
	if (skb == NULL) {
		printk("%s: unable to allocate sk_buff for write data\n",
			__FUNCTION__);
		return -ENOMEM;
	}
	skb_put(skb, count);
	if (copy_from_user((void *)skb->data, (void *)buf, count))
		return -EFAULT;

	/* hang it into the queue */
	skb_queue_tail(&sennpld_skbq, skb);

	/* try to write it - depends on whether a cbd_t is free */
#ifdef SENNPLD_DEBUGW
	if (++cnt >= 4) {
#endif
	sennpld_trywrite(1);
#ifdef SENNPLD_DEBUGW
		cnt = 0;
	}
#endif

	/* always report success since the data are safe in the sk_buff */
	return count;
}

static ssize_t
sennpld_read(struct file *filp, char *buf, size_t count, loff_t *f_pos)
{
	int ret, len, tot, left;
	unsigned char *mybuf;
#ifdef SENNPLD_TRIGX
	volatile	iop8260_t	*io;
#endif

	ret = 0;

	if (buf == NULL)
		return -EINVAL;

	if (count == 0)
		return 0;

	if (rxhead == rxtail)
		if (filp->f_flags & O_NONBLOCK) {
#ifdef SENNPLD_TRIGX
		io = &immr->im_ioport;
		io->iop_pdatd |= 0x01000000;
		io->iop_pdatd &= ~0x01000000;
#endif
			return -EWOULDBLOCK;
		}

	/*
	 * The data are always a multiple of 16 bits! This assumes
	 * 8 bits per byte!
	 */
	if (count < 2 || count % 2 != 0)
		return -EINVAL;

	ret = wait_event_interruptible(sennpld_wait, rxhead != rxtail);
	if (ret)
		return ret; /* ERESTARTSYS */

	/* copy out as much data as possible up to count bytes */
	mybuf = (unsigned char *)kmalloc(count, GFP_KERNEL);
	tot = 0;
	left = count;
	while (tot < count) {
		if (rxtail < rxhead) {
			/* rxtail wrqpped around - copy from rxhead to rxend */
			len = rxend - rxhead;
			if (left < len)
				len = left;
			memcpy((void *)(mybuf + tot), rxhead, len);
			rxhead += len;
			if (rxhead == rxend)
				rxhead = rxbuf;
			if (rxhead == rxtail) {
				tot += len;
				break;
			}
		} else {
			/* copy from rxhead to rxtail */
			len = rxtail - rxhead;
			if (left < len)
				len = left;
			memcpy((void *)(mybuf + tot), rxhead, len);
			rxhead += len;
			if (rxhead == rxtail) {
				tot += len;
				break;
			}
		}
		tot += len;
		left -= len;
	}
	if (tot != 0) {
		/* NOTE: this may cause data loss! */
		if (copy_to_user((void *)buf, mybuf, tot)) {
			kfree(mybuf);
			return -EFAULT;
		}
	}
	kfree(mybuf);
	return tot;
}

static unsigned int
sennpld_poll(struct file *filp, struct poll_table_struct *ptab)
{
	unsigned int res;

	/* only sleep if there are no pending data */
	if (rxhead == rxtail)
		poll_wait(filp, &sennpld_wait, ptab);

	/* writing always works - this is only woken when there are readable data */
	res = POLLOUT|POLLWRNORM;
	/* double check - may be redundant */
	if (rxhead != rxtail)
		res |= POLLIN|POLLRDNORM;
	/* double check - may be redundant */
	if (lostchars != 0)
		res |= POLLIN|POLLPRI;

	return res;
}

static int
sennpld_ioctl(struct inode *inode, struct file *filp, unsigned int cmd,
	unsigned long arg)
{
#ifdef SENNPLD_TRIG
	volatile	iop8260_t	*io;
#endif
#ifdef SENN_BUFFERED_DEBUG
	int len;
	void *targ;
#endif
	switch (cmd) {
#ifdef SENN_BUFFERED_DEBUG
		case SENN_GET_DEBUG:
			targ = (void *)arg;
			len = 0;
			/*
			 * Copy from the pointer to the end of the buffer. Note that
			 * senn_debug[] is filled with 0 in the open routine!
			 * Only do this if senn_dptr has wrapped around!
			 */
			if (*(senn_dptr + 1) != '\0') {
				len = (&senn_debug[(SENN_BUF_SIZE - 1)] - senn_dptr) + 1;
				if (copy_to_user(targ, senn_dptr, len))
					return -EFAULT;
			}
			/* no need to copy it twice */
			if (len == SENN_BUF_SIZE)
				break;
			/*
			 * Copy from the start of the buffer to the pointer. Note that
			 * senn_debug[] is filled with 0 in the open routine!
			 */
			targ += len;
			len = senn_dptr - senn_debug;
			/* senn_dptr may have wrapped */
			if (len == 0)
				break;
			if (copy_to_user(targ, senn_debug, len))
				return -EFAULT;
			break;
#endif
#ifdef SENNPLD_TRIG
		case SENN_SET_PD7:
			io = &immr->im_ioport;
			io->iop_pdatd |= 0x01000000;
			io->iop_pdatd &= ~0x01000000;
			break;
#endif
		case SENN_CLR_RXB:
			rxhead = rxtail;
			break;
		case SENN_GET_LOST:
			if (copy_to_user((void *)arg, &lostchars, sizeof(lostchars)))
				return -EFAULT;
			lostchars = 0;
			break;
		default:
			return -ENOIOCTLCMD;
	}
	return 0;
}

static void
receive_chars(void)
{
	unsigned char *cp;
	int	i, avail;
	ushort	status;
	volatile cbd_t	*bdp;
#ifdef RXCNT
	int rxcnt;
#endif

	/*
	 * Just loop through the closed BDs and copy the characters into
	 * the ring buffer as far as possible.
	 */
	bdp = rx_cur;
#ifdef RXCNT
	rxcnt = 0;
#endif

	for (;;) {
		if (bdp->cbd_sc & BD_SC_EMPTY)	/* If this one is empty */
			break;			/*   we are all done */

		/*
		 * Handle the case where there is no application waiting for data
		 * but the PLD is sending us data anyway. Just throw away the data.
		 */
		if (already_open == 0)
			goto nxt;

		/*
		 * Get the number of characters and the buffer pointer.
		 * What would normally be the CRC is actually the last data byte,
		 * so bump the length by one.
		*/
		i = bdp->cbd_datlen;
		cp = (unsigned char *)__va(bdp->cbd_bufaddr);
		status = bdp->cbd_sc;

		/* check for errors */
		if (status & BD_SC_CD) {
			printk("%s: CD\n", __FUNCTION__);
		}
		if (status & BD_SC_OV) {
			printk("%s: OV\n", __FUNCTION__);
			goto nxt;
		}
#if 0 /* we always see this error because the PLD sends no CRC - ignore it */
		if (status & BD_SC_CR) {
			printk("%s: CR\n", __FUNCTION__);
		}
#endif
		if (status & BD_SC_NO) {
			printk("%s: NO\n", __FUNCTION__);
			goto nxt;
		}
		if (status & BD_SC_DE) {
			printk("%s: DE\n", __FUNCTION__);
			goto nxt;
		}
#ifdef SENNPLD_DEBUGR
		if (status & BD_SC_LAST) {
			printk("%s: LAST len %d\n", __FUNCTION__, i);
		}
#endif

		/* figure total space available in rxbuf */
		if (rxhead <= rxtail)
			avail = (rxend - rxtail) + (rxhead - rxbuf);
		else
			avail = rxhead - rxtail;

		/* if only 2 bytes are left then the ring buffer is full */
		/* must be a multiple of 2! */
		if (avail <= 2) {
			lostchars += i;
			goto nxt;
		}

		/*
		 * this never seems to be printed out due to the if above, but
		 * leave it in.
		 */
		if (avail < i) {
			printk("%s: lost data, need %d only have %d free!\n", __FUNCTION__,
				i, avail);
			lostchars += i - avail;
			i = avail;
		}

		/* don't try to get fancy */
		if (rxhead <= rxtail) {
			avail = rxend - rxtail;
			if (avail > i)
				avail = i;
			memcpy(rxtail, cp, avail);
			i -= avail;
			cp += avail;
			rxtail += avail;
			if (i != 0) {
				avail = rxhead - rxbuf;
				if (avail > i)
					avail = i;
				/* don't want tail == head after this! */
				if ((rxbuf + avail) == rxhead)
					avail -= 2;
				memcpy(rxbuf, cp, avail);
				rxtail = rxbuf + avail;
			}
			goto done;
		}
		if (rxhead > rxtail) {
			/* don't want tail == head after this! */
			avail = rxhead - rxtail;
			if (avail > i)
				avail = i;
			/* don't want tail == head after this! */
			if ((rxtail + avail) == rxhead)
				avail -= 2;
			memcpy(rxtail, cp, avail);
			rxtail += avail;
		}

	done:
#ifdef RXCNT
		rxcnt++;
#endif
		/* wake up any waiters */
		if (status & BD_SC_LAST)
			wake_up_interruptible(&sennpld_wait);

		/*
		 * This BD is ready to be used again.  Clear status.
		 * Get next BD.
		 */
	nxt:
		bdp->cbd_sc |= BD_SC_EMPTY;
		bdp->cbd_sc &= ~(BD_SC_CD | BD_SC_OV | BD_SC_CR | BD_SC_NO | BD_SC_DE);

		if (bdp->cbd_sc & BD_SC_WRAP)
			bdp = rx_bd_base;
		else
			bdp++;
	}

	rx_cur = (cbd_t *)bdp;
#ifdef RXCNT
	if (rxcnt > 5)
		printk("RX %d\n", rxcnt);
#endif
}

/*
 * This is the driver's interrupt routine
 */
static void
sennpld_interrupt(int irq, void * dev_id, struct pt_regs * regs)
{
	u_char	events;
	volatile scc_t *sccp;
	volatile cpm8260_t *cp;
#ifdef SENNPLD_INTDEBUG
	volatile cbd_t *bdp;
	int i;
#endif

	sccp = &immr->im_scc[0];
	events = sccp->scc_scce;
	if (events & SCCM_TXE) {
		SENNDEBUG("%s: TXError!\n", __FUNCTION__);
#ifdef SENNPLD_INTDEBUG
		bdp = tx_bd_base;
		for (i = 0; i < TX_NUM_FIFO; i++) {
			if (bdp->cbd_sc & (BD_SC_CT | BD_SC_UN))
				printk("BD_SC[%d] %#x\n", i, bdp->cbd_sc);
			bdp++;
		}
#endif
		/* the 8260 handbook syas that a TX reset is required */
		cp = &immr->im_cpm;
		cp->cp_cpcr = mk_cr_cmd(CPM_CR_SCC1_PAGE, CPM_CR_SCC1_SBLOCK, 0,
				CPM_CR_RESTART_TX) | CPM_CR_FLG;
		while (cp->cp_cpcr & CPM_CR_FLG);
	}

	if (events & SCCM_BSY) {
		SENNDEBUG("%s: RX BSY!\n", __FUNCTION__);
		/* the 8260 handbook syas that an ENTER HUNT MODE is required */
		cp = &immr->im_cpm;
		cp->cp_cpcr = mk_cr_cmd(CPM_CR_SCC1_PAGE, CPM_CR_SCC1_SBLOCK, 0,
				CPM_CR_HUNT_MODE) | CPM_CR_FLG;
		while (cp->cp_cpcr & CPM_CR_FLG);
		receive_chars();
	}

	if (events & SCCM_RX)
		receive_chars();
	if (events & SCCM_TX) {
		XSENNDEBUG("%s: TX\n", __FUNCTION__);
		sennpld_trywrite(2);
	}
	sccp->scc_scce = events;
}

int __init
sennpld_init(void)
{
	uint		mem_addr, dp_addr;
	int			i, j;
	volatile	cbd_t		*bdp;
	volatile	cpm8260_t	*cp;
	volatile	scc_t		*scp;
	volatile	scc_trans_t	*stp;
	volatile	immap_t		*immap;
	volatile	iop8260_t	*io;
	uint		iobits;


	lostchars = 0;

	immap = immr;
	cp = &immap->im_cpm;
	io = &immap->im_ioport;

	/*
	 * Configure SCC1
	 * 	Port D, 31 - SCC1 RxD
	 * 	Port D, 30 - SCC1 TxD
	 * 	Port D, 29 - SCC1 RTS
	 */
	io->iop_ppard |=  0x00000007;
	io->iop_pdird |=  0x00000006;
	io->iop_psord |=  0x00000002;
	io->iop_pdird &= ~0x00000001;
	io->iop_psord &= ~0x00000005;
#ifdef SENNPLD_TRIG
	/* make PD7 an output */
	io->iop_psord &= ~0x01000000;
	io->iop_pdird |= 0x01000000;
	/* make PA22 an output */
	io->iop_pdira |= 0x200;
	io->iop_psora &= ~0x200;
#endif

	/*
	 * XXX - CTS?!?
	 * CD (PC14), CLK3 (PC29) and CLK4 (PC28).
	 * Note the SCC1:CTS is being used for SMC2:SMTXD. Hopefully CTS will
	 * be pulled to ground internally.
	 */
#ifdef ONLY_CLK4
	/* TX and RX clocks are both on CLK4 */
	iobits = 0x00020008;	/* PC14, PC28 */
#else
	iobits = 0x0002000c;	/* PC14, PC28, PC29 */
#endif
	io->iop_pparc |= iobits;
	io->iop_pdirc &= ~iobits;
	io->iop_psorc &= ~iobits;

	immap->im_cpmux.cmx_scr &= ~0xff000000;
#ifdef ONLY_CLK4
	/* Connect SCC1 to NMSI.  Connect CLK4 to SCC1. */
	immap->im_cpmux.cmx_scr |= 0x3f000000;
#else
	/* Connect SCC1 to NMSI.  Connect CLK3 and CLK4 to SCC1. */
	immap->im_cpmux.cmx_scr |= 0x37000000;
#endif

	/*
	 * We need to allocate transmit and receive buffer
	 * descriptors from dual port ram, and a character
	 * buffer area from host mem.
	 */
	dp_addr = m8260_cpm_dpalloc(sizeof(cbd_t) * RX_NUM_FIFO, 8);

	if (dp_addr == CPM_DP_NOSPACE) {
		printk("%s: cannot allocate RX cbd_t\n", __FUNCTION__);
		return -ENOMEM;
	}

	/* Allocate space for FIFOs in the host memory.
	*/
	mem_addr = m8260_cpm_hostalloc(RX_NUM_FIFO * RX_BUF_SIZE, 1);

	if (mem_addr == 0) {
		printk("%s: cannot allocate RX buffers\n", __FUNCTION__);
		return -ENOMEM;
	}

	memset((void *)mem_addr, 0, RX_BUF_SIZE);

	/*
	 * Set the physical address of the host memory
	 * buffers in the buffer descriptors, and the
	 * virtual address for us to work with.
	 */
	bdp = (cbd_t *)&immap->im_dprambase[dp_addr];
	rx_cur = rx_bd_base = (cbd_t *)bdp;

	for (j=0; j<(RX_NUM_FIFO-1); j++) {
		bdp->cbd_bufaddr = __pa(mem_addr);
		bdp->cbd_sc = BD_SC_EMPTY | BD_SC_INTRPT;
#ifdef SENNPLD_DEBUG
		memset((void *)mem_addr, 'a' + j, RX_BUF_SIZE);
#endif
		mem_addr += RX_BUF_SIZE;
		bdp++;
	}
	bdp->cbd_bufaddr = __pa(mem_addr);
	bdp->cbd_sc = BD_SC_WRAP | BD_SC_EMPTY | BD_SC_INTRPT;

	scp = &immap->im_scc[0];
	stp = (scc_trans_t *)&immap->im_dprambase[PROFF_SCC1];
	scp->scc_gsmrl = 0;
	stp->st_genscc.scc_rbase = dp_addr;

	dp_addr = m8260_cpm_dpalloc(sizeof(cbd_t) * TX_NUM_FIFO, 8);

	if (dp_addr == CPM_DP_NOSPACE) {
		printk("%s: cannot allocate TX cbd_t\n", __FUNCTION__);
		return -ENOMEM;
	}

	/*
	 * Allocate space for FIFOs in the host memory.
	*/
	mem_addr = m8260_cpm_hostalloc(TX_NUM_FIFO * TX_BUF_SIZE, 1);

	if (mem_addr == 0) {
		printk("%s: cannot allocate TX buffers\n", __FUNCTION__);
		return -ENOMEM;
	}

	memset((void *)mem_addr, 0, TX_BUF_SIZE);

	/* allocate space for rxbuf ring buffer */
	rxbuf = (unsigned char *)kmalloc(RXBUF_SIZE, GFP_KERNEL);
	if (rxbuf == NULL) {
		printk("%s: cannot allocate %d bytes for RX ring buffer\n",
			__FUNCTION__, RXBUF_SIZE);
		return -ENOMEM;
	}
	rxend = rxbuf + RXBUF_SIZE;
	rxhead = rxtail = rxbuf;

	/*
	 * Set the physical address of the host memory
	 * buffers in the buffer descriptors, and the
	 * virtual address for us to work with.
	 */
	bdp = (cbd_t *)&immap->im_dprambase[dp_addr];
	tx_cur = tx_bd_base = (cbd_t *)bdp;

	for (j=0; j<(TX_NUM_FIFO-1); j++) {
		bdp->cbd_bufaddr = __pa(mem_addr);
		bdp->cbd_sc = BD_SC_INTRPT;
		mem_addr += TX_BUF_SIZE;
		bdp++;
	}
	bdp->cbd_bufaddr = __pa(mem_addr);
	bdp->cbd_sc = (BD_SC_WRAP | BD_SC_INTRPT);

	stp->st_genscc.scc_tbase = dp_addr;

	/*
	 * Set up the transparent mode parameters in the
	 * parameter ram.
	 */
	stp->st_genscc.scc_rfcr = CPMFCR_GBL | CPMFCR_EB;
	stp->st_genscc.scc_tfcr = CPMFCR_GBL | CPMFCR_EB;

	stp->st_genscc.scc_mrblr = RX_BUF_SIZE;
	stp->st_cpres = 0xffff;
	stp->st_cmask = 0xf0b8;

	/*
	 * Send the CPM an initialize command.
	*/
	cp->cp_cpcr = mk_cr_cmd(CPM_CR_SCC1_PAGE, CPM_CR_SCC1_SBLOCK, 0,
			CPM_CR_INIT_TRX) | CPM_CR_FLG;
	while (cp->cp_cpcr & CPM_CR_FLG);

	/*
	 * Put SCC1 into bi-directional, transparent mode. The MBP is written
	 * first (REVD). Note that 16-bit CRC is enabled but not sent.
	 */
	scp->scc_gsmrh = (SCC_GSMRH_REVD|SCC_GSMRH_TRX|SCC_GSMRH_TTX|SCC_GSMRH_CDS|
		SCC_GSMRH_CTSS);

	/*
	 * Disable all interrupts and clear all pending
	 * events.
	 */
	scp->scc_sccm = 0;
	scp->scc_scce = 0xffff;
	/* do not care about DSR since we use CD/CTS as SYNC */

	/*
	 * Install interrupt handler.
	*/
	i = request_irq(SIU_INT_SCC1, sennpld_interrupt, 0, "sennpld", NULL);
	if (i) {
		printk("Unable to get irq %d for the SENNPLD!\n", SIU_INT_SCC1);
		return -EIO;
	}

	i = register_chrdev(SENNPLD_MAJ, "sennpld", &sennpld_fops);
	if (i < 0) {
		printk("sennpld: Couldn't register device (major=%d)\n", SENNPLD_MAJ);
		return i;
	}

	/* initialize wait queue head and sk_buff queue heads */
	init_waitqueue_head(&sennpld_wait);
	skb_queue_head_init(&sennpld_skbq);

	printk("Sennheiser PLD driver vers. %s initialized\n", SENNPLD_VER);
	/* turn on RX/TX */
	scp->scc_sccm = (SCCM_TXE | SCCM_BSY | SCCM_TX | SCCM_RX);
	scp->scc_gsmrl = (SCC_GSMRL_ENR | SCC_GSMRL_ENT);
	return 0;
}

module_init(sennpld_init);
