/*
 * Generic SRAM driver.
 * Copyright (C) 2004-2006 DENX Software Engineering, Wolfgang Denk, wd@denx.de
 */
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/vmalloc.h>
#include <linux/errno.h>

#include <asm/uaccess.h>
#include <asm/io.h>
#ifdef CONFIG_CPC45
#include <platforms/cpc45.h>
#endif
#ifdef CONFIG_HMI1001
#include <platforms/hmi1001.h>
#endif
#ifdef CONFIG_INKA4X0
#include <platforms/inka4x0.h>
#endif
#ifdef CONFIG_NC650
#include <platforms/nc650.h>
#endif
#ifdef CONFIG_SMMACO4
#include <platforms/tqm5200.h>
#endif
#ifdef CONFIG_TQM5200
#include <platforms/tqm5200.h>
#endif
#ifdef CONFIG_UC101
#include <platforms/uc101.h>
#endif

#if 0	/* Debugging turned off */
# define debugk(fmt,args...)	printk(fmt ,##args)
#else
# define debugk(fmt,args...)
#endif

#define SRAM_MAJOR	32

struct sram_region {
	ulong start_virt;
	ulong start_phys;
	ulong size;
};

/* Default SRAM region list, for backward compatibility
 * with platforms using one region. */
#ifndef SRAM_REGIONS
#define SRAM_REGIONS			\
{					\
	.start_phys = CFG_SRAM_PADDR,	\
	.size = CFG_SRAM_SIZE		\
}
#endif

/* Main region list, driver attempts to ioremap those regions and
 * each successfully mapped region is assigned a minor numer.
 * Thus first region on the list is represented as a character
 * device with (32,0) major and minor numbers, etc. */
static struct sram_region region_table[] = {
	SRAM_REGIONS
};

#define REGION_NUM (sizeof(region_table)/sizeof(region_table[0]))

static ssize_t sram_read(struct file *file, char *buf,
			size_t count, loff_t *ppos)
{
	unsigned long p = *ppos;
	kdev_t rdev = file->f_dentry->d_inode->i_rdev;
	int reg_num = MINOR(rdev);

	debugk("SRAM_DRV: sram_read: rdev major = %d, rdev minor = %d\n",
			MAJOR(rdev), MINOR(rdev));

	if (p >= region_table[reg_num].size)
		return 0;

	if (count > region_table[reg_num].size - p)
		count = region_table[reg_num].size - p;

	p += region_table[reg_num].start_virt;

	if (copy_to_user(buf, (void*) p, count))
		return -EFAULT;

	*ppos += count;

	return count;
}

static ssize_t sram_write(struct file *file, const char *buf,
			 size_t count, loff_t *ppos)
{
	unsigned long p = *ppos;
	kdev_t rdev = file->f_dentry->d_inode->i_rdev;
	int reg_num = MINOR(rdev);

	debugk("SRAM_DRV: sram_write: rdev major = %d, rdev minor = %d\n",
			MAJOR(rdev), MINOR(rdev));

	if (p >= region_table[reg_num].size)
		return -ENOSPC;

	if (count > region_table[reg_num].size - p)
		count = region_table[reg_num].size - p;

	p += region_table[reg_num].start_virt;

	if (copy_from_user((void*) p, buf, count))
		return -EFAULT;

	*ppos += count;

	return count;
}

static int sram_mmap(struct file *file, struct vm_area_struct *vma)
{
	unsigned long offset = vma->vm_pgoff << PAGE_SHIFT;
	unsigned long size = vma->vm_end - vma->vm_start;
	kdev_t rdev = file->f_dentry->d_inode->i_rdev;
	int reg_num = MINOR(rdev);

	debugk("SRAM_DRV: sram_mmap: rdev major = %d, rdev minor = %d\n",
			MAJOR(rdev), MINOR(rdev));

	if (offset + size > region_table[reg_num].size)
		return -EINVAL;

	offset += region_table[reg_num].start_phys;

	if (remap_page_range(vma->vm_start, offset, size, vma->vm_page_prot))
		return -EAGAIN;

	return 0;
}

static loff_t sram_lseek(struct file *file, loff_t offset, int orig)
{
	kdev_t rdev = file->f_dentry->d_inode->i_rdev;
	int reg_num = MINOR(rdev);

	debugk("SRAM_DRV: sram_lseek: rdev major = %d, rdev minor = %d\n",
			MAJOR(rdev), MINOR(rdev));

	switch (orig) {
		case 0:
			file->f_pos = offset;
			return file->f_pos;
		case 1:
			file->f_pos += offset;
			return file->f_pos;
		case 2:
			file->f_pos = 
				region_table[reg_num].size - offset;
			return file->f_pos;
		default:
			return -EINVAL;
	}
}

static int sram_open(struct inode *inode, struct file *file)
{
	kdev_t rdev = file->f_dentry->d_inode->i_rdev;
	int reg_num = MINOR(rdev);

	debugk("SRAM_DRV: sram_open: rdev major = %d, rdev minor = %d\n",
			MAJOR(rdev), MINOR(rdev));

	if (reg_num >= REGION_NUM)
		return -ENODEV;

	if (region_table[reg_num].start_virt == 0)
		return -ENODEV;

	return capable(CAP_SYS_RAWIO) ? 0 : -EPERM;
}

static struct file_operations sram_fops = {
	llseek:		sram_lseek,
	read:		sram_read,
	write:		sram_write,
	mmap:		sram_mmap,
	open:		sram_open,
};

int __init sram_drv_init(void)
{
	int i, reg_cnt;

	debugk("SRAM_DRV: regions to initialize = %d\n", REGION_NUM);

	reg_cnt = 0;
	for (i = 0; i < REGION_NUM; i++) {
		debugk("SRAM_DRV: region %d, phys = %08lx, size = %08lx\n",
				i, region_table[i].start_phys,
				region_table[i].size);

#ifdef CONFIG_UC101
		region_table[i].start_virt = 
			(ulong)__ioremap(region_table[i].start_phys,
					region_table[i].size, _PAGE_GUARDED | _PAGE_COHERENT);
#else
		region_table[i].start_virt = 
			(ulong)ioremap(region_table[i].start_phys,
					region_table[i].size);
#endif

		if (region_table[i].start_virt == 0) {
			printk(KERN_ERR "SRAM_DRV: ioremap(0x%08lx, 0x%08lx) failed\n",
					region_table[i].start_phys,
					region_table[i].size);
			continue;
		}
		debugk("SRAM_DRV: region %d, virt = 0x%08lx\n",
					i, region_table[i].start_virt);
		reg_cnt++;
	}

	debugk("SRAM_DRV: regions properly initialized = %d\n", REGION_NUM);
	if (reg_cnt == 0)
		return -EIO;

	if (register_chrdev(SRAM_MAJOR, "sram", &sram_fops)) {
		printk(KERN_ERR "SRAM_DRV: register_chrdev failed\n");
		for (i = 0; i < REGION_NUM; i++) {
			if (region_table[i].start_virt)
				iounmap((void *)region_table[i].start_virt);
		}
		return -EIO;
	}

	printk("SRAM_DRV: initialized\n");
	return 0;
}

/******************************
 **** Module Declarations *****
 ******************************/

module_init (sram_drv_init);

#ifdef MODULE

static void sram_drv_cleanup (void)
{
	int i;

	for (i = 0; i < REGION_NUM; i++) {
		if (region_table[i].start_virt)
			iounmap((void *)region_table[i].start_virt);
	}

	if (unregister_chrdev(SRAM_MAJOR, "sram"))
		printk(KERN_ERR "SRAM_DRV: failed to unregister character device\n");
}

module_exit (sram_drv_cleanup);

#endif	/* MODULE */

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Wolfgang Denk <wd@denx.de>");
MODULE_DESCRIPTION("SRAM driver");
