/*
 * (C) Copyright 2005 DENX Software Engineering
 * Wolfgang Denk, wd@denx.de
 * All rights reserved.
 *
 * (C) Copyright 2005 TQ-Systems GbmH
 * Martin Krause, martin.krause@tqs.de
 *
 * Derived from vl_adc12048.c
 *
 * Device dirver for digital IO (digio) control on TQM5200 modules. Actually
 * PWM signalling on GPIO29, GPIO30 and GPIO31 is supported.
 *
 * 06/05:  Created, <martin.krause@tqs.de>
 *
 * $Id:$
 */

#include <linux/kernel.h>	/* We're doing kernel work	*/
#include <linux/module.h>	/* Specifically, a module	*/
#include <asm/uaccess.h>	/* for put_user			*/
#include <linux/init.h>		/* for __initfunc		*/
#include <asm/io.h>		/* for ioremap_nocache		*/
#include <linux/delay.h>
#include <asm/mpc5xxx.h>

#if CONFIG_MODVERSIONS==1
# include <linux/modversions.h>
#endif
#include <linux/fs.h>		/* character device definitions	*/
#include <linux/proc_fs.h>	/* proc definitions */
#include <linux/wrapper.h>	/* wrapper for compatibility with future
				 * versions */
#include <linux/tqm5200_digio.h>

#ifndef KERNEL_VERSION
# define KERNEL_VERSION(a,b,c)	((a)*65536+(b)*256+(c))
#endif

#if (defined(MODULE) && defined(DEBUG))
MODULE_PARM(dbg_level, "i");
MODULE_PARM_DESC(dbg_level, "Debug trace level (0-3)");
#endif

#ifdef	DEBUG

static int dbg_level;

#define dbg(dbg_lvl,fmt,args...)					\
	do {								\
		if (dbg_level >= dbg_lvl)				\
			printk(KERN_DEBUG "" 				\
			__FILE__ ": "__MODULE_STRING(__LINE__) ": "	\
			fmt ,##args);					\
	} while (0)
#else
#define dbg(dbg_lvl,fmt,args...)
#endif /* DEBUG */

/*
 * The name of our device, as it will appear in /proc/devices
 */
#define DEVICE_NAME		"tqm5200_digio"

#ifndef	TQM5200_DIGIO_MAJOR
#define TQM5200_DIGIO_MAJOR	240	/* free for local/experimental use */
#endif	/* TQM5200_DIGIO_MAJOR */

/*
 * Hardware Settings
 */
#define	SM501_MMIO_PHYSBASE	0xE3E00000 /* physical base address of SM501
					    * MMIO space */
#define	SM501_MMIO_SIZE		0x20000    /* size to map in virtual address
					    * space */
#define PWM_CHANNELS		3	   /* number of PWM channels */

/*
 * Pointer to start of SM501 MMIO space (in virtual address space)
 */
static unsigned char *mmio;

/*
 * prevent concurent access of the same device
 */
static int digio_used = 0;

static int digio_major;

/*
 * default configuration, could be overwritten by kernel command line
 */
static int init_pwm_enabled = 1;
static int init_duty_cycle = PWM0_DUTY_CYCLE_DEFAULT;

/*
 * data structure to store the actual driver state
 */
typedef struct
{
	unsigned int state;
	unsigned int duty_cycle;
	unsigned int freq;
} pwm_data_t;

static pwm_data_t pwm_data[PWM_CHANNELS];

/*
 * Prototypes
 */
int __init tqm5200_digio_init (void);
static int tqm5200_digio_open(struct inode *, struct file *);
static int tqm5200_digio_release(struct inode *, struct file *);
static int tqm5200_digio_ioctl(struct inode*, struct file*, unsigned int,
			      unsigned long);
static void pwm_pin_enable(void);
#ifdef MODULE
static void pwm_pin_disable(void);
#endif /* MODULE */
static int pwm_start(unsigned int channel);
static int pwm_stop(unsigned int channel);
static int pwm_set_duty_cycle(struct pwm_setdc *dc);

static unsigned int smi_read32(unsigned int offset);
static void smi_write32(unsigned int offset, unsigned int data);

#ifdef CONFIG_TB5200
static void led_init (void);
static void led_cleanup (void);
static void led_set (unsigned int);
#ifdef CONFIG_PROC_FS
static int led_proc_read (char *buf, char **start, off_t offset,
			  int count, int *eof, void *data);
static int led_proc_write (struct file *file, const char *buffer,
			   unsigned long count, void *data);
#endif /* CONFIG_PROC_FS */
#endif /* CONFIG_TB5200 */

struct file_operations tqm5200_digio_ops =
{
	owner:		THIS_MODULE,
	open:		tqm5200_digio_open,
	release:	tqm5200_digio_release,
	ioctl:		tqm5200_digio_ioctl,
};


/*
 * smi_read32
 *
 * Read the value of the 32-bit register specified by offset from the SM501
 * MMIO space. A big to little endian swap is done, if the SM501 is running in
 * big endian mode.
 *
 * Parameters
 *    [in]
 *        offset - offset of register to read relative to the mmio space
 *
 *    [out]
 *        None
 *
 * Returns
 *    Value of 32-bit register specified by offset
 */
unsigned int smi_read32(unsigned int offset)
{
	if (*(volatile unsigned int *)(mmio + SM501_ENDIAN_CTRL)) {
		/*
		 * big endian mode -> swap bytes
		 */
		unsigned int data1;
		unsigned int data2;
		char *pdata1 = (char *)&data1;
		char *pdata2 = (char *)&data2;

		data1 = *(volatile unsigned int *)(mmio + offset);
		pdata2[0] = pdata1[3];
		pdata2[1] = pdata1[2];
		pdata2[2] = pdata1[1];
		pdata2[3] = pdata1[0];

		return data2;
	} else {
		return *(volatile unsigned int *)(mmio + offset);
	}
}

/*
 * smi_write32
 *
 * Write the value of the 32-bit register specified by offset from the SM501
 * MMIO space. A big to little endian swap is done, if the SM501 is running in
 * big endian mode.
 *
 * Parameters
 *    [in] offset - offset of register to write relative to the mmio
 *         data   - value to write to the specified register
 * space
 *
 *    [out]
 *        None
 *
 * Returns
 *    none
 */
void smi_write32(unsigned int offset, unsigned int data)
{
	if (*(volatile unsigned int *)(mmio + SM501_ENDIAN_CTRL)) {
		/*
		 * big endian mode -> swap bytes
		 */
		unsigned int data1;
		unsigned int data2;
		char *pdata1 = (char *)&data1;
		char *pdata2 = (char *)&data2;

		data1 = data;
		pdata2[0] = pdata1[3];
		pdata2[1] = pdata1[2];
		pdata2[2] = pdata1[1];
		pdata2[3] = pdata1[0];
		*(volatile unsigned int *)(mmio + offset) = data2;
	} else {
		*(volatile unsigned int *)(mmio + offset) = data;
	}
}

/*
 * Enable and configure the pins GPIO29, GPIO30 and GPIO31 for PWM output.
 */
static void pwm_pin_enable(void)
{
	unsigned int data;

	/* enable GPIO, PWM and I2C clock control */
	data = smi_read32 (SM501_POWER_MODE0);
	data |= 0x00000040;
	smi_write32 (SM501_POWER_MODE0, data);
	data = smi_read32 (SM501_POWER_MODE1);
	data |= 0x00000040;
	smi_write32 (SM501_POWER_MODE1, data);

	/* configure GPIO29, GPIO30 and GPIO31 for PWM */
	data = smi_read32 (SM501_GPIO_CONTROL);
	data |= 0xE0000000;
	smi_write32 (SM501_GPIO_CONTROL, data);

	/* configure default PWM values, but don't enable pwm */
	smi_write32 (SM501_PWM0, PWM_REG_DEFAULT);
	pwm_data[0].duty_cycle = 50;
	smi_write32 (SM501_PWM1, PWM_REG_DEFAULT);
	pwm_data[1].duty_cycle = 50;
	smi_write32 (SM501_PWM2, PWM_REG_DEFAULT);
	pwm_data[2].duty_cycle = 50;

	return;
}

/*
 * disable PWM and ocnfigure the pins GPIO29, GPIO30 and GPIO31 as GPIO
 * inputs.
 */
#ifdef MODULE
static void pwm_pin_disable(void)
{
	unsigned int data;

	/* clear PWM registers */
	smi_write32 (SM501_PWM0, 0);
	smi_write32 (SM501_PWM1, 0);
	smi_write32 (SM501_PWM2, 0);

	/* configure GPIO29, GPIO30 and GPIO31 as GPIO */
	data = smi_read32 (SM501_GPIO_CONTROL);
	data &= ~0xE0000000;
	smi_write32 (SM501_GPIO_CONTROL, data);

	/* disable GPIO, PWM and I2C clock control */
	data = smi_read32 (SM501_POWER_MODE0);
	data &= ~0x00000040;
	smi_write32 (SM501_POWER_MODE0, data);

	data = smi_read32 (SM501_POWER_MODE1);
	data &= ~0x00000040;
	smi_write32 (SM501_POWER_MODE1, data);

	return;
}
#endif /* MODULE */

static int pwm_start(unsigned int channel)
{
	unsigned int data;

	if (channel >= PWM_CHANNELS)
		return -ENOTTY;

	/* start PWM on spezified channel */
	data = smi_read32 (SM501_PWM_BASE + (channel * 4));
	data |= 0x00000001;
	smi_write32 (SM501_PWM_BASE + (channel * 4), data);

	pwm_data[channel].state = 1;

	return 0;
}

static int pwm_stop(unsigned int channel)
{
	unsigned int data;

	if (channel >= PWM_CHANNELS)
		return -ENOTTY;

	/* stop PWM on spezified channel */
	data = smi_read32 (SM501_PWM_BASE + (channel * 4));
	data &= ~0x00000001;
	smi_write32 (SM501_PWM_BASE + (channel * 4), data);

	pwm_data[channel].state = 0;

	return 0;
}

static int pwm_set_duty_cycle(struct pwm_setdc *dc)
{
	unsigned int data;
	unsigned int pwm_high, pwm_low;

	if (dc->channel >= PWM_CHANNELS)
		return -ENOTTY;
	if (dc->duty_cycle > 100)
		return -ENOTTY;

	pwm_high = (PWM_CYCLE_CNT * dc->duty_cycle) / 100;
	pwm_low = (PWM_CYCLE_CNT - pwm_high);

	data = smi_read32 (SM501_PWM_BASE + (dc->channel * 4));
	data &=	~0xFFFFFF00;
	data |= (pwm_high << 20);
	data |=	(pwm_low << 8);
	smi_write32 (SM501_PWM_BASE + (dc->channel * 4), data);

	pwm_data[dc->channel].duty_cycle = dc->duty_cycle;

	return 0;
}


#ifdef CONFIG_PROC_FS
int digio_read_procmem(char *buf, char **start, off_t offset,
		      int count, int *eof, void *data)
{
	int i, len = 0;
	int limit = count - 120;

	dbg (2, "/proc: buf %p, *start %p, offset %i, count %i, *eof %d, "
	     "data %p\n", buf, *start, (int)offset, count, *eof, data);

	for (i = 0; i < PWM_CHANNELS && len <= limit; i++)
	{
		len += sprintf (buf + len, "channel %d: %s duty cycle %d\n",
				i, pwm_data[i].state ? "enabled, " : "disabled,",
				pwm_data[i].duty_cycle);
	}
	*eof = 1;
	return len;
}
#endif /* CONFIG_PROC_FS */


#ifdef CONFIG_TB5200
static void led_init (void)
{
	struct mpc5xxx_gpt *gpt = (struct mpc5xxx_gpt *)(MPC5xxx_GPT + 0x40);

	/* configure timer 4 for simple GPIO output */
	gpt->emsr &=  ~0x00000737;
	gpt->emsr |=   0x00000024;
}

static void led_cleanup (void)
{
	struct mpc5xxx_gpt *gpt = (struct mpc5xxx_gpt *)(MPC5xxx_GPT + 0x40);

	/* configure timer 4 for simple GPIO input */
	gpt->emsr &=  ~0x00000737;
}

static void led_set (unsigned int state)
{
	struct mpc5xxx_gpt *gpt = (struct mpc5xxx_gpt *)(MPC5xxx_GPT + 0x40);

	if (state == 0) {
		gpt->emsr &=  ~(1 << 4);
	} else {
		gpt->emsr |=  (1 << 4);
	}
	dbg (2, "&(gpt->emsr): %#x; *(gpt->emsr): %#x\n", &(gpt->emsr),
	     gpt->emsr);
}

static unsigned int led_get (void)
{
	struct mpc5xxx_gpt *gpt = (struct mpc5xxx_gpt *)(MPC5xxx_GPT + 0x40);

	return ((gpt->emsr >> 4) & 0x1);
}

#ifdef CONFIG_PROC_FS
static int led_proc_read (char *buf, char **start, off_t offset,
			 int count, int *eof, void *data)
{
	int len = 0;

	dbg (2, "/proc: buf %p, *start %p, offset %i, count %i, *eof %d, "
	     "data %p\n", buf, *start, (int)offset, count, *eof, data);

	len += sprintf (buf, "%s\n", led_get () ? "on" : "off");

	*eof = 1;
	return len;
}

static int led_proc_write (struct file *file, const char *buffer,
			  unsigned long count, void *data)
{
	char buf[10], val;

	if (count > sizeof(buf))
		return -EINVAL;

	if (copy_from_user(buf, buffer, count)) {
		return -EFAULT;
	}

	val = simple_strtol(buf, NULL, 0) & 0x01;
	dbg (2, "/proc: set led to %d\n", val);
	led_set (val);

	return count;
}
#endif /* CONFIG_PROC_FS */

#endif /* CONFIG_TB5200 */


/*
 * Initialize the driver, register us as a character device
 */
int __init tqm5200_digio_init (void)
{
	struct pwm_setdc dc;
	struct proc_dir_entry *entry;

	/*
	 * Remap I/O space into kernel memory
	 */
	mmio = (unsigned char *)ioremap_nocache (SM501_MMIO_PHYSBASE,
						 SM501_MMIO_SIZE);
	if (mmio == NULL) {
		printk (DEVICE_NAME ": unable to remap io memory\n");
		return -ENODEV;
	}

	digio_major = register_chrdev (TQM5200_DIGIO_MAJOR, DEVICE_NAME,
				      &tqm5200_digio_ops);

	if (digio_major < 0) {
		printk (DEVICE_NAME ": register_chrdev failed with %d\n",
			digio_major);
		iounmap ((void *)mmio);
		return digio_major;
	}

	digio_major = TQM5200_DIGIO_MAJOR;

	/* Enable PWM DIGIO pins */
	pwm_pin_enable ();

	/*
	 * Start PWM on channel 0, if not deactivated via command line. Use
	 * a default duty cycle or the value specified at command line.
	 */
	dc.channel = 0;
	/*
	 * the backlight converter is low active, so inverte pwm logic
	 */
	dc.duty_cycle = 100 - init_duty_cycle;
	pwm_set_duty_cycle (&dc);
	if (init_pwm_enabled)
	    pwm_start (0);

#ifdef CONFIG_TB5200
	led_init ();
#endif /* CONFIG_TB5200 */

#ifdef CONFIG_PROC_FS
	entry = create_proc_entry("driver/pwm", 0, NULL);
	if (entry == NULL) {
		printk("Unable to create /proc/driver/pwm entry\n");
		return -EINVAL;
	}
	entry->read_proc = digio_read_procmem;
	entry->write_proc = NULL;
	entry->data = NULL;
#ifdef CONFIG_TB5200
	entry = create_proc_entry("driver/led", 0, NULL);
	if (entry == NULL) {
		printk("Unable to create /proc/driver/led entry\n");
		return -EINVAL;
	}
	entry->read_proc = led_proc_read;
	entry->write_proc = led_proc_write;
	entry->data = NULL;
#endif /* CONFIG_TB5200 */
#endif /* CONFIG_PROC_FS */

	printk (DEVICE_NAME ": driver initialized, major: %d\n",
		digio_major);
	return 0;
}


/*
 * Called whenever a process attempts to open the device
 */
static int tqm5200_digio_open (struct inode *inode, struct file *file)
{
	int minor  = MINOR(inode->i_rdev);

	dbg (1, "tqm5200_digio_open(%p,%p): minor %d\n", inode, file, minor);

	if (minor != 0)
		return -ENXIO;

	/* exclusive open only */
	if (digio_used) {
		return -EBUSY;
	}

	/* only open for read is allowed */
	if((file->f_flags & O_ACCMODE) != O_RDONLY)
		return -EROFS;

	digio_used++;

	MOD_INC_USE_COUNT;

	return 0;
}


/*
 * Called when a process closes the device.
 */
static int tqm5200_digio_release (struct inode *inode, struct file *file)
{

	dbg (1, "tqm5200_digio_release(%p,%p)\n", inode, file);

	digio_used--;

	MOD_DEC_USE_COUNT;

	return 0;
}


/*
 * no read or write entry point for this driver
 *
 */


/*
 * ioctl entry point.
 */
static int tqm5200_digio_ioctl(struct inode* inode, struct file* file,
		  unsigned int cmd, unsigned long arg )
{
	int ret = 0;
	struct pwm_setdc dc;

	dbg (1, "act.cmd %#x, arg %#lx\n", cmd, arg);

	/* known command type? */
	if (_IOC_TYPE (cmd) != TQM_DIGIO_IOCTL_MGC)
		return -ENOTTY;

	if (_IOC_DIR (cmd) & _IOC_READ)
		ret = !access_ok (VERIFY_WRITE, (void *) arg, _IOC_SIZE (cmd));
	if (_IOC_DIR (cmd) & _IOC_WRITE)
		ret |= !access_ok (VERIFY_READ, (void *) arg, _IOC_SIZE (cmd));
	if (ret)
		return -EFAULT;

	switch (cmd) {
	case DIGIO_CMD_PWM_START:
		ret = pwm_start (arg);
		break;

	case DIGIO_CMD_PWM_STOP:
	        ret = pwm_stop (arg);
		break;

	case DIGIO_CMD_SET_DUTY_CYCLE:
		if (copy_from_user((void *)&dc, (void *)arg, sizeof(dc)))
			return -EFAULT;
		ret = pwm_set_duty_cycle (&dc);
		break;

	case DIGIO_CMD_SET_PWM_FREQ:
		return -ENOTTY;	/* not implemented yet */
		break;

	default:
		return -ENOTTY;
		break;
	}

	return ret;
}

module_init(tqm5200_digio_init);

int __init tqm5200_digio_setup (char *options)
{
	while (options && *options) {
		if (strncmp (options, "pwmoff", 6) == 0) {
			options += 6;
			init_pwm_enabled = 0;
			if (*options != ',')
				return 0;
			options++;
		}

		if (strncmp (options, "pwm:", 4) == 0) {
			options += 4;
			if (!*options)
				return 0;

			init_duty_cycle = simple_strtoul (options, &options, 0);
			if (*options != ',')
				return 0;
			options++;
		}
	}

	return 0;
}

__setup ("digio=", tqm5200_digio_setup);


#ifdef MODULE
void tqm5200_digio_exit (void)
{
	int ret;

	pwm_pin_disable ();

	iounmap ((void *)mmio);

#ifdef CONFIG_TB5200
	led_cleanup ();
#endif /* CONFIG_TB5200 */

#ifdef CONFIG_PROC_FS
	remove_proc_entry("driver/pwm", NULL);
#ifdef CONFIG_TB5200
	remove_proc_entry("driver/led", NULL);
#endif /* CONFIG_TB5200 */
#endif /* CONFIG_PROC_FS */
	ret = unregister_chrdev (digio_major, DEVICE_NAME);
	if (ret < 0) {
		printk ("unregister_chrdev: error %d\n", ret);
	}
}

module_exit(tqm5200_digio_exit);

MODULE_AUTHOR("TQ-Systems GmbH");
MODULE_DESCRIPTION("digio driver for TQM5200 based boards");
MODULE_LICENSE("GPL");
EXPORT_NO_SYMBOLS;
#endif	/* MODULE */
