/*
 * Driver for UC101 GPIO.
 * (C) Copyright 2006 Heiko Schocher, DENX <hs@denx.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/version.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/poll.h>

#include <linux/proc_fs.h>

#include <linux/miscdevice.h>
#include <asm/mpc5xxx.h>

#define UC101_GPIO_VERSION "1.0"
#define UC101_GPIO_NAME "uc101_gpio"

#define UC101_GPIO_MINOR	250

/* Outputs: */
/* simple GPIO Interrupt as Output */
#define GPIO_PSC3_8	2	/* LED H2 */
#define GPIO_USB1_9	3	/* LED H3 */

/* Inputs: */
/* simple GPIO */
#define GPIO_PSC3_7	(31-18)		/* Switch restart */
#define GPIO_PSC3_6	(31-19)		/* State Warm/Coldstart */
#define GPIO_PSC3_1	(31-22)		/* Switch Menue B */
#define GPIO_PSC3_0	(31-23)		/* Switch Menue A */

#define GPIO_USB_8	(31-12)		/* Bit 3 */
#define GPIO_USB_7	(31-13)		/* Bit 2 */
#define GPIO_USB_6	(31-14)		/* Bit 1 */
#define GPIO_USB_0	(31-15)		/* Bit 0 */

/* Timer GPIO configurations */
#define GPT_OUT_0	0x00000027	/* Simple GPIO output driving 0 */
#define	GPT_OUT_1	0x00000037	/* Simple GPIO output driving 1 */
#define	GPT_DISABLE	0x00000000	/* GPT pin disabled */

#define GPT_OUTPUT	1		/* driven as an output */
#define GPT_INPUT	2		/* driven as an input */

/* proc interface for driver */
#define UC101_GPIO_PROC	"driver/uc101_gpio"

/* Froward declarations */
static inline void	uc101_gpt_disable(int pin);
static inline void	uc101_gpt_set(int pin, int data);
static inline void	uc101_gpt_set_ip(int pin);
static inline int	uc101_gpt_get(int pin, int dir);
static inline void	uc101_simp_enable(int pin, int data, int dir);
static inline void	uc101_simp_disable(int n);
static inline void	uc101_simp_set(int pin, int data);
static inline int	uc101_simp_get(int pin, int dir);
static inline void	uc101_simpir_enable(int pin, int data, int dir);
static inline void	uc101_simpir_disable(int n);
static inline void	uc101_simpir_set(int pin, int data);
static inline int	uc101_simpir_get(int pin, int dir);

/* Driver's banner */
static char uc101_gpio_banner[] __initdata = KERN_INFO UC101_GPIO_NAME ": "
						UC101_GPIO_VERSION "\n";

/* Disable and reset Timer GPIO */
static inline void uc101_gpt_disable(int n)
{
	volatile struct mpc5xxx_gpt *gpt =
		(volatile struct mpc5xxx_gpt *)MPC5xxx_GPT + n;
	gpt->emsr = GPT_DISABLE;
}

/* Configure Timer GPIO as a Simple GPIO output and set output value */
static inline void uc101_gpt_set(int n, int data)
{
	volatile struct mpc5xxx_gpt *gpt =
		(volatile struct mpc5xxx_gpt *)MPC5xxx_GPT + n;
	gpt->emsr = data ? GPT_OUT_1 : GPT_OUT_0;
}

static inline void uc101_gpt_set_ip(int n)
{
	volatile struct mpc5xxx_gpt *gpt =
		(volatile struct mpc5xxx_gpt *)MPC5xxx_GPT + n;
	gpt->emsr = 0;
}

/* Get latched Timer GPIO output value */
static inline int uc101_gpt_get(int n, int dir)
{
	volatile struct mpc5xxx_gpt *gpt =
		(volatile struct mpc5xxx_gpt *)MPC5xxx_GPT + n;
	if (dir == GPT_OUTPUT) {
		return (gpt->emsr & 0x10) >> 4;
	} else {
		return (gpt->sr & 0x100) >> 8;
	}
}

/* Enable Simple Interrupt GPIO */
static inline void uc101_simpir_enable (int n, int data, int dir)
{
	volatile struct mpc5xxx_gpio *pgpio = 
		(struct mpc5xxx_gpio *)MPC5xxx_GPIO;

	if (dir == GPT_OUTPUT) {
		pgpio->sint_ode &= ~(1 << n);
		/* Set initial value */
		uc101_simpir_set(n, data);
		/* Set GPIO direction as an output */
		pgpio->sint_ddr |= (1 << n);
	} else {
		/* Set GPIO direction as an input */
		pgpio->sint_ddr &= ~(1 << n);
	}
	/* Enable Simple GPIO pin */
	pgpio->sint_gpioe |= (1 << n);
}

/* Disable and reset Simple Interrupt GPIO */
static inline void uc101_simpir_disable(int n)
{
	volatile struct mpc5xxx_gpio *pgpio = 
		(struct mpc5xxx_gpio *)MPC5xxx_GPIO;
	/* Disable Simple GPIO pin */
	pgpio->sint_gpioe &= ~(1 << n);
}

/* Set Simple Interrupt GPIO output value */
static inline void uc101_simpir_set(int n, int data)
{
	volatile struct mpc5xxx_gpio *pgpio =
		(struct mpc5xxx_gpio *)MPC5xxx_GPIO;
	pgpio->sint_dvo = data ? pgpio->sint_dvo | (1 << n) : 
					pgpio->sint_dvo & ~(1 << n);
}

/* Get latched Simple Interrupt GPIO output value */
static inline int uc101_simpir_get(int n, int dir)
{
	volatile struct mpc5xxx_gpio *pgpio =
		(struct mpc5xxx_gpio *)MPC5xxx_GPIO;
	return ((pgpio->sint_ival) >> (n)) & 1;
}
/* Enable Simple GPIO */
static inline void uc101_simp_enable (int n, int data, int dir)
{
	volatile struct mpc5xxx_gpio *pgpio = 
		(struct mpc5xxx_gpio *)MPC5xxx_GPIO;
	if (dir == GPT_OUTPUT) {
		/* Set initial value */
		pgpio->simple_dvo |= data << n;
		/* Set GPIO direction as an output */
		pgpio->simple_ddr |= 1 << n;
	} else {
		/* Set GPIO direction as an input */
		pgpio->simple_ddr &= ~(1 << n);
	}
	/* Enable Simple GPIO pin */
	pgpio->simple_gpioe |= 1 << n;
}

/* Disable and reset Simple GPIO */
static inline void uc101_simp_disable(int n)
{
	volatile struct mpc5xxx_gpio *pgpio = 
		(struct mpc5xxx_gpio *)MPC5xxx_GPIO;
	/* Disable Simple GPIO pin */
	pgpio->simple_gpioe &= ~(1 << n);
	/* Reset GPIO value */
	pgpio->simple_dvo &= ~(1 << n);
	/* Reset GPIO direction to default */
	pgpio->simple_ddr &= ~(1 << n);
}

/* Set Simple GPIO output value */
static inline void uc101_simp_set(int n, int data)
{
	volatile struct mpc5xxx_gpio *pgpio =
		(struct mpc5xxx_gpio *)MPC5xxx_GPIO;
	pgpio->simple_dvo = data ? pgpio->simple_dvo | (1 << n) : 
					pgpio->simple_dvo & ~(1 << n);
}

/* Get latched Simple GPIO output value */
static inline int uc101_simp_get(int n, int dir)
{
	volatile struct mpc5xxx_gpio *pgpio =
		(struct mpc5xxx_gpio *)MPC5xxx_GPIO;
	if (dir == GPT_OUTPUT) {
		return ((pgpio->simple_dvo) >> (n)) & 1;
	} else {
		return ((pgpio->simple_ival) >> (n)) & 1;
	}
}

static int uc101_get_state (unsigned long *state)
{
	*state = (uc101_simp_get(GPIO_PSC3_7, GPT_INPUT) << 15) |
		 (uc101_simp_get(GPIO_PSC3_6, GPT_INPUT) << 14) |
		 (uc101_simp_get(GPIO_PSC3_1, GPT_INPUT) << 13) |
		 (uc101_simp_get(GPIO_PSC3_0, GPT_INPUT) << 12) |
		 (uc101_simp_get(GPIO_USB_8, GPT_INPUT) << 11) |
		 (uc101_simp_get(GPIO_USB_7, GPT_INPUT) << 10) |
		 (uc101_simp_get(GPIO_USB_6, GPT_INPUT) << 9) |
		 (uc101_simp_get(GPIO_USB_0, GPT_INPUT) << 8) |
		 (uc101_simpir_get(GPIO_USB1_9, GPT_OUTPUT) << 7) | /* H3 */
		 (uc101_simpir_get(GPIO_PSC3_8, GPT_OUTPUT) << 6) | /* H2 */
		 (uc101_gpt_get(5, GPT_OUTPUT) << 5) |	/* H5 */
		 (uc101_gpt_get(4, GPT_OUTPUT) << 4) |	/* H4 */
		 (uc101_gpt_get(3, GPT_OUTPUT) << 3) |	/* HB */
		 (uc101_gpt_get(6, GPT_OUTPUT) << 2) |	/* Loc. Pow. enable */
		 (uc101_gpt_get(1, GPT_OUTPUT) << 1) |	/* reset Coldstart */
		 (uc101_gpt_get(0, GPT_OUTPUT));	/* WDT */
	return 0;
}

static int uc101_write_state (unsigned char state)
{
	uc101_simpir_set(GPIO_USB1_9, state & 0x80); /* LED H3 */
	uc101_simpir_set(GPIO_PSC3_8, state & 0x40); /* LED H2 */
	uc101_gpt_set(5, state & 0x20);	/* LED H5 */
	uc101_gpt_set(4, state & 0x10);	/* LED H4 */
	uc101_gpt_set(3, state & 0x08);	/* LED Heartbeat */
#if 0
	/* only activate, if the driver is OK */
	uc101_gpt_set(6, state & 0x04);	/* Interbus */
	uc101_gpt_set(1, state & 0x02);	/* reset Coldstart */
	uc101_gpt_set(0, state & 0x01);	/* trigger WDT */
#endif

	return 0;
}

/*
 * Device file operations
 */
static ssize_t uc101_gpio_write(struct file *file, const char *buf, size_t bufsz, loff_t *off)
{
	unsigned char state;

	if (copy_from_user((void *) &state, buf, 1))
		return -EFAULT;
	uc101_write_state (state);

	return sizeof (state);
}

static ssize_t uc101_gpio_read(struct file *file, char *buf, size_t bufsz, loff_t *off)
{
	unsigned long state = 0;

	uc101_get_state (&state);
	if (copy_to_user(buf, (void *) &state, sizeof(unsigned long)))
		return -EFAULT;

	return sizeof (state);
}

static unsigned int uc101_gpio_poll(struct file *file, struct poll_table_struct *ptbl)
{
	/* It is always possible to read pin states */
	return POLLIN | POLLRDNORM;
}

#ifdef MODULE
static int uc101_gpio_open(struct inode *inode, struct file *file)
{
	MOD_INC_USE_COUNT;
	return 0;
}

static int uc101_gpio_close(struct inode *inode, struct file *file)
{
	MOD_DEC_USE_COUNT;
	return 0;
}
#endif

static struct file_operations uc101_gpio_fops = {
	.owner = THIS_MODULE,
	.read	= uc101_gpio_read,
	.write	= uc101_gpio_write,
	.poll	= uc101_gpio_poll,
#ifdef MODULE
	.open	= uc101_gpio_open,
	.release = uc101_gpio_close
#endif
};

static struct miscdevice uc101_gpio_miscdev = {
	UC101_GPIO_MINOR,
	UC101_GPIO_NAME,
	&uc101_gpio_fops
};

/*
 * procfs operations
 */
#ifdef CONFIG_PROC_FS

static int uc101_gpio_proc_read(char *page, char **start, off_t off,
					int count, int *eof, void *data)
{
	unsigned long	state = 0;
	char *p = page;

	uc101_get_state (&state);
	p += sprintf (p, "%x", (unsigned int)state);
	return p - page;
}

static int uc101_gpio_proc_write(struct file *file, const char *buffer,
	unsigned long count, void *data)
{
	unsigned char	state;

	if (count == 5) {
		unsigned char buf[10];
		unsigned char *p = (unsigned char *)buf;
		state = 0;
		if (count > 10) count = 10;
		if ( copy_from_user(buf, buffer, count) )
			return -EFAULT;
		if ((*p == '0') && (p[1] == 'x')) {
			p+=2;
			while (1) {
				if (*p >= '0' && *p <= '9') {
					state *= 16;
					state += *p - '0';
				} else if (*p >= 'a' && *p <= 'f') {
					state *= 16;
					state += *p - 'a' + 10;
				} else if (*p >= 'A' && *p <= 'F') {
					state *= 16;
					state += *p - 'A' + 10;
				} else {
					break;
				}
				p++;
			}
		} else {
			return -EFAULT;
		}
	} else {
		return -EFAULT;
	}
	uc101_write_state (state);

	return 5;
}
#endif

/*
 * Initialization/cleanup
 */
static int __devinit uc101_gpio_init(void)
{
	printk(uc101_gpio_banner);

#ifdef CONFIG_PROC_FS
	struct proc_dir_entry * proc;
#endif

	if (misc_register(&uc101_gpio_miscdev) < 0) {
		printk(KERN_ERR "UC101 GPIO: can't misc_register on minor=%d\n",
			UC101_GPIO_MINOR);
		goto error;
	}

#ifdef CONFIG_PROC_FS
	if (!(proc = create_proc_entry(UC101_GPIO_PROC, S_IFREG|S_IRUGO, NULL))) {
		printk(KERN_ERR "failed to create /proc/" UC101_GPIO_PROC "\n");
		goto error_unreg;
	}

	proc->read_proc	 = uc101_gpio_proc_read;
	proc->write_proc = uc101_gpio_proc_write;
#endif

	/* Inputs */
	uc101_simp_enable (GPIO_USB_8, 0, GPT_INPUT);
	uc101_simp_enable (GPIO_USB_7, 0, GPT_INPUT);
	uc101_simp_enable (GPIO_USB_6, 0, GPT_INPUT);
	uc101_simp_enable (GPIO_USB_0, 0, GPT_INPUT);
	uc101_simp_enable (GPIO_PSC3_6, 0, GPT_INPUT);
	uc101_simp_enable (GPIO_PSC3_7, 0, GPT_INPUT);
	uc101_simp_enable (GPIO_PSC3_1, 0, GPT_INPUT);
	uc101_simp_enable (GPIO_PSC3_0, 0, GPT_INPUT);

	/* Timer GPIO */
	uc101_simpir_enable (GPIO_PSC3_8, 0, GPT_OUTPUT);
	uc101_simpir_enable (GPIO_USB1_9, 0, GPT_OUTPUT);
	uc101_gpt_set(3, 0);	/* LED Heartbeat */
	uc101_gpt_set(4, 0);	/* LED H2 */
	uc101_gpt_set(5, 0);	/* LED H3 */

	return 0;

error_unreg:
	misc_deregister(&uc101_gpio_miscdev);
error:
	return -ENODEV;
}

static void __devexit uc101_gpio_cleanup(void)
{
	uc101_simp_disable (GPIO_USB_0);
	uc101_simp_disable (GPIO_USB_6);
	uc101_simp_disable (GPIO_USB_7);
	uc101_simp_disable (GPIO_USB_8);
	uc101_simp_disable (GPIO_PSC3_7);
	uc101_simp_disable (GPIO_PSC3_8);

	uc101_gpt_disable (0);
	uc101_gpt_disable (1);
	uc101_gpt_disable (3);
	uc101_gpt_disable (4);
	uc101_gpt_disable (5);
	uc101_gpt_disable (6);

#ifdef CONFIG_PROC_FS
	remove_proc_entry(UC101_GPIO_PROC, NULL);
#endif

	misc_deregister(&uc101_gpio_miscdev);
}

module_init(uc101_gpio_init)
module_exit(uc101_gpio_cleanup)

MODULE_AUTHOR("Heiko Schocher <hs@denx.de>");
MODULE_DESCRIPTION("UC101 GPIO Driver");
MODULE_LICENSE("GPL");
EXPORT_NO_SYMBOLS;
