/*
 * This driver provides access to the VMEbus using the Tundra Universe.
 *
 * (C) Copyright 2005 Wolfgang Denk, wd@denx.de
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 */

/*
 * Standard in kernel modules
 */
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/version.h>
#include <linux/mm.h>			/* for mem_map		*/
#include <linux/wrapper.h>		/* for mem_map_reserve	*/
#include <linux/pci.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>
#include <asm/types.h>
#include <asm/uaccess.h>
#include <asm/io.h>

#include <asm/universe.h>

#ifndef KERNEL_VERSION
# define KERNEL_VERSION(a,b,c)	(((a) << 16) + ((b) << 8) + (c))
#endif

#if (!defined(LINUX_VERSION_CODE) || LINUX_VERSION_CODE < KERNEL_VERSION(2,2,0))
#error "You need to use at least Linux kernel version 2.2.0"
#endif


/***************************************************************************
 * Global stuff
 ***************************************************************************
 */
#undef	DEBUG

/*
 * Deal with CONFIG_MODVERSIONS
 */
#if CONFIG_MODVERSIONS==1
/* # define MODVERSIONS */
# include <linux/modversions.h>
#endif

/*
 * For character devices
 */
#include <linux/fs.h>		/* character device definitions	*/
#include <linux/wrapper.h>	/* wrapper for compatibility with future versions */

#define	VME_MAJOR	221	/* reserved for VME -- see http://www.lanana.org/ */

#define MAX_PCI_MAPS	64	/* max available different pci spaces for mapping */

/*
 * Device Declarations
 */

/*
 * The name for our device, as it will appear in /proc/devices
 */
#define DEVICE_NAME	"universe"
#define	DRV_VERSION	"1.1"


/***************************************************************************
 * Local stuff
 ***************************************************************************
 */

typedef struct _UNI_DEV UNI_DEV;

struct _UNI_DEV {
	struct pci_dev *pdev;
	u32            addr;
	UNIVERSE       *uregs;
	spinlock_t     lock;
};

static UNI_DEV   *dev = NULL;

static u32 universe_upper_memspc;
static struct universe_img universe_pci_maps[MAX_PCI_MAPS];
static int universe_last_map_index;

static int Major;

#ifdef	DEBUG
# define debugk(fmt,args...)	printk(fmt ,##args)
# define DEBUG_OPTARG(arg)	arg,
#else
# define debugk(fmt,args...)
# define DEBUG_OPTARG(arg)
#endif


/*
 * The following array is indexed using the device minor number.
 *
 * A "name" value of NULL indicates unused entries.
 */
#define	VME_NOT_USED	{ NULL, 0, 0, 0, 0, 0, NULL, }

static vme_t vme_memory[] = {
	/* Minor Number 0 */
	{ "vme_pti0",	0, 0, 0, 0, NULL,	},

	/* Minor Number 1 */
	{ "vme_pti1",	0, 0, 0, 0, NULL,	},

	/* Minor Number 2 */
	{ "vme_pti2",	0, 0, 0, 0, NULL,	},

	/* Minor Number 3 */
	{ "vme_pti3",	0, 0, 0, 0, NULL,	},

	/* Minor Number 4 */
	{ "vme_vsi0",	0, 0, 0, 0, NULL,	},

	/* Minor Number 5 */
	{ "vme_vsi1",	0, 0, 0, 0, NULL,	},

	/* Minor Number 6 */
	{ "vme_vsi2",	0, 0, 0, 0, NULL,	},

	/* Minor Number 7 */
	{ "vme_vsi3",	0, 0, 0, 0, NULL,	},

	/* Minor Number 8 */
	{ "vme_ctl",	0, 0, 0, 0, NULL,	},
};

#define	VME_MAX_MINOR	(sizeof(vme_memory) / sizeof(vme_t))


/*
 * Prototypes
 */
static int	universe_init (void) __init;
static void     init_image(int minor, SLAVE_IMAGE *img);
static int	universe_open(struct inode *, struct file *);
static ssize_t	universe_read (struct file *file,
			       char *buf, size_t len, loff_t *off);
static ssize_t	universe_write(struct file *file,
			       const char *buf, size_t len, loff_t *off);
static loff_t	universe_lseek(struct file *file, loff_t offset, int orig);
static int	universe_release(struct inode *, struct file *);
static int	universe_mmap(struct file *file, struct vm_area_struct *vma);
static int	universe_ioctl(struct inode *, struct file *,
			       unsigned int, unsigned long);


#if CONFIG_PROC_FS

#define DEBUG_DEVNAME        "debug"
#define DEBUG_USB_DIR        "universe"

typedef struct {
	int count;
	int size;
	wait_queue_head_t wq;
} debug_dev_t;

static debug_dev_t debug_dev;

static devfs_handle_t devfs_handle = NULL;

static struct proc_dir_entry *debug_proc_dir;
static struct proc_dir_entry *debug_proc_usb_dir;

struct file_operations debug_fops = {
};

#define SAY( fmt, args... )  p += sprintf(p, fmt, ## args )
#define SAYV(  num )         p += sprintf(p, "%25.25s: %8.8lX\n", "Value", num )
#define SAYC( label, yn )    p += sprintf(p, "%25.25s: %s\n", label, yn )
#define SAYS( label, v )     p += sprintf(p, "%25.25s: %lu\n", label, v )

static int universe_read_proc(char *page, char **start, off_t off,
			      int count, int *eof, void *data);
static int universe_write_proc(struct file *file, const char *buffer,
			       unsigned long count, void *data);

#endif /* CONFIG_PROC_FS */


/***************************************************************************
 * Driver interface
 ***************************************************************************
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)

static struct file_operations universe_fops = {
	universe_lseek,	/* lseek	*/
	universe_read,		/* read		*/
	universe_write,	/* write	*/
	NULL,			/* readdir	*/
	NULL,			/* select	*/
	universe_ioctl,	/* ioctl	*/
	universe_mmap,		/* mmap		*/
	universe_open,		/* open		*/
	NULL,			/* flush	*/
	universe_release	/* close	*/
};

#else	/* Linux kernel >= 2.3.x */

static struct file_operations universe_fops = {
	owner:			THIS_MODULE,
	open:			universe_open,
	release:		universe_release,
	read:			universe_read,
	write:			universe_write,
	llseek:			universe_lseek,
	ioctl: 			universe_ioctl,
	mmap:			universe_mmap,
};
#endif


/***************************************************************************
 *
 * Convert minor number to pointer to device info
 */
vme_t *vme_dev (int minor)
{
	vme_t *mp;

	switch (minor) {
	case VME_PTI_0:
	case VME_PTI_1:
	case VME_PTI_2:
	case VME_PTI_3:
	case VME_VSI_0:
	case VME_VSI_1:
	case VME_VSI_2:
	case VME_VSI_3:
	case VME_CTL:
		mp = &vme_memory[minor];
		if (mp->name)
			return (mp);
		/* else: empty entry */
		return (NULL);
	default:
		return (NULL);
	}
}


/*************************************************************************
 *
 * Interrupt related stuff
 *
 */

static struct int_hndlr vme_int_table[256];
static int vme_int_count = 0;


static void universe_int_hndlr(int irq, void *dev_id, struct pt_regs *regs)
{
	u32 val;
	int i;
	u8 vector = 0;
	struct int_hndlr *ih;

	val = readl(&dev->uregs->lint_stat);
	debugk("Universe IRQ: irq=%d (lint_stat=%x)!!!\n", irq, val);

	/*
	 * Test all irq levels from high prio to low
	 */
	for (i=7; i>=1; i--) {
		if (val & (1 << i)) {
			vector = readl(&dev->uregs->vx_statid[i-1]) & 0x000000ff;
			debugk("Level=%d Vector=%02x!\n", i, vector);
			break;
		}
	}

	if ((i >= 1) && (i <= 7)) {

		if (vector == 0) {
			printk("Bad vme interrupt vector 0!\n");
		} else {
			ih = &vme_int_table[vector];

			/*
			 * Call connected interrupt service routine
			 */
			if (ih->handler) {
				ih->handler(vector, ih->arg);
			} else {
				printk("Uninitialized vme interrupt vector 0x%02x!\n", vector);
			}
		}

		/*
		 * Clear interrupt source
		 */
		writel(1 << i, &dev->uregs->lint_stat);
	}

	debugk("Universe IRQ done!!!\n");
}


/*
 * Enable VMEbus interrupt level
 */
int universe_int_enable(int level)
{
	u32 mask;
	u32 val;
	u32 flags;

	if ((level < 1) || (level > 7))
		return (-1);

	spin_lock_irqsave(&dev->lock, flags);

	/*
	 * Clear any pending intr. for this level
	 */
	mask = 1 << level;
	writel(mask, &dev->uregs->lint_stat);

	/*
	 * Enable the interrupt
	 */
	val = readl(&dev->uregs->lint_en);
	writel(val | mask, &dev->uregs->lint_en);

	spin_unlock_irqrestore(&dev->lock, flags);

	return 0;
}


/*
 * Disable VMEbus interrupt level
 */
int universe_int_disable(int level)
{
	u32 mask;
	u32 val;
	u32 flags;

	if ((level < 1) || (level > 7))
		return (-1);

	spin_lock_irqsave(&dev->lock, flags);

	/*
	 * Clear any pending intr. for this level
	 */
	mask = 1 << level;
	writel(mask, &dev->uregs->lint_stat);

	/*
	 * Disable the interrupt
	 */
	val = readl(&dev->uregs->lint_en);
	writel(val & ~mask, &dev->uregs->lint_en);

	spin_unlock_irqrestore(&dev->lock, flags);

	return 0;
}


/*
 * Register a VMEbus interrupt handler
 */
int universe_request_irq(unsigned int irq,
			 void (*handler)(int, void *),
			 void *argument)
{
	int rc;

	debugk ("Install Universe handler for IRQ %d\n", irq);

	if ((irq >= 0) && (irq <= 256)) {
		struct int_hndlr *ih = &vme_int_table[irq];

		if (ih->handler) {
			printk ("Nested VMEbus interrupts not supported (yet)"
				" [VMEbus vector=%d]\n", irq);
			return (-EBUSY);
		}

		ih->handler = handler;
		ih->arg = argument;

		debugk ("VMEbus IRQ %d installed\n", irq);

		/*
		 * Enable VME Bus interrupt when the first handler gets installed
		 */
		if (vme_int_count == 0) {

			debugk ("Enabling Universe VME interrupt handler\n");

			rc = request_irq(UNIVERSE_IRQ0, universe_int_hndlr, 0, "Universe", NULL);
			if (rc) {
				printk("Universe: Couldn't request irq0 rc=%d\n", rc);
				return (-EBUSY);
			}

			writel(0x00000000, &dev->uregs->lint_map0);   /* map VME interrupts */
		}
		++vme_int_count;	/* increment reference count */

		return (0);
	}

	debugk ("Invalid Universe IRQ %d (0...256 only)\n", irq);
	return (-EINVAL);
}


/*
 * De-register and free a VMEbus interrupt handler
 */
int universe_free_irq(unsigned int irq)
{
	debugk ("Un-install handler for Universe IRQ %d\n", irq);

	if ((irq >= 0) && (irq <= 256)) {
		struct int_hndlr *ih = &vme_int_table[irq];

		if (ih->handler == NULL) {
			printk ("Can't uninstall NULL VMEBus interrupt %d\n", irq);
			return (-EINVAL);
		}

		if (vme_int_count == 0) {
			printk ("Error: dead Universe VMEbus interrupt %d\n", irq);
			return (0);
		}

		--vme_int_count;	/* decrement reference count */

		/*
		 * Disable VMEbus interrupt upon removal of last handler
		 */
		if (vme_int_count == 0) {
			debugk ("Disabling Universe VME interrupt handler\n");

			free_irq (UNIVERSE_IRQ0, NULL);
		}

		ih->handler = NULL;
		ih->arg = 0L;

		debugk ("VMEbus IRQ %d uninstalled\n", irq);

		return (0);
	}

	debugk ("Invalid Universe IRQ %d (0...256 only)\n", irq);
	return (-EINVAL);
}


/*
 * Fire VMEbus interrupt
 */
int universe_int_generate(int level, int vector)
{
	u32 flags;
	u32 vint_stat;
	u32 vint_en;
	u32 statid;

	if (level < 1 || level > 7 || vector > 255 || vector < 2)
		return -1;

	spin_lock_irqsave(&dev->lock, flags);

        /* Ensure bus interrupt is not already asserted */

        vint_stat = readl(&dev->uregs->vint_stat);
        if (vint_stat & ((1 << level) << 24)) {
		spin_unlock_irqrestore(&dev->lock, flags);
		return -1;
	}

        /*
         * Clear Software Interrupt bit before setting it
         * It needs a 0 to 1 transition to generate an interrupt
         */

        vint_en = readl(&dev->uregs->vint_en);
	vint_en &= ~((1 << level) << 24);
	writel(vint_en, &dev->uregs->vint_en);

	/*
	 * store vector number, wiping out LSB of vector number
	 * This implies that only even number vectors should be used
	 * (Universe FEATURE ?)
	 */

        statid = readl(&dev->uregs->statid);
	vector >>= 1;
	statid = (statid & STATID_MASK) | (vector << 25);
	writel(statid, &dev->uregs->statid);

	/* Generate a SW interrupt on the VMEbus at the requested level */

        vint_en = readl(&dev->uregs->vint_en);
	vint_en |= ((1 << level) << 24);
	writel(vint_en, &dev->uregs->vint_en);

	spin_unlock_irqrestore(&dev->lock, flags);

	return 0;
}


/***************************************************************************
 *
 * Initialize the driver - Register the character device
 */
static int __init universe_init (void)
{
	struct pci_dev *pdev = NULL;
	u32 bar;
	int j;
	int i;
	vme_t *mp;
	u32 flags;

	debugk("%s: enter\n", __FUNCTION__);

	printk(KERN_INFO "UNIVERSE: VMEbus Driver, version " DRV_VERSION ".\n");

	if ((pdev = pci_find_device(PCI_VENDOR_ID_UNIVERSE, PCI_DEVICE_ID_UNIVERSE, pdev)) != NULL) {
		pci_read_config_dword(pdev, PCI_BASE_ADDRESS_0, &bar);
		if (bar & 1) {
			pci_read_config_dword(pdev, PCI_BASE_ADDRESS_1, &bar);
		}
		bar &= PCI_BASE_ADDRESS_MEM_MASK;
		debugk("Tundra Universe found at 0x%08x!\n", bar);

		/* Lets turn Latency off */
		pci_write_config_byte(pdev, PCI_LATENCY_TIMER, 0x00);

		dev = kmalloc(sizeof(*dev), GFP_KERNEL);
		if (NULL == dev) {
			printk("UNIVERSE: No memory!\n");
			return -ENOMEM;
		}

		memset(dev, 0, sizeof(*dev));
		dev->pdev = pdev;

		dev->uregs = (UNIVERSE *)ioremap_nocache(bar, UNIVERSE_REG_SIZE);
		if (dev->uregs == NULL) {
			printk("UNIVERSE: Can't map universe memory!\n");
			kfree(dev);
			return -ENOMEM;
		}

		if (((PCI_DEVICE_ID_UNIVERSE <<16) | PCI_VENDOR_ID_UNIVERSE) !=
		    readl(&dev->uregs->pci_id)) {
			printk("UNIVERSE: Cannot read PCI-ID via Mapping: %08x\n",
			       readl(&dev->uregs->pci_id));
			iounmap(dev->uregs);
			kfree(dev);
			return -ENODEV;
		}

		spin_lock_irqsave(&dev->lock, flags);

		/* turn off windows */
		for (j=0; j <4; j ++) {
			writel(0x00800000, &dev->uregs->pti[j].ctl);
			writel(0x00800000, &dev->uregs->vsi[j].ctl);
		}

		/*
		 * Write to Misc Register
		 * Set VME Bus Time-out
		 *   Arbitration Mode
		 *   DTACK Enable
		 */
		writel(0x15040000 | (readl(&dev->uregs->misc_ctl) & 0x00020000), &dev->uregs->misc_ctl);

		if (readl(&dev->uregs->misc_ctl) & 0x00020000) {
			printk(KERN_INFO "UNIVERSE: System Controller!\n");
		} else {
			printk(KERN_INFO "UNIVERSE: Not System Controller!\n");
		}

		/*
		 * Lets turn off interrupts
		 */
		writel(0x00000000,&dev->uregs->lint_en);   /* Disable interrupts in the Universe first */
		writel(0x0000FFFF,&dev->uregs->lint_stat); /* Clear Any Pending Interrupts */
		eieio();
		writel(0x0000, &dev->uregs->lint_map0);  /* Map all ints to 0 */
		writel(0x0000, &dev->uregs->lint_map1);  /* Map all ints to 0 */
		eieio();

		spin_unlock_irqrestore(&dev->lock, flags);

		if ((i = register_chrdev(VME_MAJOR, DEVICE_NAME, &universe_fops)) < 0) {
			printk("Unable to get major %d for VME devices: rc=%d\n", VME_MAJOR, i);
			return (i);
		}

		Major = VME_MAJOR;

		debugk (DEVICE_NAME " registered: major = %d\n", Major);

		/*
		 * Init vme_memory structure
		 */
		vme_memory[VME_PTI_0].img = &dev->uregs->pti[0];
		vme_memory[VME_PTI_1].img = &dev->uregs->pti[1];
		vme_memory[VME_PTI_2].img = &dev->uregs->pti[2];
		vme_memory[VME_PTI_3].img = &dev->uregs->pti[3];
		vme_memory[VME_VSI_0].img = &dev->uregs->vsi[0];
		vme_memory[VME_VSI_1].img = &dev->uregs->vsi[1];
		vme_memory[VME_VSI_2].img = &dev->uregs->vsi[2];
		vme_memory[VME_VSI_3].img = &dev->uregs->vsi[3];
		vme_memory[VME_CTL].addr = (u32)dev->uregs;

		/*
		 * Linux allocates the pci mem spaces top down, so we have to find the
		 * last pci device with the last mem space. This is the lowest pci mem
		 * space we have to use to allocate pci target images for the Universe.
		 *
		 * Right now, we know we are running on the P3G4 and SCCB where the Universe
		 * itself is the last pci device. So just use its pci mem space. This has
		 * to be fixed for other platforms!
		 */
		universe_upper_memspc = bar;
		debugk("UNIVERSE: upper pci memspace = %08x\n", universe_upper_memspc);

		for (i=0; i<MAX_PCI_MAPS; i++) {
			universe_pci_maps[i].vme_addr = 0xffffffff; /* mark as not mapped */
			universe_pci_maps[i].pci_addr = 0xffffffff; /* mark as not mapped */
			universe_pci_maps[i].size = 0xffffffff; /* mark as not mapped */
			universe_pci_maps[i].ctl = 0xffffffff; /* mark as not mapped */
		}
		universe_last_map_index = 0;

		/*
		 * Map CTL device space (Universe registers)
		 */
		mp = vme_dev(VME_CTL);
		mp->flags = 0;
		mp->phys = bar;
		mp->size = UNIVERSE_REG_SIZE;
		mp->addr = (unsigned long)dev->uregs;

#if CONFIG_PROC_FS
		devfs_handle = devfs_register(NULL, DEBUG_DEVNAME, DEVFS_FL_AUTO_DEVNUM,
					      0, 0,
					      S_IFCHR | S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP,
					      &debug_fops, &debug_dev);

		memset(&debug_dev, 0, sizeof(debug_dev_t));

		debug_proc_dir = proc_mkdir(DEBUG_DEVNAME, NULL);
		if (debug_proc_dir == NULL) {
			printk(KERN_ERR "debug: Unable to create /proc/%s entry\n", DEBUG_DEVNAME);
			return -ENOMEM;
		}

		debug_proc_usb_dir = create_proc_entry(DEBUG_USB_DIR, 0644, debug_proc_dir);
		if (debug_proc_usb_dir == NULL) {
			printk(KERN_ERR "DEBUG: Unable to create /proc/%s/%s entry\n", DEBUG_DEVNAME, DEBUG_USB_DIR);
			remove_proc_entry("debug", NULL);
			return -ENOMEM;
		}
		debug_proc_usb_dir->read_proc = universe_read_proc;
		debug_proc_usb_dir->write_proc = universe_write_proc;
#endif
	}

	debugk("%s: leave\n", __FUNCTION__);
	return 0;
}


static void init_image(int minor, SLAVE_IMAGE *img)
{
	vme_t *mp;
	u32 flags;

	if ((mp = vme_dev(minor)) == NULL) {
		return;
	}

	if (mp->name == NULL)		/* unused entry */
		return;

	if (((minor >= VME_PTI_0) && (minor <= VME_PTI_3)) ||
	    ((minor >= VME_VSI_0) && (minor <= VME_VSI_3))) {
		spin_lock_irqsave(&dev->lock, flags);
		writel(0, &mp->img->ctl);
		writel(img->bs, &mp->img->bs);
		writel(img->bd, &mp->img->bd);
		writel(img->to, &mp->img->to);
		writel(img->ctl, &mp->img->ctl);
		spin_unlock_irqrestore(&dev->lock, flags);
		debugk ("Minor=%d: bs=%08x bd=%08x to=%08x ctl=%08x\n",
			minor, img->bs, img->bd, img->to, img->ctl);
	}

	mp->flags = 0;
	mp->phys = img->bs;
	mp->size = img->bd - img->bs;

	/*
	 * Only map pci target images
	 */
	if ((minor >= VME_PTI_0) && (minor <= VME_PTI_3)) {
		mp->addr = (unsigned long)ioremap_nocache(mp->phys, mp->size);
		debugk ("Mapped: phys=%08lx virt=%08lx size=%08lx\n",
			mp->phys, mp->addr, mp->size);
	}
}


/***************************************************************************
 *
 * called whenever a process attempts to open the device
 */
static int universe_open (struct inode *inode, struct file *file)
{
	int minor;
	vme_t *mp;

	minor = MINOR(inode->i_rdev);

	debugk ("%s: minor %d\n", __FUNCTION__, minor);

	if ((mp = vme_dev(minor)) == NULL) {
		return -ENXIO;
	}

	/*
	 * exclusive open only
	 */
	if (mp->flags & UNIVERSE_FLAGS_BUSY) {
		return -EBUSY;
	}

	mp->flags |= UNIVERSE_FLAGS_BUSY;

	/*
	 * Make sure that the module isn't removed while
	 * the file is open by incrementing the usage count
	 */
	MOD_INC_USE_COUNT;

	debugk ("%s: \"%s\" done\n", __FUNCTION__, mp->name);

	return 0;
}



/***************************************************************************
 *
 * called when a process closes the device.
 */
static int universe_release (struct inode *inode, struct file *file)
{
	int minor;
	vme_t *mp;

	minor = MINOR(inode->i_rdev);

	debugk ("%s: minor %d\n", __FUNCTION__, minor);

	if ((mp = vme_dev(minor)) == NULL) {
		return -ENXIO;
	}

	/* We're now ready for our next caller */
	mp->flags &= ~UNIVERSE_FLAGS_BUSY;

	MOD_DEC_USE_COUNT;

	return 0;
}


/***************************************************************************
 *
 * lseek entry point:
 */
static loff_t universe_lseek(struct file *file, loff_t offset, int orig)
{
	int minor;
	vme_t *mp;

	minor = MINOR(file->f_dentry->d_inode->i_rdev);

	debugk ("%s: minor %d\n", __FUNCTION__, minor);

	if ((mp = vme_dev(minor)) == NULL) {
		return -ENXIO;
	}

	switch (orig) {
	case 0:
		if (offset > mp->size)
			return -EINVAL;
		file->f_pos = offset;
		return file->f_pos;
	case 1:
		if (file->f_pos + offset > mp->size)
			return -EINVAL;
		file->f_pos += offset;
		return file->f_pos;
	default:
		return -EINVAL;
	}
}


/***************************************************************************
 *
 * read entry point:
 */
static ssize_t universe_read(struct file *file, char *buf, size_t len, loff_t *off)
{
	int minor, error;
	ulong offset = *off;
	uint count;
	vme_t *mp;

	minor = MINOR(file->f_dentry->d_inode->i_rdev);

	debugk ("%s: minor %d len %d off %#x\n",
		__FUNCTION__, minor, len, (unsigned int)*off);

	/* POSIX: mtime/ctime may not change for 0 count */
	if (!len)
		return 0;

	if ((mp = vme_dev(minor)) == NULL) {
		return -ENXIO;
	}

	/*
	 * Check that the file was opened for reading ?
	 * Seems to be done in sys_read.
	 */
	if (!(file->f_mode & FMODE_READ)) {
		debugk ("%s: minor %d (%s) not readable\n",
			__FUNCTION__, minor, mp->name);
		return -EBADF;
	}

	/*
	 * make sure that the user isn't trying to read past end of memory
	 */
	if (offset > mp->size) {
		debugk ("%s: minor %d (%s) off %08lx > size %08lx\n",
			__FUNCTION__, minor, mp->name, offset, mp->size);
		return -EINVAL;
	}

	count = mp->size - (unsigned int)offset;
	if (count > len)
		count = (unsigned int)len;

	/* Copy out */
	if ((error = verify_area(VERIFY_WRITE, buf, count)) != 0) {
		printk("%s: (%s) verify_area returned %d\n",
		       __FUNCTION__, mp->name, error);
		return error;
	}

	copy_to_user((void *)buf, (void*) (mp->addr + offset),  count);
	file->f_pos += count;

	return count;
}


/***************************************************************************
 *
 * write entry point:
 */
static ssize_t universe_write(struct file *file, const char *buf, size_t len, loff_t *off)
{
	int minor, error;
	ulong offset = *off;
	uint count;
	vme_t *mp;
	minor = MINOR(file->f_dentry->d_inode->i_rdev);

	debugk ("%s: minor %d len %d off %lx\n",
		__FUNCTION__, minor, len, offset);

	/* POSIX: mtime/ctime may not change for 0 count */
	if (!len)
		return 0;

	if ((mp = vme_dev(minor)) == NULL) {
		return -ENXIO;
	}

	/*
	 * Check that the file was opened for writing ?
	 * Seems to be done in sys_write.
	 */
	if (!(file->f_mode & FMODE_WRITE)) {
		debugk ("%s: minor %d (%s) not writeable\n",
			__FUNCTION__, minor, mp->name);
		return -EBADF;
	}
	/*
	 * make sure that the user isn't trying to write past end of memory
	 */
	if (offset >= mp->size) {
		debugk ("universe_write: minor %d off %08lx > size %08lx\n",
			minor, offset, mp->size);
		return (offset == mp->size) ? -ENOSPC : -EINVAL;
	}

	count = mp->size - (unsigned int)offset;
	if (count > len)
		count = (unsigned int)len;

	/* Copy in */
	if ((error = verify_area(VERIFY_READ, (const void *) buf, count)) != 0) {
		printk("%s: (%s) verify_area returned %d\n",
		       __FUNCTION__, mp->name, error);
		return error;
	}

	copy_from_user((void *)(mp->addr + offset), (const void*) buf, count);
	file->f_pos += count;

	return count;
}

/***************************************************************************
 *
 * ioctl entry point:
 */
static int universe_ioctl (struct inode *inode, struct file *file,
			   unsigned int cmd, unsigned long arg)
{
	int minor;
	SLAVE_IMAGE img;
	struct universe_img uimg;
	int i;

	minor = MINOR(file->f_dentry->d_inode->i_rdev);

	debugk ("IOCTL: minor %d, cmd=%x, arg=%ld\n", minor, cmd, arg);

	switch (cmd) {
	case IOCTL_CONFIG_PTI:
		if (copy_from_user(&uimg, (void*)arg, sizeof(uimg))) {
			return -EFAULT;
		}

		/*
		 * Search, if this mem space if already mapped
		 */
		for (i=0; i<universe_last_map_index; i++) {
			if ((universe_pci_maps[i].vme_addr == uimg.vme_addr) &&
			    (universe_pci_maps[i].ctl == uimg.ctl) &&
			    (universe_pci_maps[i].size == uimg.size)) {
				/*
				 * Found already mapped image -> use it
				 */
				uimg.pci_addr = universe_pci_maps[i].pci_addr;
				break;
			}
		}

		if (i == MAX_PCI_MAPS) {
			return -ENOMEM;
		} else if (i == universe_last_map_index) {
			/*
			 * Calc pci base address for this mem space
			 */
			uimg.pci_addr = (universe_upper_memspc - uimg.size) & ~(uimg.size - 1);
			universe_upper_memspc = uimg.pci_addr; /* save new upper limit */

			/*
			 * Save image in mapped array
			 */
			universe_pci_maps[i].vme_addr = uimg.vme_addr;
			universe_pci_maps[i].pci_addr = uimg.pci_addr;
			universe_pci_maps[i].size = uimg.size;
			universe_pci_maps[i].ctl = uimg.ctl;
			universe_last_map_index = i;
			universe_last_map_index++;
		}

		img.bs = uimg.pci_addr;
		img.bd = uimg.pci_addr + uimg.size;
		img.to = uimg.vme_addr - uimg.pci_addr;
		img.ctl = uimg.ctl;
		init_image(minor, &img);

		/*
		 * Copy resulting pci address back to user space
		 */
		if (copy_to_user((void*)arg, &uimg, sizeof(uimg))) {
			return -EFAULT;
		}

		break;

	case IOCTL_CONFIG_VSI:
		if (copy_from_user(&uimg, (void*)arg, sizeof(uimg))) {
			return -EFAULT;
		}

		img.bs = uimg.vme_addr;
		img.bd = uimg.vme_addr + uimg.size;
		img.to = uimg.pci_addr - uimg.vme_addr;
		img.ctl = uimg.ctl;
		init_image(minor, &img);

		break;

	default:
		return -EFAULT;
	}

	return 0;
}


/***************************************************************************
 *
 * mmap entry point:
 */
static int universe_mmap(struct file *file, struct vm_area_struct *vma)
{
	int minor, size;
	unsigned long error;
	unsigned long offset;
	vme_t *mp;

	minor = MINOR(file->f_dentry->d_inode->i_rdev);

	debugk ("%s: minor %d\n",__FUNCTION__,minor);

	if ((mp = vme_dev(minor)) == NULL) {
		return -ENXIO;
	}

	size   = vma->vm_end - vma->vm_start;
	offset = vma->vm_pgoff << PAGE_SHIFT;

	if (size > mp->size) {
		debugk ("%s: minor %d (%s) bad size %d > %ld\n",
			__FUNCTION__, minor, mp->name, size, mp->size);
		return -EFAULT;
	}

	pgprot_val(vma->vm_page_prot) |= _PAGE_NO_CACHE|_PAGE_GUARDED;

	/* remapping */
	if ((error = io_remap_page_range(vma->vm_start, mp->phys+offset, size,
					 vma->vm_page_prot)) != 0) {
		printk ("%s: minor %d (%s) remap_page_range() error %ld\n",
			__FUNCTION__, minor, mp->name, error);
		return error;
	}

	return 0;
}


/*************************************************************************
 *
 * PROC-FS related stuff
 *
 */

#if CONFIG_PROC_FS

static int universe_read_proc(char *page, char **start, off_t off,
			      int count, int *eof, void *data)
{
	 char * p = page;
	 int len;

 	 SAY( "Universe:\n" );

	 len = ( p - page ) - off;
	 if ( len < 0 )
		  len = 0;
	 *eof = ( len <=count ) ? 1 : 0;
	 *start = page + off;
	 return len;
}

static int universe_write_proc(struct file *file, const char *buffer,
			       unsigned long count, void *data)
{
	char mydata[0x100];
	int rc;

	/* TODO: Check for count > len */
	if (copy_from_user(mydata, buffer, count) ) {
		return -EFAULT;
	}

	if (strncmp(mydata, "virq", 4) == 0) {
		printk("Universe: Trigger VME IRQ request!\n");

		rc = universe_int_generate(2, 0x72); /* generate VMEbus interrupt */
		printk("universe_int_generate returned %d!\n", rc);

		printk("Generate IRQ done!\n");
	}

	return count;
}

#endif /* CONFIG_PROC_FS */


/***************************************************************************
 *
 * Module Declarations
 */

module_init (universe_init);

#ifdef MODULE
/*
 * Cleanup - unregister the appropriate file from /proc
 */
void universe_cleanup (void)
{
	int i, rc;

	printk(KERN_INFO "Unloading Universe VMEbus driver.\n");

	for (i=0; i<VME_MAX_MINOR; ++i) {
		debugk ("%s[%d]: cleanup minor %d - \"%s\"\n",
			__FUNCTION__,__LINE__,i,
			vme_memory[i].name ? vme_memory[i].name : "NULL");
		if (vme_memory[i].name == NULL)
			continue;
		if (vme_memory[i].addr == 0)
			continue;
		iounmap ((void *)vme_memory[i].addr);
		vme_memory[i].phys  = 0;
		vme_memory[i].addr  = 0;
		vme_memory[i].size  = 0;
		vme_memory[i].flags = 0;
	}

#if CONFIG_PROC_FS
	devfs_unregister(devfs_handle);

	remove_proc_entry(DEBUG_USB_DIR, debug_proc_usb_dir);
	remove_proc_entry(DEBUG_DEVNAME, NULL);
#endif

	/*
	 * Unregister the device
	 */
	rc = unregister_chrdev (Major, DEVICE_NAME);

	/*
	 * If there's an error, report it
	 */
	if (rc < 0) {
		printk (DEVICE_NAME "unregister_chrdev: error %d\n", rc);
	}
}

module_exit (universe_cleanup);

MODULE_AUTHOR("Stefan Roese <sr@denx.de>");
MODULE_DESCRIPTION("Tundra Universe VMEbus driver");
MODULE_LICENSE ("GPL");

#endif	/* MODULE */

EXPORT_SYMBOL(universe_int_enable);
EXPORT_SYMBOL(universe_int_disable);
EXPORT_SYMBOL(universe_request_irq);
EXPORT_SYMBOL(universe_free_irq);
EXPORT_SYMBOL(universe_int_generate);
