/*
    (C) Copyright 2006
    Wolfgang Denk, DENX Software Engineering, wd@denx.de.

    Driver for the lm73 sensor, based on lm75.c - which has:
    Copyright (c) 1998, 1999  Frodo Looijaard <frodol@dds.nl>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


/*
 * Partial support for the LM73 chip - supported are temperature readout,
 * t_high and t_low readout and setting. Only one, largest temperature
 * resolution is supported (11 bits: 10 for the temp + 1 for the sign,
 * LSb = 0.25 C.
 * Not supported:
 * - shutdown mode & 1-shot conversion
 * - other temperature resolutions
 */
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/i2c.h>
#include <linux/i2c-proc.h>
#include <linux/init.h>
#include "lm73.h"


/* Addresses to scan */
static unsigned short normal_i2c[] = { SENSORS_I2C_END };
static unsigned short normal_i2c_range[] = { 0x48, 0x4f, SENSORS_I2C_END };
static unsigned int normal_isa[] = { SENSORS_ISA_END };
static unsigned int normal_isa_range[] = { SENSORS_ISA_END };

/* Insmod parameters */
SENSORS_INSMOD_1(lm73);


/* Each client has this additional data */
struct lm73_data {
	struct i2c_client client;
	int sysctl_id;
	struct semaphore update_lock;
	char valid;			/* !=0 if following fields are valid */
	unsigned long last_updated;	/* In jiffies */
	u16 temp, temp_low, temp_high;	/* Register values */
};


static int lm73_attach_adapter(struct i2c_adapter *adapter);
static int lm73_detect(struct i2c_adapter *adapter, int address,
			unsigned short flags, int kind);
static void lm73_init_client(struct i2c_client *client);
static int lm73_detach_client(struct i2c_client *client);
static int lm73_read_value(struct i2c_client *client, u8 reg);
static int lm73_write_value(struct i2c_client *client, u8 reg, u16 value);
static void lm73_temp(struct i2c_client *client, int operation,
			int ctl_name, int *nrels_mag, long *results);
static void lm73_update_client(struct i2c_client *client);


/* This is the driver that will be inserted */
static struct i2c_driver lm73_driver = {
	.name		= "LM73 sensor chip driver",
	.id		= I2C_DRIVERID_LM73,
	.flags		= I2C_DF_NOTIFY,
	.attach_adapter	= lm73_attach_adapter,
	.detach_client	= lm73_detach_client,
};

/* -- SENSORS SYSCTL START -- */

#define LM73_SYSCTL_TEMP 1200	/* Degrees Celsius * 100 */

/* -- SENSORS SYSCTL END -- */

/* These files are created for each detected LM73. This is just a template;
   though at first sight, you might think we could use a statically
   allocated list, we need some way to get back to the parent - which
   is done through one of the 'extra' fields which are initialized
   when a new copy is allocated. */
static ctl_table lm73_dir_table_template[] = {
	{LM73_SYSCTL_TEMP, "temp", NULL, 0, 0644, NULL, &i2c_proc_real,
	 &i2c_sysctl_real, NULL, &lm73_temp}, {0}
};

static int lm73_attach_adapter(struct i2c_adapter *adapter)
{
	return i2c_detect(adapter, &addr_data, lm73_detect);
}

/* This function is called by i2c_detect */
int lm73_detect(struct i2c_adapter *adapter, int address,
		unsigned short flags, int kind)
{
	struct i2c_client *new_client;
	struct lm73_data *data;
	u16 chip_id;
	int ret;

#ifdef DEBUG
	/*
	 * Make sure we aren't probing the ISA bus!! This is just a safety
	 * check at this moment; i2c_detect really won't call us.
	 */

	if (i2c_is_isa_adapter(adapter)) {
		printk("lm73.o: lm73_detect called for an ISA bus adapter?\n");
		return -1;
	}
#endif

	if (!i2c_check_functionality(adapter, I2C_FUNC_SMBUS_BYTE_DATA |
					I2C_FUNC_SMBUS_WORD_DATA))
		return -1;


	/* Create and populate the client structure */
	if (!(data = kmalloc(sizeof(struct lm73_data), GFP_KERNEL)))
		return -ENOMEM;

	new_client = &data->client;
	new_client->addr = address;
	new_client->data = data;
	new_client->adapter = adapter;
	new_client->driver = &lm73_driver;
	new_client->flags = 0;
	strcpy(new_client->name, LM73_CLIENT_NAME);
	data->valid = 0;
	init_MUTEX(&data->update_lock);


	/* detect if we indeed have a LM73 chip */
	chip_id = lm73_read_value(new_client, LM73_REG_ID);
	if ((chip_id & LM73_MANUFACTURER_MASK) != LM73_MANUFACTURER) {
		printk(KERN_WARNING "lm73.o: error: wrong manufacturer id\n");
		kfree(data);
		return -1;
	}
	if ((chip_id & LM73_PRODUCT_MASK) != LM73_PRODUCT) {
		printk(KERN_WARNING "lm73.o: error: wrong product id\n");
		kfree(data);
		return -1;
	}
	pr_debug("lm73.o: detected National Semiconductor LM73 "\
			"sensor, rev. 0x%02hx\n",
			chip_id & LM73_REVISION_MASK);

	/* Tell the I2C layer a new client has arrived */
	if ((ret = i2c_attach_client(new_client))) {
		kfree(data);
		return ret;
	}

	/* Register a new directory entry with module sensors */
	if ((ret = i2c_register_entry(new_client, LM73_TYPE_NAME,
					lm73_dir_table_template,
					THIS_MODULE)) < 0) {
		i2c_detach_client(new_client);
		kfree(data);
		return ret;
	}
	data->sysctl_id = ret;

	lm73_init_client(new_client);
	return 0;
}

static int lm73_detach_client(struct i2c_client *client)
{
	struct lm73_data *data = client->data;

	i2c_deregister_entry(data->sysctl_id);
	i2c_detach_client(client);
	kfree(client->data);
	return 0;
}

/* All registers are word-sized, except for the configuration register.
   LM73 uses a high-byte first convention, which is exactly opposite to
   the usual practice. */
static int lm73_read_value(struct i2c_client *client, u8 reg)
{
	if (reg == LM73_REG_CONF)
		return i2c_smbus_read_byte_data(client, reg);
	else
		return swab16(i2c_smbus_read_word_data(client, reg));
}

/* All registers are word-sized, except for the configuration register.
   LM73 uses a high-byte first convention, which is exactly opposite to
   the usual practice. */
static int lm73_write_value(struct i2c_client *client, u8 reg, u16 value)
{
	if (reg == LM73_REG_CONF)
		return i2c_smbus_write_byte_data(client, reg, value);
	else
		return i2c_smbus_write_word_data(client, reg, swab16(value));
}

static void lm73_init_client(struct i2c_client *client)
{
	u8 tmp;

	/* Put into normal mode */
	tmp = (u8)lm73_read_value(client, LM73_REG_CONF);
	lm73_write_value(client, LM73_REG_CONF, tmp & ~(1 << 7));
}

static void lm73_update_client(struct i2c_client *client)
{
	struct lm73_data *data = client->data;

	down(&data->update_lock);

	if ((jiffies - data->last_updated > HZ + HZ / 2) ||
		(jiffies < data->last_updated) || !data->valid) {
		pr_debug("Starting lm73 update\n");

		data->temp = lm73_read_value(client, LM73_REG_TEMP);
		data->temp_low = lm73_read_value(client, LM73_REG_T_LOW);
		data->temp_high = lm73_read_value(client, LM73_REG_T_HIGH);
		data->last_updated = jiffies;
		data->valid = 1;
	}

	up(&data->update_lock);
}


void lm73_temp(struct i2c_client *client, int operation, int ctl_name,
		int *nrels_mag, long *results)
{
	struct lm73_data *data = client->data;
	if (operation == SENSORS_PROC_REAL_INFO)
		*nrels_mag = 2; /* divide by 10^2 = 100 */
	else if (operation == SENSORS_PROC_REAL_READ) {
		lm73_update_client(client);
		results[0] = LM73_TEMP_FROM_REG(data->temp_low);
		results[1] = LM73_TEMP_FROM_REG(data->temp_high);
		results[2] = LM73_TEMP_FROM_REG(data->temp);
		*nrels_mag = 3;
	} else if (operation == SENSORS_PROC_REAL_WRITE) {
		if (*nrels_mag >= 1) {
			data->temp_low = LM73_TEMP_TO_REG(results[0]);
			lm73_write_value(client, LM73_REG_T_LOW,
						data->temp_low);
		}
		if (*nrels_mag >= 2) {
			data->temp_high = LM73_TEMP_TO_REG(results[1]);
			lm73_write_value(client, LM73_REG_T_HIGH,
						data->temp_high);
		}
	}
}

static int __init lm73_init(void)
{
	printk(KERN_INFO "lm73.o version %s (%s)\n", LM73_VERSION, LM73_DATE);
	return i2c_add_driver(&lm73_driver);
}

static void __exit lm73_exit(void)
{
	i2c_del_driver(&lm73_driver);
}

MODULE_AUTHOR("Wolfgang Denk wd@denx.de");
MODULE_DESCRIPTION("LM73 driver");
MODULE_LICENSE("GPL");

module_init(lm73_init);
module_exit(lm73_exit);
