/*
 * drivers/mtd/maps/cam5200.c
 * MTD mapping driver for NIOS cpu flash on CAM5200 board.
 *
 * 2006 (c) Wolfgang Denk, DENX Software Engineering, <wd@denx.de>.
 *
 * This file is licensed under the terms of the GNU General Public License
 * version 2.  This program is licensed "as is" without any warranty of
 * any kind, whether express or implied.
 *
 * CAM5200 is a TQM5200B based board. Additionally it also features
 * a NIOS cpu. The NIOS CPU peripherals are accessible through MPC5xxx
 * Local Bus on CS5. This includes 32 bit wide RAM and SRAM as well as
 * 16 bit wide flash device. Big Endian order on a 32 bit CS5 makes
 * access to flash chip slightly more complicated as additional byte
 * swapping is necessary within each 16 bit wide flash 'word'. The
 * swapping is performed in the cam5200_read16, cam5200_memcpy_fromio,
 * cam5200_write16 and and cam5200_memcpy_toio routines.
 *
 */

#include <linux/config.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/io.h>
#include <asm/mpc5xxx.h>
#ifdef CONFIG_UBOOT
#include <asm/ppcboot.h>
#endif

#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>

#if 0	/* Debugging turned off */
# define debugk(fmt,args...)	printk(fmt ,##args)
#else
# define debugk(fmt,args...)
#endif

#define FLASH_BANK_MAX 1

/* trivial struct to describe partition information */
struct mtd_part_def
{
	int nums;
	unsigned char *type;
	struct mtd_partition* mtd_part;
};

static struct mtd_info* mtd_banks[FLASH_BANK_MAX];
static struct map_info* map_banks[FLASH_BANK_MAX];
#ifdef CONFIG_MTD_PARTITIONS
static struct mtd_part_def part_banks[FLASH_BANK_MAX];
#endif
static unsigned long num_banks;
static unsigned long start_scan_addr;

static map_word cam5200_read16(struct map_info *map, unsigned long ofs)
{
	map_word val;
	__u16 tmp;

	tmp = *((__u16 *)(map->map_priv_1 + ofs));

	val.x[0] = ((tmp & 0xff00) >> 8) | ((tmp & 0x00ff) << 8);
	return val;
}

static void *cam5200_memcpy_fromio(char *to, char *from, ssize_t len)
{
	void *ptr = to;

	if(((u32)from % 2) == 1) {
		*to++ = *(from++ - 1);
		len--;
	}

	while(len >= 2) {
		*to = *(from + 1);
		*(to + 1) = *from;

		to += 2;
		from += 2;
		len -= 2;
	}

	if (len == 1) {
		*to = *(from + 1);
		len --;
	}

	return  ptr;
}

static void cam5200_copy_from(struct map_info *map,
		       void *to, unsigned long from, ssize_t len)
{
	cam5200_memcpy_fromio(to, (void *)(map->map_priv_1 + from), len);
}

static void cam5200_write16(struct map_info *map, const map_word d, unsigned long adr)
{
	__u16 tmp;

	tmp = ((d.x[0] & 0xff00) >> 8) | ((d.x[0] & 0x00ff) << 8);

	*((__u16 *)(map->map_priv_1 + adr)) = tmp;
}

 
static void *cam5200_memcpy_toio(char *to, const char *from, ssize_t len)
{
	void *ptr = to;

	if(((u32)from % 2) == 1) {
		*(to++ - 1) = *from++;
		len--;
	}

	while(len >= 2) {
		*to = *(from + 1);
		*(to + 1) = *from;

		to += 2;
		from += 2;
		len -= 2;
	}

	if (len == 1) {
		*(to + 1) = *from;
		len--;
	}

	return  ptr;
}
static void cam5200_copy_to(struct map_info *map,
		     unsigned long to, const void *from, ssize_t len)
{
	cam5200_memcpy_toio((void *)(map->map_priv_1 + to), from, len);
}

/*
 * The following defines the partition layout of NIOS flash on a CAM5200 boards.
 *
 * See include/linux/mtd/partitions.h for definition of the
 * mtd_partition structure.
 */

#ifdef CONFIG_MTD_PARTITIONS
/* partition definition for first (and only) flash bank
 */
static struct mtd_partition cam5200_partitions[] = {
	{
		name:	"user",
		offset:	0x00000000,
		size:	0x00400000,
	},
};
#endif	/* CONFIG_MTD_PARTITIONS */

#define NB_OF(x)  (sizeof(x)/sizeof(x[0]))

int __init init_cam5200_mtd(void)
{
	int idx = 0, ret = 0;
	unsigned long flash_addr, flash_size, mtd_size = 0;
#ifdef CONFIG_MTD_CMDLINE_PARTS
	int n;
	char mtdid[4];
	const char *part_probes[] = { "cmdlinepart", NULL };
#endif
	flash_addr = CAM5200_NIOS_FLASH_VIRT_START;
	flash_size = CAM5200_NIOS_FLASH_SIZE;

	/* request maximum flash size address space */
	start_scan_addr = flash_addr; /* already mapped */

	for(idx = 0 ; idx < FLASH_BANK_MAX ; idx++) {
		if (mtd_size >= flash_size)
			break;

		debugk ("%s: chip probing count %d\n", __FUNCTION__, idx);

		map_banks[idx] = (struct map_info *)kmalloc(
					sizeof(struct map_info), GFP_KERNEL);
		if (map_banks[idx] == NULL) {
			ret = -ENOMEM;
			goto error_mem;
		}
		memset((void *)map_banks[idx], 0, sizeof(struct map_info));
		map_banks[idx]->name = (char *)kmalloc(16, GFP_KERNEL);
		if (map_banks[idx]->name == NULL) {
			ret = -ENOMEM;
			goto error_mem;
		}
		memset((void *)map_banks[idx]->name, 0, 16);

		sprintf(map_banks[idx]->name, "CAM5200-%d", idx);
		map_banks[idx]->size      = flash_size;
		map_banks[idx]->bankwidth = 2;
		map_banks[idx]->read      = cam5200_read16;
		map_banks[idx]->copy_from = cam5200_copy_from;
		map_banks[idx]->write     = cam5200_write16;
		map_banks[idx]->copy_to   = cam5200_copy_to;
		map_banks[idx]->map_priv_1=
			start_scan_addr + ((idx > 0) ?
			(mtd_banks[idx-1] ? mtd_banks[idx-1]->size : 0) : 0);

		/* start to probe flash chips */
		mtd_banks[idx] = do_map_probe("cfi_probe", map_banks[idx]);
		if (mtd_banks[idx]) {
			mtd_banks[idx]->owner = THIS_MODULE;
			mtd_size += mtd_banks[idx]->size;
			num_banks++;
			debugk ("%s: bank %ld, name: %s, size: %d bytes \n",
				__FUNCTION__,
				num_banks,
				mtd_banks[idx]->name,
				mtd_banks[idx]->size);
		}
	}

	/* no supported flash chips found */
	if (!num_banks) {
		printk("CAM5200: No supported flash chips found!\n");
		ret = -ENXIO;
		goto error_mem;
	}

#ifdef CONFIG_MTD_PARTITIONS
	/*
	 * Select static partition definitions
	 */
	part_banks[0].mtd_part	= cam5200_partitions;
	part_banks[0].nums	= NB_OF(cam5200_partitions);
	part_banks[0].type	= "static image";

	for(idx = 0; idx < num_banks ; idx++) {
#ifdef CONFIG_MTD_CMDLINE_PARTS
		sprintf(mtdid, "%d", idx);
		n = parse_mtd_partitions(mtd_banks[idx],
						part_probes,
						&part_banks[idx].mtd_part,
						0);
		debugk ("%s: %d command line partitions on bank %s\n",
			__FUNCTION__, n, mtdid);
		if (n > 0) {
			part_banks[idx].type = "command line";
			part_banks[idx].nums = n;
		}
#endif	/* CONFIG_MTD_CMDLINE_PARTS */

		if (part_banks[idx].nums == 0) {
		    printk (KERN_NOTICE
			    "CAM5200 NIOS flash bank %d: no partition info "
			    "available, registering whole device\n", idx);
		    add_mtd_device(mtd_banks[idx]);
		} else {
		    printk (KERN_NOTICE
			    "CAM5200 NIOS flash bank %d: Using %s partition "
			    "definition\n", idx, part_banks[idx].type);
		    add_mtd_partitions (mtd_banks[idx],
					part_banks[idx].mtd_part,
					part_banks[idx].nums);
		}
	}
#else	/* ! CONFIG_MTD_PARTITIONS */
	printk (KERN_NOTICE "CAM5200 NIOS flash: registering %ld flash banks "
		"at once\n", num_banks);

	for(idx = 0 ; idx < num_banks ; idx++) {
		add_mtd_device(mtd_banks[idx]);
	}
#endif	/* CONFIG_MTD_PARTITIONS */

	return 0;
error_mem:
	for (idx = 0 ; idx < FLASH_BANK_MAX ; idx++) {
		if (map_banks[idx] != NULL) {
			if (map_banks[idx]->name != NULL) {
				kfree(map_banks[idx]->name);
				map_banks[idx]->name = NULL;
			}
			kfree(map_banks[idx]);
			map_banks[idx] = NULL;
		}
	}

	iounmap((void *)start_scan_addr);

	return ret;
}

static void __exit cleanup_cam5200_mtd(void)
{
	unsigned int idx = 0;
	for(idx = 0 ; idx < num_banks ; idx++) {
		/* destroy mtd_info previously allocated */
		if (mtd_banks[idx]) {
#ifdef CONFIG_MTD_PARTITIONS
			del_mtd_partitions(mtd_banks[idx]);
#endif
			map_destroy(mtd_banks[idx]);
		}
		/* release map_info not used anymore */
		kfree(map_banks[idx]->name);
		kfree(map_banks[idx]);
	}
	if (start_scan_addr) {
		iounmap((void *)start_scan_addr);
		start_scan_addr = 0;
	}
}

module_init(init_cam5200_mtd);
module_exit(cleanup_cam5200_mtd);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Wolfgang Denk <wd@denx.de>");
MODULE_DESCRIPTION("MTD map driver for NIOS flash on CAM5200 board");
