/*
 * Handle mapping of the flash memory access routines
 * on CC1000 boards
 *
 * $Id: tqm8xxl.c,v 1.13 2004/10/20 22:21:53 dwmw2 Exp $
 *
 * Copyright 2001-2005 Wolfgang Denk, <wd@denx.de>
 *
 * based on rpxlite.c
 * Copyright(C) 2001 Kirk Lee <kirk@hpc.ee.ntu.edu.tw>
 *
 * This code is GPLed
 *
 */

/*
 * The CC1000 series has the following flash memory organization:
 *	| capacity |	| chip type |	| bank0 |	| bank1 |
 *	    4MiB	   1Mx16	  4MiB		   0		or
 *	    8MiB	   2Mx16	  8MiB		   0
 * Thus, we choose CONFIG_MTD_CFI_I1 & CONFIG_MTD_CFI_B2 at
 * kernel configuration.
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <asm/io.h>

#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#ifdef CONFIG_MTD_PARTITIONS
#include <linux/mtd/partitions.h>
#endif

/* SCH: turn debug on */
#if 0	/* Debugging turned off */
# define debugk(fmt,args...)	printk(fmt ,##args)
#else
# define debugk(fmt,args...)
#endif

#define FLASH_BANK_MAX 1

/* trivial struct to describe partition information */
struct mtd_part_def
{
	int nums;
	unsigned char *type;
	struct mtd_partition* mtd_part;
};

static struct mtd_info* mtd_banks[FLASH_BANK_MAX];
static struct map_info* map_banks[FLASH_BANK_MAX];
static struct mtd_part_def part_banks[FLASH_BANK_MAX];
static unsigned long num_banks;
static void __iomem *start_scan_addr;

/*
 * Here are partition information for all known CC1000 boards.
 * See include/linux/mtd/partitions.h for definition of the mtd_partition
 * structure.
 *
 * The *_max_flash_size is the maximum possible mapped flash size which
 * is not necessarily the actual flash size.  It must correspond to the
 * value specified in the mapping definition defined by the
 * "struct map_desc *_io_desc" for the corresponding machine.
 */

#ifdef CONFIG_MTD_PARTITIONS

	/* 4MB flash */
static struct mtd_partition cc1000_4mb_partitions[] = {
	/*
	 * partition definition for first flash bank
	 * also ref. to "drivers/char/flash_config.c"
	 */
	{
	  .name	  = "rootfs",		/* Root file system (ext2)*/
	  .offset = 0x00000000,
	  .size	  = 0x00300000,		/* 3 MB */
/*	  .mask_flags = MTD_WRITEABLE,	/ * force read-only */
	},
	{
	  .name	  = "u-boot",		/* U-Boot */
	  .offset = 0x00300000,
	  .size	  = 0x00040000,		/* 256 KB */
/*	  .mask_flags = MTD_WRITEABLE,	/ * force read-only */
	},
	{
	  .name	  = "kernel",
	  .offset = 0x00340000,		/* Kernel */
	  .size	  = 0x000c0000,		/* 768 KB */
/*	  .mask_flags = MTD_WRITEABLE,	/ * force read-only */
	},
};

	/* 8MB flash */
static struct mtd_partition cc1000_8mb_partitions[] = {
	/*
	 * partition definition for first flash bank
	 * also ref. to "drivers/char/flash_config.c"
	 */
	{
	  .name	  = "data",		/* Data file system (jffs2)*/
	  .offset = 0x00000000,
	  .size	  = 0x00400000,		/* 4 MB */
/*	  .mask_flags = MTD_WRITEABLE,	/ * force read-only */
	},
	{
	  .name	  = "rootfs",		/* Root file system (ext2)*/
	  .offset = 0x00400000,
	  .size	  = 0x00300000,		/* 3 MB */
/*	  .mask_flags = MTD_WRITEABLE,	/ * force read-only */
	},
	{
	  .name	  = "u-boot",		/* U-Boot */
	  .offset = 0x00700000,
	  .size	  = 0x00040000,		/* 256 KB */
/*	  .mask_flags = MTD_WRITEABLE,	/ * force read-only */
	},
	{
	  .name	  = "kernel",
	  .offset = 0x00740000,		/* Kernel */
	  .size	  = 0x000c0000,		/* 768 KB */
/*	  .mask_flags = MTD_WRITEABLE,	/ * force read-only */
	},
};

#endif	/* CONFIG_MTD_PARTITIONS */

int __init init_cc1000_mtd(void)
{
	int idx = 0, ret = 0;
	unsigned long flash_addr, flash_size, mtd_size = 0;
	/* pointer to CC1000 board info data */
	bd_t *bd = (bd_t *)__res;
#ifdef CONFIG_MTD_CMDLINE_PARTS
	int n;
	char mtdid[4];
	const char *part_probes[] = { "cmdlinepart", NULL };
#endif

	flash_addr = bd->bi_flashstart;
	flash_size = bd->bi_flashsize;

	/* request maximum flash size address space */
	start_scan_addr = ioremap(flash_addr, flash_size);
	if (!start_scan_addr) {
		printk(KERN_WARNING "%s: Failed to ioremap address:0x%lx\n",
			__FUNCTION__, flash_addr);
		return -EIO;
	}

	for (idx = 0 ; idx < FLASH_BANK_MAX ; idx++) {
		if (mtd_size >= flash_size)
			break;

		debugk ("%s: chip probing count %d\n", __FUNCTION__, idx);

		map_banks[idx] = (struct map_info *)kmalloc(sizeof(struct map_info),
							GFP_KERNEL);
		if (map_banks[idx] == NULL) {
			ret = -ENOMEM;
			/* FIXME: What if some MTD devices were probed already? */
			goto error_mem;
		}

		memset((void *)map_banks[idx], 0, sizeof(struct map_info));
		map_banks[idx]->name = (char *)kmalloc(16, GFP_KERNEL);

		if (map_banks[idx]->name == NULL) {
			ret = -ENOMEM;
			/* FIXME: What if some MTD devices were probed already? */
			goto error_mem;
		}
		memset((void *)map_banks[idx]->name, 0, 16);
		sprintf(map_banks[idx]->name, "CC1000-%d", idx);

		map_banks[idx]->size	   = flash_size;
		map_banks[idx]->bankwidth  = 2;

		simple_map_init(map_banks[idx]);

		map_banks[idx]->virt = start_scan_addr;
		map_banks[idx]->phys = flash_addr;
		/* FIXME: This looks utterly bogus, but I'm trying to
		   preserve the behaviour of the original (shown here)...

		map_banks[idx]->map_priv_1 =
		start_scan_addr + ((idx > 0) ?
		(mtd_banks[idx-1] ? mtd_banks[idx-1]->size : 0) : 0);
		*/

		if (idx && mtd_banks[idx-1]) {
			map_banks[idx]->virt += mtd_banks[idx-1]->size;
			map_banks[idx]->phys += mtd_banks[idx-1]->size;
		}

		/* start to probe flash chips */
		mtd_banks[idx] = do_map_probe("cfi_probe", map_banks[idx]);

		if (mtd_banks[idx]) {
			mtd_banks[idx]->owner = THIS_MODULE;
			mtd_size += mtd_banks[idx]->size;
			num_banks++;

			debugk (KERN_INFO "%s: bank %d, name: %s, size: %d bytes\n",
				__FUNCTION__,
				num_banks,
				mtd_banks[idx]->name,
				mtd_banks[idx]->size);
		}
	}

	/* no supported flash chips found */
	if (!num_banks) {
		printk (KERN_NOTICE "CC1000: No supported flash chips found!\n");
		ret = -ENXIO;
		goto error_mem;
	}

#ifdef CONFIG_MTD_PARTITIONS
	/*
	 * Select static partition definitions
	 */
	if (flash_size == 0x00400000) {
		part_banks[0].mtd_part	= cc1000_4mb_partitions;
		part_banks[0].type	= "static image";
		part_banks[0].nums	= ARRAY_SIZE(cc1000_4mb_partitions);
	} else {
		part_banks[0].mtd_part	= cc1000_8mb_partitions;
		part_banks[0].type	= "static image";
		part_banks[0].nums	= ARRAY_SIZE(cc1000_8mb_partitions);
	}

	for(idx = 0; idx < num_banks ; idx++) {
#ifdef CONFIG_MTD_CMDLINE_PARTS
		sprintf(mtdid, "%d", idx);
		n = parse_mtd_partitions(mtd_banks[idx],
					 part_probes,
					 &part_banks[idx].mtd_part,
					 0);
		debugk ("%s: %d command line partitions on bank %s\n",
			__FUNCTION__, n, mtdid);
		if (n > 0) {
			part_banks[idx].type = "command line";
			part_banks[idx].nums = n;
		}
#endif	/* CONFIG_MTD_CMDLINE_PARTS */

		if (part_banks[idx].nums == 0) {
		    printk (KERN_NOTICE
			"CC1000 flash bank %d: no partition info available, "
			"registering whole device\n",
			idx);
		    add_mtd_device(mtd_banks[idx]);
		} else {
		    printk (KERN_NOTICE
			"CC1000 flash bank %d: Using %s partition definition\n",
			idx,
			part_banks[idx].type);
		    add_mtd_partitions (mtd_banks[idx],
					part_banks[idx].mtd_part,
					part_banks[idx].nums);
		}
	}
#else	/* ! CONFIG_MTD_PARTITIONS */
	printk (KERN_NOTICE "CC1000 flash: registering %d flash banks at once\n",
		num_banks);

	for(idx = 0 ; idx < num_banks ; idx++) {
		add_mtd_device(mtd_banks[idx]);
	}
#endif	/* CONFIG_MTD_PARTITIONS */

	return 0;
error_mem:
	for (idx = 0 ; idx < FLASH_BANK_MAX ; idx++) {
		if (map_banks[idx] != NULL) {
			if (map_banks[idx]->name != NULL) {
				kfree(map_banks[idx]->name);
				map_banks[idx]->name = NULL;
			}
			kfree(map_banks[idx]);
			map_banks[idx] = NULL;
		}
	}

	iounmap(start_scan_addr);

	return ret;
}

static void __exit cleanup_cc1000_mtd(void)
{
	unsigned int idx = 0;
	for(idx = 0 ; idx < num_banks ; idx++) {
		/* destroy mtd_info previously allocated */
		if (mtd_banks[idx]) {
			del_mtd_partitions(mtd_banks[idx]);
			map_destroy(mtd_banks[idx]);
		}
		/* release map_info not used anymore */
		kfree(map_banks[idx]->name);
		kfree(map_banks[idx]);
	}

	if (start_scan_addr) {
		iounmap(start_scan_addr);
		start_scan_addr = 0;
	}
}

module_init(init_cc1000_mtd);
module_exit(cleanup_cc1000_mtd);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Kirk Lee <kirk@hpc.ee.ntu.edu.tw>");
MODULE_DESCRIPTION("MTD map driver for CC1000 boards");
