/*
 * Copyright(C) 2005
 * Heiko Schocher, DENX Software Engineering, <hs@denx.de>
 *
 * This code is GPLed
 * 
 */

/*
 * IDS8247 uses 1 x 28F640J3
 * for a total of 8 MB flash in one bank
 *
 */
#include <linux/config.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <asm/io.h>

#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>


#if 0	/* Debugging turned off */
# define debugk(fmt,args...)	printk(fmt ,##args)
#else
# define debugk(fmt,args...)
#endif

#define FLASH_BANK_MAX 1

/* trivial struct to describe partition information */
struct mtd_part_def
{
	int nums;
	unsigned char *type;
	struct mtd_partition* mtd_part;
};

static struct mtd_info* mtd_banks[FLASH_BANK_MAX];
static struct map_info* map_banks[FLASH_BANK_MAX];
static struct mtd_part_def part_banks[FLASH_BANK_MAX];
static unsigned long num_banks;
static unsigned long start_scan_addr;

static map_word ids8247_read8(struct map_info *map, unsigned long ofs)
{
        map_word val;
        val.x[0] = *((__u8 *)(map->map_priv_1 + ofs));
        return val;
}

static void *flash_memcpy_fromio(void *to, void *from, ssize_t len)
{
        void *ptr = to;
        while(len--) {
                *(char *)to++ = *(char *)from++;
        }
        return  ptr;
}
static void ids8247_copy_from(struct map_info *map,
                       void *to, unsigned long from, ssize_t len)
{
        /* workaround for misaligned memory access: use byte copy. */
        flash_memcpy_fromio(to, (void *)(map->map_priv_1 + from), len);
}

static void ids8247_write8(struct map_info *map, const map_word d, unsigned long adr)
{
        *((__u8 *)(map->map_priv_1 + adr)) = d.x[0];
}

static void ids8247_copy_to(struct map_info *map,
                     unsigned long to, const void *from, ssize_t len)
{
        memcpy_toio((void *)(map->map_priv_1 + to), from, len);
}

/*
 * The following defines the partition layout of IDS8247 boards.
 *
 * See include/linux/mtd/partitions.h for definition of the
 * mtd_partition structure.
 */

#ifdef CONFIG_MTD_PARTITIONS

/* partition definition for first flash bank
 * also ref. to "drivers/char/flash_config.c" 
 */

static struct mtd_partition ids8247_partitions[] = {
    {
        name:   "kernel1",                      /* Linux kernel image   */
        offset: 0,
        size:   0x100000,                       /*  1MiB                */
/*      mask_flags: MTD_WRITEABLE,              / * force read-only     */
    },
    {
        name:   "kernel2",                      /* Linux kernel image1  */
        offset: 0x100000,
        size:   0x100000,                       /*  1MiB                */
/*      mask_flags: MTD_WRITEABLE,              / * force read-only     */
    },
    {
        name:   "cramfs1",                      /* Rootfs 1             */
        offset: 0x200000,
        size:   0x280000,                       /*  2.5 MiB             */
    },
    {
        name:   "cramfs2",                      /* Rootfs 2             */
        offset: 0x480000,
        size:   0x280000,                       /*  2.5 MiB             */
    },
    {
        name:   "u-boot",                       /* U-Boot               */
        offset: 0x700000,
        size:   0x060000,                       /*  256kiB              */
    },
    {
        name:   "env",                          /* U-Boot Environment   */
        offset: 0x760000,
        size:   0x020000,                       /*  256kiB              */
    }
};

#endif	/* CONFIG_MTD_PARTITIONS */

#define NB_OF(x)  (sizeof(x)/sizeof(x[0]))

int __init init_ids8247_mtd(void)
{
	int idx = 0, ret = 0;
	unsigned long flash_addr, flash_size, mtd_size = 0;
	/* pointer to IDS8247 board info data */
	bd_t *bd = (bd_t *)__res;
#ifdef CONFIG_MTD_CMDLINE_PARTS
	int n;
	char mtdid[4];
	const char *part_probes[] = { "cmdlinepart", NULL };
#endif

	flash_addr = bd->bi_flashstart;
	flash_size = bd->bi_flashsize;

	/* request maximum flash size address space */
	start_scan_addr = (unsigned long)ioremap(flash_addr, flash_size);
	if (!start_scan_addr) {
		printk("%s: Failed to ioremap address: 0x%lx\n",
			__FUNCTION__, flash_addr);
		return -EIO;
	}

	for(idx = 0 ; idx < FLASH_BANK_MAX ; idx++) {
		if (mtd_size >= flash_size)
			break;
		
		debugk ("%s: chip probing count %d\n", __FUNCTION__, idx);
		
		map_banks[idx] = (struct map_info *)kmalloc(sizeof(struct map_info),
							GFP_KERNEL);
		if (map_banks[idx] == NULL) {
			ret = -ENOMEM;
			goto error_mem;
		}
		memset((void *)map_banks[idx], 0, sizeof(struct map_info));
		map_banks[idx]->name = (char *)kmalloc(16, GFP_KERNEL);
		if (map_banks[idx]->name == NULL) {
			ret = -ENOMEM;
			goto error_mem;
		}
		memset((void *)map_banks[idx]->name, 0, 16);
		
		sprintf(map_banks[idx]->name, "IDS8247-%d", idx);
		map_banks[idx]->size      = flash_size;
		map_banks[idx]->bankwidth = 1;
                map_banks[idx]->read      = ids8247_read8;
                map_banks[idx]->copy_from = ids8247_copy_from;
                map_banks[idx]->write     = ids8247_write8;
                map_banks[idx]->copy_to   = ids8247_copy_to;
                map_banks[idx]->map_priv_1=
                        start_scan_addr + ((idx > 0) ?
                        (mtd_banks[idx-1] ? mtd_banks[idx-1]->size : 0) : 0);

		/* start to probe flash chips */
		mtd_banks[idx] = do_map_probe("cfi_probe", map_banks[idx]);
		if (mtd_banks[idx]) {
			mtd_banks[idx]->owner = THIS_MODULE;
			mtd_size += mtd_banks[idx]->size;
			num_banks++;
			debugk ("%s: bank %ld, name: %s, size: %d bytes \n",
				__FUNCTION__,
				num_banks, 
				mtd_banks[idx]->name,
				mtd_banks[idx]->size);
		}
	}

	/* no supported flash chips found */
	if (!num_banks) {
		printk("IDS8247: No supported flash chips found!\n");
		ret = -ENXIO;
		goto error_mem;
	}

#ifdef CONFIG_MTD_PARTITIONS
	/*
	 * Select static partition definitions
	 */
	part_banks[0].mtd_part	= ids8247_partitions;
	part_banks[0].type	= "static image";
	part_banks[0].nums	= NB_OF(ids8247_partitions);

	for(idx = 0; idx < num_banks ; idx++) {
#ifdef CONFIG_MTD_CMDLINE_PARTS
		sprintf(mtdid, "%d", idx);
		n = parse_mtd_partitions(mtd_banks[idx],
					 part_probes,
					 &part_banks[idx].mtd_part,
					 0);
		debugk ("%s: %d command line partitions on bank %s\n",
			__FUNCTION__, n, mtdid);
		if (n > 0) {
			part_banks[idx].type = "command line";
			part_banks[idx].nums = n;
		}
#endif	/* CONFIG_MTD_CMDLINE_PARTS */

		if (part_banks[idx].nums == 0) {
		    printk (KERN_NOTICE
			    "IDS8247 flash bank %d: no partition info "
			    "available, registering whole device\n", idx);
		    add_mtd_device(mtd_banks[idx]);
		} else {
		    printk (KERN_NOTICE 
			    "IDS8247 flash bank %d: Using %s partition "
			    "definition\n", idx, part_banks[idx].type);
		    add_mtd_partitions (mtd_banks[idx],
					part_banks[idx].mtd_part, 
					part_banks[idx].nums);
		}
	}
#else	/* ! CONFIG_MTD_PARTITIONS */
	printk (KERN_NOTICE "IDS8247 flash: registering %d flash banks "
	        "at once\n", num_banks);

	for(idx = 0 ; idx < num_banks ; idx++) {
		add_mtd_device(mtd_banks[idx]);
	}
#endif	/* CONFIG_MTD_PARTITIONS */

	return 0;
error_mem:
	for (idx = 0 ; idx < FLASH_BANK_MAX ; idx++) {
		if (map_banks[idx] != NULL) {
			if (map_banks[idx]->name != NULL) {
				kfree(map_banks[idx]->name);
				map_banks[idx]->name = NULL;
			}
			kfree(map_banks[idx]);
			map_banks[idx] = NULL;
		}
	}

	iounmap((void *)start_scan_addr);

	return ret;
}

static void __exit cleanup_ids8247_mtd(void)
{
	unsigned int idx = 0;
	for(idx = 0 ; idx < num_banks ; idx++) {
		/* destroy mtd_info previously allocated */
		if (mtd_banks[idx]) {
			del_mtd_partitions(mtd_banks[idx]);
			map_destroy(mtd_banks[idx]);
		}
		/* release map_info not used anymore */
		kfree(map_banks[idx]->name);
		kfree(map_banks[idx]);
	}
	if (start_scan_addr) {
		iounmap((void *)start_scan_addr);
		start_scan_addr = 0;
	}
}

module_init(init_ids8247_mtd);
module_exit(cleanup_ids8247_mtd);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Heiko Schocher <hs@denx.de>");
MODULE_DESCRIPTION("MTD map driver for IDS8247 boards");
