/*
 * drivers/mtd/maps/mcc200.c
 *
 * MTD mapping for the MCC200 board.
 *
 */
#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <asm/io.h>
#include <asm/mpc5xxx.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>

#ifdef CONFIG_PRS200
# define BOARD_NAME_L	"prs200"
# define BOARD_NAME_U	"PRS200"
#else
# define BOARD_NAME_L	"mcc200"
# define BOARD_NAME_U	"MCC200"
#endif

#ifdef CONFIG_UBOOT
#include <asm/ppcboot.h>

extern unsigned char __res[];
bd_t *bd = (bd_t *)__res;

#define FLASH_ADDR  (bd->bi_flashstart)
#define FLASH_SIZE  (bd->bi_flashsize)
#else
#define FLASH_ADDR  CONFIG_PPC_5xxx_FLASH_ADDR
#define FLASH_SIZE  CONFIG_PPC_5xxx_FLASH_SIZE
#endif

#ifndef CONFIG_MTD_PARTITIONS
#error need to define CONFIG_MTD_PARTITIONS
#endif
static struct mtd_info *mymtd;

#define BANK_SELECT_BIT		(1 << 6)

static inline void set_high_address_bit(int set)
{
	if (set)
		*(volatile int *)MPC5xxx_CS0CFG |= BANK_SELECT_BIT;
	else
		*(volatile int *)MPC5xxx_CS0CFG &= ~BANK_SELECT_BIT;
}

static void *flash_memcpy_fromio(void *to, void *from, ssize_t len)
{
	void *ptr = to;
	while(len--) {
		*(char *)to++ = *(char *)from++;
	}
	return  ptr;
}

static map_word mcc200_map_read(struct map_info *map, unsigned long ofs)
{
	set_high_address_bit(ofs >= (32 << 20));
	return inline_map_read(map, ofs);
}

static void mcc200_map_write(struct map_info *map, const map_word datum, unsigned long ofs)
{
	set_high_address_bit(ofs >= (32 << 20));
	inline_map_write(map, datum, ofs);
}

static void mcc200_map_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
	if (map->cached)
		memcpy(to, (char *)map->cached + from, len);
	else {
		if (from < (32 << 20)) {
			ssize_t bytes = min(len, (ssize_t)((32 << 20) - from));

			set_high_address_bit(0);
#if 0
			memcpy_fromio(to, map->virt + from, bytes);
#else
			flash_memcpy_fromio(to, map->virt + from, bytes);
#endif
			from += bytes;
			len -= bytes;
		}
		if (len) {
			set_high_address_bit(1);
			memcpy_fromio(to, map->virt + from, len);
		}
	}
}

static void mcc200_map_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len)
{
	if (to < (32 << 20)) {
		ssize_t bytes = min(len, (ssize_t)((32 << 20) - to));
			
		set_high_address_bit(0);
		memcpy_toio(map->virt + to, from, bytes);
		to += bytes;
		len -= bytes;
	}
	if (len) {
		set_high_address_bit(1);
		memcpy_toio(map->virt + to, from, len);
	}
}


struct map_info mcc200_map = {
	name:		BOARD_NAME_L,
	bankwidth:	4,
	read:		mcc200_map_read,
	write:		mcc200_map_write,
	copy_from:	mcc200_map_copy_from,
	copy_to:	mcc200_map_copy_to,
};

static struct mtd_partition mcc200_partition[] =
{
	{
		name:		"U-Boot",
		size:		(768 << 10),		/*  0.75 MiB */
		offset:		0
	},
	{
		name:		"kernel",
                size:           (1 << 20),		/*  1.00 MiB */
		offset:		MTDPART_OFS_APPEND
	},
	{
		name:		"jffs1",
                size:           0x1E40000,		/* 30.25 MiB */
		offset:		MTDPART_OFS_APPEND
	},
	{
		name:		"jffs2",
		size:		(32 << 20),		/* 32.00 MiB */
		offset:		MTDPART_OFS_APPEND
	},
};

#define NB_OF(x)	(sizeof(x)/sizeof(x[0]))

static int __init init_mcc200 (void)
{
#if defined(CONFIG_MTD_CMDLINE_PARTS)
	const char *part_probes[] = { "cmdlinepart", NULL };
	int parts_nb;
	struct mtd_partition* mtd_parts;
#endif

	printk (KERN_ALERT BOARD_NAME_U " flash device: %lx at %lx\n", FLASH_SIZE,
		FLASH_ADDR);
	mcc200_map.phys = FLASH_ADDR;
	mcc200_map.virt = (void *)mcc200_map.phys;
	mcc200_map.size = FLASH_SIZE;

	mymtd = do_map_probe ("cfi_probe", &mcc200_map);
	if (!mymtd) {
		printk ("cfi_probe failed\n");
		return -ENXIO;
	}

	mymtd->owner = THIS_MODULE;

#if defined(CONFIG_MTD_CMDLINE_PARTS)
	parts_nb = parse_mtd_partitions(mymtd, part_probes, &mtd_parts, 0);
	if (parts_nb > 0) {
		printk (KERN_NOTICE "Using command line partition definition\n");
		add_mtd_partitions(mymtd, mtd_parts, parts_nb);
	} else 
#endif
	{ /* Continuation of the 'else' two lines above */
		printk (KERN_NOTICE "Using static partition definition\n");
		add_mtd_partitions (mymtd, mcc200_partition, NB_OF (mcc200_partition));
	}

	return 0;
}


static void __exit cleanup_mcc200 (void)
{
	if (mymtd) {
		del_mtd_device (mymtd);
		map_destroy (mymtd);
	}
}
module_init(init_mcc200);
module_exit(cleanup_mcc200);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Dmitry Konyshev");
MODULE_DESCRIPTION("MTD map driver for MCC200/PRS200 boards");
