/*
 * drivers/mtd/maps/top5200.c
 *
 * MTD mapping driver for EMK TOP5200 CPU module
 *
 * 2003 (c) Wolfgang Denk, DENX Software Engineering, <wd@denx.de>.
 * 2006 (c) Reinhard Meyer, EMK Elektronik GmbH, <info@emk-elektronik.de>
 * This file is licensed under the terms of the GNU General Public License
 * version 2.  This program is licensed "as is" without any warranty of
 * any kind, whether express or implied.
 */

/*
 * TOP5200 uses different FLASH chips (order option)
 * currently ranging from 2 Megabytes to 64 Megabytes
 * *** but ***
 * since only 16 Megabytes can be directly mapped on a MPC5200
 * the larger FLASH's extra address lines are connected
 * to 2 GPIO Ports
 */

/*
 * ATTENTION:
 * 1. complex mapping MUST be enabled
 * 2. u-boot will never report more than 16MB since no complex mapping is
 * supported in u-boot. However the kernel badly panics if it finds more
 * FLASH than u-boot reported (why we have not investigated). Anyway the
 * u-boot reported value is fixed to prevent this.
 * 3. currently FLASH sizes > 16MB MUST BE partitioned
 */

/*
 * revisions:
 * 2006-08-01 EMK-RM added spinlocking around bank select and access
 */

#include <linux/config.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/io.h>
#ifdef CONFIG_UBOOT
#include <asm/ppcboot.h>
#endif

#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>

#ifndef CONFIG_MTD_COMPLEX_MAPPINGS
#error CONFIG_MTD_COMPLEX_MAPPINGS must be set for this driver to work!!
#endif

#if 0				/* Debugging turned off */
#define debugk(fmt,args...)	printk(fmt ,##args)
#else
#define debugk(fmt,args...)
#endif

#define FLASH_BANK_MAX 1

static unsigned long flash_addr = 0, flash_size = 0;
static struct mtd_info *flash_mtd;
static struct map_info *flash_map;

#ifdef CONFIG_HMI5200_MTD_SRAM
static unsigned long sram_addr = 0, sram_size = 0;
static struct mtd_info *sram_mtd;
static struct map_info *sram_map;
#endif

/* spinlock to atomize bank selection and access to FLASH */
static spinlock_t lock;

#include <asm/mpc5xxx.h>
#define	BANKBIT0 0x60
#define BANKBIT1 0x70
#define	BANKSETBIT 0x34
#define	BANKCLRBIT 0x24

/*
 * this function drives the topmost 2 address lines to the FLASH on TOP5200
 * selecting one out of four 16 MB banks
 */
static void select_bank (unsigned bank)
{
	if (bank & 1)
		out_be32 ((u32 *) (MPC5xxx_GPT + BANKBIT0), BANKSETBIT);
	else
		out_be32 ((u32 *) (MPC5xxx_GPT + BANKBIT0), BANKCLRBIT);
	if (bank & 2)
		out_be32 ((u32 *) (MPC5xxx_GPT + BANKBIT1), BANKSETBIT);
	else
		out_be32 ((u32 *) (MPC5xxx_GPT + BANKBIT1), BANKCLRBIT);
}

/*
 * the following 4 methods of accessing the FLASH all use a logical adress
 * in the 0..64MB range.
 * the real start address of the FLASH must be in the map_priv_1 .
 * selecting the bank and accessing the FLASH must be done with IRQs
 * disabled.
 */

/*
 * read one byte of data from FLASH
 */
static map_word flash_banked_read8 (struct map_info *map, unsigned long ofs)
{
	map_word val;
	unsigned long flags;

	spin_lock_irqsave (&lock, flags);
	select_bank (ofs >> 24);
	val.x[0] = *((__u8 *) (map->map_priv_1 + (ofs & 0x00ffffff)));
	spin_unlock_irqrestore (&lock, flags);
	return val;
}

/*
 * read a string of bytes from FLASH
 * note: a bank boundary MUST NOT be crossed
 */
static void flash_banked_copy_from (struct map_info *map,
				    void *to, unsigned long from, ssize_t len)
{
	unsigned long flags;

	spin_lock_irqsave (&lock, flags);
	select_bank (from >> 24);
	memcpy_fromio (to, (void *) (map->map_priv_1 + (from & 0x00ffffff)),
		       len);
	spin_unlock_irqrestore (&lock, flags);
}

/*
 * write one byte to FLASH
 */
static void flash_banked_write8 (struct map_info *map, const map_word d,
				 unsigned long adr)
{
	unsigned long flags;

	spin_lock_irqsave (&lock, flags);
	select_bank (adr >> 24);
	*((__u8 *) (map->map_priv_1 + (adr & 0x00ffffff))) = d.x[0];
	spin_unlock_irqrestore (&lock, flags);
}

/*
 * write a string of bytes to FLASH
 * note: a bank boundary MUST NOT be crossed
 */
static void flash_banked_copy_to (struct map_info *map,
				  unsigned long to, const void *from,
				  ssize_t len)
{
	unsigned long flags;

	spin_lock_irqsave (&lock, flags);
	select_bank (to >> 24);
	memcpy_toio ((void *) (map->map_priv_1 + (to & 0x00ffffff)), from,
		     len);
	spin_unlock_irqrestore (&lock, flags);
}

#ifdef CONFIG_HMI5200_MTD_SRAM
/*
 * read one byte of data from SRAM
 */
static map_word sram_banked_read8 (struct map_info *map, unsigned long ofs)
{
	map_word val;

	val.x[0] = *((__u8 *) (map->map_priv_1 + ofs));
	return val;
}

/*
 * read a string of bytes from SRAM
 */
static void sram_banked_copy_from (struct map_info *map,
				   void *to, unsigned long from, ssize_t len)
{
	memcpy_fromio (to, (void *) (map->map_priv_1 + from), len);
}

/*
 * write one byte to SRAM
 */
static void sram_banked_write8 (struct map_info *map, const map_word d,
				unsigned long adr)
{
	*((__u8 *) (map->map_priv_1 + adr)) = d.x[0];
}

/*
 * write a string of bytes to SRAM
 */
static void sram_banked_copy_to (struct map_info *map,
				 unsigned long to, const void *from,
				 ssize_t len)
{
	memcpy_toio ((void *) (map->map_priv_1 + to), from, len);
}
#endif

/*
 * The following defines the partition layout of Top5200 boards.
 *
 * See include/linux/mtd/partitions.h for definition of the
 * mtd_partition structure.
 *
 * IMPORTANT: no partitions must cross any 16 Megabyte boundary !!!
 * otherwise adapt flash_banked_copy_*() accordingly to handle bank switches
 * mid-copy
 *
 * note: TOP5200 with upto 32MB of FLASH have 64k sectors, above it is 128k
 * sectors. It might be wise to have partitions observe 128k boundaries
 */

#ifdef CONFIG_MTD_PARTITIONS

#ifdef CONFIG_LMS5200
/* special customers baseboard with FPGA-video and very large boot picture */
/* partition definition for first (and only) flash bank */
static struct mtd_partition flash_partitions[] = {
      {name: "UBoot", offset: 0x00000000, size: 0x00080000, mask_flags:MTD_WRITEABLE,},
      {name: "FPGAimg", offset: 0x00080000, size:0x000c0000,},
      {name: "BOOTbmp", offset: 0x00140000, size:0x000a0000,},
      {name: "I2CTable", offset: 0x001e0000, size:0x00020000,},
      {name: "kernel", offset: 0x00200000, size:0x00200000,},
      {name: "storage1", offset: 0x00400000, size:0x00c00000,},
      {name: "storage2", offset: 0x01000000, size:0x01000000,},
      {name: "storage3", offset: 0x02000000, size:0x01000000,},
      {name: "storage4", offset: 0x03000000, size:0x01000000,},
};
#else
/* all other baseboards */
/* partition definition for first (and only) flash bank */
static struct mtd_partition flash_partitions[] = {
      {name: "UBoot", offset: 0x00000000, size: 0x00080000, mask_flags:MTD_WRITEABLE,},
      {name: "kernel", offset: 0x00080000, size:0x00180000,},
      {name: "root", offset: 0x00200000, size:0x00e00000,},
      {name: "bank1", offset: 0x01000000, size:0x01000000,},
      {name: "bank2", offset: 0x02000000, size:0x01000000,},
      {name: "bank3", offset: 0x03000000, size:0x01000000,},
};
#endif

#ifdef CONFIG_HMI5200_MTD_SRAM
/* partition definition for the SRAM in M48T37 */
/* - use depreciated, handled by RTC driver now */
static struct mtd_partition sram_partitions[] = {
      {name: "sram", offset: 0x00000000, size:0x00007ff0,},
};
#endif

#endif /* CONFIG_MTD_PARTITIONS */

#define NB_OF(x)  (sizeof(x)/sizeof(x[0]))

static int __init init_flash (void)
{
	int ret = 0;

#ifdef CONFIG_UBOOT
	/* pointer to Top5200 board info data */
	extern unsigned char __res[];
	bd_t *bd = (bd_t *) __res;

	flash_addr = bd->bi_flashstart;
	flash_size = bd->bi_flashsize;
	debugk ("%s: UBOOT says: %lx %lx\n", __FUNCTION__, flash_addr,
		flash_size);
	/* fix u-boot report for FLASH >= 16MB */
	if (flash_size == (1 << 24))
		flash_size = (1 << 26);
#else
	flash_addr = CONFIG_PPC_5xxx_FLASH_ADDR;
	flash_size = CONFIG_PPC_5xxx_FLASH_SIZE;
	debugk ("%s: CONFIG says: %lx %lx\n", __FUNCTION__, flash_addr,
		flash_size);
#endif

	spin_lock_init (&lock);

	/* probe the FLASH */
	debugk ("%s: probing FLASH\n", __FUNCTION__);

	/* alloc structure for FLASH mapping */
	flash_map =
		(struct map_info *) kmalloc (sizeof (struct map_info),
					     GFP_KERNEL);
	if (flash_map == NULL) {
		ret = -ENOMEM;
		goto error_mem;
	}

	/* fill structure */
	memset ((void *) flash_map, 0, sizeof (struct map_info));
	flash_map->name = "TOP5200-flash";
	flash_map->size = flash_size;
	flash_map->bankwidth = 1;
	flash_map->read = flash_banked_read8;
	flash_map->copy_from = flash_banked_copy_from;
	flash_map->write = flash_banked_write8;
	flash_map->copy_to = flash_banked_copy_to;
	flash_map->map_priv_1 = flash_addr;

	/* probe the FLASH */
	flash_mtd = do_map_probe ("cfi_probe", flash_map);
	/* found it? */
	if (flash_mtd) {
		flash_mtd->owner = THIS_MODULE;
		debugk ("TOP5200 %s, size: %d bytes\n", flash_mtd->name,
			flash_mtd->size);
	} else {
		printk ("Top5200: No supported flash chip found!\n");
		ret = -ENXIO;
		goto error_mem;
	}

#ifdef CONFIG_MTD_PARTITIONS
	/* Select static partition definitions */
	if (NB_OF (flash_partitions) == 0) {
		printk (KERN_NOTICE
			"TOP5200 FLASH: no partition info available, registering whole device\n");
		add_mtd_device (flash_mtd);
	} else {
		printk (KERN_NOTICE
			"TOP5200 FLASH: Using static partition definition\n");
		add_mtd_partitions (flash_mtd, flash_partitions,
				    NB_OF (flash_partitions));
	}
#else  /* ! CONFIG_MTD_PARTITIONS */
	printk (KERN_NOTICE "TOP5200 FLASH: registering\n");
	add_mtd_device (flash_mtd);
#endif /* CONFIG_MTD_PARTITIONS */

	return 0;
      error_mem:
	if (flash_map != NULL) {
		kfree (flash_map);
		flash_map = NULL;
	}

	iounmap ((void *) flash_addr);

	return ret;
}

#ifdef CONFIG_HMI5200_MTD_SRAM
static int __init init_sram (void)
{
	int ret = 0;

	sram_addr = 0xf0010000;
	sram_size = 0x7ff0;

	/* probe the SRAM */
	debugk ("%s: probing SRAM\n", __FUNCTION__);

	/* alloc structure for SRAM mapping */
	sram_map =
		(struct map_info *) kmalloc (sizeof (struct map_info),
					     GFP_KERNEL);
	if (sram_map == NULL) {
		ret = -ENOMEM;
		goto error_mem;
	}

	/* fill structure */
	memset ((void *) sram_map, 0, sizeof (struct map_info));
	sram_map->name = "HMI5200-SRAM ";
	sram_map->size = sram_size;
	sram_map->bankwidth = 1;
	sram_map->read = sram_banked_read8;
	sram_map->copy_from = sram_banked_copy_from;
	sram_map->write = sram_banked_write8;
	sram_map->copy_to = sram_banked_copy_to;
	sram_map->map_priv_1 = sram_addr;

	/* probe the FLASH */
	sram_mtd = do_map_probe ("map_ram", sram_map);
	/* found it? */
	if (sram_mtd) {
		sram_mtd->owner = THIS_MODULE;
		debugk ("TOP5200 %s, size: %d bytes\n", sram_mtd->name,
			sram_mtd->size);
	} else {
		printk ("Top5200: No supported flash chip found!\n");
		ret = -ENXIO;
		goto error_mem;
	}

#ifdef CONFIG_MTD_PARTITIONS
	/* Select static partition definitions */
	if (NB_OF (sram_partitions) == 0) {
		printk (KERN_NOTICE
			"TOP5200 SRAM: no partition info available, registering whole device\n");
		add_mtd_device (sram_mtd);
	} else {
		printk (KERN_NOTICE
			"TOP5200 SRAM: Using static partition definition\n");
		add_mtd_partitions (sram_mtd, sram_partitions,
				    NB_OF (sram_partitions));
	}
#else  /* ! CONFIG_MTD_PARTITIONS */
	printk (KERN_NOTICE "TOP5200 SRAM: registering\n");
	add_mtd_device (sram_mtd);
#endif /* CONFIG_MTD_PARTITIONS */

	return 0;
      error_mem:
	if (sram_map != NULL) {
		kfree (sram_map);
		sram_map = NULL;
	}

	iounmap ((void *) sram_addr);

	return ret;
}
#endif

static void __exit cleanup_top5200_mtd (void)
{
	/* destroy mtd_info previously allocated */
	if (flash_mtd) {
#ifdef CONFIG_MTD_PARTITIONS
		del_mtd_partitions (flash_mtd);
#endif
		map_destroy (flash_mtd);
	}
	/* release map_info not used anymore */
	kfree (flash_map);
	if (flash_addr) {
		iounmap ((void *) flash_addr);
		flash_addr = 0;
	}
#ifdef CONFIG_HMI5200_MTD_SRAM
	/* destroy mtd_info previously allocated */
	if (sram_mtd) {
#ifdef CONFIG_MTD_PARTITIONS
		del_mtd_partitions (sram_mtd);
#endif
		map_destroy (sram_mtd);
	}
	/* release map_info not used anymore */
	kfree (sram_map);
	if (sram_addr) {
		iounmap ((void *) sram_addr);
		sram_addr = 0;
	}
#endif
}

static int __init init_top5200_mtd (void)
{
	int res = init_flash ();

	if (res)
		return res;
#ifdef CONFIG_HMI5200_MTD_SRAM
	res = init_sram ();
	if (res)
		return res;
#endif
	return 0;
}

module_init (init_top5200_mtd);
module_exit (cleanup_top5200_mtd);

MODULE_LICENSE ("GPL");
MODULE_AUTHOR ("Reinhard Meyer <info@emk-elektronik.de>");
MODULE_DESCRIPTION ("MTD map driver for Top5200 board");
