/*
 *  drivers/mtd/nand/ids8247.c
 *
 *  Copyright (C) 2005 Heiko Schocher, DENX Software Engineering, <hs@denx.de>
 *
 *  Derived from drivers/mtd/nand/edb7312.c
 *       Copyright (C) 2002 Marius Grger (mag@sysgo.de)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *  Overview:
 *   This is a device driver for the NAND flash device found on the
 *   IDS8247 board which utilizes the Samsung K9F5608 part. This is
 *   a 256Mibit (32MiB x 8 bits) NAND flash device.
 */

#include <linux/slab.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/partitions.h>
#include <asm/io.h>
#include <asm/delay.h>
#include <platforms/ids8247.h>

/* handy sizes */
#define SZ_1K                           0x00000400



/*
 * MTD structure for IDS8247 board
 */
static struct mtd_info *ids8247_mtd = NULL;

/* Internal buffers. Page buffer and oob buffer for one block*/
static u_char data_buf[512 + 16];
static u_char oob_buf[16 * 32];

/*
 * Module stuff
 */

static int ids8247_fio_pbase = CFG_NAND_PADDR;

#ifdef MODULE
MODULE_PARM(ids8247_fio_pbase, "i");

__setup("ids8247_fio_pbase=", ids8247_fio_pbase);
#endif

#if ((defined CONFIG_MTD_PARTITIONS) || (defined CONFIG_MTD_PARTITIONS_MODULE))
/*
 * Define static partitions for flash device
 */
static struct mtd_partition partition_info[] = {
	{ name: "IDS8247 Nand partition 1",
		  offset: 0,
   	          size: 4*1024*1024 },
	{ name: "IDS8247 Nand partition 2",
		  offset: 4*1024*1024,
   	          size: 28*1024*1024 },
};
#define NUM_PARTITIONS 2
#endif


/*
 *	hardware specific access to control-lines
 */
static void ids8247_hwcontrol(struct mtd_info *mtd, int cmd)
{
	struct nand_chip *this = mtd->priv;

	switch(cmd) {
		case NAND_CTL_SETCLE:
 			*(((volatile __u8 *) this->IO_ADDR_W) + 0xa) = 0; 
			break;
		case NAND_CTL_CLRCLE:
 			*(((volatile __u8 *) this->IO_ADDR_W) + 0x8) = 0; 
			break;
		case NAND_CTL_SETALE:
 			*(((volatile __u8 *) this->IO_ADDR_W) + 0x9) = 0; 
			break;
		case NAND_CTL_CLRALE:
 			*(((volatile __u8 *) this->IO_ADDR_W) + 0x8) = 0; 
			break;
		case NAND_CTL_SETNCE:
 			*(((volatile __u8 *) this->IO_ADDR_W) + 0x8) = 0; 
			break;
		case NAND_CTL_CLRNCE:
 			*(((volatile __u8 *) this->IO_ADDR_W) + 0xc) = 0; 
			break;
	}
}

#ifdef CONFIG_MTD_PARTITIONS
static const char *part_probes[] = { "cmdlinepart", NULL };
#endif

/*
 * Main initialization routine
 */
static int __init ids8247_init (void)
{
	struct nand_chip *this;
	const char *part_type = 0;
	int mtd_parts_nb = 0;
	struct mtd_partition *mtd_parts = 0;
	int ids8247_fio_base;

#ifdef CONFIG_MTD_DEBUG
	printk(  "[ids8247_init]\n");
#endif

	/* Allocate memory for MTD device structure and private data */
	ids8247_mtd = kmalloc(sizeof(struct mtd_info) + sizeof(struct nand_chip),
			    GFP_KERNEL);
	if (!ids8247_mtd) {
		printk("Unable to allocate IDS8247 NAND MTD device structure.\n");
		return -ENOMEM;
	}

	/* map physical address */
	ids8247_fio_base = (unsigned long)ioremap(ids8247_fio_pbase, SZ_1K);
	if(!ids8247_fio_base) {
		printk("ioremap IDS8247 NAND flash failed\n");
		kfree(ids8247_mtd);
		return -EIO;
	}

	/* Get pointer to private data */
	this = (struct nand_chip *) (&ids8247_mtd[1]);

	/* Initialize structures */
	memset((char *) ids8247_mtd, 0, sizeof(struct mtd_info));
	memset((char *) this, 0, sizeof(struct nand_chip));

	/* Link the private data with the MTD structure */
	ids8247_mtd->priv = this;

	/* insert callbacks */
	this->IO_ADDR_R = (void *) ids8247_fio_base;
	this->IO_ADDR_W = (void *) ids8247_fio_base;
	this->hwcontrol = ids8247_hwcontrol;
        this->dev_ready = NULL;
	/* 12 us command delay time */
	this->chip_delay = 12;
	/* ECC mode */
	this->eccmode = NAND_ECC_SOFT;

	/* Set internal data buffer */
	this->data_buf = data_buf;
	this->oob_buf = oob_buf;

	/* Scan to find existence of the device */
	if (nand_scan (ids8247_mtd, 1)) {
		iounmap((void *)ids8247_fio_base);
		kfree (ids8247_mtd);
		return -ENXIO;
	}

#ifdef CONFIG_MTD_PARTITIONS
	ids8247_mtd->name = "ids8247-nand";
	mtd_parts_nb = parse_mtd_partitions(ids8247_mtd, part_probes, &mtd_parts, 0);
	if (mtd_parts_nb > 0)
	  part_type = "command line";
	else
	  mtd_parts_nb = 0;
#endif
	if (mtd_parts_nb == 0)
	{
#if ((defined CONFIG_MTD_PARTITIONS) || (defined CONFIG_MTD_PARTITIONS_MODULE))
		mtd_parts = partition_info;
		mtd_parts_nb = NUM_PARTITIONS;
#endif
		part_type = "static";
	}

	/* Register the partitions */
	printk(KERN_NOTICE "Using %s partition definition\n", part_type);
	add_mtd_partitions(ids8247_mtd, mtd_parts, mtd_parts_nb);

	/* Return happy */
	return 0;
}
module_init(ids8247_init);

/*
 * Clean up routine
 */
static void __exit ids8247_cleanup (void)
{
#ifdef CONFIG_MTD_DEBUG
	printk(  "[ids8247_cleanup]\n");
#endif

	/* Unregister the device */
	del_mtd_device (ids8247_mtd);

	/* Free the MTD device structure */
	kfree (ids8247_mtd);
}
module_exit(ids8247_cleanup);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Heiko Schocher <hs@denx.de>");
MODULE_DESCRIPTION("MTD map driver for IDS8247 board");
