/*
 * drivers/spi/hmi1001-kbd.c
 *
 * HMI1001 keyboard/LED driver for MPC5xxx based board
 *
 * 2005 (c) Wolfgang Denk, DENX Software Engineering, <wd@denx.de>
 *
 * This file is licensed under the terms of the GNU Genral Public License
 * version 2. This program is licensed "as is" without any warranty of
 * any kind, whether expressed or implied.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/string.h>
#include <linux/poll.h>
#include <linux/spinlock.h>
#include <linux/proc_fs.h>
#include <linux/interrupt.h>
#include <asm/mpc5xxx.h>
#include <asm/io.h>
#include <asm/ppcboot.h>

#include <linux/spi.h>
#include <linux/hmi1001-kbd.h>
#include "spi-algo-mpc5xxx.h"

#define SPIDEV_MAJOR	11	/* Raw keyboard device */
#define SPIDEV_NAME		"spi-kbd"
#define SPILED_MINOR	1
#define SPIKBD_MINOR	2

#define TIMER_ID	7  /* Timer_7 */
#define TIMER_IRQ	MPC5xxx_MAIN_IRQ_BASE + 16

	/* comment it if you're not going use it */
#undef DEBUG
#define PROC_SPIKBD_ENABLE	1
#define PROC_SPILED_ENABLE	1

#define PROC_SPIKBD_FILE	"driver/hmi1001-kbd"
#define PROC_SPILED_FILE	"driver/hmi1001-led"

#define	MAX_MSG_LEN	3

#ifdef DEBUG
#define debugk(arg...)	do { printk("%s: ", __FUNCTION__); \
	    printk(arg); } while (0)
#else
#define debugk(arg...)
#endif

	/* forward declarations */

static int spi_kbd_open(struct inode *inode, struct file *file);
static int spi_kbd_close(struct inode *inode, struct file *file);
static ssize_t spi_kbd_read(struct file *file, char *buf,
		size_t count, loff_t *offset);
static ssize_t spi_led_write(struct file *file, const char *buf,
		size_t count, loff_t *offset);
static ssize_t spi_led_read(struct file *file, char *buf,
		size_t count, loff_t *offset);
static int spi_kbd_ioctl(struct inode *inode, struct file *file,
		unsigned int cmd, unsigned long arg);
static unsigned int spi_kbd_poll(struct file *file, poll_table *wait);
static int transfer_bytes(const char *buf_out, char *buf_in,
		unsigned int count);
static int hmi1001_attach_adapter(struct spi_adapter *adap);
static void hmi1001_detach_client(struct spi_client* client);
static int hmi1001_command(struct spi_client *client,
		unsigned int cmd, void *arg);

	/* global variables */

static long spi_key_state	= 0;	/* buffered keyboard state */
static long spi_led_state	= 0;	/* buffered led state */
static int  spi_key_num 	= 8;	/* number of used keys/leds */

static int spi_kbd_opened	= 0;	/* kbd device is opened for read */
static int spi_led_opened	= 0;	/* led device is opened for write */ 
	/* need to read keyboard state */
static atomic_t spi_kbd_intr	= ATOMIC_INIT(0);
static int spi_key_changed	= 0;	/* keyboard state changed */

	/* lock for accessing kbd/led states */
static spinlock_t dev_lock	= SPIN_LOCK_UNLOCKED;
	/* wait queue for keyboard event */
static DECLARE_WAIT_QUEUE_HEAD(spi_key_wait); 
	/* HMI1001 SPI adapter for transferring bytes */
static struct spi_adapter *hmi1001_adapter	= NULL;
	/* HMI1001 SPI driver */
static struct spi_driver hmi1001_driver = {
	.name		= "hmi1001",
	.flags		= SPI_DF_NOTIFY,
	.attach_adapter	= hmi1001_attach_adapter,
	.detach_client	= hmi1001_detach_client,
	.command	= hmi1001_command,
	.owner		= THIS_MODULE,
};
	/* interrupt handling task */
static struct tq_struct spi_kbd_task;
	/* spi-kbd device fops */    
static struct file_operations spi_kbd_fops = {
	.open		= spi_kbd_open,
	.release	= spi_kbd_close,
	.read		= spi_kbd_read,
	.poll		= spi_kbd_poll,
	.ioctl		= spi_kbd_ioctl,
	.owner		= THIS_MODULE,
};
	/* spi-led device fops */
static struct file_operations spi_led_fops = {
	.open		= spi_kbd_open,
	.release	= spi_kbd_close,
	.read		= spi_led_read,
	.write		= spi_led_write,
	.ioctl		= spi_kbd_ioctl,
	.owner		= THIS_MODULE,
};

	/* macroses for device locking */

#define lock_dev()		spin_lock_bh(&dev_lock)
#define unlock_dev()		spin_unlock_bh(&dev_lock)
#define lock_dev_irq(flags)	spin_lock_irqsave(&dev_lock, flags)
#define unlock_dev_irq(flags)	spin_unlock_irqrestore(&dev_lock, flags)

/*
 * get message length to tranfer to the kbd/led device
 */
static inline int msg_len(void)
{
 	return (spi_key_num + 7 ) / 8;
}

/*
 * mask off unused state bits
 */
static inline unsigned long mask_state(unsigned long state)
{
	return state & ~((unsigned long)(-1) >> spi_key_num);
}

/*
 * Perform on-kbd-event handling
 */
void spi_kbd_task_routine(void *data)
{
	int old;

	do {
		old  = atomic_read(&spi_kbd_intr);

		lock_dev();

			/* update keyboard state */
		transfer_bytes((char *)&spi_led_state,
			(char *)&spi_key_state, msg_len());

		spi_key_changed = 1;
			/* notify waiters of key change */
		wake_up_interruptible(&spi_key_wait);  

		unlock_dev();

		/* if any keyboard event happened, then repeat */
	} while (!atomic_sub_and_test(old, &spi_kbd_intr));

	debugk("led_state=0x%08X kbd_state=0x%08X\n",
		spi_led_state, spi_key_state);
}

/*
 * IRQ handler
 */
static void spi_kbd_interrupt(int irq, void *data, struct pt_regs *pt_regs)
{
	struct mpc5xxx_gpt *gpt = (struct mpc5xxx_gpt *)MPC5xxx_GPT + TIMER_ID;
	u32 val;
	
	debugk("interrupt irq=%d\n", irq);
	
	/* clear IC event bit */
	val = in_be32(&gpt->sr);
	val |= 0x0F;
	out_be32(&gpt->sr, val);
	
	if (atomic_inc_return(&spi_kbd_intr) == 1) {

		queue_task(&spi_kbd_task, &tq_immediate);

		mark_bh(IMMEDIATE_BH);
	}
}

/*
 * Perform actions on spi-led/kbd device opening
 */
static int spi_kbd_open(struct inode *inode, struct file *file)
{
	int led = MINOR(inode->i_rdev) == SPILED_MINOR;
	int wr  = file->f_flags & O_ACCMODE;

	debugk("min_maj=%d_%d flags=0x%08X(led=%d, wr=%d)\n",
		MAJOR(inode->i_rdev), MINOR(inode->i_rdev),
		file->f_flags, led, wr);

	if (!led) {
		if (spi_kbd_opened)
			return -EBUSY;
		spi_kbd_opened = 1;
		spi_key_changed = 1; /* allow the first read,
					subsequent calls will succeed
					only if key press/release happens */
	}
	else {
		if (wr) {
			if (spi_led_opened) /* already opened for write */
				return -EBUSY;

			spi_led_opened = 1;
		}

			/* replace fops for led device */
		inode->i_fop = &spi_led_fops;
		file->f_op   = &spi_led_fops;    
	}

	return 0;
}

/*
 * Perform action on spi-kbd/led device closing
 */
static int spi_kbd_close(struct inode *inode, struct file *file)
{
	int led = MINOR(inode->i_rdev) == SPILED_MINOR;
	int wr  = file->f_flags & O_ACCMODE; 

	debugk("led = %d, wr = %d\n", led, wr);

	if (!led)
		spi_kbd_opened = 0;
	else if (wr)
		spi_led_opened = 0;

	return 0;
}

/*
 * Read current kbd state into supplied buffer
 */
static ssize_t spi_kbd_read(struct file *file, char *buf,
		size_t count, loff_t *offset)
{
	unsigned long state;
	int len;

	debugk("buf=0x%08X, count=%u, offset=%lld\n", buf, count, *offset);

	if (verify_area(VERIFY_WRITE, buf, count))
		return -EFAULT;

	if (!spi_key_changed)
    	/* no user input happened since last call */
		return 0;
  
	lock_dev();

	len = msg_len();
	state = mask_state(spi_key_state);

	spi_key_changed = 0;

	unlock_dev();

	debugk("kbd_state=0x%08X\n", spi_key_state);

	count = count > len ? len : count;

	if (copy_to_user(buf, &state, count))
		return -EFAULT;

	return count;
}

/*
 * Write new led state to device
 */
static ssize_t spi_led_write(struct file *file, const char *buf,
		size_t count, loff_t *offset)
{
	unsigned long state = 0;
	int len, ret;

	debugk("buf=0x%08X, count=%d, offset=%lld\n", buf, count, *offset);

	count = count > MAX_MSG_LEN ? MAX_MSG_LEN : count;

	if (copy_from_user(&state, buf, count))
		return -EFAULT;

	lock_dev();

	len	= msg_len();
	count	= count > len ? len : count;

	state		= mask_state(state);
	spi_led_state	&= (unsigned long)(-1) >> (count * 8);
	spi_led_state	|= state;

	ret = transfer_bytes((char *)&spi_led_state,
		(char *)&spi_key_state, count);

	unlock_dev();

	debugk("led_state=0x%08X kbd_state=0x%08X\n",
		spi_led_state, spi_key_state);

	return ret;
}

/*
 * Read current led state into buffer
 */
static ssize_t spi_led_read(struct file *file, char *buf,
		size_t count, loff_t *offset)
{
	long state;
	int len;

	debugk("buf=0x%08X, count=%d, offset=%lld\n", buf, count, *offset);

	if (verify_area(VERIFY_WRITE, buf, count))
		return -EFAULT;

	lock_dev();

	len	= msg_len();
	state	= mask_state(spi_led_state);

	unlock_dev();

	debugk("led_state=0x%08X\n", spi_led_state);

	count = count > len ? len : count;  

	if (copy_to_user(buf, &state, count))
		return -EFAULT;

	return count;
}

/*
 * Set currently used number of keys/leds
 */
static int spi_kbd_ioctl(struct inode *inode, struct file *file,
		unsigned int cmd, unsigned long arg)
{
	debugk("cmd=0x%x, arg=0x%lx\n", cmd, arg);

	if (cmd == SPI_KBD_SET_USED_KEYS) {

		if (arg > 24)
			return -EINVAL;

		lock_dev();

		spi_key_num = arg;

		unlock_dev();

		debugk("key_num=%d\n", spi_key_num);

		return 0;
	}

	return -EINVAL;
}

/*
 * Wait for user input
 */
static unsigned int spi_kbd_poll(struct file *file, poll_table *wait)
{
	if (!spi_key_changed)
		poll_wait(file, &spi_key_wait, wait);

	return spi_key_changed ? POLLIN | POLLRDNORM : 0;
}

	/* SPI layer */

/*
 * Transfer byte to/from kbd/led device through HMI1001 adapter
 */
static int transfer_bytes(const char *buf_out, char *buf_in,
		unsigned int count)
{
	char buf[MAX_MSG_LEN];
	struct spi_msg msg = {
		.addr	= 0,
		.flags	= SPI_M_RD|SPI_M_WR,
		.len	= count,
		.buf	= buf
	};
	int ret, i;

	count = count > MAX_MSG_LEN ? MAX_MSG_LEN : count;

	memcpy(buf, buf_out, count);

	ret = spi_transfer(hmi1001_adapter, &msg, 1);

	if (ret < 0)
		return ret;

		/* reversed copy to input buffer */
	for (i = 0; i < count; ++i)
		buf_in[count - i - 1] = buf[i];

	return count;
}

/*
 * Notify spi driver about the addition of an adapter
 */
static int hmi1001_attach_adapter(struct spi_adapter *adap)
{
	struct mpc5xxx_psc *psc = (struct mpc5xxx_psc *)MPC5xxx_PSC3;
	
	debugk("attached adapter 0x%08X\n", adap);
		/* Consider that HMI1001 is the only SPI adapter on board */
	hmi1001_adapter = adap;
	
		/* Correct the DSCKL to be equal 2us */
	out_be16(&psc->ccr, 0x1f0f);	/* FrameSyncDiv = 31, BitClkDiv = 15 */
	return 0;
}

/*
 * Notify spi driver about client detached
 */
static void hmi1001_detach_client(struct spi_client *client)
{
	/* do nothing specific */
}

/*
 * Notify spi driver about command for spi client
 */
static int hmi1001_command(struct spi_client *client,
		unsigned int cmd, void *arg)
{
		/* none of commands are allowed */
	return -EINVAL;
}

/*
 * Configure Timer_N pin for receiving interrupts
 * from external device at input on falling edge
 */
static int setup_gpt(int timer)
{
	struct mpc5xxx_gpt *gpt = (struct mpc5xxx_gpt *)MPC5xxx_GPT + timer;

	if (timer < 0 || timer > 7)
		return -EINVAL;

	out_be32(&gpt->emsr, 0x00020101);	/* IC at input falling edge */
						/* IntEn = 1 */
						/* Timer_Ms = 1 - IC */
	out_be32(&gpt->cr, 0);

	return 0;
} 

	/* proc interface */

static int print_state(long state, int num, char *buf)
{
	int i;
	char *cur = buf;

	for (i=0; i < num; ++i) {
		if (state < 0)
			cur += sprintf(cur, "%d ", i);
		state <<= 1;
	}
	
	*(cur - 1) = '\n';

	return cur - buf;	
}

#ifdef PROC_SPIKBD_ENABLE

/*
 * Read current kbd state into buffer
 */
static int proc_kbd_read(char *page, char **start, off_t offset,
		int count, int *eof, void *data)
{
	long state;
	int num, len;

	lock_dev();

	state	= mask_state(spi_key_state);
	num		= spi_key_num;
	
	unlock_dev();

	len = print_state(state, num, page);
	
	if (offset > len)
		return -EINVAL;
	
	count = count > len - offset ? len - offset : count;
	
	*start = page + offset;
	
	return count;	
}

#endif/*  PROC_SPIKBD_ENABLE */

#ifdef PROC_SPILED_ENABLE

/*
 * Read current led state into buffer
 */
static int proc_led_read(char *page, char **start, off_t offset,
		int count, int *eof, void *data)
{
	long state;
	int num, len;

	lock_dev();
	
	state = mask_state(spi_led_state);
	num = spi_key_num;
	
	unlock_dev();
	
	len = print_state(state, num, page);
		
	if (offset > len)
		return -EINVAL;

	count = count > len - offset ? len - offset : count;
	*start = page + offset;

	return count;
}

/*
 * Write new led state to the led device
 */
static int proc_led_write(struct file *file, const char *buffer,
		unsigned long count, void *data)
{
	unsigned long state = 0;
	char *page, *cur;
	int ret, led;

	if (verify_area(VERIFY_READ, buffer, count))
		return -EFAULT;

	page = (char *)__get_free_page(GFP_KERNEL);

	if (!page)
		return -ENOMEM;

	count = count > PAGE_SIZE - 1 ? PAGE_SIZE - 1 : count;

	ret = -EFAULT;
		
	if (copy_from_user(page, buffer, count))
		goto ex_free;
	
	page[count] = '\0';
	
	cur = page;
	
	ret = 0;
	
	while ( (cur = strtok(cur, " ")) ) {
		
		if (sscanf(cur, "%d", &led))
			state |= (1 << (31 - led)), ++ret;
		else
			break;

		cur += strlen(cur) + 1;
	}

	if (ret)
		count = cur - page;
		
	ret = 0;

ex_free:
	free_page((long)page);
	
	if (ret)
		return ret;
		
	lock_dev();

	spi_led_state = mask_state(state);

	ret = transfer_bytes((char *)&spi_led_state,
		(char *)&spi_key_state, msg_len());

	if (ret < 0)
		count = ret;

	unlock_dev();

	debugk("led_state=0X%08X kbd_state=0x%08X\n",
		spi_led_state, spi_key_state);

	return count;
}

#endif /* PROC_SPILED_ENABLE */

/*
 * Initialize proc interface
 */
static int proc_init(void)
{
	int ret = 0;

#ifdef PROC_SPIKBD_ENABLE
	ret = create_proc_read_entry(PROC_SPIKBD_FILE, 0, NULL,
		proc_kbd_read, NULL) ? 0 : -1;

	if (!ret)
#endif

#ifdef PROC_SPILED_ENABLE
	{
		struct proc_dir_entry *entry =
			create_proc_entry(PROC_SPILED_FILE, 0, NULL);

		if (entry) {
			entry->read_proc  = proc_led_read;
			entry->write_proc = proc_led_write;
		}
		else
			goto err;
	}    
#endif
	return ret;

err:
#ifdef PROC_SPIKBD_ENABLE
	remove_proc_entry(PROC_SPIKBD_FILE, NULL);
#endif
	return -1;
}

/*
 * Uninitialize proc interface
 */
static void proc_cleanup(void)
{
#ifdef PROC_SPIKBD_ENABLE
	remove_proc_entry(PROC_SPIKBD_FILE, NULL);
#endif

#ifdef PROC_SPILED_ENABLE
	remove_proc_entry(PROC_SPILED_FILE, NULL);
#endif
}

/*
 * Driver entry point - initialize all layers
 */
static int __init spi_kbd_init(void)
{
	int ret;

	printk("HMI1001 keyboard/led driver: Revision 0.1\n");

	setup_gpt(TIMER_ID);

	INIT_TQUEUE(&spi_kbd_task, spi_kbd_task_routine, NULL);

	ret = request_irq(TIMER_IRQ, spi_kbd_interrupt, SA_INTERRUPT,
		"spi-kbd", NULL);

	if (ret) {
		printk("can't install keyboard IRQ for GPT7 (ret=%d)\n", ret);

		goto ex;
	}

	ret = spi_add_driver(&hmi1001_driver);

	if (ret) {
		printk("cannot init spi layer (ret=%d)\n", ret);

		goto ex_irq;
	}

	debugk("spi layer initialized\n");

	ret = register_chrdev(SPIDEV_MAJOR, SPIDEV_NAME, &spi_kbd_fops);

	if (ret) {
		printk("can't register character device (ret=%d)\n", ret);

		goto ex_spi_irq;
	}

	debugk("registered character device at major=%d\n", SPIDEV_MAJOR);

	ret = proc_init();

	if (ret) {
		printk("can't init proc interface (ret=%d)\n", ret);

		goto ex_dev_spi_irq;
	}

ex:

	return ret;

ex_dev_spi_irq:

	unregister_chrdev(SPIDEV_MAJOR, SPIDEV_NAME);

ex_spi_irq:

	spi_del_driver(&hmi1001_driver);

ex_irq:

	free_irq(TIMER_IRQ, NULL);

	return ret;
}

/*
 * Uninitialize driver
 */
static void __exit spi_kbd_exit(void)
{
	proc_cleanup();

	unregister_chrdev(SPIDEV_MAJOR, SPIDEV_NAME);

	spi_del_driver(&hmi1001_driver);

	free_irq(TIMER_IRQ, NULL);

	printk("HMI1001 keyboard/LED driver unloaded\n");
}

module_init(spi_kbd_init);
module_exit(spi_kbd_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Wolfgang Denk, <wd@denx.de>");
MODULE_DESCRIPTION("HMI1001 keyboardd/led driver");
