/* ------------------------------------------------------------------------- */
/* spi-algo-bit.c i2c driver algorithms for bit-shift adapters		     */
/* ------------------------------------------------------------------------- */
/*   Copyright (C) 1995-2000 Simon G. Vogl

    Adapted for SPI by Kate Alhola  2004

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.		     */
/* ------------------------------------------------------------------------- */

/* With some changes from Kysti Mlkki <kmalkki@cc.hut.fi> and even
   Frodo Looijaard <frodol@dds.nl> */

/* $Id: i2c-algo-bit.c,v 1.30 2001/07/29 02:44:25 mds Exp $ */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <linux/ioport.h>
#include <linux/errno.h>
#include <linux/sched.h>

#include <linux/spi.h>
#include <linux/spi-algo-bit.h>

/* ----- global defines ----------------------------------------------- */
#define DEB(x) if (spi_debug>=1) x;
#define DEB2(x) if (spi_debug>=2) x;
#define DEBSTAT(x) if (spi_debug>=3) x; /* print several statistical values*/
#define DEBPROTO(x) if (spi_debug>=9) { x; }
 	/* debug the protocol by showing transferred bits */

/* debugging - slow down transfer to have a look at the data .. 	*/
/* I use this with two leds&resistors, each one connected to sda,scl 	*/
/* respectively. This makes sure that the algorithm works. Some chips   */
/* might not like this, as they have an internal timeout of some mils	*/
/*
#define SLO_IO      jif=jiffies;while(time_before_eq(jiffies, jif+spi_table[minor].veryslow))\
                        if (need_resched) schedule();
*/


/* ----- global variables ---------------------------------------------	*/

#ifdef SLO_IO
	int jif;
#endif

/* module parameters:
 */
static int spi_debug;
static int bit_test;	/* see if the line-setting functions work	*/
static int bit_scan;	/* have a look at what's hanging 'round		*/

/* --- setting states on the bus with the right timing: ---------------	*/

#define setmosi(adap,val) adap->setsda(adap->data, val)
#define setscl(adap,val) adap->setscl(adap->data, val)
#define getmiso(adap) adap->getsda(adap->data)
#define getscl(adap) adap->getscl(adap->data)

static inline void sdalo(struct spi_algo_bit_data *adap)
{
	setsda(adap,0);
	udelay(adap->udelay);
}

static inline void sdahi(struct spi_algo_bit_data *adap)
{
	setsda(adap,1);
	udelay(adap->udelay);
}

static inline void scllo(struct spi_algo_bit_data *adap)
{
	setscl(adap,0);
	udelay(adap->udelay);
#ifdef SLO_IO
	SLO_IO
#endif
}

/*
 * Raise scl line, and do checking for delays. This is necessary for slower
 * devices.
 */
static inline int sclhi(struct spi_algo_bit_data *adap)
{
	int start=jiffies;

	setscl(adap,1);

	udelay(adap->udelay);

	/* Not all adapters have scl sense line... */
	if (adap->getscl == NULL )
		return 0;

 	while (! getscl(adap) ) {	
 		/* the hw knows how to read the clock line,
 		 * so we wait until it actually gets high.
 		 * This is safer as some chips may hold it low
 		 * while they are processing data internally. 
 		 */
		setscl(adap,1);
		if (time_after_eq(jiffies, start+adap->timeout)) {
			return -ETIMEDOUT;
		}
		if (current->need_resched)
			schedule();
	}
	DEBSTAT(printk(KERN_DEBUG "needed %ld jiffies\n", jiffies-start));
#ifdef SLO_IO
	SLO_IO
#endif
	return 0;
} 




/* send a byte without start cond., look for arbitration, 
   check ackn. from slave */
/* returns:
 * 1 if the device acknowledged
 * 0 if the device did not ack
 * -ETIMEDOUT if an error occurred (while raising the scl line)
 */
static int spi_byte(struct spi_adapter *spi_adap, char c)
{
	int i;
	int sb;
	int ack;
	struct spi_algo_bit_data *adap = spi_adap->algo_data;

	/* assert: scl is low */
	DEB2(printk(KERN_DEBUG " spi_outb:%2.2X\n",c&0xff));
	for ( i=7 ; i>=0 ; i-- ) {
		sb = c & ( 1 << i );
		setsda(adap,sb);
		udelay(adap->udelay);
		DEBPROTO(printk(KERN_DEBUG "%d",sb!=0));
		if (sclhi(adap)<0) { /* timed out */
			sdahi(adap); /* we don't want to block the net */
			return -ETIMEDOUT;
		};
		/* do arbitration here: 
		 * if ( sb && ! getsda(adap) ) -> ouch! Get out of here.
		 */
		setscl(adap, 0 );
		udelay(adap->udelay);
	}
	sdahi(adap);
	if (sclhi(adap)<0){ /* timeout */
		return -ETIMEDOUT;
	};
	/* read ack: SDA should be pulled down by slave */
	ack=getsda(adap);	/* ack: sda is pulled low ->success.	 */
	DEB2(printk(KERN_DEBUG " spi_outb: getsda() =  0x%2.2x\n", ~ack ));

	DEBPROTO( printk(KERN_DEBUG "[%2.2x]",c&0xff) );
	DEBPROTO(if (0==ack){ printk(KERN_DEBUG " A ");} else printk(KERN_DEBUG " NA ") );
	scllo(adap);
	return 0==ack;		/* return 1 if device acked	 */
	/* assert: scl is low (sda undef) */
}




static int sendbytes(struct spi_adapter *spi_adap,const char *buf, int count)
{
	struct spi_algo_bit_data *adap = spi_adap->algo_data;
	char c;
	const char *temp = buf;
	int retval;
	int wrcount=0;

	while (count > 0) {
		c = *temp;
		DEB2(printk(KERN_DEBUG "spi-algo-bit.o: %s sendbytes: writing %2.2X\n",
			    spi_adap->name, c&0xff));
		retval = spi_outb(spi_adap,c);
		if (retval>0) {
			count--; 
			temp++;
			wrcount++;
		} else { /* arbitration or no acknowledge */
			printk(KERN_ERR "spi-algo-bit.o: %s sendbytes: error - bailout.\n",
			       spi_adap->name);
			spi_stop(adap);
			return (retval<0)? retval : -EFAULT;
			        /* got a better one ?? */
		}
#if 0
		/* from asm/delay.h */
		__delay(adap->mdelay * (loops_per_sec / 1000) );
#endif
	}
	return wrcount;
}

static inline int readbytes(struct spi_adapter *spi_adap,char *buf,int count)
{
	char *temp = buf;
	int inval;
	int rdcount=0;   	/* counts bytes read */
	struct spi_algo_bit_data *adap = spi_adap->algo_data;

	while (count > 0) {
		inval = spi_inb(spi_adap);
/*printk("%#02x ",inval); if ( ! (count % 16) ) printk("\n"); */
		if (inval>=0) {
			*temp = inval;
			rdcount++;
		} else {   /* read timed out */
			printk(KERN_ERR "spi-algo-bit.o: readbytes: spi_inb timed out.\n");
			break;
		}

		if ( count > 1 ) {		/* send ack */
			sdalo(adap);
			DEBPROTO(printk(" Am "));
		} else {
			sdahi(adap);	/* neg. ack on last byte */
			DEBPROTO(printk(" NAm "));
		}
		if (sclhi(adap)<0) {	/* timeout */
			sdahi(adap);
			printk(KERN_ERR "spi-algo-bit.o: readbytes: Timeout at ack\n");
			return -ETIMEDOUT;
		};
		scllo(adap);
		sdahi(adap);
		temp++;
		count--;
	}
	return rdcount;
}

/* doAddress initiates the transfer by generating the start condition (in
 * try_address) and transmits the address in the necessary format to handle
 * reads, writes as well as 10bit-addresses.
 * returns:
 *  0 everything went okay, the chip ack'ed
 * -x an error occurred (like: -EREMOTEIO if the device did not answer, or
 *	-ETIMEDOUT, for example if the lines are stuck...) 
 */
static inline int bit_doAddress(struct spi_adapter *spi_adap,
                                struct spi_msg *msg, int retries) 
{
	unsigned short flags = msg->flags;
	struct spi_algo_bit_data *adap = spi_adap->algo_data;

	unsigned char addr;
	int ret;
	if ( (flags & SPI_M_TEN)  ) { 
		/* a ten bit address */
		addr = 0xf0 | (( msg->addr >> 7) & 0x03);
		DEB2(printk(KERN_DEBUG "addr0: %d\n",addr));
		/* try extended address code...*/
		ret = try_address(spi_adap, addr, retries);
		if (ret!=1) {
			printk(KERN_ERR "died at extended address code.\n");
			return -EREMOTEIO;
		}
		/* the remaining 8 bit address */
		ret = spi_outb(spi_adap,msg->addr & 0x7f);
		if (ret != 1) {
			/* the chip did not ack / xmission error occurred */
			printk(KERN_ERR "died at 2nd address code.\n");
			return -EREMOTEIO;
		}
		if ( flags & SPI_M_RD ) {
			spi_repstart(adap);
			/* okay, now switch into reading mode */
			addr |= 0x01;
			ret = try_address(spi_adap, addr, retries);
			if (ret!=1) {
				printk(KERN_ERR "died at extended address code.\n");
				return -EREMOTEIO;
			}
		}
	} else {		/* normal 7bit address	*/
		addr = ( msg->addr << 1 );
		if (flags & SPI_M_RD )
			addr |= 1;
		if (flags & SPI_M_REV_DIR_ADDR )
			addr ^= 1;
		ret = try_address(spi_adap, addr, retries);
		if (ret!=1) {
			return -EREMOTEIO;
		}
	}
	return 0;
}

static int bit_xfer(struct spi_adapter *spi_adap,
		    struct spi_msg msgs[], int num)
{
	struct spi_msg *pmsg;
	struct spi_algo_bit_data *adap = spi_adap->algo_data;
	
	int i,ret;

	spi_start(adap);
	for (i=0;i<num;i++) {
		pmsg = &msgs[i];
		if (!(pmsg->flags & SPI_M_NOSTART)) {
			if (i) {
				spi_repstart(adap);
			}
			ret = bit_doAddress(spi_adap,pmsg,spi_adap->retries);
			if (ret != 0) {
				DEB2(printk(KERN_DEBUG "spi-algo-bit.o: NAK from device addr %2.2x msg #%d\n",
					msgs[i].addr,i));
				return (ret<0) ? ret : -EREMOTEIO;
			}
		}
		if (pmsg->flags & SPI_M_RD ) {
			/* read bytes into buffer*/
			ret = readbytes(spi_adap,pmsg->buf,pmsg->len);
			DEB2(printk(KERN_DEBUG "spi-algo-bit.o: read %d bytes.\n",ret));
			if (ret < pmsg->len ) {
				return (ret<0)? ret : -EREMOTEIO;
			}
		} else {
			/* write bytes from buffer */
			ret = sendbytes(spi_adap,pmsg->buf,pmsg->len);
			DEB2(printk(KERN_DEBUG "spi-algo-bit.o: wrote %d bytes.\n",ret));
			if (ret < pmsg->len ) {
				return (ret<0) ? ret : -EREMOTEIO;
			}
		}
	}
	spi_stop(adap);
	return num;
}

static int algo_control(struct spi_adapter *adapter, 
	unsigned int cmd, unsigned long arg)
{
	return 0;
}

static u32 bit_func(struct spi_adapter *adap)
{
	return SPI_FUNC_SMBUS_EMUL | SPI_FUNC_10BIT_ADDR | 
	       SPI_FUNC_PROTOCOL_MANGLING;
}


/* -----exported algorithm data: -------------------------------------	*/

static struct spi_algorithm spi_bit_algo = {
	"Bit-shift algorithm",
	SPI_ALGO_BIT,
	bit_xfer,
	NULL,
	NULL,				/* slave_xmit		*/
	NULL,				/* slave_recv		*/
	algo_control,			/* ioctl		*/
	bit_func,			/* functionality	*/
};

/* 
 * registering functions to load algorithms at runtime 
 */
int spi_bit_add_bus(struct spi_adapter *adap)
{
	int i;
	struct spi_algo_bit_data *bit_adap = adap->algo_data;

	if (bit_test) {
		int ret = test_bus(bit_adap, adap->name);
		if (ret<0)
			return -ENODEV;
	}

	DEB2(printk(KERN_DEBUG "spi-algo-bit.o: hw routines for %s registered.\n",
	            adap->name));

	/* register new adapter to i2c module... */

	adap->id |= spi_bit_algo.id;
	adap->algo = &spi_bit_algo;

	adap->timeout = 100;	/* default values, should	*/
	adap->retries = 3;	/* be replaced by defines	*/

#ifdef MODULE
	MOD_INC_USE_COUNT;
#endif
	spi_add_adapter(adap);

	return 0;
}


int spi_bit_del_bus(struct spi_adapter *adap)
{
	int res;

	if ((res = spi_del_adapter(adap)) < 0)
		return res;

	DEB2(printk("spi-algo-bit.o: adapter unregistered: %s\n",adap->name));

#ifdef MODULE
	MOD_DEC_USE_COUNT;
#endif
	return 0;
}

int __init spi_algo_bit_init (void)
{
	printk(KERN_INFO "spi-algo-bit.o: spi bit algorithm module\n");
	return 0;
}



EXPORT_SYMBOL(spi_bit_add_bus);
EXPORT_SYMBOL(spi_bit_del_bus);

#ifdef MODULE
MODULE_AUTHOR("Kate Alhola / Simon G. Vogl <simon@tk.uni-linz.ac.at>");
MODULE_DESCRIPTION("Spi-Bus bit-banging algorithm");
MODULE_LICENSE("GPL");

MODULE_PARM(spi_debug,"i");

MODULE_PARM_DESC(spi_debug,
            "debug level - 0 off; 1 normal; 2,3 more verbose; 9 bit-protocol");

int init_module(void) 
{
	return spi_algo_bit_init();
}

void cleanup_module(void) 
{
}
#endif
