/*
 * drivers/spi/spi-algo-mpc5200psc.c
 *
 * PSC SPI  algorithm driver for MPC5200
 *
 *
 * 2004 (c) Kate Alhola
 *  MPC5200 spi driver original version, deriver from i2c-algo-mpc5200.c
 *
 * 2003-2004 (C) Wolfgang Denk, DENX Software Engineering, <wd@denx.de>.
 *
 * This file is licensed under the terms of the GNU General Public License
 * version 2.  This program is licensed "as is" without any warranty of
 * any kind, whether express or implied.
 */

/**
 * @file spi-algo-mpc5200psc.c
 * @brief  SPI algorithm driver for MPC5200
 *
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/string.h>
#include <asm/io.h>
#include <asm/mpc5xxx.h>
#ifdef CONFIG_UBOOT
#include <asm/ppcboot.h>
#endif

#include <linux/spi.h>
#include "spi-algo-mpc5xxx-psc.h"

#define MPC5xxx_PSC_SICR ( \
	MPC5xxx_PSC_SICR_MSTR | \
	MPC5xxx_PSC_SICR_SPI | \
	MPC5xxx_PSC_SICR_GENCLK | \
	MPC5xxx_PSC_SICR_SIM(1))
#define MPC5xxx_PSC_RFALARM	0x0c
#define MPC5xxx_PSC_TFALARM	0x10

#undef DEBUG

#ifdef DEBUG
#define debug(s...) do { printk("%s: ", __FUNCTION__); printk(s); } while (0)
#else
#define debug(s...)
#endif

static int speed = 1000;

MODULE_PARM(speed, "i");

/* Dynamically update the alarm level so interrupts don't occur too fast */
static void mpc5xxx_psc_spi_adjust_alarm(struct spi_algo_mpc5xxx_psc_data *algo_data)
{
	struct mpc5xxx_psc *psc = algo_data->regs;
	int alarm;

	alarm = 0x200 - algo_data->rx_count;
	if (alarm < 0) alarm = 0;
	if (alarm > 0x1ff) alarm = 0x1ff;
	out_be16(&psc->rfalarm, alarm);
}

static void
mpc5xxx_psc_spi_isr(int irq, void *dev_id, struct pt_regs *regs)
{
	struct spi_algo_mpc5xxx_psc_data *algo_data = 
		(struct spi_algo_mpc5xxx_psc_data*)dev_id;
	struct mpc5xxx_psc *psc = algo_data->regs;
	int num, i;

	if (in_be16(&psc->mpc5xxx_psc_isr) & MPC5xxx_PSC_IMR_RXRDY) {
		num = in_be16(&psc->rfnum);

		/* Discard unexpected rx data */
		if ((!algo_data->rx_count) || (!algo_data->rx_ptr)) {
			for (i = 0; i < num; i++)
				readb(&psc->mpc5xxx_psc_buffer_8);
			algo_data->rx_count = 0;
			wake_up_interruptible(&algo_data->wait_queue);
			return;
		}

		/* Don't overflow the buffer */
		if (num > algo_data->rx_count)
			num = algo_data->rx_count;

		/* Transfer data into rx buffer */
		for (i = 0; i < num; i++)
			if (algo_data->rx_ptr)
				*algo_data->rx_ptr++ = 
					readb(&psc->mpc5xxx_psc_buffer_8);
			else
				readb(&psc->mpc5xxx_psc_buffer_8);
		algo_data->rx_count -= num;

		/* Dynamically update the alarm level so interrupts don't
		 * occur too frequently */
		mpc5xxx_psc_spi_adjust_alarm(algo_data);

		/* Only wakeup process if data transfer is complete */
		if (algo_data->rx_count == 0)
			wake_up_interruptible(&algo_data->wait_queue);
	}

	return;
}

/**
 * mpc5xxx_psc_spi_init
 * @brief Initialize PSC as SPI
 * @param algo_data 
 */

static int mpc5xxx_psc_spi_init(struct spi_adapter *adap)
{
	struct spi_algo_mpc5xxx_psc_data	*algo_data = adap->algo_data;
	struct mpc5xxx_psc	*psc = algo_data->regs;
	struct mpc5xxx_cdm      *cdmregs=(struct mpc5xxx_cdm *)MPC5xxx_CDM;
	struct mpc5xxx_gpio     *gpio=(struct mpc5xxx_gpio *)MPC5xxx_GPIO;
	int psc_num=MPC5xxx_PSCn(psc);
	volatile u16* cdm_mclken_div_pscx=
		psc_num == 1 ? &(cdmregs->mclken_div_psc1) :
		psc_num == 2 ? &(cdmregs->mclken_div_psc2) :
		psc_num == 3 ? &(cdmregs->mclken_div_psc3) :
		psc_num == 6 ? &(cdmregs->mclken_div_psc6) : NULL;
	int irq = 
		psc_num == 1 ? MPC5xxx_PSC1_IRQ :
		psc_num == 2 ? MPC5xxx_PSC2_IRQ :
		psc_num == 3 ? MPC5xxx_PSC3_IRQ :
		psc_num == 6 ? MPC5xxx_PSC6_IRQ : 0;
	
	if (!cdm_mclken_div_pscx)
		return -1;

	if (algo_data->flags & MPC5XXX_PSC_SPI_USE_FIFO) {
		if (request_irq(irq, mpc5xxx_psc_spi_isr,
				SA_INTERRUPT | SA_SAMPLE_RANDOM, 
				"mpc5xxx_psc_spi", algo_data)) {
			printk(KERN_ALERT "mpc5xxx_psc_spi: request_irq failed\n");
			return -1;
		}
		
		init_MUTEX(&algo_data->sema);
		init_waitqueue_head(&algo_data->wait_queue);
		algo_data->rx_count = 0;
		algo_data->rx_ptr = NULL;
	}

	debug("psc=%08x cdm=%08x gpio=%08x portcfg=%08x\n",
	      (int)psc, (int)cdmregs, (int)gpio, gpio->port_config);

	/* Reset to default.
	 */
	if (psc_num != 6) {
		gpio->port_config &= ~(0x7 << ((psc_num - 1) * 4));
		cdmregs->clk_enables |= (1 << (psc_num + 4)); /* PSC clock enable */
	} else
		cdmregs->clk_enables |= 1 << 4;

	psc->command = MPC5xxx_PSC_RST_RX;
	psc->command = MPC5xxx_PSC_RST_TX;
	psc->command = MPC5xxx_PSC_SEL_MODE_REG_1;
	psc->command = MPC5xxx_PSC_RST_ERR_STAT;
	psc->mode = 0;
	if (psc_num != 6)
		gpio->port_config |= (0x6 << ((psc_num - 1) * 4));
	else
		gpio->port_config |= 0x7 << 20;
	psc->command = MPC5xxx_PSC_RX_DISABLE | MPC5xxx_PSC_TX_DISABLE;
	psc->sicr = MPC5xxx_PSC_SICR |
#ifdef CONFIG_SPI_5xxx_USE_EOF
		(adap->set_spi_adap_cs ? 0 : MPC5xxx_PSC_SICR_USEEOF) |
#endif
		(adap->bus_params & SPI_CLOCK_ACTIVE_LOW ? 0 : MPC5xxx_PSC_SICR_CPHA) | 
		(adap->bus_params & SPI_CLOCK_DATA_START ? 0 : MPC5xxx_PSC_SICR_CPOL);
	*cdm_mclken_div_pscx = 0x8007;	/* Divide Fvco from 528 to 66Mhz */
	psc->ccr=0x0700 | ((66000 / speed - 1) & 0xff);                 /* set SCK and DSCKL delay */
	psc->ctur=0x00;                  /* Set DTL delay 2us */
	psc->ctlr=0x84;
	psc->rfalarm = MPC5xxx_PSC_RFALARM;
	psc->tfalarm = MPC5xxx_PSC_TFALARM;
	psc->rfcntl = 1; /* granularity */
	psc->tfcntl = 1;
	psc->command = MPC5xxx_PSC_RST_RX;
	psc->command = MPC5xxx_PSC_RST_TX;
	psc->command = MPC5xxx_PSC_SEL_MODE_REG_1;
	psc->command = MPC5xxx_PSC_RST_ERR_STAT;
	psc->command = MPC5xxx_PSC_RX_ENABLE | MPC5xxx_PSC_TX_ENABLE;  /* Enable rx & tx */

	debug("port_config=%08x sicr=%08x clk_enables=%08x div_psc%d=%04x\n",
	      gpio->port_config, psc->sicr, cdmregs->clk_enables, psc_num, *cdm_mclken_div_pscx);

	/* Select an arbitrary address.
	 */

	/* Enable module and disable interrupts.
	 */
	debug("mpc5xxx_psc_spi_init:exit\n");

	return 0;
}

/**
 * mpc5xxx_xferbytes
 * @brief transfer bytes via SPI bus. Send data out via MOSI and inputs data from MISO
 * @param adap  spi adapter structure
 * @param buf   message buffer
 * @param count number of bytes to transfer
 * @param flafs read / write mode flag 
 * @return number of bytes transfered or 0>error code
 *
 */
static int mpc5xxx_psc_xferbytes(struct spi_adapter	*adap,char *buf,int count,int flags)
{
	struct spi_algo_mpc5xxx_psc_data	*algo_data	= adap->algo_data;
	struct mpc5xxx_psc		*psc		= algo_data->regs;
	int				wrcount;
	int				rdcount;
	unsigned int                    rd=0;
	unsigned int                    wd;
        int                             to;

	debug("adap=%x buf=%x count=%d flags=%x psc=%x status=%04x tfstat=%04x mode=%02x\n",
	       (int)adap,(int)buf,count,flags,(int)psc,psc->mpc5xxx_psc_status,psc->tfstat,psc->mode);

	psc->command = MPC5xxx_PSC_RST_RX;
	psc->command = MPC5xxx_PSC_RST_TX;
	psc->command = MPC5xxx_PSC_SEL_MODE_REG_1;
	psc->command = MPC5xxx_PSC_RST_ERR_STAT;
	psc->command = MPC5xxx_PSC_RX_ENABLE | MPC5xxx_PSC_TX_ENABLE;  /* Enable rx & tx */
	wrcount = 0;
	rdcount = 0;
	to=10000;
	while( (to>=0) && ((wrcount < count) || (rdcount<count))) {
	    if((wrcount < count) && (psc->mpc5xxx_psc_status & MPC5xxx_PSC_SR_TXRDY)) {
		wd=flags & SPI_M_WR  ? buf[wrcount]:0;
		/* Mark last buffer byte as last frame byte.
		 * CS became inactive after transmitting this byte */
		if (!adap->set_spi_adap_cs && (wrcount >= count - 1))
			psc->ircr2 |= 1;
		psc->mpc5xxx_psc_buffer_8=wd;
		wrcount++;
		to=10000;
	    }
	    if((rdcount < count) && (psc->mpc5xxx_psc_status & MPC5xxx_PSC_SR_RXRDY)) {
		rd=psc->mpc5xxx_psc_buffer_8;
		if(flags & SPI_M_RD) buf[rdcount]=rd;
		rdcount++;
		to=10000;
	    }
	    psc->command=0x05;               /* Enable rx & tx */
//	    if(to==10000) printk("mpc5xxx_psc_xferbytes:wn=%d rn=%d wd=%02x rd=%02x tfnum=%d rfnum=%d  tfstat=%04x\n",wrcount,rdcount,wd,rd,psc->tfnum,psc->rfnum,psc->tfstat);
	    to--;
	}
	if(to<=0) {
	    debug("mpc5xxx_psc_xferbytes:timeout psc_status=%04x wn=%d rn=%d wr=%02x rd=%02x tfnum=%d rfnum=%d  tfstat=%04x\n",
		   psc->mpc5xxx_psc_status,wrcount,rdcount,wd,rd,psc->tfnum,psc->rfnum,psc->tfstat);
	    return -1;
		}
	return wrcount;
}

/**
 * mpc5xxx_psc_fifo_xferbytes
 * @brief transfer bytes via SPI bus using the FIFO.
 * @param adap  spi adapter structure
 * @param buf   message buffer
 * @param count number of bytes to transfer
 * @param flafs read / write mode flag 
 * @return number of bytes transfered or 0>error code
 *
 */
static int mpc5xxx_psc_fifo_xferbytes(struct spi_adapter	*adap,char *buf,int count,int flags)
{
	struct spi_algo_mpc5xxx_psc_data	*algo_data	= adap->algo_data;
	struct mpc5xxx_psc		*psc		= algo_data->regs;
	int err, idx;

	if (count < 1)
		return -1;

	/* Grab the semaphore, only one transfer allowed at a time */
	err = down_interruptible(&algo_data->sema);
	if (err)
		return err;

	/* Disable SPI interrupts so rx buffer can be manipulated */
	out_be16(&psc->mpc5xxx_psc_imr, 0);

#if 0
	/* Set the clock polarity */
	sicr = in_be32(&psc->sicr);
	if (clkedge == SPI_CLKEDGE_RISING)
		sicr &= ~MPC5xxx_PSC_SICR_CPOL; /* Clear means rising edge */
	else
		sicr |= MPC5xxx_PSC_SICR_CPOL; /* Set means falling edge */
	out_be32(&psc->sicr, sicr);
#endif

	/* Initialize the receive buffer; this can be done without locking
	 * because the ISR doesn't touch it until after interrupts are
	 * enabled
	 */
	algo_data->rx_ptr = flags & SPI_M_RD ? buf : NULL;
	algo_data->rx_count = count;
	mpc5xxx_psc_spi_adjust_alarm(algo_data);

	/* Start the engine and transfer data to fifo */
	writeb(MPC5xxx_PSC_RX_ENABLE | MPC5xxx_PSC_TX_ENABLE, &psc->command);
#ifdef CONFIG_SPI_5xxx_USE_EOF
	for (idx = 0; idx < count - 1; idx++)
		writeb((flags & SPI_M_WR ? buf[idx] : 0), 
			&psc->mpc5xxx_psc_buffer_8);

	/* Signal EOF for the last byte to terminate the transfer */
	writeb(MPC5xxx_PSC_IRCR2_NXTEOF, &psc->ircr2);
	writeb((flags & SPI_M_WR ? buf[idx] : 0), &psc->mpc5xxx_psc_buffer_8);

#else
	for (idx = 0; idx < count; idx++)
		writeb((flags & SPI_M_WR ? buf[idx] : 0), 
			&psc->mpc5xxx_psc_buffer_8);
#endif
	/* Enable interrupts and go to sleep until transfer is complete*/
	out_be16(&psc->mpc5xxx_psc_imr, MPC5xxx_PSC_IMR_RXRDY);
	err = wait_event_interruptible(algo_data->wait_queue, 
		algo_data->rx_count == 0);

	/* Terminate the transfer */
	algo_data->rx_ptr = NULL;
	writeb(MPC5xxx_PSC_TX_DISABLE | MPC5xxx_PSC_RX_DISABLE, &psc->command);

	/* Release the semaphore to allow other transfers */
	up(&algo_data->sema);
	return err < 0 ? err : count;
}

/**
 * mpc5xxx_xfer
 * @brief transfer multiple frames via SPI bus. Coltroll CS as specified in frame strucs
 * @param adap  spi adapter structure
 * @param msgs[]  message frame array
 * @param num number of frames
 * @return number of bytes transfered or 0> error code
 *
 */

static int mpc5xxx_psc_xfer(struct spi_adapter	*adap,
			struct spi_msg		msgs[],
			int			num)
{
	struct spi_algo_mpc5xxx_psc_data *algo_data = adap->algo_data;
	struct spi_msg *pmsg;
	int i;
	int cs;
        int ret=0;

	debug("mpc5xxx_xfers:adap=%x msgs=%x num=%d\n",(int)adap,(int)msgs,num);
	i=0;
	while((i < num) && (ret>=0)) {
		pmsg = &msgs[i];
		if (adap->set_spi_adap_cs)
			adap->set_spi_adap_cs(adap, cs=pmsg->addr, pmsg->flags & SPI_M_CS);
		if (algo_data->flags & MPC5XXX_PSC_SPI_USE_FIFO)
			ret = mpc5xxx_psc_fifo_xferbytes(adap, pmsg->buf, pmsg->len,pmsg->flags);
		else
			ret = mpc5xxx_psc_xferbytes(adap, pmsg->buf, pmsg->len,pmsg->flags);
		if (adap->set_spi_adap_cs)
			adap->set_spi_adap_cs(adap, cs=pmsg->addr, pmsg->flags & SPI_M_CSREL);
		i++;
	}
	return (ret > 0) ? ret : -EREMOTEIO;
}

static int mpc5xxx_psc_set_bitrate(struct spi_adapter	*adap,
			       unsigned long		bitrate)	
{
	struct spi_algo_mpc5xxx_psc_data *algo_data = adap->algo_data;
	struct mpc5xxx_psc *psc = algo_data->regs;

	if (bitrate > 16000 || bitrate < 256)
		return -EINVAL;

	psc->ccr=0x0700 | ((66000 / bitrate - 1) & 0xff);                 /* set SCK and DSCKL delay */
	return 0;
}

static struct spi_algorithm mpc5xxx_psc_algo = {
	"MPC5xxx PSC algorithm",
	mpc5xxx_psc_xfer,
	mpc5xxx_psc_set_bitrate
};

/**
 * spi_mpc5xxx_add_bus
 * @brief registers spi bus adapter with our algorithm to spi subsystem
 * @param adap  spi adapter structure
 * @return 0 in sucess -1 in error
 *
 */

int spi_mpc5xxx_psc_add_bus(struct spi_adapter *adap)
{
	adap->algo = &mpc5xxx_psc_algo;
	if (mpc5xxx_psc_spi_init(adap))
		return -1;

	if (spi_add_adapter(adap))
		return -1;

	return 0;
}

int spi_mpc5xxx_psc_del_bus(struct spi_adapter *adap)
{
	if (spi_del_adapter(adap))
		return -1;

	return 0;
}

EXPORT_SYMBOL(spi_mpc5xxx_psc_add_bus);
EXPORT_SYMBOL(spi_mpc5xxx_psc_del_bus);
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Kate Alhola kate(at)katix.org");
MODULE_DESCRIPTION("SPI-Bus MPC5xxx psc spi algorithm module");
