/*
 * drivers/spi/spi-algo-mpc5xxx.c
 *
 * SPI  algorithm driver for MPC5200
 *
 *
 * 2004 (c) Kate Alhola
 *  MPC5200 spi driver original version, deriver from i2c-algo-mpc5200.c
 *
 * 2003-2006 (C) Wolfgang Denk, DENX Software Engineering, <wd@denx.de>.
 *
 * This file is licensed under the terms of the GNU General Public License
 * version 2.  This program is licensed "as is" without any warranty of
 * any kind, whether express or implied.
 */

/**
 * @file spi-algo-mpc5xxx.c
 * @brief  SPI algorithm driver for MPC5200
 *
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/string.h>
#include <asm/io.h>
#include <asm/mpc5xxx.h>
#ifdef CONFIG_UBOOT
#include <asm/ppcboot.h>
#endif

#include <linux/spi.h>
#include "spi-algo-mpc5xxx.h"

#undef DEBUG

#ifdef DEBUG
#define debug(s...) do { printk("%s: ", __FUNCTION__); printk(s); } while (0)
#else
#define debug(s...)
#endif

static int speed = 1000;

MODULE_PARM(speed, "i");

/**
 * mpc5xxx_spi_init
 * @brief Initialize SPI
 * @param algo_data 
 */

static int mpc5xxx_spi_init(struct spi_adapter *adap)
{
	struct mpc5xxx_spi *spi_regs = (struct mpc5xxx_spi*)MPC5xxx_SPI;
	struct mpc5xxx_cdm *cdmregs=(struct mpc5xxx_cdm *)MPC5xxx_CDM;
	struct mpc5xxx_gpio *gpio=(struct mpc5xxx_gpio *)MPC5xxx_GPIO;

	gpio->port_config |= 0x8 << 8;
	cdmregs->clk_enables |= 1 << 11;
	spi_regs->cr1 = MPC5xxx_CR1_SPE | MPC5xxx_CR1_MSTR |
		(adap->bus_params & SPI_CLOCK_ACTIVE_LOW ? 0 : MPC5xxx_CR1_CPHA) | 
		(adap->bus_params & SPI_CLOCK_DATA_START ? MPC5xxx_CR1_CPOL : 0);
	spi_regs->brr = 3;
	spi_regs->ddr = 0xe; 

	debug("mpc5xxx_spi_init:exit\n");

	return 0;
}
/**
 * mpc5xxx_xferbytes
 * @brief transfer bytes via SPI bus. Send data out via MOSI and inputs data from MISO
 * @param adap  spi adapter structure
 * @param buf   message buffer
 * @param count number of bytes to transfer
 * @param flafs read / write mode flag 
 * @return number of bytes transfered or 0>error code
 *
 */
static int mpc5xxx_xferbytes(struct spi_adapter	*adap,char *buf,int count,int flags)
{
	struct mpc5xxx_spi		*spi_regs = (struct mpc5xxx_spi*)MPC5xxx_SPI;
	int				cur = 0;
	unsigned char                   val = 0;

	debug("adap=%x buf=%x count=%d flags=%x\n", (int)adap, (int)buf, count, flags);

	while (cur < count) {
		spi_regs->dr = (flags & SPI_M_WR) ? buf[cur] : 0;
		while (!(spi_regs->sr & MPC5xxx_SR_SPIF));
		val = spi_regs->dr;
		if (flags & SPI_M_RD)
			buf[cur] = val;
		cur++;
	}

	return count;
}

/**
 * mpc5xxx_xfer
 * @brief transfer multiple frames via SPI bus. Coltroll CS as specified in frame strucs
 * @param adap  spi adapter structure
 * @param msgs[]  message frame array
 * @param num number of frames
 * @return number of bytes transfered or 0> error code
 *
 */

static int mpc5xxx_xfer(struct spi_adapter	*adap,
			struct spi_msg		msgs[],
			int			num)
{
	struct spi_msg			*pmsg;
	int				i;
	int                             cs;
        int                             ret=0;

	i = 0;
	while((i < num) && (ret >= 0)) {
		pmsg = &msgs[i];
		if (adap->set_spi_adap_cs)
			adap->set_spi_adap_cs(adap, cs=pmsg->addr, pmsg->flags & SPI_M_CS);
		ret = mpc5xxx_xferbytes(adap, pmsg->buf, pmsg->len,pmsg->flags);
		if (adap->set_spi_adap_cs)
			adap->set_spi_adap_cs(adap, cs=pmsg->addr, pmsg->flags & SPI_M_CSREL);
		i++;
	}
	return (ret > 0) ? ret : -EREMOTEIO;
}

static int mpc5xxx_set_bitrate(struct spi_adapter	*adap,
			       unsigned long		bitrate)	
{
#ifdef CONFIG_UBOOT
	extern unsigned char __res[];
	bd_t *bd = (bd_t *)__res;
	unsigned long ipb_freq = bd->bi_ipbfreq;
#else
	unsigned long ipb_freq = CONFIG_PPC_5xxx_IPBFREQ;
#endif
	struct mpc5xxx_spi * spi_regs = (struct mpc5xxx_spi*)MPC5xxx_SPI;
	unsigned long divisor = ipb_freq / bitrate;
	int spr = 0, sppr = 0;

	if (divisor > 2048)
		return -EINVAL;

	if (divisor > 256) {
		sppr = divisor >> 8;
		if ((sppr << 8) == divisor)
			sppr--;
		divisor /= (sppr + 1);
	}

	spr = ffs(divisor);
	if ((1 << spr) == divisor)
		spr--;

	return -ENOSYS;
}

static struct spi_algorithm mpc5xxx_algo = {
	"MPC5xxx algorithm",
	mpc5xxx_xfer,
	mpc5xxx_set_bitrate
};

/**
 * spi_mpc5xxx_add_bus
 * @brief registers spi bus adapter with our algorithm to spi subsystem
 * @param adap  spi adapter structure
 * @return 0 in sucess -1 in error
 *
 */

int spi_mpc5xxx_add_bus(struct spi_adapter *adap)
{
	adap->algo = &mpc5xxx_algo;
	if (mpc5xxx_spi_init(adap))
		return -1;

	if (spi_add_adapter(adap))
		return -1;

	return 0;
}

int spi_mpc5xxx_del_bus(struct spi_adapter *adap)
{
	if (spi_del_adapter(adap))
		return -1;

	return 0;
}

EXPORT_SYMBOL(spi_mpc5xxx_add_bus);
EXPORT_SYMBOL(spi_mpc5xxx_del_bus);
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Dmitry Konyshev <probables@emcraft.com>");
MODULE_DESCRIPTION("SPI-Bus MPC5xxx spi algorithm module");
