/*
 *  linux/drivers/spi/spi-core.c
 *
 *  Copyright (C) 2004 Kate Alhola
 *  Copyright (C) 2001 Russell King
 *  Copyright (C) 2002 Compaq Computer Corporation
 *
 *  Adapted from l3-core.c by Jamey Hicks <jamey.hicks@compaq.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License.
 *
 *  See linux/Documentation/spi for further documentation.
 */
/**
 * @file spi-core.c
 *
 * @brief SPI ( Serial Periphereal Interface ) subsystem core module.
 * This module is responsible to registering and controlling spi 
 * algorithm drivers, adapters and target drivers
 *
 */

#include <linux/module.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/proc_fs.h>
#include <linux/kmod.h>
#include <linux/init.h>
#include <linux/spi.h>
#include <asm/uaccess.h>


#define SPI_DATE "20041011"
#define SPI_VERSION "0.0.1"

static DECLARE_MUTEX(adapter_lock);
static LIST_HEAD(adapter_list);

static DECLARE_MUTEX(driver_lock);
static LIST_HEAD(driver_list);

/* ---------------------------------------------------
 * /proc entry declarations
 *----------------------------------------------------
 */


static int spiproc_init(void);
static int spiproc_cleanup(void);

static ssize_t spiproc_bus_read(struct file * file, char * buf,size_t count, 
                                loff_t *ppos);
static int read_bus_spi(char *buf, char **start, off_t offset, int len,
                           int *eof , void *private);

/* To implement the dynamic /proc/bus/spi-? files, we need our own 
   implementation of the read hook */
static struct file_operations spiproc_operations = {
	read:		spiproc_bus_read,
};

static int spiproc_initialized = 0;
static int adap_id=0;  /* adapter_id, incremented when new adapter is added */

/**
 * spi_add_adapter - register a new SPI bus adapter
 * @param adap: spi_adapter structure for the registering adapter
 *
 * @brief Make the adapter available for use by clients using name adap->name.
 * The adap->adapters list is initialised by this function.
 *
 * @returns 0;
 */
int spi_add_adapter(struct spi_adapter *adap)
{
    int res=0;
        printk("spi_add_adapter:%8x %s\n",(int)adap,adap->name);
	INIT_LIST_HEAD(&adap->clients);
	down(&adapter_lock);
        init_MUTEX(&adap->lock);
	list_add(&adap->adapters, &adapter_list);
	up(&adapter_lock);

	adap->id=adap_id;
	if (spiproc_initialized) {
		char name[8];
		struct proc_dir_entry *proc_entry;

		sprintf(name,"spi-%d", adap_id);

		proc_entry = create_proc_entry(name,0,proc_bus);
		if (! proc_entry) {
			printk("spi-core.o: Could not create /proc/bus/%s\n",
			       name);
			res = -ENOENT;
		} else
		{
		    proc_entry->proc_fops = &spiproc_operations;
		    proc_entry->owner = THIS_MODULE;
		    adap->inode = proc_entry->low_ino;
		}
	}
	adap_id++;
	return res;	
}

/**
 * spi_del_adapter - unregister a SPI bus adapter
 * @param adap: spi_adapter structure to unregister
 *
 * @brief Remove an adapter from the list of available SPI Bus adapters.
 *
 * @returns 0;
 */
int spi_del_adapter(struct spi_adapter *adap)
{
    int i;
    down(&adapter_lock);
    list_del(&adap->adapters);
    up(&adapter_lock);
	if (spiproc_initialized) {
		char name[8];
		sprintf(name,"spi-%d", i);
		remove_proc_entry(name,proc_bus);
	}
	return 0;
}

static struct spi_adapter *__spi_get_adapter(const char *name)
{
	struct list_head *l;

	list_for_each(l, &adapter_list) {
		struct spi_adapter *adap = list_entry(l, struct spi_adapter, adapters);

		if (strcmp(adap->name, name) == 0)
			return adap;
	}

	return NULL;
}

/**
 * spi_get_adapter - get a reference to an adapter
 * @param name: driver name
 * @returns pointer to spi_adapter struct or NULL if no adapter found
 *
 * @brief Obtain a spi_adapter structure for the specified adapter.  If the adapter
 * is not currently load, then load it.  The adapter will be locked in core
 * until all references are released via spi_put_adapter.
 */
struct spi_adapter *spi_get_adapter(const char *name)
{
	struct spi_adapter *adap = NULL;
	int try;

	for (try = 0; try < 2; try ++) {
		down(&adapter_lock);
		adap = __spi_get_adapter(name);
		if (adap && !try_inc_mod_count(adap->owner))
			adap = NULL;
		up(&adapter_lock);

		if (adap)
			break;

		if (try == 0)
			request_module(name);
	}

	return adap;
}

/**
 * spi_put_adapter - release a reference to an adapter
 * @param adap: driver to release reference
 *
 * Indicate to the SPI core that you no longer require the adapter reference.
 * The adapter module may be unloaded when there are no references to its
 * data structure.
 *
 * You must not use the reference after calling this function.
 */
void spi_put_adapter(struct spi_adapter *adap)
{
	if (adap && adap->owner)
		__MOD_DEC_USE_COUNT(adap->owner);
}

/**
 * spi_add_driver - register a new SPI device driver
 * @param driver - driver structure to make available
 *
 * Make the driver available for use by clients using name driver->name.
 * The driver->drivers list is initialised by this function.
 *
 * Returns 0;
 */
int spi_add_driver(struct spi_driver *driver)
{
        printk("spi_add_driver:%8x %s\n", (int)driver,driver->name);
	down(&driver_lock);
	list_add(&driver->drivers, &driver_list);
	up(&driver_lock);

	/* now look for instances of driver on our adapters
	 */
	if (driver->flags& (SPI_DF_NOTIFY|SPI_DF_DUMMY)) {
	    struct list_head *l;

	    list_for_each(l, &adapter_list) {
		struct spi_adapter *adap = list_entry(l, struct spi_adapter, adapters);
		if(adap) driver->attach_adapter(adap);
	    }
	}
	return 0;
}

/**
 * spi_del_driver - unregister a SPI device driver
 * @param driver: driver to remove
 *
 * Remove an driver from the list of available SPI Bus device drivers.
 *
 * Returns 0;
 */
int spi_del_driver(struct spi_driver *driver)
{
	down(&driver_lock);
	list_del(&driver->drivers);
	up(&driver_lock);
	return 0;
}

static struct spi_driver *__spi_get_driver(const char *name)
{
	struct list_head *l;

	list_for_each(l, &driver_list) {
		struct spi_driver *drv = list_entry(l, struct spi_driver, drivers);

		if (strcmp(drv->name, name) == 0)
			return drv;
	}

	return NULL;
}

/**
 * spi_get_driver - get a reference to a driver
 * @param name: driver name
 *
 * Obtain a spi_driver structure for the specified driver.  If the driver is
 * not currently load, then load it.  The driver will be locked in core
 * until all references are released via spi_put_driver.
 */
struct spi_driver *spi_get_driver(const char *name)
{
	struct spi_driver *drv = NULL;
	int try;

	for (try = 0; try < 2; try ++) {
		down(&adapter_lock);
		drv = __spi_get_driver(name);
		if (drv && !try_inc_mod_count(drv->owner))
			drv = NULL;
		up(&adapter_lock);

		if (drv)
			break;

		if (try == 0)
			request_module(name);
	}

	return drv;
}

/**
 * spi_put_driver - release a reference to a driver
 * @param drv: driver to release reference
 *
 * Indicate to the SPI core that you no longer require the driver reference.
 * The driver module may be unloaded when there are no references to its
 * data structure.
 *
 * You must not use the reference after calling this function.
 */
void spi_put_driver(struct spi_driver *drv)
{
	if (drv && drv->owner)
		__MOD_DEC_USE_COUNT(drv->owner);
}

/**
 * spi_attach_client - attach a client to an adapter and driver
 * @param client: client structure to attach
 * @param adap: adapter (module) name
 * @param drv: driver (module) name
 *
 * Attempt to attach a client (a user of a device driver) to a particular
 * driver and adapter.  If the specified driver or adapter aren't registered,
 * request_module is used to load the relevant modules.
 *
 * Returns 0 on success, or negative error code.
 */
int spi_attach_client(struct spi_client *client, const char *adap, const char *drv)
{
	struct spi_adapter *adapter = spi_get_adapter(adap);
	struct spi_driver  *driver = spi_get_driver(drv);
	int ret = -ENOENT;

        printk("SPI: attaching client %p\n", client);

	if (!adapter)
		printk(KERN_ERR  ": unable to get adapter: %s\n",adap);
	if (!driver)
		printk(KERN_ERR  ": unable to get driver: %s\n", drv);

	if (adapter && driver) {
		ret = 0;

		client->adapter = adapter;
		client->driver  = driver;

		list_add(&client->__adap, &adapter->clients);

		if (driver->attach_adapter)
		ret = driver->attach_adapter(client);
	}

	if (ret) {
		spi_put_driver(driver);
		spi_put_adapter(adapter);
	}
	return ret;
}

/**
 * spi_detach_client - detach a client from an adapter and driver
 * @param client: client structure to detach
 *
 * Detach the client from the adapter and driver.
 */
void spi_detach_client(struct spi_client *client)
{
	struct spi_adapter *adapter = client->adapter;
	struct spi_driver  *driver = client->driver;

	driver->detach_client(client);

	client->adapter = NULL;
	client->driver  = NULL;

	spi_put_driver(driver);
	spi_put_adapter(adapter);

	list_del(&client->__adap);

	return;
}

/**
 * spi_transfer - transfer information on an SPI bus
 * @param adap: adapter structure to perform transfer on
 * @param msgs: array of spi_msg structures describing transfer
 * @param num: number of spi_msg structures
 *
 * Transfer the specified messages to/from a device on the SPI bus.
 *
 * Returns number of messages successfully transferred, otherwise negative
 * error code.
 */
int spi_transfer(struct spi_adapter *adap, struct spi_msg msgs[], int num)
{
	int ret = -ENOSYS;

	if (adap->algo->xfer) {
		down(&adap->lock);
		ret = adap->algo->xfer(adap, msgs, num);
		up(&adap->lock);
	}
	return ret;
}

/**
 * spi_read_write - send and receive data to/from a device on an SPI bus
 * @param client: registered client structure
 * @param addr: SPI bus address
 * @param buf: buffer for bytes to send/receive
 * @param len: number of bytes to send/receive
 *
 * @brief Send len bytes pointed to by buf to device address addr and at the
 * same time receive len bytes from device address addr on the SPI bus
 * described by client to a buffer pointed to by buf.
 *
 * @returns the number of bytes transferred, or negative error code.
 */
int spi_read_write(struct spi_client *client, int addr, const char *buf, int len)
{
	struct spi_adapter *adap = client->adapter;
	struct spi_msg msg;
	int ret;

	msg.addr = addr;
	msg.flags = SPI_M_WR | SPI_M_RD;
	msg.buf = (char *)buf;
	msg.len = len;

	ret = spi_transfer(adap, &msg, 1);
	return ret == 1 ? len : ret;
}

/**
 * spi_write - send data to a device on an SPI bus
 * @param client: registered client structure
 * @param addr: SPI bus address
 * @param buf: buffer for bytes to send
 * @param len: number of bytes to send
 *
 * @brif Send len bytes pointed to by buf to device address addr on the SPI bus
 * described by client.
 *
 * @returns the number of bytes transferred, or negative error code.
 */
int spi_write(struct spi_client *client, int addr, const char *buf, int len)
{
	struct spi_adapter *adap = client->adapter;
	struct spi_msg msg;
	int ret;

	msg.addr = addr;
	msg.flags = SPI_M_WR;
	msg.buf = (char *)buf;
	msg.len = len;

	ret = spi_transfer(adap, &msg, 1);
	return ret == 1 ? len : ret;
}

/**
 * spi_read - receive data from a device on an SPI bus
 * @param client: registered client structure
 * @param addr: SPI bus address
 * @param buf: buffer for bytes to receive
 * @param len: number of bytes to receive
 *
 * Receive len bytes from device address addr on the SPI bus described by
 * client to a buffer pointed to by buf.
 *
 * Returns the number of bytes transferred, or negative error code.
 */
int spi_read(struct spi_client *client, int addr, char *buf, int len)
{
	struct spi_adapter *adap = client->adapter;
	struct spi_msg msg;
	int ret;

	msg.addr = addr;
	msg.flags = SPI_M_RD;
	msg.buf = buf;
	msg.len = len;

	ret = spi_transfer(adap, &msg, 1);
	return ret == 1 ? len : ret;
}

/**
 * read_bus_spi - generates the output for /proc/bus/spi 
 * @param buf: buffer for bytes to send
 * @param offset: offset from begining of the file
 * @param len: number of bytes to send
 * @param eof: end of file flag ( not used in current version )
 * @param private: not user
 *
 * This function generates the output for /proc/bus/spi 
 *
 * Returns the number of bytes transferred, or negative error code.
 */

int read_bus_spi(char *buf, char **start, off_t offset, int len, int *eof, 
                 void *private)
{
	int i = 0;
	struct list_head *l;
	int nr = 0;
	/* Note that it is safe to write a `little' beyond len. Yes, really. */

	list_for_each(l, &adapter_list) {
		struct spi_adapter *adap = list_entry(l, struct spi_adapter, adapters);
		if (adap) {
		    nr += sprintf(buf+nr, "spi-%d\t", i);
		    nr += sprintf(buf+nr,"\t%-32s\t%-32s\n",
			              adap->name,
			              adap->algo->name);
		    i++;
		}
	}
	return nr;
}

/**
 * spiproc_bus_read - generates the output for /proc/bus/spi-? 
 * @param file: 
 * @param buf: buffer for bytes to send
 * @param offser: offset from begining of the file
 * @param count: number of bytes to send
 * @param ppos: 
 * This function generates the output for /proc/bus/spi-? 
 *
 * Returns the number of bytes transferred, or negative error code.
 */


ssize_t spiproc_bus_read(struct file * file, char * buf,size_t count, 
                         loff_t *ppos)
{
	struct inode * inode = file->f_dentry->d_inode;
	char *kbuf;
	int len=0;
	size_t len_total;
	struct list_head *l,*cl;

	if (count > 8192)
		return -EINVAL; 
	len_total = file->f_pos + count;
	/* Too bad if this gets longer (unlikely) */
	if (len_total > 4000)
		len_total = 4000;
	list_for_each(l, &adapter_list) {
	    struct spi_adapter *adap = list_entry(l, struct spi_adapter, adapters);
	    if (adap->inode == inode->i_ino) {
		/* We need a bit of slack in the kernel buffer; this makes the
		   sprintf safe. */
		if (! (kbuf = kmalloc(count + 80,GFP_KERNEL)))
		    return -ENOMEM;
		list_for_each(cl, &(adap->clients)) {
		    struct spi_client *client = list_entry(cl, struct spi_client, __adap);
	    		if (client ) 
			    len += sprintf(kbuf+len,"%-32s\t%-32s\n",
				              client->name,
				              client->driver->name);
		};
	    }
	}
	len = len - file->f_pos;
	if (len > count)
	    len = count;
	if (len < 0) 
	    len = 0;
	if (copy_to_user (buf,kbuf+file->f_pos, len)) {
	    kfree(kbuf);
	    return -EFAULT;
	}
	file->f_pos += len;
	kfree(kbuf);
	return len;
}
/**
 * spiproc_init
 * @brief initialize spi sybsystems entries in proc filesystem 
 *
 */

int spiproc_init(void)
{

	struct proc_dir_entry *proc_bus_spi;

	spiproc_initialized = 0;

	if (! proc_bus) {
		printk(KERN_ERR "spi-core.o: /proc/bus/ does not exist");
		spiproc_cleanup();
		return -ENOENT;
 	} 
	proc_bus_spi = create_proc_entry("spi",0,proc_bus);
	if (!proc_bus_spi) {
		printk(KERN_ERR "spi-core.o: Could not create /proc/bus/spi");
		spiproc_cleanup();
		return -ENOENT;
 	}
	proc_bus_spi->read_proc = &read_bus_spi;
	proc_bus_spi->owner = THIS_MODULE;
	spiproc_initialized += 2;
	return 0;
}

int spiproc_cleanup(void)
{

	if (spiproc_initialized >= 1) {
		remove_proc_entry("spi",proc_bus);
		spiproc_initialized -= 2;
	}
	return 0;
}

/**
 * spi_adapter_id  This call returns a unique low identifier for each registered adapter
 * @param adap:spi_adapter 
 * @returns unique id or -1 if the adapter was not registered. 
 */

int spi_adapter_id(struct spi_adapter *adap)
{
    return adap->id;
}


static int __init spi_init(void)
{
	printk(KERN_INFO "spi-core: SPI core module version %s (%s)\n", SPI_VERSION, SPI_DATE);

	init_MUTEX(&driver_lock);
	
	spiproc_init();
	
	return 0;
}


EXPORT_SYMBOL(spi_add_adapter);
EXPORT_SYMBOL(spi_del_adapter);
EXPORT_SYMBOL(spi_get_adapter);
EXPORT_SYMBOL(spi_put_adapter);

EXPORT_SYMBOL(spi_add_driver);
EXPORT_SYMBOL(spi_del_driver);
EXPORT_SYMBOL(spi_get_driver);
EXPORT_SYMBOL(spi_put_driver);

EXPORT_SYMBOL(spi_attach_client);
EXPORT_SYMBOL(spi_detach_client);

EXPORT_SYMBOL(spi_transfer);
EXPORT_SYMBOL(spi_read_write);
EXPORT_SYMBOL(spi_write);
EXPORT_SYMBOL(spi_read);
EXPORT_SYMBOL(spi_adapter_id);

#ifdef MODULE

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Kate Alhola kate(at)katix.org");
MODULE_DESCRIPTION("SPI-Bus main module");

#endif

int init_module(void) 
{
	return spi_init();
}

void cleanup_module(void) 
{
	spiproc_cleanup();
}

#ifndef MODULE

module_init(init_module);
module_exit(cleanup_module);

#endif
