/*
 * drivers/spi/spi-eval-dev.c
 *
 * SPI device node driver for mpc5200 board with MPC5xxx
 *
 * 2005 (c) Wolfgang Denk, DENX Software Engineering, <wd@denx.de>.
 *
 * This file is licensed under the terms of the GNU General Public License
 * version 2.  This program is licensed "as is" without any warranty of
 * any kind, whether express or implied.
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <asm/mpc5xxx.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <linux/spi.h>
#include <linux/spi-eval.h>
#include <linux/proc_fs.h>
#include <linux/miscdevice.h>

#undef DEBUG
#undef TRACE

#ifdef DEBUG
#define debug(s...)	do { printk("%s: ", __FUNCTION__); printk(s); } while (0)
#else
#define debug(s...)
#endif

#ifdef TRACE
#define ENTER()		do { printk("%s: enter\n", __FUNCTION__); } while (0)
#define LEAVE()		do { printk("%s: leave\n", __FUNCTION__); } while (0)
#else
#define ENTER()
#define LEAVE()
#endif

#define GPIO_PSC2_0		(1 << 4)
#define DEV_NAME		"spi_eval"

static ssize_t spi_eval_read(struct file *file, char *buf, size_t count, loff_t *offset);
static ssize_t spi_eval_write(struct file *file, const char *buf, size_t count, loff_t *offset);
static int spi_eval_open(struct inode *inode, struct file *file);
static int spi_eval_release(struct inode *inode, struct file *file);

static int spi_eval_attach_adapter(struct spi_adapter *adap);
static int spi_eval_detach_client(struct spi_client *client);

static int spi_eval_command(struct spi_client *client, unsigned int cmd, void *arg);
static int spi_eval_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);

static struct file_operations spi_eval_fops = {
	owner:		THIS_MODULE,
	read:		spi_eval_read,
	write:		spi_eval_write,
	ioctl:		spi_eval_ioctl,
	open:		spi_eval_open,
	release:	spi_eval_release,
};

static struct miscdevice spi_eval_misc_dev = {
	MISC_DYNAMIC_MINOR,
	DEV_NAME,
	&spi_eval_fops
};

static char	*read_buf	= NULL;
static size_t	buf_size	= 0;
static int	busy		= 0;

static int	proc_open	= 0;
static int	proc_read	= 0;
static int	proc_write	= 0;

static struct proc_dir_entry *proc_entry	= NULL;
static struct spi_adapter *spi_eval_adaps	= NULL;

static struct spi_driver spi_eval_driver = {
	name:		"SPI-Evaluator device driver",
	flags:		SPI_DF_DUMMY,
	attach_adapter:	spi_eval_attach_adapter,
	detach_client:	spi_eval_detach_client,
	command:	spi_eval_command,
	inc_use:	NULL,
	dec_use:	NULL,
};

static struct spi_client spi_eval_client_template = {
	name:		"SPI-Evaluator device entry",
	adapter:	NULL,
	driver:		&spi_eval_driver,
	addr:           0,
};

static void gpio_init(void)
{
	struct mpc5xxx_gpio *gpio = (struct mpc5xxx_gpio *)MPC5xxx_GPIO;

	gpio->port_config &= ~(7 << 4);
	gpio->simple_gpioe |= GPIO_PSC2_0;
	gpio->simple_ode &= ~GPIO_PSC2_0;
	gpio->simple_ddr |= GPIO_PSC2_0;
	gpio->simple_dvo &= ~GPIO_PSC2_0;
}

static void gpio_set(int slave)
{
	struct mpc5xxx_gpio *gpio = (struct mpc5xxx_gpio *)MPC5xxx_GPIO;

	if (slave)
		gpio->simple_dvo |= GPIO_PSC2_0;
	else
		gpio->simple_dvo &= ~GPIO_PSC2_0;
};

static inline char* alloc_read_buf(size_t size)
{
	read_buf = kmalloc(size, GFP_KERNEL);

	if (read_buf)
		buf_size = size;
	else
		buf_size = 0;
	
	return read_buf;
}

static inline void free_read_buf(void)
{
	kfree(read_buf);
	read_buf = NULL;
	buf_size = 0;
}

static ssize_t spi_eval_read(struct file *file, char *buf, size_t count, loff_t *offset)
{
	ENTER();

	int ret;

	if (!read_buf) {
		proc_read++;
		return 0;
	}
		
	ret = min(count, buf_size);
	if (copy_to_user(buf, read_buf, ret))
		ret = -EFAULT;

	free_read_buf();
	
	LEAVE();
	return ret;
}

static ssize_t spi_eval_write(struct file *file, const char *buf, size_t count, loff_t *offset)
{
	ENTER();

	int ret;
	struct spi_client *client = (struct spi_client*)file->private_data;
	
	if (read_buf) {
		proc_write++;
		free_read_buf();
	}

	read_buf = alloc_read_buf(count);
	if (!read_buf)
		return -ENOMEM;

	if (copy_from_user(read_buf, buf, count)) {
		ret = -EFAULT;
		goto Fail;
	}

	ret = spi_read_write(client, client->addr, read_buf, count);
	if (ret < 0)
		goto Fail;

	buf_size = ret;

	LEAVE();
	return ret;
	
 Fail:
	free_read_buf();
	return ret;
}

int spi_eval_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	ENTER();
	
	if (SPI_EVAL_SLAVE_SELECT != cmd)
		return -ENOTTY;

	if (arg > 1)
		return -EINVAL;
		
	free_read_buf();
	gpio_set(arg);

	LEAVE();	
	return 0;
}

int spi_eval_open(struct inode *inode, struct file *file)
{
	ENTER();

	struct spi_client *client;
	
	if (busy) {
		proc_open++;
		return -EBUSY;
	}
	
	if (!spi_eval_adaps)
		return -ENODEV;

	client = kmalloc(sizeof(struct spi_client), GFP_KERNEL);
	if (!client)
		return -ENOMEM;

	memcpy(client, &spi_eval_client_template, sizeof(struct spi_client));
	client->adapter = spi_eval_adaps;
	file->private_data = client;

	busy++;

	LEAVE();
	return 0;
}

static int spi_eval_release(struct inode *inode, struct file *file)
{
	ENTER();

	struct spi_cliente *client = file->private_data;

	kfree(client);
	file->private_data = NULL;
	
	free_read_buf();

	busy--;

	LEAVE();
	return 0;
}

int spi_eval_attach_adapter(struct spi_adapter *adap)
{
	ENTER();

	int id;

	if ((id = spi_adapter_id(adap)) < 0)
		return -ENODEV;
		
	if (!spi_eval_adaps)
		spi_eval_adaps = adap;
	else
		spi_eval_adaps = NULL;

	LEAVE();
	return 0;
}

int spi_eval_detach_client(struct spi_client *client)
{
	ENTER();
	LEAVE();
	return 0;
}

static int spi_eval_command(struct spi_client *client, unsigned int cmd, void *arg)
{
	ENTER();
	LEAVE();
	return -1;
}

static inline int read_bus_spi(char *buf, char **start, off_t offset, int len, int *eof, void *private)
{
	ENTER();

	int ret = 0;

	ret += sprintf(buf + ret, "%d\tattempts to open already opened device\n", proc_open);
	ret += sprintf(buf + ret, "%d\twrites overwriting the read buffer\n", proc_write);
	ret += sprintf(buf + ret, "%d\treads without preceding write\n", proc_read);

	LEAVE();
	return ret;
}

static inline int proc_init(void)
{
	if (!proc_bus)
		return -ENOENT;

	proc_entry = create_proc_entry(DEV_NAME, 0, proc_bus);

	if (!proc_entry)
		return -ENOENT;

	proc_entry->read_proc = &read_bus_spi;
	proc_entry->owner = THIS_MODULE;

	return 0;
}

static inline void proc_cleanup(void)
{
	if (proc_entry)
		remove_proc_entry(DEV_NAME, proc_bus);
}

static int __init spi_eval_init(void)
{
	ENTER();

	int ret;
	
	if (misc_register(&spi_eval_misc_dev))
		return -EIO;

	ret = spi_add_driver(&spi_eval_driver);
	if (ret) {
		misc_deregister(&spi_eval_misc_dev);
		return ret;
	}

	gpio_init();
	proc_init();

	LEAVE();
	return 0;
}

static void __exit spi_eval_cleanup(void)
{
	ENTER();

	free_read_buf();
	proc_cleanup();
	spi_del_driver(&spi_eval_driver);
	misc_deregister(&spi_eval_misc_dev);

	LEAVE();
}

EXPORT_NO_SYMBOLS;

MODULE_DESCRIPTION("SPI-Evaluator device node driver");
MODULE_LICENSE("GPL");

module_init(spi_eval_init);
module_exit(spi_eval_cleanup);

