/*
 * drivers/spi/spi_iti5200.c
 *
 * SPI adapter driver for iti5200 board with MPC5xxx
 *
 * 2003 (c) Wolfgang Denk, DENX Software Engineering, <wd@denx.de>.  
 * 2004 (c) Kate Alhola
 * This
 * file is licensed under the terms of the GNU General Public License
 * version 2.  This program is licensed "as is" without any warranty of
 * any kind, whether express or implied.
 */
/**
 * @file spi-iti5200.c
 * @author Kate Alhola 
 * @brief  SPI adapter driver for iti5200 board with MPC5xxx
 *
 */


/*
SPI assigments for ITI5200 and Link400 boards
	Link400		ITI 1.0
PSC1_0	Uart_TXD	Uart_TXD
PCS1_1  Uart_RXD	Uart_RXD
PSC1_2  Uart_RTS(nc)	Uart_RTS
PSC1_3  Uart_CTS(nc)	Uart_CTS	
PSC1_4	Codec_CS	Codec_CS  SPI_CS0  TSC2301_CS GPIO_WKUP_0

PSC2_0	I2S_TXD/TXD2	I2S_TXD
PSC2_1	I2S_RXD/RXD2	I2S_RXD
PSC2_2	I2S_MCLK/RTS2	I2S_MCLK
PSC2_3	I2S_CLK/CTS2	I2S_CLK
PSC2_4	I2S_FRAME	I2S_FRAME

PSC3_0	SPI_MOSI	SPI_MOSI
PSC3_1	SPI_MISO	SPI_MISO
PSC3_2	SPI_CLK		SPI_CLK
PSC3_3	SPI_FRAME	SPI_FRAME
PSC3_4	MMC_CS		MMC_CS    SPI_CS1  MMC_CS    GPIO_SINT_0
PSC3_5	LCD_CD		LCD_CS    SPI_CS2  LCD_CS    GPIO_SINT_1
PSC3_6	LCD_CD	        LCD_CD    SPI_CS3 /LCD_CD    GPIO_PSC3_6
PSC3_7	RS2/DSR		J5.32                        GPIO_PSC3_7
PSC3_8	RS2/DTR		J5.34                        GPIO_SINT_2
PSC3_9	NAND_BSY        NAND_BSY                     GPIO_WKUP_2



*/


#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/string.h>
#include <asm/mpc5xxx.h>

#include <linux/spi.h>
#include "spi-algo-mpc5xxx-psc.h"

#define MPC5xxx_SPI_ENABLE	1	/* Disable  */

#define GPIO_PSC3_4 0x1000
#define GPIO_SINT0 1
#define GPIO_SINT1 2


struct mpc5xxx_gpio *itipower5200_gpio=(struct mpc5xxx_gpio *)MPC5xxx_GPIO;
struct mpc5xxx_gpw  *itipower5200_gpw =(struct mpc5xxx_gpw  *)MPC5xxx_GPW;

int set_spi_iti5200_cs(struct spi_adapter *,int id,int val);
static int __init spi_iti5200_init(void);

/** iti5200_spi_data_1 spi adapter algorithm specific data */
static struct spi_algo_mpc5xxx_psc_data iti5200_spi_data_1 = {
    (struct mpc5xxx_psc *)MPC5xxx_PSC3, /** use PSC3 in iti5200 and derivatives */
};

/** SPI CS assigment table */
struct spi_cs spi_cs_iti5200[] = { 
	{"tsc2301",0}, 
	{"mmc",1},
	{"lcd",2},
	{"lcd_cd",3},
	    };



/** spi adapter specific data */
static struct spi_adapter iti5200_spi_ops_1 = {
    name:            "iti5200 SPI interface",
    algo:   NULL,
    algo_data:       &iti5200_spi_data_1, /**< algorithm specific data */
    set_spi_adap_cs: &set_spi_iti5200_cs, /**< chip select controll function */
    cs:          NULL,                    /**< chip select asigment table */
};
 
/**
 * set_spi_iti5200_cs set adapter spi cs line
 * @param adap   spi_adapter structure
 * @param id     chip select line id
 * @param val    chip select line value
 * @returns 0
 */

 int set_spi_iti5200_cs(struct spi_adapter *adap,int id,int val)
 {
     printk("set_spi_iti5200_cs: adap=%x id=%d val=%d\n",(int)adap,id,val);
 	switch(id)
	{
		case 0:
		    if(val)
			itipower5200_gpw->wdvo|= GPIO_WKUP_0;
		    else
			itipower5200_gpw->wdvo&=~GPIO_WKUP_0;
		break;
		case 1:
		    if(val)
			itipower5200_gpio->sint_dvo|= GPIO_SINT0;
		    else
			itipower5200_gpio->sint_dvo&=~GPIO_SINT0;
		break;
		case 2:
		    if(val)
			itipower5200_gpio->sint_dvo|= GPIO_SINT1;
		    else
			itipower5200_gpio->sint_dvo&=~GPIO_SINT1;
		break;
		case 3:
		    if(val)
			itipower5200_gpio->simple_dvo|= GPIO_PSC3_4;
		    else
			itipower5200_gpio->simple_dvo&=~GPIO_PSC3_4;
		break;
	}
 
	return 0;
 
 };

/**
 * spi_iti5200_init
 * @param void
 * @returns 0 in success or error status
 * @brief initializes chip select lines for iti5200 board and then registers 
 * to spi_mpc5xxxx_add_bus to mpc5200psc spi driver ( spi-algo-mpc5200psc)
 *
 */

static int __init spi_iti5200_init(void)
{
	if (MPC5xxx_SPI_ENABLE) {
	    itipower5200_gpio=(struct mpc5xxx_gpio *)MPC5xxx_GPIO;
            itipower5200_gpw=(struct mpc5xxx_gpw *)MPC5xxx_GPW;
	    /* SPI_CS0 / TSC2301_CS */
	    itipower5200_gpw->wgpioe|=GPIO_WKUP_0; /* Make it GPIO */  
	    itipower5200_gpw->wdvo  |=GPIO_WKUP_0; /* Initially 1 */
	    itipower5200_gpw->wddr  |=GPIO_WKUP_0; /* make it output */
            /* SPI_CS3 / PSC3_6 GPIO_PSC3_4 */ 
	    itipower5200_gpio->simple_gpioe|=GPIO_PSC3_4; /* Make it GPIO */  
	    itipower5200_gpio->simple_dvo  |=GPIO_PSC3_4; /* Initially 1 */
	    itipower5200_gpio->simple_ddr  |=GPIO_PSC3_4; /* make it output */
	    /* SPI_CS1 and SPI_CS1 / PSC3_4 and PSC3_5 / GPIO_SINT0 and GPIO_SINT1 */
	    itipower5200_gpio->sint_gpioe|=GPIO_SINT0 | GPIO_SINT1; /* Make it GPIO */  
	    itipower5200_gpio->sint_dvo  |=GPIO_SINT0 | GPIO_SINT1; /* Initially 1 */
	    itipower5200_gpio->sint_ddr  |=GPIO_SINT0 | GPIO_SINT1; /* make it output */

	    if (spi_mpc5xxx_psc_add_bus(&iti5200_spi_ops_1))
		return -ENODEV;
	    printk(KERN_INFO "spi-iti5200: SPI module #1 installed PSC=%x\n",(int)iti5200_spi_data_1.regs);
	}
	
	
	
	return 0;
}
/**
 * spi_iti5200_exit plugh
 * @returns 0 in success or error status
 * @brief unregisters adapter with spi_mpc5xxxx_del_bus to mpc5200psc spi driver ( spi-algo-mpc5200psc)
 *
 */

static void __exit spi_iti5200_exit(void)
{
	if (MPC5xxx_SPI_ENABLE) {
		if (spi_mpc5xxx_psc_del_bus(&iti5200_spi_ops_1))
			return;
		printk(KERN_INFO "spi_iti5200.o: SPI module #1 uninstalled\n");
	}
}

module_init(spi_iti5200_init);
module_exit(spi_iti5200_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Kate Alhola kate(at)katix.org");
MODULE_DESCRIPTION("iti5200 SPI adapter");
