/*
 * drivers/spi/spi_mcc200.c
 *
 * SPI adapter driver for MPC5200 based board
 *
 * 2005 (c) Wolfgang Denk, DENX Software Engineering, <wd@denx.de>
 *
 * This file is licensed under the term of  the GNU General Public License
 * version 2. The program licensed "as is" without any warranty of any
 * kind, whether express or implied.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <asm/mpc5xxx.h>

#include <linux/spi.h>
#include "spi-algo-mpc5xxx-psc.h"

#define MCC200_SPI_ACCEL_ADAPTER_NAME	"mcc200-accel"
#define MCC200_SPI_GYROES_ADAPTER_NAME	"mcc200-gyroes"
#define MCC200_SPI_PSOC_ADAPTER_NAME	"mcc200-psoc"

	/* algorithm-specific data */
static struct spi_algo_mpc5xxx_psc_data mcc200_accel_algo_data = {
	(struct mpc5xxx_psc *)MPC5xxx_PSC3, /* use PSC3 */
#ifdef	CONFIG_SPI_5xxx_PSC3_FIFO
	MPC5XXX_PSC_SPI_USE_FIFO,
#else
	0,
#endif
};
static struct spi_algo_mpc5xxx_psc_data mcc200_gyroes_algo_data = {
	(struct mpc5xxx_psc *)MPC5xxx_PSC6, /* use PSC6 */
#ifdef	CONFIG_SPI_5xxx_PSC6_FIFO
	MPC5XXX_PSC_SPI_USE_FIFO,
#else
	0,
#endif
};
static struct spi_algo_mpc5xxx_psc_data mcc200_psoc_algo_data = {
	(struct mpc5xxx_psc *)MPC5xxx_PSC2, /* use PSC2 */
#ifdef	CONFIG_SPI_5xxx_PSC2_FIFO
	MPC5XXX_PSC_SPI_USE_FIFO,
#else
	0,
#endif
};

static int mcc200_adap_accel_cs(struct spi_adapter *adap, int id, int val)
{
	if (id != 0) {
		printk(KERN_ERR "%s: Unknown SPI device ID %i on PCS3\n",
		       __FUNCTION__, id);
		return -ENODEV;
	}

	if (val & SPI_M_CS)
		*(volatile int*)MPC5xxx_GPT5_ENABLE &= ~0x10;
	if (val & SPI_M_CSREL)
		*(volatile int*)MPC5xxx_GPT5_ENABLE |= 0x10;
	udelay(1);

	return 0;
}

static int mcc200_adap_gyro_cs(struct spi_adapter *adap, int id, int val)
{
	volatile int * port;

	switch (id) {
	case 0: /* GyroX */
		port = (volatile int*)MPC5xxx_GPT4_ENABLE;
		break;
	case 1: /* GyroY */
		port = (volatile int*)MPC5xxx_GPT3_ENABLE;
		break;
	case 2: /* GyroZ */
		port = (volatile int*)MPC5xxx_GPT2_ENABLE;
		break;
	default:
		printk(KERN_ERR "%s: Unknown SPI device ID %i on PSC6\n",
		       __FUNCTION__, id);
		return -ENODEV;
	}
		
	if (val & SPI_M_CS)
		*port &= ~0x10;
	if (val & SPI_M_CSREL)
		*port |= 0x10;
	udelay(1);

	return 0;
}

static int mcc200_adap_psoc_cs(struct spi_adapter *adap, int id, int val)
{
	volatile struct mpc5xxx_psc *ppsc = (struct mpc5xxx_psc *)(MPC5xxx_PSC2);

	if (id != 0) {
		printk(KERN_ERR "%s: Unknown SPI device ID %i on PCS2\n", __FUNCTION__, id);
		return -ENODEV;
	}

	if (val & SPI_M_CS)
		ppsc->op1 = MPC5xxx_PSC_OP1_RTS; /* assert ouput port #RTS(PSC2_2) (#SS SPI signal) */

	if (val & SPI_M_CSREL)
		ppsc->op0 = MPC5xxx_PSC_OP0_RTS; /* negate output port #RTS(PSC2_2) (#SS SPI signal) */

	udelay(1);

	return 0;
}


	/* adapter-specific data */
static struct spi_adapter mcc200_spi_accel_adapter = {
	.name		= MCC200_SPI_ACCEL_ADAPTER_NAME,
	.algo		= NULL,
	.algo_data	= &mcc200_accel_algo_data,  /* algorithm specific data */
	.set_spi_adap_cs = NULL,
	.bus_params	= SPI_CLOCK_ACTIVE_LOW | SPI_CLOCK_DATA_START,
};
static struct spi_adapter mcc200_spi_gyroes_adapter = {
	.name		= MCC200_SPI_GYROES_ADAPTER_NAME,
	.algo		= NULL,
	.algo_data	= &mcc200_gyroes_algo_data,  /* algorithm specific data */
	.set_spi_adap_cs = mcc200_adap_gyro_cs,
	.bus_params	= SPI_CLOCK_ACTIVE_LOW | SPI_CLOCK_DATA_START,
};
static struct spi_adapter mcc200_spi_psoc_adapter = {
	.name		= MCC200_SPI_PSOC_ADAPTER_NAME,
	.algo		= NULL,
	.algo_data	= &mcc200_psoc_algo_data,  /* algorithm specific data */
	.set_spi_adap_cs = mcc200_adap_psoc_cs,
	.bus_params	= SPI_CLOCK_ACTIVE_LOW,
};


	/* module initialization function */	
static int __init mcc200_spi_init(void)
{
	int ret, ret2, ret3;

	printk("MCC200 SPI adapter driver: Revision 0.1\n");

	/* Setup chip selects */
	*(volatile int*)MPC5xxx_GPT5_ENABLE = 0x34;
#ifndef CONFIG_PRS200
	*(volatile int*)MPC5xxx_GPT4_ENABLE = 0x34;
	*(volatile int*)MPC5xxx_GPT3_ENABLE = 0x34;
	*(volatile int*)MPC5xxx_GPT2_ENABLE = 0x34;
#endif
	/* PSoC CS setup */
	mcc200_adap_psoc_cs (NULL, 0, SPI_M_CSREL);

	/* reset the Accel device */
	*(volatile int*)MPC5xxx_GPT6_ENABLE = 0x24;
	udelay(1000);
	*(volatile int*)MPC5xxx_GPT6_ENABLE = 0x34;

	ret = spi_mpc5xxx_psc_add_bus(&mcc200_spi_accel_adapter);
#ifndef CONFIG_PRS200
	ret2 = spi_mpc5xxx_psc_add_bus(&mcc200_spi_gyroes_adapter);
#endif
	ret3 = spi_mpc5xxx_psc_add_bus(&mcc200_spi_psoc_adapter);

	return !ret || !ret2 || !ret3;
}

	/* module uninitialization function */
static void __exit mcc200_spi_cleanup(void)
{
	spi_mpc5xxx_psc_del_bus(&mcc200_spi_accel_adapter);
#ifndef CONFIG_PRS200
	spi_mpc5xxx_psc_del_bus(&mcc200_spi_gyroes_adapter);
#endif
	spi_mpc5xxx_psc_del_bus(&mcc200_spi_psoc_adapter);
	printk("MCC200 SPI adapter driver unloaded\n");
}

module_init(mcc200_spi_init);
module_exit(mcc200_spi_cleanup);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Dmitry Konyshev, <probables@emcraft.com>");
MODULE_DESCRIPTION("MCC200 SPI adapter");
