/*
 * drivers/spi/spi_mcc200.c
 *
 * SPI device driver for the MCC200 board, based on the
 * "spi-dev" driver by Kate Alhola, kate (at) katix.org
 * 
 * Main features:
 * - Controls multiple SPI devices using device addressing (see SPIDEV,
 *    SPIBUS, SPIMKID in mcc200spi.h).
 * - Runs timed SPI write operations and collects device data
 *    for subsequent read.
 *
 * 2006 (c) Wolfgang Denk, DENX Software Engineering, <wd@denx.de>
 *
 * This file is licensed under the term of  the GNU General Public License
 * version 2. The program licensed "as is" without any warranty of any
 * kind, whether express or implied.
 */


#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/version.h>
#include <linux/smp_lock.h>
#ifdef CONFIG_DEVFS_FS
#include <linux/devfs_fs_kernel.h>
#endif
#include <linux/init.h>
#include <asm/uaccess.h>
#include <asm/mpc5xxx.h>
#ifdef CONFIG_UBOOT
#include <asm/ppcboot.h>
#endif
#include <linux/mcc200spi.h>
#include <linux/tqueue.h>

/* If you want debugging uncomment: */
#undef DEBUG

#define SPI_MAX_MSG_SIZE	8192
#define SPI_MAX_MSGS		42
#define SPI_ADAP_MAX		4
#define SPIDEV_ADAPS_MAX	SPI_ADAP_MAX
#define SPI_CMDQ_MAX		1024
#define SPI_DATAQ_MAX		4096

#define MCC200_GPT (MPC5xxx_GPT + 0x50)	/* GPT5 */
#define TIMER5_IRQ  (MPC5xxx_MAIN_IRQ_BASE + 14)

# define TIMEVAL_TO_TIMESPEC(tv, ts) {		\
	(ts)->tv_sec = (tv)->tv_sec;		\
	(ts)->tv_nsec = (tv)->tv_usec * 1000;	\
}

#ifdef CONFIG_UBOOT
extern unsigned char __res[];
#define MPC5xxx_IPBFREQ (((bd_t *)__res)->bi_ipbfreq)
#else
#define MPC5xxx_IPBFREQ CONFIG_PPC_5xxx_IPBFREQ	/* Assume Hz, not MHz */
#endif

/* Command queue element */
struct spi_chain {
	struct list_head list;
	int msg_num;		/* Number of messages in chain */
	struct spi_msg *rdwr;	/* ptr to message chain */
	int bus;		/* Adapter to send to */
	u8 **data_ptrs;		/* Saved here for later kfree() */
};

/* Data queue element */
struct data_chain {
	struct list_head list;
	struct spi_rd *spi_rd;	/* Ptr to data */
};

static ssize_t mcc200spidev_read (struct file *file, char *buf, size_t count, 
                            loff_t *offset);
static ssize_t mcc200spidev_write (struct file *file, const char *buf, size_t count, 
                            loff_t *offset);
static int mcc200spidev_open (struct inode *inode, struct file *file);
static int mcc200spidev_release (struct inode *inode, struct file *file);
static int mcc200spidev_attach_adapter(struct spi_adapter *adap);
static void mcc200spidev_detach_client(struct spi_client *client);
static int mcc200spidev_command(struct spi_client *client, unsigned int cmd,
                            void *arg);
static int mcc200spidev_ioctl (struct inode *inode, struct file *file,
				unsigned int cmd, unsigned long arg);
static int __init mcc200spi_dev_init(void);
static void mcc200spidev_cleanup(void);
static void mcc200spi_group_write(void *ptr);
static int  mcc200spi_data_to_queue (struct spi_msg *msg, int num, int bus,
				struct timespec *ts);
static void mcc200spi_data_free_entry(struct data_chain *entry);
static void mcc200spi_free_queues (void);
static int mcc200spi_start_timer(void);
static int mcc200spi_stop_timer(void);

static struct file_operations mcc200spidev_fops = {
	owner:		THIS_MODULE,
	llseek:		no_llseek,
	read:		mcc200spidev_read,
	write:		mcc200spidev_write,
	ioctl:		mcc200spidev_ioctl,
	open:		mcc200spidev_open,
	release:	mcc200spidev_release,
};

static struct spi_adapter *mcc200spidev_adaps[SPIDEV_ADAPS_MAX];
#ifdef CONFIG_DEVFS_FS
static devfs_handle_t devfs_spi[SPIDEV_ADAPS_MAX];
static devfs_handle_t devfs_handle = NULL;
#endif

/*
 * @brief spidev_driver 
 * struct to register us as a spi driver to spi-core
 *
 */

static struct spi_driver mcc200spidev_driver = {
	name:		"MCC200 spi-dev driver",
	flags:		SPI_DF_DUMMY,
	attach_adapter:	mcc200spidev_attach_adapter,
	detach_client:	mcc200spidev_detach_client,
	command:	mcc200spidev_command,
	inc_use:	NULL,
	dec_use:	NULL,
};

static struct spi_client mcc200spidev_client = {
	name:		"MCC200 SPI /dev entry",
	adapter:	NULL,
	driver:		&mcc200spidev_driver,
	addr:       0,
};

static int spidev_opened;
static int periodic;
/* Max number of commands in command queue 
*/
static int max_spi_chain = SPI_CMDQ_MAX;
static int in_spi_chain;

/* Max number of data chunks in list,
 * adjustable via the SPI_SET_BUF ioctl
 */
static int max_data_chain = SPI_DATAQ_MAX;
static int in_data_chain;
static int spi_timeout;

/* Input command queue
 */
static struct list_head spi_list;
/* Output data queue
 */
static struct list_head data_list;

/* Semaphore to protect the data queue
 */
static struct semaphore dataq_sema;
/* Wait queue to awake the sleeping "read"
 */
static wait_queue_head_t dataq_wq;

/* Poll timer is in milliseconds */
#define DIV 1000
#define PERIOD_TO_JIFFIES(t) ((t) * HZ / DIV)

/* Task queue to schedule the SPI transfer, used
 * to implement the timed execution of SPI commands.
 * spi_transfer() for MCC200 spi-psc adapter is 
 * interrupt-driven and uses semaphores for synchronization,
 * so the kernel timers and tasklets will not work.
 */
static struct tq_struct spi_group_task = {
	.routine = mcc200spi_group_write
};

/*
 * mcc200spidev_write - write() operation is not supported.
 */

static ssize_t mcc200spidev_write (struct file *file, const char *buf, size_t count,
                             loff_t *offset)
{
	return -EINVAL;
}

/*
 * mcc200spidev_read - read() hook for kernel
 * Returns the number of spi_rd chunks held in buffer,
 * not the read data size in bytes!
 * @param file: kernel file structure
 * @param buf: buffer for spi_rd units to send
 * @param count: number of bytes to send
 * @param offset: not used
 *
 */
static ssize_t mcc200spidev_read (struct file *file, char *buf, size_t count,
                            loff_t *offset)
{
	struct spi_rd *spi_rd;
	struct data_chain *entry;
	struct list_head *ptr, *n;
	int buflen, pktsz;
	char *p;

	if (!periodic)
		return -EINVAL;

	if (file->f_flags & O_NONBLOCK && !in_data_chain) {
		return -EAGAIN;
	}

	if (wait_event_interruptible(dataq_wq, in_data_chain)) {
		return -ERESTARTSYS;
	}

	down(&dataq_sema);

	buflen = count;
	p = buf;
	/* Read the data chunks from the list head
	 * (oldest first) and pack them into the user buffer.
	 */
	for (ptr = data_list.next; ptr != &data_list; ptr = n) {
		n = ptr->next;
		entry = list_entry(ptr, struct data_chain, list);
		spi_rd = entry->spi_rd;
		pktsz = spi_rd->nbytes + sizeof(struct spi_rd);
		if (pktsz > buflen) {
			/* Not enough space in user buffer */
			break;
		}
		if (copy_to_user(p, spi_rd, pktsz)) {
			up(&dataq_sema);
			return -EFAULT;
		}
		/* Free queue element */
		list_del(&entry->list);
		mcc200spi_data_free_entry(entry);

		p += pktsz;
		buflen -= pktsz;
	}

	up(&dataq_sema);
	return (count - buflen);
}

/*
 * mcc200spidev_ioctl - read() hook for kernel
 * @param inode: kernel inode structure
 * @param file: kernel file structure
 * @param cmd: ioctl command code
 * @param arg: ioctl command argument
 *
 */
static int mcc200spidev_ioctl (struct inode *inode, struct file *file,
				unsigned int cmd, unsigned long arg)
{
	struct spi_client *client = (struct spi_client *)file->private_data;
	struct spi_rdwr_ioctl_data rdwr_arg;
	struct spi_msg *rdwr_pa;
	u8 **data_ptrs;
	int i,j,datasize,res,bus;
	struct spi_setclock setclk_arg;

#ifdef DEBUG
	printk(KERN_DEBUG "spi-mcc200dev: spi-%d ioctl, cmd: 0x%x, arg: %lx.\n", 
	       MINOR(inode->i_rdev),cmd, arg);
#endif /* DEBUG */

	switch ( cmd ) {
	case SPI_RDWR:
		if (copy_from_user(&rdwr_arg, 
				   (struct spi_rdwr_ioctl_data *)arg, 
				   sizeof(rdwr_arg))) 
			return -EFAULT;

		/* Put an arbritrary limit on the number of messages that can
		 * be sent at once */
		if (rdwr_arg.nmsgs > SPI_MAX_MSGS)
			return -EINVAL;

		rdwr_pa = (struct spi_msg *)
			kmalloc(rdwr_arg.nmsgs * sizeof(struct spi_msg), 
			GFP_KERNEL);

		if (rdwr_pa == NULL) return -ENOMEM;
		if (copy_from_user(rdwr_pa, rdwr_arg.msgs,
				   rdwr_arg.nmsgs * sizeof(struct spi_msg))) {
			kfree(rdwr_pa);
			return -EFAULT;
		}
		data_ptrs = (u8 **) kmalloc(rdwr_arg.nmsgs * sizeof(u8 *),
					    GFP_KERNEL);
		if (data_ptrs == NULL) {
			kfree(rdwr_pa);
			return -ENOMEM;
		}
		res = 0;
		for( i=0; i<rdwr_arg.nmsgs; i++ ) {
			/* Limit the size of the message to a sane amount */
			if (rdwr_pa[i].len > SPI_MAX_MSG_SIZE) {
				res = -EINVAL;
				break;
			}
			data_ptrs[i] = rdwr_pa[i].buf;
			rdwr_pa[i].buf = kmalloc(rdwr_pa[i].len, GFP_KERNEL);
			if(rdwr_pa[i].buf == NULL) {
				res = -ENOMEM;
				break;
			}
			if(copy_from_user(rdwr_pa[i].buf,
				data_ptrs[i],
				rdwr_pa[i].len)) {
				++i; /* Needs to be kfreed too */
				res = -EFAULT;
				break;
			}
		}
		if (res < 0) {
			for (j = 0; j < i; ++j)
				kfree(rdwr_pa[j].buf);
			kfree(data_ptrs);
			kfree(rdwr_pa);
			return res;
		}

		/* Prepare the messages
		 * It is assumed that all SPI messages in one ioctl call
		 * are to be sent to one device.
		 */
		bus = SPIBUS(rdwr_pa[0].addr);
		for (j = 0, res = 0; j < rdwr_arg.nmsgs && res >= 0; j++) {
			rdwr_pa[j].addr = SPIDEV(rdwr_pa[j].addr);
		}
		if (!periodic) {
			/* Process immediately */
			res = spi_transfer(mcc200spidev_adaps[bus], rdwr_pa,
				rdwr_arg.nmsgs);
			while(i-- > 0) {
				if (res >= 0 && (rdwr_pa[i].flags & SPI_M_RD)) {
					if(copy_to_user(data_ptrs[i],
							rdwr_pa[i].buf,
							rdwr_pa[i].len)) {
						res = -EFAULT;
					}
				}
				kfree(rdwr_pa[i].buf);
			}
			kfree(data_ptrs);
			kfree(rdwr_pa);
		} else {
			/* For the periodic mode, compose the SPI command group
			 * by storing the commands in the linked list.
			 */
			struct spi_chain *entry;
			if ((in_spi_chain > max_spi_chain) ||
				!(entry = kmalloc(sizeof(struct spi_chain), 
					GFP_KERNEL))) {
				for (j = 0; j < i; ++j)
					kfree(rdwr_pa[j].buf);
				kfree(data_ptrs);
				kfree(rdwr_pa);
				return -ENOMEM;
			}
			entry->msg_num = rdwr_arg.nmsgs;
			entry->rdwr = rdwr_pa;
			entry->bus = bus;
			entry->data_ptrs = data_ptrs;
			list_add_tail(&entry->list, &spi_list);
			in_spi_chain++;
#ifdef DEBUG
			printk("%s:%d:added to cmdchain\n", __FUNCTION__, 
				in_spi_chain);
#endif
		}
		return res;
	case SPI_SET_CLOCK:
		if (copy_from_user(&setclk_arg, 
				(struct spi_setclock *)arg, 
				sizeof(setclk_arg)))
			return -EFAULT;

		client->adapter = mcc200spidev_adaps[setclk_arg.adap];

		if (!client->adapter->algo->set_bitrate)
			return -ENOSYS;
		return client->adapter->algo->set_bitrate(client->adapter, 
			setclk_arg.speed);

	case SPI_SET_GROUP: /* Enable group composition */
		periodic = 1;
		return 0;

	case SPI_SET_BUF: /* set the size of data linked list */
		if (copy_from_user(&datasize, (int *)arg, sizeof(int)))
			return -EFAULT;
		down(&dataq_sema);
		max_data_chain = datasize;
		up(&dataq_sema);
		return 0;

	case SPI_SET_PERIOD:
		if (copy_from_user(&spi_timeout, (int *)arg, sizeof(int)))
			return -EFAULT;
		if (!spi_timeout) {
			periodic = 0; /* stop periodic mode */
			mcc200spi_stop_timer();
			mcc200spi_free_queues(); /* Free all structures */
		} else {
			if (!periodic) { /* Periodic mode not enabled */
				return -EINVAL;
			}
			mcc200spi_start_timer();
			mcc200spi_group_write(NULL);
		}
		return 0;
	default:
		return 0;
	}
	return 0;
}

/* Periodic interrupt service function.
 */
static void mcc200spi_timer_isr(int irq, void *data, struct pt_regs *pt_regs)
{
	struct mpc5xxx_gpt *gpt = (struct mpc5xxx_gpt *)MCC200_GPT;

	/* clear event bit */
	gpt->sr |= 0x0F;

	/* We cannot directly use spi_transfer() here because the SPI algorithm
	 * driver for MC200 is interrupt-based with wake_up()/sleep_on() 
	 * synchronization. The latter functions use rescheduling which cannot be 
	 * performed in interrupt context.
	 */
	if (periodic) {
		schedule_task(&spi_group_task);
	}
}

/* Configure and start the timer counter.
*/
static int mcc200spi_start_timer(void)
{
	struct mpc5xxx_gpt* gpt = (struct mpc5xxx_gpt*)MCC200_GPT;
	int ret;

#ifdef DEBUG
	printk("%s: emsr %#x\n", __FUNCTION__, gpt->emsr);
#endif
	gpt->emsr &= ~0x1000; /* Disable timer counter */
	gpt->emsr |= (0x100 | 0x400); /* Interrupt enable, Continuous mode */
	/* spi_timeout is in milliseconds */
	gpt->cir = ((MPC5xxx_IPBFREQ / 1000 / 10) << 16) | (spi_timeout * 10);
#ifdef DEBUG
	printk("%s: after: emsr %#x, cir %#x\n", __FUNCTION__, gpt->emsr, gpt->cir);
#endif
	ret = request_irq(TIMER5_IRQ, mcc200spi_timer_isr, SA_INTERRUPT,
					  "mcc200 spi timer", NULL);
	if (ret ) {
		printk("Can't request %d interrupt for GPT5\n", TIMER5_IRQ);
		return 1;
	}
	gpt->emsr |= 0x1000; /* Enable counter */
	return 0;
}

/* Stop the timer counter
*/
static int mcc200spi_stop_timer(void)
{
	struct mpc5xxx_gpt* gpt = (struct mpc5xxx_gpt*)MCC200_GPT;

	gpt->emsr &= ~0x1000; /* Disable counter */
	free_irq(TIMER5_IRQ, NULL);
	return 0;
}

/*
 * mcc200spidev_open - open() hook for kernel
 * @param inode: kernel inode structure
 * @param file: kernel file structure
 *
 */

static int mcc200spidev_open (struct inode *inode, struct file *file)
{
	if (spidev_opened)
		return -EBUSY;

	file->private_data = &mcc200spidev_client;

#if USECOUNT
	for (i = 0; i < num_adaps; i++) {
		if (mcc200spidev_adaps[i]->inc_use)
			mcc200spidev_adaps[i]->inc_use(mcc200spidev_adaps[i]);
	}
#endif

#ifdef DEBUG
	printk("spi-mcc200dev: opened\n");
#endif
	spidev_opened = 1;
	init_MUTEX(&dataq_sema);
	init_waitqueue_head(&dataq_wq);
	INIT_LIST_HEAD(&spi_list);
	INIT_LIST_HEAD(&data_list);
	in_data_chain = 0;
	in_spi_chain = 0;
	return 0;
}
/*
 * mcc200spidev_release - close() hook for kernel
 * @param inode: kernel inode structure
 * @param file: kernel file structure
 *
 */

static int mcc200spidev_release (struct inode *inode, struct file *file)
{
#ifdef DEBUG
	printk("spi-mcc200dev: Closed\n");
#endif
#if USECOUNT
	lock_kernel();
	for (i = 0; i < num_adaps; i++) {
		if (mcc200spidev_adaps[i]->dec_use)
			mcc200spidev_adaps[i]->dec_use(mcc200spidev_adaps[i]);
	}
	unlock_kernel();
#endif

	mcc200spi_stop_timer();
	periodic = 0;
	mcc200spi_free_queues();
	spidev_opened = 0;
	file->private_data = NULL;
	return 0;
}
/*
 * mcc200spidev_attach_adapter - notify that new adapter has been attached
 * @param adap: mcc200spi_adapter structure of new adapter
 *
 */

static int mcc200spidev_attach_adapter(struct spi_adapter *adap)
{
	int i;

#ifdef DEBUG
	printk("%s:adap=%x id=%d %s\n",__FUNCTION__,
		   adap,spi_adapter_id(adap),adap->name);
#endif

	if ((i = spi_adapter_id(adap)) < 0) {
		printk("spi-mcc200dev: Unknown adapter ?!?\n");
		return -ENODEV;
	}
	if (i >= SPIDEV_ADAPS_MAX) {
		printk("spi-mcc200dev: Adapter number too large?!? (%d)\n",i);
		return -ENODEV;
	}

	if (! mcc200spidev_adaps[i]) {
		mcc200spidev_adaps[i] = adap;
	} else {
		mcc200spidev_adaps[i] = NULL;
	}

	return 0;
}

static void mcc200spidev_detach_client(struct spi_client *client)
{
	return;
}

static int mcc200spidev_command(struct spi_client *client, unsigned int cmd,
                           void *arg)
{
	return -1;
}

/* Executes the SPI commands from the command queue.
 * For the SPI messages marked with SPI_M_RD, collect
 * the data to tha data queue.
 */

static void mcc200spi_group_write(void *arg)
{
	struct spi_msg *rdwr_pa;
	struct list_head *ptr;
	struct spi_chain *entry;
	int i,res,bus;

	for (ptr = spi_list.next; periodic && ptr != &spi_list; ptr = ptr->next) {
		entry = list_entry(ptr, struct spi_chain, list);

		rdwr_pa = entry->rdwr;

		bus = entry->bus;
		for (i = 0; i < entry->msg_num; i++) {
			char *saved;
#ifdef DEBUG
			printk("%s:%d:bus %d, dev %d, cmd %#x, flags %s:%s\n", 
				__FUNCTION__, i, 
				bus, SPIDEV(rdwr_pa[i].addr), rdwr_pa[i].buf[0],
				(rdwr_pa[i].flags & SPI_M_RD) ? "RD" : "  ",
				(rdwr_pa[i].flags & SPI_M_WR) ? "WR" : "  ");
#endif
			/* Save the SPI commands for subsequent use. The buffers
			 * may be overwritten by the SPI device data
			 */
			saved = kmalloc(rdwr_pa[i].len, GFP_KERNEL);
			memcpy(saved, rdwr_pa[i].buf, rdwr_pa[i].len);
			entry->data_ptrs[i] = saved;
		}
		res = spi_transfer(mcc200spidev_adaps[bus], rdwr_pa, entry->msg_num);
#ifdef DEBUG
		printk("%s: transfer OK: %d\n", __FUNCTION__, res);
#endif
		if (res >= 0) {
			struct timeval tv;
			struct timespec ts;
			do_gettimeofday(&tv);
			TIMEVAL_TO_TIMESPEC(&tv, &ts);
			mcc200spi_data_to_queue(rdwr_pa, entry->msg_num, bus, &ts);
		}

		/* Restore the commands in message buffers
		 */
		for (i = 0; i < entry->msg_num; i++) {
			memcpy(rdwr_pa[i].buf, entry->data_ptrs[i], rdwr_pa[i].len);
			kfree(entry->data_ptrs[i]);
		}		
	}
	if (in_data_chain) {
		wake_up_interruptible(&dataq_wq);
	}
}

/* 
 * Pack all read data (if any) into one queue structure.
 * It is assumed that each entry in command queue is to be sent 
 * to one SPI device, so the reply device data are packed
 * into one buffer.
 */
static int mcc200spi_data_to_queue (struct spi_msg *msg, int num, int bus,
					struct timespec *ts)
{
	struct spi_rd *spi_rd;
	struct data_chain *new;
	int i, totlen, res = 0;
	char *ptr;
	
	for (i = 0, totlen = 0; i < num; i++) {
		if (msg[i].flags & SPI_M_RD) {
			totlen += msg[i].len;
		}
	}

	if (!totlen) { /* No data */
		return res;
	}
#ifdef DEBUG
	printk("%s: total data len %d\n", __FUNCTION__, totlen);
#endif
	/* Protect from concurrent access in the "read" function
	 */
	down(&dataq_sema);

	if (in_data_chain >= max_data_chain) {
		struct data_chain *entry;
#ifdef DEBUG
		printk("%s: queue overflow\n", __FUNCTION__);
#endif
		/* Queue overflow, remove the oldest first 
		 */
		entry = list_entry(data_list.next, struct data_chain, list);
		list_del(&entry->list);
		mcc200spi_data_free_entry(entry);
	}

	spi_rd = kmalloc(sizeof(struct spi_rd) + totlen, GFP_KERNEL);
	if (!spi_rd) {
		res = -ENOMEM;
		goto xit;
	}

	/* Copy the data into one buffer. It is assumed that data are from the
	 * one SPI device.
	 */
	for (i = 0, ptr = spi_rd->data; i < num; i++) {
		if (msg[i].flags & SPI_M_RD) {
#ifdef DEBUG
			printk("%s:%d:buf[0] %#x\n", __FUNCTION__, i, msg[i].buf[0]);
#endif
			memcpy(ptr, msg[i].buf, msg[i].len);
			ptr += msg[i].len;
		}
	}
	spi_rd->nbytes = totlen;
	*(&spi_rd->ts) = *ts;
	spi_rd->id = SPIMKID(bus, msg->addr);

	/* Allocate the new queue element 
	 */
	new = kmalloc(sizeof(struct data_chain), GFP_KERNEL);
	if (!new) {
		res = -ENOMEM;
		goto xit;
	}

	new->spi_rd = spi_rd;

	list_add_tail(&new->list, &data_list);
	in_data_chain++;
#ifdef DEBUG
	printk("%s: %d in data chain\n", __FUNCTION__, in_data_chain);
#endif

 xit:
	up(&dataq_sema);
	return res;
}

/* Free one data queue entry
 */
static void mcc200spi_data_free_entry(struct data_chain *entry)
{
	kfree(entry->spi_rd);
	kfree(entry);
	in_data_chain--;
}

/* Free all allocated memory.
 */
static void mcc200spi_free_queues (void)
{
	struct list_head *ptr, *n;
	struct spi_chain *entry;
	struct data_chain *data;
	int i;

	/* Command list. */
	for (ptr = spi_list.next; ptr != &spi_list; ptr = n) {	
		n = ptr->next;
		entry = list_entry(ptr, struct spi_chain, list);
		for (i = 0; i < entry->msg_num; i++) {
			kfree(entry->rdwr[i].buf);
		}
		kfree(entry->rdwr);
		kfree(entry->data_ptrs);
		kfree(entry);
	}
	in_spi_chain = 0;
	INIT_LIST_HEAD(&spi_list);

	/* Data list */
	for (ptr = data_list.next; ptr != &data_list; ptr = n) {	
		n = ptr->next;
		data = list_entry(ptr, struct data_chain, list);
		ptr = ptr->next;
		mcc200spi_data_free_entry(data);
	}
	in_data_chain = 0;
	INIT_LIST_HEAD(&data_list);
}
/* 
 * mcc200spi_dev_init(void)  module init function
 *
 */

static int __init mcc200spi_dev_init(void)
{
	int res;

	printk(KERN_DEBUG "MCC200 spi device: spi /dev entries driver module\n");

#ifdef CONFIG_DEVFS_FS
	if (devfs_register_chrdev(MCC200SPI_MAJOR, "mcc200spi", &mcc200spidev_fops)) {
#else
	if (register_chrdev(MCC200SPI_MAJOR,"mcc200spi",&mcc200spidev_fops)) {
#endif
		printk("spi-mcc200dev: unable to get major %d for spi bus\n",
		       MCC200SPI_MAJOR);
		return -EIO;
	}
#ifdef CONFIG_DEVFS_FS
	devfs_handle = devfs_mk_dir(NULL, "spi", NULL);
#endif

	if ((res = spi_add_driver(&mcc200spidev_driver))) {
		printk("spi-mcc200dev: Driver registration failed,"
		"  module not inserted.\n");
		mcc200spidev_cleanup();
		return res;
	}

	return 0;
}

static void mcc200spidev_cleanup(void)
{
	int res;

	if ((res = spi_del_driver(&mcc200spidev_driver))) {
		printk("spi-mcc200dev: Driver deregistration failed, "
			   "module not removed.\n");
		return;
	}

#ifdef CONFIG_DEVFS_FS
	devfs_unregister(devfs_handle);
	if ((res = devfs_unregister_chrdev(MCC200SPI_MAJOR, "mcc200spi"))) {
#else
	if ((res = unregister_chrdev(MCC200SPI_MAJOR,"mcc200spi"))) {
#endif
		printk("spi-mcc200dev: unable to release major %d for spi bus\n",
			MCC200SPI_MAJOR);
		return;
	}
}

EXPORT_NO_SYMBOLS;

MODULE_DESCRIPTION("MCC200 SPI /dev entries driver");
MODULE_LICENSE("GPL");

module_init(mcc200spi_dev_init);
module_exit(mcc200spidev_cleanup);

