/*
 *  drivers/video/sm712fb.c -- frame buffer device for
 *  Silicon Motion Lynx3DM chip.
 *
 *  This was written to drive the Lynx board in the Embedded Planet
 *  Icebox (embedded PPC machine) and is only tested on that machine.
 *  It should really be generalised quite a bit - for one thing it is
 *  hard-wired to 640x480.
 *
 *  Copyright (C) 2002 David Gibson, IBM Corporation.
 *
 *  Based on chipsfb.c:
 *  Copyright (C) 1998 Paul Mackerras,
 *  which in turn is derived from the Powermac "chips" driver:
 *  Copyright (C) 1997 Fabio Riccardi.
 *  And from the frame buffer device for Open Firmware-initialized devices:
 *  Copyright (C) 1997 Geert Uytterhoeven.
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License. See the file COPYING in the main directory of this archive for
 *  more details.  */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/fb.h>
#include <linux/selection.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <asm/io.h>

#include <video/fbcon.h>
#include <video/fbcon-cfb8.h>
#include <video/fbcon-cfb16.h>
#include <video/fbcon-cfb24.h>
#include <video/macmodes.h>

#include "sm712fb.h"

MODULE_LICENSE("GPL");

#define SM712_PCI_MEMORY_SIZE			0x00800000

struct fb_info_lynx {
	struct fb_info info;
	struct fb_fix_screeninfo fix;
	struct fb_var_screeninfo var;
	struct display disp;
	struct {
		u8 red, green, blue;
	} palette[256];
	struct pci_dev *pdev;

	u8 chiprev;
	unsigned long ramsize;

	unsigned long aperture_phys;
	u8 *aperture;
	unsigned long frame_buffer_phys;
	u8 *frame_buffer;

	struct fb_info_lynx *next;
#ifdef FBCON_HAS_CFB16
	u16 fbcon_cfb16_cmap[16];
#endif
#ifdef FBCON_HAS_CFB24
	u32 fbcon_cfb24_cmap[16];
#endif
};
static struct fb_info_lynx *all_lynx;
static int currcon; /* = 0 */
static GraphicDevice smi;
static int videomode = 0x301;
static int monitor_type = 1;

#define MMIO_DPR_OFFSET	(0x000)
#define MMIO_VPR_OFFSET	(0x800)
#define MMIO_IO_OFFSET	(0xc0000)

const struct fb_videomode res_mode_init[] = {
	/*name, refresh,  x	 y pixclk   le	ri  up	lo   hs vs  s  vmode */
	{NULL, 0,        800, 600, },   /* TODO: DUAL_800_600 */
	{NULL, 0,        1024, 768, },  /* TODO: DUAL_1024_768 */
	{NULL, 0,        800, 600, 27778, 64, 24, 22, 1, 72, 2, 0, FB_VMODE_NONINTERLACED}, /* SINGLE_800_600 */
	{NULL, 0,        1024, 768, 15384, 168, 8, 29, 3, 144, 4, 0, FB_VMODE_NONINTERLACED}, /* SINGLE_1024_768 */
	{NULL, 0,        1280, 1024, 9090, 200, 48, 26, 1, 184, 3, 0, FB_VMODE_NONINTERLACED}, /* SINGLE_1280_1024 */
	{NULL, 0,        720, 576, },   /* TODO: TV_MODE_CCIR */
	{NULL, 0,        720, 484, },  /* TODO: TV_MODE_EIA */
	{NULL, 0,        640, 480, 39721, 40, 24, 32, 11, 96, 2, 0, FB_VMODE_NONINTERLACED}, /* SINGLE_640_480 */
};

#if 0
#define write_ind(lp, num, val, ap, dp)	do { \
		writeb((num), (lp)->aperture + MMIO_IO_OFFSET + (ap)); \
		writeb((val), (lp)->aperture + MMIO_IO_OFFSET + (dp)); \
	} while (0)
#define read_ind(lp, num, ap, dp)	({ writeb((num), (lp)->aperture + MMIO_IO_OFFSET + (ap)); \
	readb((lp)->aperture + MMIO_IO_OFFSET + (dp));})
#else
#define write_ind(lp, num, val, ap, dp) do { \
		outb((num), (ap)); outb((val), (dp)); \
	} while (0)
#define read_ind(lp, num, ap, dp)	({ outb((num), (ap)); inb((dp)); })
#endif


/* CRTC registers */
#define write_crt(lp, num, val)	write_ind(lp, num, val, 0x3d4, 0x3d5)
#define read_crt(lp, num)	read_ind(lp, num, 0x3d4, 0x3d5)
/* graphics registers */
#define write_grx(lp, num, val)	write_ind(lp, num, val, 0x3ce, 0x3cf)
#define read_grx(lp, num)	read_ind(lp, num, 0x3ce, 0x3cf)
/* sequencer registers */
#define write_seq(lp, num, val)	write_ind(lp, num, val, 0x3c4, 0x3c5)
#define read_seq(lp, num)	read_ind(lp, num, 0x3c4, 0x3c5)
/* system control registers */
#define write_scr write_seq
#define read_scr read_seq
/* flat panel registers */
#define write_fpr write_seq
#define read_fpr read_fpr
/* memory control registers */
#define write_mcr write_seq
#define read_mcr read_seq

/* attribute controller */
#define write_atr(lp, num, val) do { \
		inb(0x3da); write_ind(lp, num, val, 0x3c0, 0x3c0); \
	} while (0)
#define read_atr(lp, num) ({inb(0x3da); read_ind(lp, num, 0x3c0, 0x3c1); })

/* Drawing processor */
#define write_dprw(lp, off, val) writew((val), (lp)->aperture + MMIO_DPR_OFFSET + (off))
#define read_dprw(lp, off) readw((lp)->aperture + MMIO_DPR_OFFSET + (off))
#define write_dprl(lp, off, val) writel((val), (lp)->aperture + MMIO_DPR_OFFSET + (off))
#define read_dprl(lp, off) readl((lp)->aperture + MMIO_DPR_OFFSET + (off))
/* Video processor */
#define write_vprw(lp, off, val) writew((val), (lp)->aperture + MMIO_VPR_OFFSET + (off))
#define read_vprw(lp, off) readw((lp)->aperture + MMIO_VPR_OFFSET + (off))
#define write_vprl(lp, off, val) writel((val), (lp)->aperture + MMIO_VPR_OFFSET + (off))
#define read_vprl(lp, off) readl((lp)->aperture + MMIO_VPR_OFFSET + (off))

/*
 * Register values for common video modes
 */
static char SMI_SCR[] = {
	/* all modes */
	0x10, 0xff, 0x11, 0xff, 0x12, 0xff, 0x13, 0xff, 0x15, 0x90,
	0x17, 0x20, 0x18, 0xb1, 0x19, 0x00,
};
static char SMI_EXT_CRT[] = {
	0x31, 0x00, 0x32, 0x00, 0x33, 0x00, 0x34, 0x00, 0x35, 0x00,
	0x36, 0x00, 0x3b, 0x00, 0x3c, 0x00, 0x3d, 0x00, 0x3e, 0x00, 0x3f, 0x00
};
static char SMI_ATTR [] = {
	0x00, 0x00, 0x01, 0x01, 0x02, 0x02, 0x03, 0x03, 0x04, 0x04, 0x05, 0x05,
	0x06, 0x06, 0x07, 0x07, 0x08, 0x08, 0x09, 0x09, 0x0a, 0x0a, 0x0b, 0x0b,
	0x0c, 0x0c, 0x0d, 0x0d, 0x0e, 0x0e, 0x0f, 0x0f, 0x10, 0x41, 0x11, 0x00,
	0x12, 0x0f, 0x13, 0x00, 0x14, 0x00,
};
static char SMI_GCR[18] = {
	0x00, 0x00, 0x01, 0x00, 0x02, 0x00, 0x03, 0x00, 0x04, 0x00, 0x05, 0x40,
	0x06, 0x05, 0x07, 0x0f, 0x08, 0xff
};
static char SMI_SEQR[] = {
	0x00, 0x00, 0x01, 0x01, 0x02, 0x0f, 0x03, 0x03, 0x04, 0x0e, 0x00, 0x03
};
static char SMI_PCR [] = {
	0x20, 0x04, 0x21, 0x30, 0x22, 0x00, 0x23, 0x00, 0x24, 0x00
};
static char SMI_MCR[] = {
	0x60, 0x01, 0x61, 0x00
#ifdef CONFIG_HMI1001 
	, 0x62, 0x74 /* Memory type is not configured by pins on HMI1001 */
#endif
};

static char SMI_HCR[] = {
	0x80, 0xff, 0x81, 0x07, 0x82, 0x00, 0x83, 0xff, 0x84, 0xff, 0x88, 0x00,
	0x89, 0x02, 0x8a, 0x80, 0x8b, 0x01, 0x8c, 0xff, 0x8d, 0x00
};

static int __init sm712fb_init_one(struct pci_dev *dp, const struct pci_device_id *ent);
static void sm712_init_hw(struct fb_info_lynx *lynx);
static int sm712_get_fix(struct fb_fix_screeninfo *fix, int con, struct fb_info *info);
static int sm712_get_var(struct fb_var_screeninfo *var, int con, struct fb_info *info);
static int sm712_set_var(struct fb_var_screeninfo *var, int con, struct fb_info *info);
static int sm712fb_switch_con(int con, struct fb_info *info);
static int sm712fb_updatevar(int con, struct fb_info *info);
static void sm712fb_blank(int blank, struct fb_info *info);
static int sm712_get_cmap(struct fb_cmap *cmap, int kspc, int con,
			 struct fb_info *info);
static int sm712_set_cmap(struct fb_cmap *cmap, int kspc, int con,
			 struct fb_info *info);
static void sm712_set_bitdepth(struct fb_info_lynx *p, struct display *disp, int con, int bpp);
static void my_sm712_set_bitdepth(struct fb_info_lynx *p, struct display *disp, int con, int bpp);
static void do_install_cmap(int con, struct fb_info *info);


static struct fb_ops sm712fb_ops = {
	owner:		THIS_MODULE,
	fb_get_fix:	sm712_get_fix,
	fb_get_var:	sm712_get_var,
	fb_set_var:	sm712_set_var,
	fb_get_cmap:	sm712_get_cmap,
	fb_set_cmap:	sm712_set_cmap,
};

static struct pci_device_id sm712fb_pci_tbl[] __devinitdata = {
	{PCI_VENDOR_ID_SILICONMOTION, PCI_DEVICE_ID_SILICONMOTION_SM712,
	 	PCI_ANY_ID, PCI_ANY_ID,},
	{0,},
};

static struct pci_driver sm712fb_pci_driver = {
	name:		"sm712fb",
	id_table:	sm712fb_pci_tbl,
	probe:		sm712fb_init_one,
	remove:		NULL,
	suspend:	NULL,
	resume:		NULL,
};

int __init sm712_init(void)
{
	printk(KERN_DEBUG "sm712fb.c: sm712_init()\n");
	return pci_module_init(&sm712fb_pci_driver);
}

module_init(sm712_init);

#if 0

#define in8(addr) readb(addr)
#define out8(addr, val) writeb(val, addr)

#define in16r(addr) readw(addr)
#define out16r(addr, val) writew(val, addr)

#define in32r(addr) readl(addr)
#define out32r(addr, val) writel(val, addr)

#else

#define in8(addr) inb((void*)addr)
#define out8(addr, val) outb(val, (void*)addr)

#define in16r(addr) in_le16((void*)addr)
#define out16r(addr, val) out_le16((void*)addr, val)

#define in32r(addr) in_le32((void*)addr)
#define out32r(addr, val) out_le32((void*)addr, val)

#endif
/*******************************************************************************
*
* Read SMI ISA register
*/
static char smiRead (unsigned short index, char reg)
{
    outb (reg, index);
    return (inb (index + 1));
}
/*******************************************************************************
*
* Write SMI ISA register
*/
static void smiWrite (unsigned short index, char reg, char val)
{
    outb (reg, index);
    outb (val, index + 1);
}

/*******************************************************************************
*
* Write a table of SMI ISA register
*/
static void smiLoadRegs (
    unsigned int iReg,
    unsigned int dReg,
    char         *regTab,
    unsigned int tabSize
    )
{
    register int i;

    for (i=0; i<tabSize; i+=2)
    {
	    if (iReg == SMI_INDX_ATTR) {
		    /* Reset the Flip Flop */
		    inb (SMI_ISR1);
		    outb (regTab[i], iReg);
		    outb (regTab[i+1], iReg);
	    } else {
		    outb (regTab[i], iReg);
		    outb (regTab[i+1], dReg);
	    }
    }
}

/*******************************************************************************
*
* Init capture port registers
*/
static void smiInitCapturePort (void)
{
    SmiCapturePort smiCP = { 0x01400600, 0x30, 0x40, 480, 640, 0, 0, 2560, 6 };
    register GraphicDevice *pGD  = (GraphicDevice *)&smi;
    register SmiCapturePort *pCP = (SmiCapturePort *)&smiCP;

    out32r ((pGD->cprBase + 0x0004), ((pCP->topClip<<16)   | pCP->leftClip));
    out32r ((pGD->cprBase + 0x0008), ((pCP->srcHeight<<16) | pCP->srcWidth));
    out32r ((pGD->cprBase + 0x000c), pCP->srcBufStart1/8);
    out32r ((pGD->cprBase + 0x0010), pCP->srcBufStart2/8);
    out32r ((pGD->cprBase + 0x0014), pCP->srcOffset/8);
    out32r ((pGD->cprBase + 0x0018), pCP->fifoControl);
    out32r ((pGD->cprBase + 0x0000), pCP->control);
}


/*******************************************************************************
*
* Init video processor registers
*/
static void smiInitVideoProcessor (void)
{
    SmiVideoProc smiVP = { 0x100000, 0, 0, 0, 0, 1600, 0x1200543, 4, 0xededed };
    SmiVideoWin  smiVW = { 0, 0, 599, 799, 0, 1600, 0, 0, 0 };
    register GraphicDevice *pGD = (GraphicDevice *)&smi;
    register SmiVideoProc  *pVP = (SmiVideoProc *)&smiVP;
    register SmiVideoWin *pVWin = (SmiVideoWin *)&smiVW;

    pVP->width    = pGD->plnSizeX * pGD->gdfBytesPP;
    pVP->control |= pGD->gdfIndex << 16;
    pVWin->bottom = pGD->winSizeY - 1;
    pVWin->right  = pGD->winSizeX - 1;
    pVWin->width  = pVP->width;

    /* color key */
    out32r ((pGD->vprBase + 0x0004), pVP->colorKey);

     /* color key mask */
    out32r ((pGD->vprBase + 0x0008), pVP->colorKeyMask);

    /* data src start adrs */
    out32r ((pGD->vprBase + 0x000c), pVP->start / 8);

    /* data width and offset */
    out32r ((pGD->vprBase + 0x0010),
	((pVP->offset   / 8 * pGD->gdfBytesPP) << 16) |
	 (pGD->plnSizeX / 8 * pGD->gdfBytesPP));

    /* video window 1 */
    out32r ((pGD->vprBase + 0x0014),
	((pVWin->top << 16) | pVWin->left));

    out32r ((pGD->vprBase + 0x0018),
	((pVWin->bottom << 16) | pVWin->right));

    out32r ((pGD->vprBase + 0x001c), pVWin->srcStart / 8);

    out32r ((pGD->vprBase + 0x0020),
	(((pVWin->offset / 8) << 16) | (pVWin->width / 8)));

    out32r ((pGD->vprBase + 0x0024),
	(((pVWin->hStretch) << 8) | pVWin->vStretch));

    /* video window 2 */
    out32r ((pGD->vprBase + 0x0028),
	((pVWin->top << 16) | pVWin->left));

    out32r ((pGD->vprBase + 0x002c),
	((pVWin->bottom << 16) | pVWin->right));

    out32r ((pGD->vprBase + 0x0030),
	pVWin->srcStart / 8);

    out32r ((pGD->vprBase + 0x0034),
	(((pVWin->offset / 8) << 16) | (pVWin->width / 8)));

    out32r ((pGD->vprBase + 0x0038),
	(((pVWin->hStretch) << 8) | pVWin->vStretch));

    /* fifo prio control */
    out32r ((pGD->vprBase + 0x0054), pVP->fifoPrio);

    /* fifo empty request levell */
    out32r ((pGD->vprBase + 0x0058), pVP->fifoERL);

    /* conversion constant */
    out32r ((pGD->vprBase + 0x005c), pVP->YUVtoRGB);

    /* vpr control word */
    out32r ((pGD->vprBase + 0x0000), pVP->control);
}

/******************************************************************************
 *
 * Init drawing engine registers
 */
static void smiInitDrawingEngine (void)
{
    GraphicDevice *pGD = (GraphicDevice *)&smi;
    unsigned int val;

    /* don't start now */
    out32r ((pGD->dprBase + 0x000c), 0x000f0000);

    /* set rop2 to copypen */
    val = 0xffff3ff0 & in32r ((pGD->dprBase + 0x000c));
    out32r ((pGD->dprBase + 0x000c), (val | 0x8000 | 0x0c));

    /* set clip rect */
    out32r ((pGD->dprBase + 0x002c), 0);
    out32r ((pGD->dprBase + 0x0030),
	((pGD->winSizeY<<16) | pGD->winSizeX * pGD->gdfBytesPP ));

    /* src row pitch */
    val = 0xffff0000 & (in32r ((pGD->dprBase + 0x0010)));
    out32r ((pGD->dprBase + 0x0010),
	(val | pGD->plnSizeX * pGD->gdfBytesPP));

    /* dst row pitch */
    val = 0x0000ffff & (in32r ((pGD->dprBase + 0x0010)));
    out32r ((pGD->dprBase + 0x0010),
	(((pGD->plnSizeX * pGD->gdfBytesPP)<<16) | val));

    /* window width src/dst */
    out32r ((pGD->dprBase + 0x003c),
	(((pGD->plnSizeX * pGD->gdfBytesPP & 0x0fff)<<16) |
	  (pGD->plnSizeX * pGD->gdfBytesPP & 0x0fff)));
    out16r ((pGD->dprBase + 0x001e), 0x0000);

    /* src base adrs */
    out32r ((pGD->dprBase + 0x0040),
	(((pGD->frameAdrs/8) & 0x000fffff)));

    /* dst base adrs */
    out32r ((pGD->dprBase + 0x0044),
	(((pGD->frameAdrs/8) & 0x000fffff)));

    /* foreground color */
    out32r ((pGD->dprBase + 0x0014), pGD->fg);

    /* background color */
    out32r ((pGD->dprBase + 0x0018), pGD->bg);

    /* xcolor */
    out32r ((pGD->dprBase + 0x0020), 0x00ffffff);

    /* xcolor mask */
    out32r ((pGD->dprBase + 0x0024), 0x00ffffff);

    /* bit mask */
    out32r ((pGD->dprBase + 0x0028), 0x00ffffff);

    /* load mono pattern */
    out32r ((pGD->dprBase + 0x0034), 0);
    out32r ((pGD->dprBase + 0x0038), 0);
}

/*****************************************************************************/
static void smiLoadMsr (const struct fb_videomode *mode)
{
	unsigned char h_synch_high, v_synch_high;

	h_synch_high = (mode->sync & FB_SYNC_HOR_HIGH_ACT) ? 0 : 0x40;	/* horizontal Synch High active */
	v_synch_high = (mode->sync & FB_SYNC_VERT_HIGH_ACT) ? 0 : 0x80;	/* vertical Synch High active */
	out8 (SMI_MISC_REG, (h_synch_high | v_synch_high | 0x29));
	/* upper64K==0x20, CLC2select==0x08, RAMenable==0x02!(todo), CGA==0x01
	 * Selects the upper 64KB page.Bit5=1
	 * CLK2 (left reserved in standard VGA) Bit3|2=1|0
	 * Disables CPU access to frame buffer. Bit1=0
	 * Sets the I/O address decode for ST01, FCR, and all CR registers
	 * to the 3Dx I/O address range (CGA emulation). Bit0=1
	 */
}
/*****************************************************************************/
static void smiLoadCrt (const struct fb_videomode *var, int bits_per_pixel)
{
	unsigned char cr[0x7a];
	int i;
	unsigned int hd, hs, he, ht, hbs, hbe;	/* Horizontal.	*/
	unsigned int vd, vs, ve, vt, vbs, vbe;	/* vertical */
	unsigned int bpp, wd, dblscan, interlaced;

	const int LineCompare = 0x3ff;
	unsigned int TextScanLines = 1;	/* this is in fact a vertical zoom factor   */

	/* Horizontal */
	hd = (var->xres) / 8;	/* HDisp.  */
	hs = (var->xres + var->right_margin) / 8;	/* HsStrt  */
	he = (var->xres + var->right_margin + var->hsync_len) / 8;	/* HsEnd   */
	ht = (var->left_margin + var->xres + var->right_margin + var->hsync_len) / 8;	/* HTotal  */
	/* Blank */
	hbs = hd;
	hbe = 0; /* Blank end at 0 */

	/* Vertical */
	vd = var->yres;		/* VDisplay   */
	vs = var->yres + var->lower_margin;	/* VSyncStart */
	ve = var->yres + var->lower_margin + var->vsync_len;	/* VSyncEnd */
	vt = var->upper_margin + var->yres + var->lower_margin + var->vsync_len;	/* VTotal  */
	vbs = vd;
	vbe = 0;

	bpp = bits_per_pixel;
	dblscan = (var->vmode & FB_VMODE_DOUBLE) ? 1 : 0;
	interlaced = var->vmode & FB_VMODE_INTERLACED;


	if (bpp == 15)
		bpp = 16;
	wd = var->xres * bpp / 64;	/* double words per line */
	if (interlaced) {	/* we divide all vertical timings, exept vd */
		vs >>= 1;
		vbs >>= 1;
		ve >>= 1;
		vt >>= 1;
	}

	memset (cr, 0, sizeof (cr));
	cr[0x00] = ht - 5;
	cr[0x01] = hd - 1;
	cr[0x02] = hbs - 1;
	cr[0x03] = (hbe & 0x1F);
	cr[0x04] = hs;
	cr[0x05] = ((hbe & 0x20) << 2) | (he & 0x1f);

	cr[0x06] = (vt - 2) & 0xFF;
	cr[0x07] = (((vt - 2) & 0x100) >> 8)
		| (((vd - 1) & 0x100) >> 7)
		| ((vs & 0x100) >> 6)
		| (((vbs - 1) & 0x100) >> 5)
		| ((LineCompare & 0x100) >> 4)
		| (((vt - 2) & 0x200) >> 4)
		| (((vd - 1) & 0x200) >> 3)
		| ((vs & 0x200) >> 2);

	cr[0x30] = ((vt - 2) & 0x400) >> 7
		| (((vd - 1) & 0x400) >> 8)
		| (((vbs - 1) & 0x400) >> 9)
		| ((vs & 0x400) >> 10)
		| (interlaced) ? 0x80 : 0;


	cr[0x08] = 0x00;
	cr[0x09] = (dblscan << 7)
		| ((LineCompare & 0x200) >> 3)
		| (((vbs - 1) & 0x200) >> 4)
		| (TextScanLines - 1);

	cr[0x10] = vs & 0xff;	/* VSyncPulseStart */
	cr[0x11] = (ve & 0x0f);
	cr[0x12] = (vd - 1) & 0xff;	/* LineCount  */
	cr[0x13] = wd & 0xff;
	cr[0x14] = 0x40;
	cr[0x15] = (vbs - 1) & 0xff;
	cr[0x16] = vbe & 0xff;
	cr[0x17] = 0xe3;	/* but it does not work */
	cr[0x18] = 0xff & LineCompare;
	cr[0x22] = 0x00;	/* todo? */


	/* now set the registers */
	for (i = 0; i <= 0x18; i++) {	/*CR00 .. CR18 */
		smiWrite (SMI_INDX_D4, i, cr[i]);
	}
	i = 0x22;		/*CR22 */
	smiWrite (SMI_INDX_D4, i, cr[i]);
	i = 0x30;		/*CR30 */
	smiWrite (SMI_INDX_D4, i, cr[i]);
}

/*****************************************************************************/
#define REF_FREQ	14318180
#define PMIN		1
#define PMAX		255
#define QMIN		1
#define QMAX		63

static unsigned int FindPQ (unsigned int freq, unsigned int *pp, unsigned int *pq)
{
	unsigned int n = QMIN, m = 0;
	long long int L = 0, P = freq, Q = REF_FREQ, H = P >> 1;
	long long int D = 0x7ffffffffffffffLL;

	for (n = QMIN; n <= QMAX; n++) {
		m = PMIN;	/* p/q ~ freq/ref -> p*ref-freq*q ~ 0 */
		L = P * n - m * Q;
		while (L > 0 && m < PMAX) {
			L -= REF_FREQ;	/* difference is greater as 0 subtract fref */
			m++;	/* and increment m */
		}
		/* difference is less or equal than 0 or m > maximum */
		if (m > PMAX)
			break;	/* no solution: if we increase n we get the same situation */
		/* L is <= 0 now */
		if (-L > H && m > PMIN) {	/* if difference > the half fref */
			L += REF_FREQ;	/* we take the situation before */
			m--;	/* because its closer to 0 */
		}
		L = (L < 0) ? -L : +L;	/* absolute value */
		if (D < L)	/* if last difference was better take next n */
			continue;
		D = L;
		*pp = m;
		*pq = n;	/*  keep improved data */
		if (D == 0)
			break;	/* best result we can get */
	}
	return (unsigned int) (0xffffffff & D);
}

/*****************************************************************************/
static void smiLoadCcr (const struct fb_videomode *var, unsigned short device_id)
{
	unsigned int p, q;
	long long freq;

	smiWrite (SMI_INDX_C4, 0x65, 0);
	smiWrite (SMI_INDX_C4, 0x66, 0);
	smiWrite (SMI_INDX_C4, 0x68, 0x50);
	smiWrite (SMI_INDX_C4, 0x69, 0x0);

	/* Memory clock */
	switch (device_id) {
	case PCI_DEVICE_ID_SILICONMOTION_SM712 :
		smiWrite (SMI_INDX_C4, 0x6a, 0x80);
		break;
	default :
		smiWrite (SMI_INDX_C4, 0x6a, 0x53);
		break;
	}
	smiWrite (SMI_INDX_C4, 0x6b, 0x15);

	/* VCLK */
	freq = 1000000000/*000LL*/ / var->pixclock * 1000;

	FindPQ ((unsigned int)freq, &p, &q);

	smiWrite (SMI_INDX_C4, 0x6c, p);
	smiWrite (SMI_INDX_C4, 0x6d, q);

}

static int __init sm712fb_init_one(struct pci_dev *dp, const struct pci_device_id *ent)
{
	struct fb_info_lynx *p = NULL;
	int err = 0;
	unsigned long addr, size;
	int i;
	u16 cmd;

	printk(KERN_DEBUG "sm712fb.c: sm712fb_init_one() slot=%s\n", dp->slot_name);

	if ((dp->resource[0].flags & IORESOURCE_MEM) == 0)
		return -EBUSY; /* FIXME: not sure what error codes we should use here */
	addr = dp->resource[0].start;
	size = dp->resource[0].end + 1 - addr;
	if (addr == 0)
		return -EBUSY;

	err = pci_request_regions(dp, "sm712fb");
	if (err)
		return err;

	p = kmalloc(sizeof(*p), GFP_ATOMIC);
	if (! p) {
		err = -ENOMEM;
		goto fail;
	}
	memset(p, 0, sizeof(*p));

	pci_set_drvdata(dp, p);

//#ifdef __BIG_ENDIAN
//	addr += 0x2000000;
//#endif
	p->pdev = dp;
	p->aperture_phys = addr;
	p->aperture = ioremap(addr, SM712_PCI_MEMORY_SIZE);

#if 0
	/* we should use pci_enable_device here, but,
	   the device doesn't declare its I/O ports in its BARs
	   so pci_enable_device won't turn on I/O responses */
	pci_read_config_word(dp, PCI_COMMAND, &cmd);
	cmd |= 3;	/* enable memory and IO space */
	pci_write_config_word(dp, PCI_COMMAND, cmd);
#else
	err = pci_enable_device(dp);
	if (err) {
		err = -EBUSY;
		goto fail;
	}
#ifdef CONFIG_MPC5200
	pci_read_config_word(dp, PCI_COMMAND, &cmd);
	cmd |= 1;       /* enable IO space */
	pci_write_config_word(dp, PCI_COMMAND, cmd);
#endif
#endif

	//p->frame_buffer_phys = p->aperture_phys + SM712_DISPLAY_MEMORY_SIZE;
	//p->frame_buffer = p->aperture + SM712_DISPLAY_MEMORY_SIZE;
    p->frame_buffer_phys = p->aperture_phys;
    p->frame_buffer = p->aperture;

	pci_read_config_byte(dp, PCI_REVISION_ID, &p->chiprev);
	printk(KERN_DEBUG "sm712fb.c: Chip Revision %x\n", (unsigned) p->chiprev);

	sm712_init_hw(p);

	printk(KERN_DEBUG "sm712fb.c: aperture physical addr 0x%08lx, virtual addr 0x%p\n",
	       p->aperture_phys, p->aperture);

	strcpy(p->fix.id, "LynxEM4+");
	p->fix.smem_start = p->frame_buffer_phys;
	p->fix.smem_len = p->ramsize;
	p->fix.type = FB_TYPE_PACKED_PIXELS;
	p->fix.visual = FB_VISUAL_PSEUDOCOLOR;

	p->fix.line_length = p->var.xres * ((p->var.bits_per_pixel + 7) >> 3);

	if (p->var.bits_per_pixel == 8) {
		p->var.red.offset = 0;
		p->var.green.offset = 0;
		p->var.blue.offset = 0;
		p->var.red.length = p->var.green.length = p->var.blue.length = 0;
#ifdef FBCON_HAS_CFB8
		p->disp.dispsw = &fbcon_cfb8;
#else
		p->disp.dispsw = &fbcon_dummy;
#endif
	} else if (p->var.bits_per_pixel == 15) {
		p->var.red.offset = 10;
		p->var.green.offset = 5;
		p->var.blue.offset = 0;
		p->var.red.length = p->var.green.length = p->var.blue.length = 5;
#ifdef FBCON_HAS_CFB16
		p->disp.dispsw = &fbcon_cfb16;
		p->disp.dispsw_data = p->fbcon_cfb16_cmap;
#else
		p->disp.dispsw = &fbcon_dummy;
#endif
	} else if (p->var.bits_per_pixel == 16) {
		p->var.red.offset = 11;
		p->var.green.offset = 5;
		p->var.blue.offset = 0;
		p->var.red.length = p->var.blue.length = 5;
		p->var.green.length = 6;
#ifdef FBCON_HAS_CFB16
		p->disp.dispsw = &fbcon_cfb16;
		p->disp.dispsw_data = p->fbcon_cfb16_cmap;
#else
		p->disp.dispsw = &fbcon_dummy;
#endif
	} else if (p->var.bits_per_pixel == 24) {
		p->var.red.offset = 16;
		p->var.green.offset = 8;
		p->var.blue.offset = 0;
		p->var.red.length = p->var.green.length = p->var.blue.length = 8;
#ifdef FBCON_HAS_CFB24
		p->disp.dispsw = &fbcon_cfb24;
		p->disp.dispsw_data = p->fbcon_cfb24_cmap;
#else
		p->disp.dispsw = &fbcon_dummy;
#endif
	}
	p->var.height = p->var.width = -1;
	p->var.vmode = FB_VMODE_NONINTERLACED;
	p->var.pixclock = 10000;
	p->var.left_margin = p->var.right_margin = 16;
	p->var.upper_margin = p->var.lower_margin = 16;
	p->var.hsync_len = p->var.vsync_len = 8;

	p->disp.var = p->var;
	p->disp.cmap.red = NULL;
	p->disp.cmap.green = NULL;
	p->disp.cmap.blue = NULL;
	p->disp.cmap.transp = NULL;
	p->disp.screen_base = p->frame_buffer;
	p->disp.visual = p->fix.visual;
	p->disp.type = p->fix.type;
	p->disp.type_aux = p->fix.type_aux;
	p->disp.line_length = p->fix.line_length;
	p->disp.can_soft_blank = 1;
	p->disp.scrollmode = SCROLL_YREDRAW;

	strcpy(p->info.modename, p->fix.id);
	p->info.node = -1;
	p->info.fbops = &sm712fb_ops;
	p->info.disp = &p->disp;
	p->info.fontname[0] = 0;
	p->info.changevar = NULL;
	p->info.switch_con = &sm712fb_switch_con;
	p->info.updatevar = &sm712fb_updatevar;
	p->info.blank = &sm712fb_blank;
	p->info.flags = FBINFO_FLAG_DEFAULT;

	for (i = 0; i < 16; ++i) {
		int j = color_table[i];
		p->palette[i].red = default_red[j];
		p->palette[i].green = default_grn[j];
		p->palette[i].blue = default_blu[j];
	}

//	my_sm712_set_bitdepth(p, &p->disp, 0, 24);
	err = register_framebuffer(&p->info);
	if (err < 0)
		goto fail;

	p->next = all_lynx;
	all_lynx = p;

	return 0;

 fail:
	if (p) {
		if (p->aperture)
			iounmap(p->aperture);
		kfree(p);
	}

	pci_release_regions(dp);
	return err;
}

#if 0
static void sm712fb_remove_one(struct pci_dev *pdev)
{
	struct fb_info_lynx *lynx;

	lynx = (struct fb_info_lynx *)pci_get_drvdata(pdev);

	/* FIXME: unregister framebuffer */

	if (lynx->aperture) {
		iounmap(lynx->aperture);
	}
	kfree(lynx);
	pci_release_regions(pdev);
}
#endif

/* The memory initialization and detection code is inspired by some
   code provided by Silicon Motion Inc. */
#define MEM_BLOCK_SIZE		(512 * 1024)	/* check memory in 512kB blocks */

unsigned long __init test_memory(u8 *fb, unsigned long range)
{
	unsigned long offset;
	unsigned long memsize = 0;

	for (offset = range; offset != 0; offset -= MEM_BLOCK_SIZE) {
		*(u32 *) (fb + offset - 4) = offset ^ 0x01234567;
		*(u32 *) (fb + offset - 8) = offset ^ 0x456789AB;
		*(u32 *) (fb + offset - 12) = offset ^ 0x89ABCDEF;
		*(u32 *) (fb + offset - 16) = offset ^ 0xCDEF0123;
	}

	for (offset = MEM_BLOCK_SIZE; offset <= range; offset += MEM_BLOCK_SIZE) {
		// In each block, verify the 128 bits.
		if ((*(u32 *) (fb + offset - 4) == (offset ^ 0x01234567))
		    && (*(u32 *) (fb + offset - 8) == (offset ^ 0x456789AB))
		    && (*(u32 *) (fb + offset - 12) == (offset ^ 0x89ABCDEF))
		    && (*(u32 *) (fb + offset - 16) == (offset ^ 0xCDEF0123))) {
			memsize += MEM_BLOCK_SIZE;
		} else
			break;
	}

	return memsize;
}

void __init sm712_ram_init(struct fb_info_lynx *lynx)
{
	lynx->ramsize = VIDEO_MEM_SIZE;

#if 0
	unsigned long range;
	u8 *fb;

	lynx->ramsize = 0;

	range = 0x1000000;	// 16MB

	write_scr(lynx, 0x18, read_scr(lynx, 0x18) | 0x41);

	fb = lynx->frame_buffer;

	write_mcr(lynx, 0x62, 0xff);
	write_mcr(lynx, 0x76, 0x7f);
	write_crt(lynx, 0xff, 0x00);

	if (lynx->chiprev >= 0xc0) {
		*((u16 *)(fb + 4)) = 0x5AA5;
		if (*((u16 *)(fb + 4)) != 0x5AA5)
			write_mcr(lynx, 0x62, 0xfe);
	}

	write_mcr(lynx, 0x76, 0x7b);

	if ( (test_memory(fb, 0x100000) != 0) && (lynx->chiprev < 0xc0) )
		write_mcr(lynx, 0x62, 0xbb);
	else
		write_crt(lynx, 0xff, 0x80);

	write_mcr(lynx, 0x76, 0x7f);

	lynx->ramsize = test_memory(fb, SM712_DISPLAY_MEMORY_SIZE); /* Test a 16MB range */

	if (lynx->ramsize == 0x400000)
		write_mcr(lynx, 0x76, 0xff);
	else
		write_mcr(lynx, 0x76, 0x3f);
#endif

}

struct regval {
	u8 index, val;
};

int __init sm712_setup(char *options)
{
    char *this_opt;

    if (!options || !*options)
        return 0;

    while ((this_opt = strsep(&options, ",")) != 0) {
	    if (!strncmp(this_opt, "lcd", 3)) {
		    monitor_type |= 1;
	    } else if (!strncmp(this_opt, "crt", 3)) {
		    monitor_type |= 2;
	    } else if (!strncmp(this_opt, "mode:", 5)) {
		    videomode = simple_strtoul(this_opt+5, NULL, 0);
	    }
    }

    return 0;
}

static void __init sm712_init_hw(struct fb_info_lynx *lynx)
{
	GraphicDevice *pGD = (GraphicDevice *)&smi;
	unsigned short device_id;
	unsigned int pci_mem_base, i;
	unsigned int gdfTab[] = { 8, 15, 16, 32, 24, 8 };
	char *gdfModes[] =
		{
		"8 Bit Index Color",
		"15 Bit 5-5-5 RGB",
		"16 Bit 5-6-5 RGB",
		"32 Bit X-8-8-8 RGB",
		"24 Bit 8-8-8 RGB",
		"8 Bit 3-3-2 RGB"
		};
	int vgaModes[16][2] =
		{
		{769, 0x00007}, {771, 0x00002}, {773, 0x00003}, {775, 0x00004},
		{784, 0x10007}, {787, 0x10002}, {790, 0x10003}, {793, 0x10004},
		{785, 0x20007}, {788, 0x20002}, {791, 0x20003}, {794, 0x20004},
		{786, 0x40007}, {789, 0x40002}, {792, 0x40003}, {795, 0x40004}
		};


	/**/
	pci_mem_base = (unsigned int)lynx->aperture;

    /* Compare with common vga mode numbers */
    for (i=0; i<16; i++) {
		if (vgaModes[i][0] == videomode)
		{
	    	if (vgaModes[i][1] == -1)
	    	{
				printk("Videomode not supported !\n");
				return;                /* mode not supported */
	    	}
	    	pGD->mode = vgaModes[i][1];        /* use driver int. mode number */
	    	break;
		}
    }

    /* Extract graphic data format */
    pGD->gdfIndex = (pGD->mode & 0x00070000) >> 16;
    if (pGD->gdfIndex > 5)
	pGD->gdfIndex = 0;
    pGD->gdfBytesPP = (gdfTab[pGD->gdfIndex] + 7) >> 3;

    /* Extract graphic resolution */
    pGD->mode &= 0xf;

    /* Exit for not supported resolutions */
    if (((pGD->mode==DUAL_800_600) || (pGD->mode==DUAL_1024_768)) && (pGD->gdfBytesPP > 1)) {
		printk (KERN_INFO "Dual screen for 1BPP only !\n");
		return;
    }

    if ((pGD->mode==SINGLE_1280_1024) && (pGD->gdfBytesPP==4)) {
		printk (KERN_INFO "Out of memory !\n");
		return;
    }

    /* Set graphic parameters */
    switch (pGD->mode) {
		case DUAL_800_600:
			pGD->winSizeX = 800;
			pGD->winSizeY = 600;
			pGD->plnSizeX = 1600;
			pGD->plnSizeY = 600;
			sprintf (pGD->modeIdent, "Dual Screen 800x600 with %s", gdfModes[pGD->gdfIndex]);
			break;
		case DUAL_1024_768:
			pGD->winSizeX = 1024;
			pGD->winSizeY = 768;
			pGD->plnSizeX = 2048;
			pGD->plnSizeY = 768;
			sprintf (pGD->modeIdent, "Dual Screen 1024x768 with %s", gdfModes[pGD->gdfIndex]);
			break;
		case SINGLE_800_600:
			pGD->winSizeX = 800;
			pGD->winSizeY = 600;
			pGD->plnSizeX = 800;
			pGD->plnSizeY = 600;
			sprintf (pGD->modeIdent, "Single Screen 800x600 with %s", gdfModes[pGD->gdfIndex]);
			break;
		case SINGLE_1024_768:
			pGD->winSizeX = 1024;
			pGD->winSizeY = 768;
			pGD->plnSizeX = 1024;
			pGD->plnSizeY = 768;
			sprintf (pGD->modeIdent,"Single Screen 1024x768 with %s", gdfModes[pGD->gdfIndex]);
			break;
		case TV_MODE_CCIR:
			pGD->winSizeX = 720;
			pGD->winSizeY = 576;
			pGD->plnSizeX = 720;
			pGD->plnSizeY = 576;
			sprintf (pGD->modeIdent, "TV Mode CCIR with %s", gdfModes[pGD->gdfIndex]);
			break;
		case TV_MODE_EIA:
			pGD->winSizeX = 720;
			pGD->winSizeY = 484;
			pGD->plnSizeX = 720;
			pGD->plnSizeY = 484;
			sprintf (pGD->modeIdent, "TV Mode EIA with %s", gdfModes[pGD->gdfIndex]);
			break;
		case SINGLE_1280_1024:
			pGD->winSizeX = 1280;
			pGD->winSizeY = 1024;
			pGD->plnSizeX = 1280;
			pGD->plnSizeY = 1024;
			sprintf (pGD->modeIdent, "Single Screen 1280x1024 with %s", gdfModes[pGD->gdfIndex]);
			break;
		case SINGLE_640_480:
			pGD->winSizeX = 640;
			pGD->winSizeY = 480;
			pGD->plnSizeX = 640;
			pGD->plnSizeY = 480;
			sprintf (pGD->modeIdent, "Single Screen 648x480 with %s", gdfModes[pGD->gdfIndex]);
			break;
		default:
			printk(KERN_INFO "Videomode not supported !\n");
			return;
    }

    lynx->var.xres = lynx->var.xres_virtual = pGD->winSizeX;
    lynx->var.yres = lynx->var.yres_virtual = pGD->winSizeY;
    lynx->var.bits_per_pixel = gdfTab[pGD->gdfIndex];

    pGD->isaBase = _IO_BASE;
    pGD->pciBase = pci_mem_base;
    pGD->dprBase = (pci_mem_base + 0x400000 + 0x8000);
    pGD->vprBase = (pci_mem_base + 0x400000 + 0xc000);
    pGD->cprBase = (pci_mem_base + 0x400000 + 0xe000);
    pGD->frameAdrs = pci_mem_base;
    pGD->memSize = VIDEO_MEM_SIZE;

	/* Set up hardware : select color mode,
	   set Register base to isa 3dx for 3?x regs*/
	out8 (SMI_MISC_REG, 0x01);

	/* Turn off display */
	smiWrite (SMI_INDX_C4, 0x01, 0x20);

	/* Unlock ext. crt regs */
	out8 (SMI_LOCK_REG, 0x40);

	/* Unlock crt regs 0-7 */
	smiWrite (SMI_INDX_D4, 0x11, 0x0e);

	/* Sytem Control Register */
	smiLoadRegs (SMI_INDX_C4, SMI_DATA_C5, SMI_SCR, sizeof(SMI_SCR));

	/* extented CRT Register */
	smiLoadRegs (SMI_INDX_D4, SMI_DATA_D5, SMI_EXT_CRT, sizeof(SMI_EXT_CRT));

	/* Attributes controller registers */
	smiLoadRegs (SMI_INDX_ATTR, SMI_INDX_ATTR, SMI_ATTR, sizeof(SMI_ATTR));

	/* Graphics Controller Register */
	smiLoadRegs (SMI_INDX_CE, SMI_DATA_CF, SMI_GCR, sizeof(SMI_GCR));

	/* Sequencer Register */
	smiLoadRegs (SMI_INDX_C4, SMI_DATA_C5, SMI_SEQR, sizeof(SMI_SEQR));

	/* Power Control Register */
	smiLoadRegs (SMI_INDX_C4, SMI_DATA_C5, SMI_PCR, sizeof(SMI_PCR));

	/* Memory Control Register */
	/* Register MSR62 is a power on configurable register. We don't */
	/* modify it */
	smiLoadRegs (SMI_INDX_C4, SMI_DATA_C5, SMI_MCR, sizeof(SMI_MCR));

	/* Set misc output register */
	smiLoadMsr (&res_mode_init[pGD->mode]);

	/* Set CRT and Clock control registers */
	smiLoadCrt (&res_mode_init[pGD->mode], gdfTab[pGD->gdfIndex]);

	smiLoadCcr (&res_mode_init[pGD->mode], device_id);

	/* Hardware Cusor Register */
	smiLoadRegs (SMI_INDX_C4, SMI_DATA_C5, SMI_HCR, sizeof(SMI_HCR));

	/* Enable  Display: LCD 1, CRT 2, both 3  */
	smiWrite (SMI_INDX_C4, 0x31, monitor_type);

	/* Video processor default setup */
	smiInitVideoProcessor ();

	/* Capture port default setup */
	smiInitCapturePort ();

	/* Drawing engine default setup */
	smiInitDrawingEngine ();

	/* Turn on display */
	smiWrite (0x3c4, 0x01, 0x01);

	printk("%s mode=%x - %s\n", __FUNCTION__, videomode, pGD->modeIdent);

	/* Initialize the RAM - this should be unecessary on machines
           with a BIOS, but it's needed on the Icebox */
	sm712_ram_init(lynx);

	printk(KERN_DEBUG "sm712fb.c: Found %ldk of video memory.\n",
	       lynx->ramsize / 1024);
	memset(lynx->frame_buffer, 0, lynx->ramsize);
}

static int sm712_get_fix(struct fb_fix_screeninfo *fix, int con,
			 struct fb_info *info)
{
	struct fb_info_lynx *lp = (struct fb_info_lynx *) info;

	*fix = lp->fix;
	return 0;
}

static int sm712_get_var(struct fb_var_screeninfo *var, int con,
			struct fb_info *info)
{

	struct fb_info_lynx *lp = (struct fb_info_lynx *) info;

	*var = lp->var;
	return 0;
}

static int sm712_set_var(struct fb_var_screeninfo *var, int con,
			 struct fb_info *info)
{
	struct fb_info_lynx *lp = (struct fb_info_lynx *) info;
	struct display *disp = (con >= 0)? &fb_display[con] : &lp->disp;

#if 0
	if ((var->xres != 640 || var->yres != 480)
	    || (var->xres_virtual != 640 || var->yres_virtual != 480)
	    || (var->bits_per_pixel != 8 && var->bits_per_pixel != 16
	    	&& var->bits_per_pixel != 24)
	    || var->nonstd
	    || (var->vmode & FB_VMODE_MASK) != FB_VMODE_NONINTERLACED)
		return -EINVAL;
#endif

	if ((var->activate & FB_ACTIVATE_MASK) == FB_ACTIVATE_NOW &&
	    var->bits_per_pixel != disp->var.bits_per_pixel) {
		sm712_set_bitdepth(lp, disp, con, var->bits_per_pixel);
	}

	return 0;
}

static int sm712fb_getcolreg(u_int regno, u_int *red, u_int *green,
			     u_int *blue, u_int *transp, struct fb_info *info)
{
	struct fb_info_lynx *p = (struct fb_info_lynx *) info;

	if (regno > 255)
		return 1;


		*red = (p->palette[regno].red<<10);// | p->palette[regno].red;
		*green = (p->palette[regno].green<<10);// | p->palette[regno].green;
		*blue = (p->palette[regno].blue<<10); // | p->palette[regno].blue;


	*transp = 0;
	return 0;
}

static int sm712fb_setcolreg(u_int regno, u_int red, u_int green, u_int blue,
			     u_int transp, struct fb_info *info)
{
	struct fb_info_lynx *p = (struct fb_info_lynx *) info;

	if (regno > 255)
		return 1;

#if 0
	printk(__FUNCTION__ " num: %d , red: %d (was %d),  green: %d (was %d), blue: %d (was %d)\n",
		regno, red>>10, red, green>>10, green, blue>>10, blue);
#endif

		red >>= 10;
		green >>= 10;
		blue >>= 10;

	p->palette[regno].red = red;
	p->palette[regno].green = green;
	p->palette[regno].blue = blue;
	/* FIXME: allow mmio */
	outb(regno, 0x3c8);
	udelay(10);
	outb(red, 0x3c9);
	udelay(10);
	outb(green, 0x3c9);
	udelay(10);
	outb(blue, 0x3c9);
	udelay(10);

#ifdef FBCON_HAS_CFB16
	if (regno < 16)
		p->fbcon_cfb16_cmap[regno] = ((red & 0xf8) << 7)
			| ((green & 0xf8) << 2) | ((blue & 0xf8) >> 3);
#endif
#ifdef FBCON_HAS_CFB24
	if (regno < 16)
		p->fbcon_cfb24_cmap[regno] = (red << 16)
			| (green << 8) | (blue);
#endif

	return 0;
}

static int sm712_get_cmap(struct fb_cmap *cmap, int kspc, int con,
			  struct fb_info *info)
{
	if (con == currcon)		/* current console? */
		return fb_get_cmap(cmap, kspc, sm712fb_getcolreg, info);
	if (fb_display[con].cmap.len)	/* non default colormap? */
		fb_copy_cmap(&fb_display[con].cmap, cmap, kspc ? 0 : 2);
	else {
		int size = fb_display[con].var.bits_per_pixel == 16 ? 32 : 256;
		fb_copy_cmap(fb_default_cmap(size), cmap, kspc ? 0 : 2);
	}
	return 0;
}

static int sm712_set_cmap(struct fb_cmap *cmap, int kspc, int con,
			 struct fb_info *info)
{
	int err;
	if (!fb_display[con].cmap.len) {	/* no colormap allocated? */
		int size = fb_display[con].var.bits_per_pixel == 16 ? 32 : 256;
		if ((err = fb_alloc_cmap(&fb_display[con].cmap, size, 0)))
			return err;
	}
	if (con == currcon)			/* current console? */
		return fb_set_cmap(cmap, kspc, sm712fb_setcolreg, info);
	else
		fb_copy_cmap(cmap, &fb_display[con].cmap, kspc ? 0 : 1);
	return 0;
}

static int sm712fb_switch_con(int con, struct fb_info *info)
{
	struct fb_info_lynx *p = (struct fb_info_lynx *) info;
	int new_bpp, old_bpp;

	//printk(__FUNCTION__ "\n");

	/* Do we have to save the colormap? */
	if (fb_display[currcon].cmap.len)
		fb_get_cmap(&fb_display[currcon].cmap, 1, sm712fb_getcolreg, info);

	new_bpp = fb_display[con].var.bits_per_pixel;
	old_bpp = fb_display[currcon].var.bits_per_pixel;
	currcon = con;

	if (new_bpp != old_bpp)
		sm712_set_bitdepth(p, &fb_display[con], con, new_bpp);

	do_install_cmap(con, info);
	return 0;
}

static int sm712fb_updatevar(int con, struct fb_info *info)
{
	return 0;
}

static void sm712fb_blank(int blank, struct fb_info *info)
{
	struct fb_info_lynx *p = (struct fb_info_lynx *) info;
	int i;

	printk(KERN_DEBUG "%s\n", __FUNCTION__);

	if (blank) {
		/* get the palette from the chip */
		for (i = 0; i < 256; ++i) {
			outb(i, 0x3c7);
			udelay(1);
			p->palette[i].red = inb(0x3c9);
			udelay(10);
			p->palette[i].green = inb(0x3c9);
			udelay(10);
			p->palette[i].blue = inb(0x3c9);
			udelay(10);
		}
		for (i = 0; i < 256; ++i) {
			outb(i, 0x3c8);
			udelay(10);
			udelay(1);
			outb(0, 0x3c9);
			udelay(10);
			outb(0, 0x3c9);
			udelay(10);
			outb(0, 0x3c9);
			udelay(10);
		}
	} else {
		for (i = 0; i < 256; ++i) {
			outb(i, 0x3c8);
			udelay(1);
			outb(p->palette[i].red, 0x3c9);
			udelay(10);
			outb(p->palette[i].green, 0x3c9);
			udelay(10);
			outb(p->palette[i].blue, 0x3c9);
			udelay(10);
		}
	}
}

/*
 * Exported functions
 */
static void do_install_cmap(int con, struct fb_info *info)
{
	if (con != currcon)
		return;
	if (fb_display[con].cmap.len)
		fb_set_cmap(&fb_display[con].cmap, 1, sm712fb_setcolreg, info);
	else {
		int size = fb_display[con].var.bits_per_pixel == 16 ? 32 : 256;
		fb_set_cmap(fb_default_cmap(size), 1, sm712fb_setcolreg, info);
	}
}

static void sm712_set_bitdepth(struct fb_info_lynx *p, struct display *disp, int con, int bpp)
{
#if 0
	int err;
	struct fb_fix_screeninfo* fix = &p->fix;
	struct fb_var_screeninfo* var = &p->var;

	printk(__FUNCTION__ "console %d, bpp %d\n", con, bpp);
	if (bpp == 24) {
		if (con == currcon) {
			write_crt(p, 0x13, 300);	// Set line length (doublewords)
			write_dprw(p, 0x1e,
				   (read_dprw(p, 0x1e) & 0xffcf) | 0x30);	//set drawing engine data format
			write_vprl(p, 0x00,
				   (read_vprl(p, 0x00) & 0xfff8ffff) | 0x40000);	//set graphic data format
		}

		fix->line_length = 800*3;
		fix->visual = FB_VISUAL_TRUECOLOR;

		var->red.offset = 16;
		var->green.offset = 8;
		var->blue.offset = 0;
		var->red.length = var->green.length = var->blue.length = 8;

#ifdef FBCON_HAS_CFB24
		disp->dispsw = &fbcon_cfb24;
		disp->dispsw_data = p->fbcon_cfb24_cmap;
#else
		disp->dispsw = &fbcon_dummy;
#endif
	} else if (bpp == 16) {
		if (con == currcon) {
			write_crt(p, 0x13, 200);	// Set line length (doublewords)
			write_dprw(p, 0x1e,
				   (read_dprw(p, 0x1e) & 0xffcf) | 0x10);
			write_vprl(p, 0x00,
				   (read_vprl(p, 0x00) & 0xfff8ffff) | 0x20000);
		}

		fix->line_length = 800*2;
		fix->visual = FB_VISUAL_TRUECOLOR;

		var->red.offset = 10;
		var->green.offset = 5;
		var->blue.offset = 0;
		var->red.length = var->blue.length = 5;
		var->green.length = 6;

#ifdef FBCON_HAS_CFB16
		disp->dispsw = &fbcon_cfb16;
		disp->dispsw_data = p->fbcon_cfb16_cmap;
#else
		disp->dispsw = &fbcon_dummy;
#endif
	} else if (bpp == 8) {
		if (con == currcon) {
			write_crt(p, 0x13, 100);		// Set line length (doublewords)
			write_dprw(p, 0x1e,
				   (read_dprw(p, 0x1e) & 0xffcf) | 0x0);
			write_vprl(p, 0x00,
				   (read_vprl(p, 0x00) & 0xfff8ffff) | 0x0);
		}

		fix->line_length = 800;
		fix->visual = FB_VISUAL_PSEUDOCOLOR;

 		var->red.offset = var->green.offset = var->blue.offset = 0;
		var->red.length = var->green.length = var->blue.length = 8;

#ifdef FBCON_HAS_CFB8
		disp->dispsw = &fbcon_cfb8;
#else
		disp->dispsw = &fbcon_dummy;
#endif
	}

	var->bits_per_pixel = bpp;
	disp->line_length = p->fix.line_length;
	disp->visual = fix->visual;
	disp->var = *var;

	if (p->info.changevar)
		(*p->info.changevar)(con);

	if ((err = fb_alloc_cmap(&disp->cmap, 0, 0)))
		return;
	do_install_cmap(con, (struct fb_info *)p);
#endif
}


static void my_sm712_set_bitdepth(struct fb_info_lynx *p, struct display *disp, int con, int bpp)
{
	int err;
	struct fb_fix_screeninfo* fix = &p->fix;
	struct fb_var_screeninfo* var = &p->var;

	printk("%s console %d, bpp %d\n",__FUNCTION__, con, bpp);

	if (bpp == 24) {
//		if (con == currcon) {
			write_crt(p, 0x13, 300);	// Set line length (doublewords)
			udelay(10);
			write_dprw(p, 0x1e,
				   (read_dprw(p, 0x1e) & 0xffcf) | 0x30);	//set drawing engine data format
			udelay(10);
			write_vprl(p, 0x00,
				   (read_vprl(p, 0x00) & 0xfff8ffff) | 0x40000);	//set graphic data format
			udelay(10);
//		}

		fix->line_length = 800*3;
		fix->visual = FB_VISUAL_TRUECOLOR;

		var->red.offset = 16;
		var->green.offset = 8;
		var->blue.offset = 0;
		var->red.length = var->green.length = var->blue.length = 8;

#ifdef FBCON_HAS_CFB24
		disp->dispsw = &fbcon_cfb24;
		disp->dispsw_data = p->fbcon_cfb24_cmap;
#else
		disp->dispsw = &fbcon_dummy;
#endif
	} else if (bpp == 16) {
		if (con == currcon) {
			write_crt(p, 0x13, 200);	// Set line length (doublewords)
			udelay(10);
			write_dprw(p, 0x1e,
				   (read_dprw(p, 0x1e) & 0xffcf) | 0x10);
			udelay(10);
			write_vprl(p, 0x00,
				   (read_vprl(p, 0x00) & 0xfff8ffff) | 0x20000);
			udelay(10);
		}

		fix->line_length = 800*2;
		fix->visual = FB_VISUAL_TRUECOLOR;

		var->red.offset = 10;
		var->green.offset = 5;
		var->blue.offset = 0;
		var->red.length = var->blue.length = 5;
		var->green.length = 6;

#ifdef FBCON_HAS_CFB16
		disp->dispsw = &fbcon_cfb16;
		disp->dispsw_data = p->fbcon_cfb16_cmap;
#else
		disp->dispsw = &fbcon_dummy;
#endif
	} else if (bpp == 8) {
		if (con == currcon) {
			write_crt(p, 0x13, 100);		// Set line length (doublewords)
			write_dprw(p, 0x1e,
				   (read_dprw(p, 0x1e) & 0xffcf) | 0x0);
			write_vprl(p, 0x00,
				   (read_vprl(p, 0x00) & 0xfff8ffff) | 0x0);
		}

		fix->line_length = 800;
		fix->visual = FB_VISUAL_PSEUDOCOLOR;

 		var->red.offset = var->green.offset = var->blue.offset = 0;
		var->red.length = var->green.length = var->blue.length = 8;

#ifdef FBCON_HAS_CFB8
		disp->dispsw = &fbcon_cfb8;
#else
		disp->dispsw = &fbcon_dummy;
#endif
	}

	var->bits_per_pixel = bpp;
	disp->line_length = p->fix.line_length;
	disp->visual = fix->visual;
	disp->var = *var;
/*
	if (p->info.changevar)
		(*p->info.changevar)(con);
*/
	if ((err = fb_alloc_cmap(&disp->cmap, 0, 0)))
		return;
	do_install_cmap(con, (struct fb_info *)p);
}
