/*
 *  linux/drivers/video/vfb-mcc200.c -- MCC200 board frame buffer device
 *
 *  Based on vfb.c written by Geert Uytterhoeven
 *
 *  2006 (c) Wolfgang Denk, DENX Software Engineering, <wd@denx.de>
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License. See the file COPYING in the main directory of this archive for
 *  more details.
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <asm/uaccess.h>
#include <linux/fb.h>
#include <linux/init.h>

#include <video/fbcon.h>
#include <video/fbcon-mfb.h>

#include <asm/mpc5xxx.h>
#ifdef CONFIG_UBOOT
#include <asm/ppcboot.h>
#endif
    /*
     *  RAM we reserve for the frame buffer. This defines the maximum
     *  screen size and enough to represent 160x100 B&W LCD
     */

#define VIDEOMEMSIZE	(2080)

#define PSOC_UART_BAUDRATE      500000

    /*
     * Work-around with LCD that has swapped columns. In this case use two buffers:
     * fb_buf - interfaces to fb-subsystem;
     * videomemory - interfaces to UART.
     */
#undef SWAPPED_LCD

#ifdef SWAPPED_LCD
static char *fb_buf;
#endif

static u_long videomemory, videomemorysize = VIDEOMEMSIZE;
static int currcon = 0;
static struct display disp;
static struct fb_info fb_info;
static struct { u_char red, green, blue, pad; } palette[256];
static char mvfb_name[16] = "MCC200 FB";

static struct fb_var_screeninfo mvfb_default = {
    /* 160x100, 1 bpp */
    160, 100, 160, 100, 0, 0, 1, 0,
    {0, 1, 0}, {0, 1, 0}, {0, 1, 0}, {0, 0, 0},
    0, 0, -1, -1, 0, 20000, 64, 64, 32, 32, 64, 2,
    0, FB_VMODE_NONINTERLACED
};

static int mvfb_enable = 0;	/* disabled by default */


    /*
     *  Interface used by the world
     */

int mvfb_setup(char*);

static int mvfb_get_fix(struct fb_fix_screeninfo *fix, int con,
		       struct fb_info *info);
static int mvfb_get_var(struct fb_var_screeninfo *var, int con,
		       struct fb_info *info);
static int mvfb_set_var(struct fb_var_screeninfo *var, int con,
		       struct fb_info *info);
static int mvfb_pan_display(struct fb_var_screeninfo *var, int con,
			   struct fb_info *info);
static int mvfb_get_cmap(struct fb_cmap *cmap, int kspc, int con,
			struct fb_info *info);
static int mvfb_set_cmap(struct fb_cmap *cmap, int kspc, int con,
			struct fb_info *info);


    /*
     *  Interface to the low level console driver
     */

int mvfb_init(void);
static int mvfbcon_switch(int con, struct fb_info *info);
static int mvfbcon_updatevar(int con, struct fb_info *info);
static void mvfbcon_blank(int blank, struct fb_info *info);


    /*
     *  Internal routines
     */

static u_long get_line_length(int xres_virtual, int bpp);
static void mvfb_encode_fix(struct fb_fix_screeninfo *fix,
			   struct fb_var_screeninfo *var);
static void set_color_bitfields(struct fb_var_screeninfo *var);
static int mvfb_getcolreg(u_int regno, u_int *red, u_int *green, u_int *blue,
                         u_int *transp, struct fb_info *info);
static int mvfb_setcolreg(u_int regno, u_int red, u_int green, u_int blue,
                         u_int transp, struct fb_info *info);
static void do_install_cmap(int con, struct fb_info *info);


static struct fb_ops mvfb_ops = {
	owner:		THIS_MODULE,
	fb_get_fix:	mvfb_get_fix,
	fb_get_var:	mvfb_get_var,
	fb_set_var:	mvfb_set_var,
	fb_get_cmap:	mvfb_get_cmap,
	fb_set_cmap:	mvfb_set_cmap,
	fb_pan_display:	mvfb_pan_display,
};
    /*
     * UART tranfer routines protos & variables
     */
static void mvfb_uart_transfer (void *arg);
static inline void mvfb_transmit_char(void);
static void mvfb_disable_tx_interrupts( void );
static void mvfb_enable_tx_interrupts( void );
static void mvfb_disable_rx_interrupts( void );
static void mvfb_enable_tx_underrun_interrupts( void );
static void mvfb_disable_tx_underrun_interrupts( void );

static struct mvfb_xmitter
{
    char *xmit_buf;
    int xmit_tail;
    int xmit_cnt;
    u16 imr;            /* shadow of interrupt mask */
} mvfb_trans;
static spinlock_t mpc5xxx_serial_lock = SPIN_LOCK_UNLOCKED;
static struct mpc5xxx_psc *mvfb_psc= (struct mpc5xxx_psc *)MPC5xxx_PSC2;

    /*
     * Display memory update routines protos & variables
     */

    /*
     * HZ jiffies == 1 sec, so 0.1HZ jiffies == 100 msec
     */
#define MVFB_DM_UPDATE_TIMEOUT	(HZ/10)

#if (((VIDEOMEMSIZE*8*1000)/PSOC_UART_BAUDRATE) > ((MVFB_DM_UPDATE_TIMEOUT*1000)/HZ))
#error "PSoC update rate is too fast for current baudrate."
#endif

static void mvfb_update_timer_handler (unsigned long arg);

static struct timer_list mvfb_update_timer;
static struct timer_list *pmvfb_update_timer = &mvfb_update_timer;

    /*
     * pmvfb_update_timer timeout handler, transfers the
     *	frame buffer to the PSoC over the UART and
     *	restarts update timer.
     */
static void mvfb_update_timer_handler (unsigned long arg)
{
    /* Restart display memory update timer*/
    mod_timer(pmvfb_update_timer, jiffies + MVFB_DM_UPDATE_TIMEOUT);
    mvfb_uart_transfer(NULL);
}

static int mvfb_rs_set_real_termios (void)
{
    struct mpc5xxx_psc *psc = mvfb_psc;
    struct mpc5xxx_gpio *gpio = (struct mpc5xxx_gpio *)MPC5xxx_GPIO;
    int baud = PSOC_UART_BAUDRATE;
#ifdef CONFIG_UBOOT
    extern unsigned char __res[];
    bd_t *bd = (bd_t *)__res;
    int bbase = bd->bi_ipbfreq;
#else
    int bbase = CONFIG_PPC_5xxx_IPBFREQ;
#endif
    int divisor = ((bbase / (baud * 16)) + 1) >> 1;
    int bits;
    int parity;
    int stop;
    int mode1;
    int mode2;
    int line = 1; /* PSC2 */
    u32 val32;

    val32 = in_be32(&gpio->port_config);
    val32 &= ~(7 << (4*line));
    val32 |= (MPC5xxx_GPIO_PSC_CONFIG_UART_WITHOUT_CD << (4*line));
    out_be32(&gpio->port_config, val32);

    /* reset and enable PSC */
    out_8(&psc->command, MPC5xxx_PSC_RST_TX
	| MPC5xxx_PSC_RX_DISABLE | MPC5xxx_PSC_TX_ENABLE);
    out_8(&psc->command, MPC5xxx_PSC_RST_RX);
    /* Set PSC operation mode as 'UART, DCD ignored' */
    out_be32(&psc->sicr, 0x0);
    /* Set clocking */
    out_be16(&psc->mpc5xxx_psc_clock_select, 0xdd00);
    /* Set tx FIFO alarm level */
    out_be16(&psc->tfalarm, 0xf8);
    /* Put PSC into the operation */
    out_8(&psc->command, MPC5xxx_PSC_SEL_MODE_REG_1
	| MPC5xxx_PSC_RX_ENABLE
	| MPC5xxx_PSC_TX_ENABLE);

    out_8(&psc->ctur, divisor>>8);
    out_8(&psc->ctlr, divisor);

    /* Program hardware for parity, data bits, stop bits */
    stop = MPC5xxx_PSC_MODE_ONE_STOP;
    parity = MPC5xxx_PSC_MODE_PARNONE;
    bits = MPC5xxx_PSC_MODE_8_BITS;

    mode1 = bits | parity | MPC5xxx_PSC_MODE_ERR;
    mode2 = stop;
#ifdef DEBUG
    printk( "requested baud: %d, bits: %d, parity: %08x\n",
	baud, bits, parity);
    printk( "baud_base: %d\n", bbase);
    printk( "mode 1 register was: %08x, now: %08x\n",
	in_8(&psc->mode), mode1);
    printk(  "mode 2 register was: %08x, now: %08x\n",
	in_8(&psc->mode), mode2);
#endif
    out_8(&psc->command, MPC5xxx_PSC_SEL_MODE_REG_1);
    out_8(&psc->mode, mode1);
    out_8(&psc->mode, mode2);

    /*
     *  Configure #RTS, set it high
     */
    psc->op0 = 0x1;
    return 0;
}

static void mvfb_disable_tx_interrupts( void )
{
    unsigned long flags;

    spin_lock_irqsave(&mpc5xxx_serial_lock, flags);
    mvfb_trans.imr &= ~MPC5xxx_PSC_IMR_TXRDY;
    out_be16(&mvfb_psc->mpc5xxx_psc_imr, mvfb_trans.imr);
    spin_unlock_irqrestore(&mpc5xxx_serial_lock, flags);
}

static void mvfb_enable_tx_interrupts( void )
{
    unsigned long flags;

    spin_lock_irqsave(&mpc5xxx_serial_lock, flags);
    mvfb_trans.imr |= MPC5xxx_PSC_IMR_TXRDY;
    out_be16(&mvfb_psc->mpc5xxx_psc_imr, mvfb_trans.imr);
    /* Send a char to start TX interrupts happening */
    mvfb_transmit_char();
    spin_unlock_irqrestore(&mpc5xxx_serial_lock, flags);
}

static void mvfb_enable_tx_underrun_interrupts( void )
{
    unsigned long flags;

    spin_lock_irqsave(&mpc5xxx_serial_lock, flags);
    mvfb_trans.imr |= MPC5xxx_PSC_IMR_URERR;
    out_be16(&mvfb_psc->mpc5xxx_psc_imr, mvfb_trans.imr);
    spin_unlock_irqrestore(&mpc5xxx_serial_lock, flags);
}

static void mvfb_disable_tx_underrun_interrupts( void )
{
    unsigned long flags;

    spin_lock_irqsave(&mpc5xxx_serial_lock, flags);
    mvfb_trans.imr &= ~MPC5xxx_PSC_IMR_URERR;
    out_be16(&mvfb_psc->mpc5xxx_psc_imr, mvfb_trans.imr);
    spin_unlock_irqrestore(&mpc5xxx_serial_lock, flags);
}

static void mvfb_disable_rx_interrupts( void )
{
    unsigned long flags;

    spin_lock_irqsave(&mpc5xxx_serial_lock, flags);
    mvfb_trans.imr &= ~MPC5xxx_PSC_IMR_RXRDY;
    out_be16(&mvfb_psc->mpc5xxx_psc_imr, mvfb_trans.imr);
    spin_unlock_irqrestore(&mpc5xxx_serial_lock, flags);
}

static inline void mvfb_transmit_char(void)
{
    int count = 256;
    int status;

    /*
     *  Called either from mvfb_enable_tx_interrupts() to start
     * the transmition, or from mvfb_rs_interrupt() interrupt
     * handler to continue transmition (place next data to TxFIFO)
     */
    if (mvfb_trans.xmit_cnt <= 0)
    {
	/*
	 * Come here when Underrun occured and no more data to transmit
	 */
	status = in_be16(&mvfb_psc->mpc5xxx_psc_status);
	if (status & MPC5xxx_PSC_SR_TXEMP)
	{ /* TxFIFO is completely empty indeed */
		mvfb_disable_tx_underrun_interrupts();
		/* notify the PSoC */
		mvfb_psc->op0 = 0x1;
	}
	/* we have nothing more to transmit, return */
	return;
    }

    /* While I'm able to transmit ... */
    while (--count >= 0) 
    {
	status = in_be16(&mvfb_psc->mpc5xxx_psc_status);
	if (!(status & MPC5xxx_PSC_SR_TXRDY))
		break;
	if (mvfb_trans.xmit_cnt <= 0)
		break;
	out_8(&mvfb_psc->mpc5xxx_psc_buffer_8,
		mvfb_trans.xmit_buf[mvfb_trans.xmit_tail++]);

	if (mvfb_trans.xmit_tail == VIDEOMEMSIZE)
		mvfb_trans.xmit_tail = 0;
	if (--mvfb_trans.xmit_cnt <= 0)
		break;             
    }

    if (mvfb_trans.xmit_cnt <= 0) 
    {
	/*
	 *  We must notify the PSoC that all buffer has transmitted in
	 * wires. We have to do it NOT after placing the last byte of the
	 * buffer to TxFIFO, BUT after transmition of the last byte from 
	 * TxFIFO.
	 *  So, when mvfb_trans.xmit_cnt == 0, all data from the buffer have
	 * been placed in TxFIFO. We disable TxRDY interrupt and enable TxEMP.
	 * When TxFIFO will become completely empty, we notify the PSoC
	 */
	mvfb_disable_tx_interrupts();
	mvfb_enable_tx_underrun_interrupts();
    }
}

static void mvfb_rs_interrupt(int irq, void *dev_id, struct pt_regs * regs)
{
    unsigned long flags;
    spin_lock_irqsave(&mpc5xxx_serial_lock, flags);
    mvfb_transmit_char();
    spin_unlock_irqrestore(&mpc5xxx_serial_lock, flags);
}

    /*
     * mvfb_uart_transfer function transfers framebuffer
     *	content into the PSoC over UART
     */
static void mvfb_uart_transfer (void *arg)
{
    int retries = 10;
    unsigned long flags;

    /*
     *  Make sure of the previous tranfer is accomplished
     */
    spin_lock_irqsave(&mpc5xxx_serial_lock, flags);
#ifdef DEBUG
    printk ("%s: in with %d in queue\n", 
		__FUNCTION__, mvfb_trans.xmit_cnt);
#endif
    
    if (!mvfb_trans.xmit_cnt)
    {
#ifdef SWAPPED_LCD
	#define WIDTH  160
	#define HEIGHT 100
	#define WIDTH_BYTES ((WIDTH + 7) >> 3)
	#define HEIGHT_BYTES (HEIGHT)
	#define BUF_SIZE (WIDTH_BYTES * HEIGHT_BYTES)
	int x, y, i; 
	char *badg = (char *)videomemory;
 
	for ( y=0 ; y<HEIGHT; y++ )
	{  
		i = y*WIDTH_BYTES;
		for (x=0 ; x<WIDTH_BYTES; x+=5)
		{
			badg[i+x] = (fb_buf[i+x+2]<<4 & 0xF0) | (fb_buf[i+x+3]>>4 & 0x0F);
			badg[i+x+1] = (fb_buf[i+x+3]<<4 & 0xF0) | (fb_buf[i+x+4]>>4 & 0x0F);
			badg[i+x+2] = (fb_buf[i+x+4]<<4 & 0xF0) | (fb_buf[i+x]>>4 & 0x0F);
			badg[i+x+3] = (fb_buf[i+x]<<4 & 0xF0) | (fb_buf[i+x+1]>>4 & 0x0F);
			badg[i+x+4] = (fb_buf[i+x+1]<<4 & 0xF0) | (fb_buf[i+x+2]>>4 & 0x0F);
		}
	}
#endif
	/*
	 *  Make sure that the PSoC has accomplished 
	 * previous oprations (#CTS must be cleared)
	 */
	if (!(mvfb_psc->ip & 0x1))
		goto out_; /* #CTS is low */

	/*
	 *  Assert #RTS and waith until PSoC becomes
	 * ready to receive (#CTS will be asserted)
	 */
	mvfb_psc->op1 = 0x1;
	while (retries--)
	{
		if (!(mvfb_psc->ip & 0x1))
			break; /* #CTS is low */
		udelay (10);
	}

	if (mvfb_psc->ip & 0x1)
	{ /*
	   *  PSoC haven't become ready (#CTS still high).
	   * Restore #RTS and exit without transfer.
	   */
		mvfb_psc->op0 = 0x1;
		goto out_;
	}

	/*
	 * Initiate next transfer
	 */
	mvfb_trans.xmit_cnt = VIDEOMEMSIZE;
	mvfb_enable_tx_interrupts ();
    }
out_:
#ifdef DEBUG
    printk ("%s: out with %d in queue\n", 
		__FUNCTION__, mvfb_trans.xmit_cnt);
#endif
    spin_unlock_irqrestore(&mpc5xxx_serial_lock, flags);
    return;
}

    /*
     *  Get the Fixed Part of the Display
     */

static int mvfb_get_fix(struct fb_fix_screeninfo *fix, int con,
		       struct fb_info *info)
{
    struct fb_var_screeninfo *var;

    if (con == -1)
	var = &mvfb_default;
    else
	var = &fb_display[con].var;
    mvfb_encode_fix(fix, var);
    return 0;
}


    /*
     *  Get the User Defined Part of the Display
     */

static int mvfb_get_var(struct fb_var_screeninfo *var, int con,
		       struct fb_info *info)
{
    if (con == -1)
	*var = mvfb_default;
    else
	*var = fb_display[con].var;
    set_color_bitfields(var);
    return 0;
}


    /*
     *  Set the User Defined Part of the Display
     */

static int mvfb_set_var(struct fb_var_screeninfo *var, int con,
		       struct fb_info *info)
{
    int activate = var->activate;
    int oldxres, oldyres, oldvxres, oldvyres, oldbpp;
    u_long line_length;

    struct display *display;
    if (con >= 0)
	display = &fb_display[con];
    else
	display = &disp;	/* used during initialization */

    /*
     *  FB_VMODE_CONUPDATE and FB_VMODE_SMOOTH_XPAN are equal!
     *  as FB_VMODE_SMOOTH_XPAN is only used internally
     */

    if (var->vmode & FB_VMODE_CONUPDATE) {
	var->vmode |= FB_VMODE_YWRAP;
	var->xoffset = display->var.xoffset;
	var->yoffset = display->var.yoffset;
    }

    /*
     *  Some very basic checks
     */
    if (!var->xres)
	var->xres = 1;
    if (!var->yres)
	var->yres = 1;
    if (var->xres > var->xres_virtual)
	var->xres_virtual = var->xres;
    if (var->yres > var->yres_virtual)
	var->yres_virtual = var->yres;
    if (var->bits_per_pixel <= 1)
	var->bits_per_pixel = 1;
    else
	return -EINVAL;

    /*
     *  Memory limit
     */
    line_length = get_line_length(var->xres_virtual, var->bits_per_pixel);
    if (line_length*var->yres_virtual > videomemorysize)
	return -ENOMEM;

    set_color_bitfields(var);

    if ((activate & FB_ACTIVATE_MASK) == FB_ACTIVATE_NOW) {
	oldxres = display->var.xres;
	oldyres = display->var.yres;
	oldvxres = display->var.xres_virtual;
	oldvyres = display->var.yres_virtual;
	oldbpp = display->var.bits_per_pixel;
	display->var = *var;
	if (oldxres != var->xres || oldyres != var->yres ||
	    oldvxres != var->xres_virtual || oldvyres != var->yres_virtual ||
	    oldbpp != var->bits_per_pixel) {
	    struct fb_fix_screeninfo fix;

	    mvfb_encode_fix(&fix, var);
#ifdef SWAPPED_LCD
	    display->screen_base = fb_buf;
#else
	    display->screen_base = (char *)videomemory;
#endif
	    display->visual = fix.visual;
	    display->type = fix.type;
	    display->type_aux = fix.type_aux;
	    display->ypanstep = fix.ypanstep;
	    display->ywrapstep = fix.ywrapstep;
	    display->line_length = fix.line_length;
	    display->can_soft_blank = 1;
	    display->inverse = 0;
	    /*
	     * if here var->bits_per_pixel is always 1,
	     * so dispsw == fbcon_monochromefb
	     */
	    display->dispsw = &fbcon_mfb;
	    if (fb_info.changevar)
		(*fb_info.changevar)(con);
	}
    }
    return 0;
}


    /*
     *  Pan or Wrap the Display
     *
     *  This call looks only at xoffset, yoffset and the FB_VMODE_YWRAP flag
     */

static int mvfb_pan_display(struct fb_var_screeninfo *var, int con,
			   struct fb_info *info)
{
    if (var->vmode & FB_VMODE_YWRAP) {
	if (var->yoffset < 0 ||
	    var->yoffset >= fb_display[con].var.yres_virtual ||
	    var->xoffset)
	    return -EINVAL;
    } else {
	if (var->xoffset+fb_display[con].var.xres >
	    fb_display[con].var.xres_virtual ||
	    var->yoffset+fb_display[con].var.yres >
	    fb_display[con].var.yres_virtual)
	    return -EINVAL;
    }
    fb_display[con].var.xoffset = var->xoffset;
    fb_display[con].var.yoffset = var->yoffset;
    if (var->vmode & FB_VMODE_YWRAP)
	fb_display[con].var.vmode |= FB_VMODE_YWRAP;
    else
	fb_display[con].var.vmode &= ~FB_VMODE_YWRAP;
    return 0;
}

    /*
     *  Get the Colormap
     */

static int mvfb_get_cmap(struct fb_cmap *cmap, int kspc, int con,
			struct fb_info *info)
{
    if (con == currcon) /* current console? */
	return fb_get_cmap(cmap, kspc, mvfb_getcolreg, info);
    else if (fb_display[con].cmap.len) /* non default colormap? */
	fb_copy_cmap(&fb_display[con].cmap, cmap, kspc ? 0 : 2);
    else
	fb_copy_cmap(fb_default_cmap(1<<fb_display[con].var.bits_per_pixel),
		     cmap, kspc ? 0 : 2);
    return 0;
}

    /*
     *  Set the Colormap
     */

static int mvfb_set_cmap(struct fb_cmap *cmap, int kspc, int con,
			struct fb_info *info)
{
    int err;

    if (!fb_display[con].cmap.len) {	/* no colormap allocated? */
	if ((err = fb_alloc_cmap(&fb_display[con].cmap,
			      1<<fb_display[con].var.bits_per_pixel, 0)))
	    return err;
    }
    if (con == currcon)			/* current console? */
	return fb_set_cmap(cmap, kspc, mvfb_setcolreg, info);
    else
	fb_copy_cmap(cmap, &fb_display[con].cmap, kspc ? 0 : 1);
    return 0;
}


int __init mvfb_setup(char *options)
{
    char *this_opt;

    fb_info.fontname[0] = '\0';

    mvfb_enable = 1;

    if (!options || !*options)
	return 0;

    while ((this_opt = strsep(&options, ",")) != NULL) {
	if (!strncmp(this_opt, "font:", 5))
	    strcpy(fb_info.fontname, this_opt+5);
    }
    return 0;
}


    /*
     *  Initialisation
     */

int __init mvfb_init(void)
{
    int rval = 0;

#ifdef SWAPPED_LCD
    if (!(fb_buf = vmalloc(videomemorysize)))
    {
	printk (KERN_ERR "%s: failed vmalloc memory for fb_buf\n",
		__FUNCTION__);
	return -ENOMEM;
    }
    memset(fb_buf, 0, videomemorysize);
#endif
    if (!(videomemory = (u_long)vmalloc(videomemorysize)))
    {
	printk (KERN_ERR "%s: failed vmalloc memory for fb\n",
		__FUNCTION__);
	return -ENOMEM;
    }
    memset((char *)videomemory, 0, videomemorysize);

    strcpy(fb_info.modename, mvfb_name);
    fb_info.changevar = NULL;
    fb_info.node = -1;
    fb_info.fbops = &mvfb_ops;
    fb_info.disp = &disp;
    fb_info.switch_con = &mvfbcon_switch;
    fb_info.updatevar = &mvfbcon_updatevar;
    fb_info.blank = &mvfbcon_blank;
    fb_info.flags = FBINFO_FLAG_DEFAULT;

    mvfb_set_var(&mvfb_default, -1, &fb_info);

    if (register_framebuffer(&fb_info) < 0)
    {
		vfree((void *)videomemory);
		return -EINVAL;
    }

    memset (&mvfb_trans, 0, sizeof(struct mvfb_xmitter));
    mvfb_trans.xmit_buf = (char *)videomemory;

    mvfb_disable_rx_interrupts();
    mvfb_disable_tx_interrupts();
    mvfb_rs_set_real_termios();
    if (request_irq(MPC5xxx_PSC2_IRQ, mvfb_rs_interrupt,
                        SA_SHIRQ | SA_INTERRUPT, "mvfb-serial", &mvfb_trans)) 
    {
	printk ("%s: can't allocate irq %d\n", __FUNCTION__, MPC5xxx_PSC2_IRQ);
    }

    /* Allocate display memory update timer */
    memset (pmvfb_update_timer, 0, sizeof(struct timer_list));
    init_timer(pmvfb_update_timer);
    pmvfb_update_timer->function = mvfb_update_timer_handler;
    pmvfb_update_timer->expires = jiffies + MVFB_DM_UPDATE_TIMEOUT;
    add_timer(pmvfb_update_timer);

    printk(KERN_INFO "fb%d: MCC200 fb device, using %ld bytes of video memory\n",
	   GET_FB_IDX(fb_info.node), videomemorysize);
    return rval;
}


static int mvfbcon_switch(int con, struct fb_info *info)
{
    /* Do we have to save the colormap? */
    if (fb_display[currcon].cmap.len)
	fb_get_cmap(&fb_display[currcon].cmap, 1, mvfb_getcolreg, info);

    currcon = con;
    /* Install new colormap */
    do_install_cmap(con, info);
    return 0;
}

    /*
     *  Update the `var' structure (called by fbcon.c)
     */

static int mvfbcon_updatevar(int con, struct fb_info *info)
{
    /* Nothing */
    return 0;
}

    /*
     *  Blank the display.
     */

static void mvfbcon_blank(int blank, struct fb_info *info)
{
    /* Nothing */
}

static u_long get_line_length(int xres_virtual, int bpp)
{
    u_long length;

    length = xres_virtual*bpp;
    length = (length+31)&-32;
    length >>= 3;
    return(length);
}

static void mvfb_encode_fix(struct fb_fix_screeninfo *fix,
			   struct fb_var_screeninfo *var)
{
    memset(fix, 0, sizeof(struct fb_fix_screeninfo));
    strcpy(fix->id, mvfb_name);
#ifdef SWAPPED_LCD
    fix->smem_start = fb_buf;
#else
    fix->smem_start = videomemory;
#endif
    fix->smem_len = videomemorysize;
    fix->type = FB_TYPE_PACKED_PIXELS;
    fix->type_aux = 0;
    /*
     * always have var->bits_per_pixel == 1,
     * so visual == mono
     */
    fix->visual = FB_VISUAL_MONO01;
    fix->ywrapstep = 1;
    fix->xpanstep = 1;
    fix->ypanstep = 1;
    fix->line_length = get_line_length(var->xres_virtual, var->bits_per_pixel);
}

static void set_color_bitfields(struct fb_var_screeninfo *var)
{
    /*
     * each displayed pixel is represented with a "pixel" value, that has
     * var->bits_per_pixel bits size.
     * "pixel" value contains "levels" of RGB colors (and transparancy).
     * The format of each "pixel" value is called colormap.
     * Because of black&white LCD, we always have var->bits_per_pixel == 1,
     * so colormap is always as follows:
     * all color components have the same offset - 1st (and single) bit
     * all color components have the same length - 1 bit
     * msb_right flag isn't used (it means Most significant bit is right)
     */
    var->red.offset = 0;
    var->red.length = 1;

    var->green.offset = 0;
    var->green.length = 1;

    var->blue.offset = 0;
    var->blue.length = 1;

    var->transp.offset = 0;
    var->transp.length = 0;

    var->red.msb_right = 0;
    var->green.msb_right = 0;
    var->blue.msb_right = 0;
    var->transp.msb_right = 0;
}


    /*
     *  Read a single color register and split it into
     *  colors/transparent. Return != 0 for invalid regno.
     */

static int mvfb_getcolreg(u_int regno, u_int *red, u_int *green, u_int *blue,
                         u_int *transp, struct fb_info *info)
{
    if (regno > 255)
	return 1;
    *red = (palette[regno].red<<8) | palette[regno].red;
    *green = (palette[regno].green<<8) | palette[regno].green;
    *blue = (palette[regno].blue<<8) | palette[regno].blue;
    *transp = 0;
    return 0;
}


    /*
     *  Set a single color register. The values supplied are already
     *  rounded down to the hardware's capabilities (according to the
     *  entries in the var structure). Return != 0 for invalid regno.
     */

static int mvfb_setcolreg(u_int regno, u_int red, u_int green, u_int blue,
                         u_int transp, struct fb_info *info)
{
    if (regno > 255)
	return 1;
    red >>= 8;
    green >>= 8;
    blue >>= 8;
    palette[regno].red = red;
    palette[regno].green = green;
    palette[regno].blue = blue;
    return 0;
}


static void do_install_cmap(int con, struct fb_info *info)
{
    if (con != currcon)
	return;
    if (fb_display[con].cmap.len)
	fb_set_cmap(&fb_display[con].cmap, 1, mvfb_setcolreg, info);
    else
	fb_set_cmap(fb_default_cmap(1<<fb_display[con].var.bits_per_pixel), 1,
		    mvfb_setcolreg, info);
}


#ifdef MODULE
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Yuri Tikhonov <yur@emcraft.com>");
MODULE_DESCRIPTION("MCC200 Virtual Framebuffer module");
int init_module(void)
{
    return mvfb_init();
}

void cleanup_module(void)
{
    int flags;

    unregister_framebuffer(&fb_info);
    save_flags(flags);
    cli();
    del_timer(pmvfb_update_timer);
    mvfb_disable_tx_interrupts();
    restore_flags(flags);
    free_irq(MPC5xxx_PSC2_IRQ, &mvfb_trans);
    vfree((void *)videomemory);
}

#endif /* MODULE */
