/*
 *  linux/include/linux/spi/spi.h
 *
 *  Copyright (C) 2001 Russell King, All Rights Reserved.
 *  Copyright (C) 2002 Compaq Computer Corporation, All Rights Reserved.
 *  Copyright (C) 2004 Kate Alhola, All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License.
 *
 * Derived from l3.h by Jamey Hicks
 */
#ifndef SPI_H
#define SPI_H

struct spi_msg {
	unsigned char	addr;	/* slave address	*/
	unsigned char	flags;		
#define SPI_M_RD	0x01
#define SPI_M_WR        0x02
#define SPI_M_CSREL	0x04    /* Value to set CS at end of the frame   */
#define SPI_M_CS        0x08    /* Value to set CS active ( default low ) */
	unsigned short	len;	/* msg length		*/
	unsigned char	*buf;	/* pointer to msg data	*/
};

#ifdef __KERNEL__

#include <linux/types.h>
#include <linux/list.h>


#define SPI_MAJOR 228  /* This will change */

struct spi_client;
struct spi_adapter;

struct spi_ops {
	int	(*open)(struct spi_client *);
	int	(*command)(struct spi_client *, int cmd, void *arg);
	void	(*close)(struct spi_client *);
};

/*
 * A driver is capable of handling one or more physical devices present on
 * SPI adapters. This information is used to inform the driver of adapter
 * events.
 */
struct spi_driver {
	/*
	 * This name is used to uniquely identify the driver.
	 * It should be the same as the module name.
	 */
	char			name[32];

	unsigned int flags;		/* div., see below		*/

	/* Notifies the driver that a new bus has appeared. This routine
	 * can be used by the driver to test if the bus meets its conditions
	 * & seek for the presence of the chip(s) it supports. If found, it 
	 * registers the client(s) that are on the bus to the spi admin. via
	 * spi_attach_client.
	 */
	int			(*attach_adapter)(struct spi_adapter *);

	/*
	 * Notifies the driver that the client has finished with its
	 * services, and any memory that it allocated for this client
	 * should be cleaned up.  In addition the chip should be
	 * shut down.
	 */
	void			(*detach_client)(struct spi_client *);

	/* a ioctl like command that can be used to perform specific functions
	 * with the device.
	 */
	int (*command)(struct spi_client *client,unsigned int cmd, void *arg);
		/*
	 * Possible operations on the driver.
	 */
	struct spi_ops		*ops;

	/*
	 * Module structure, if any.	
	 */
	struct module		*owner;

	/*
	 * drivers list
	 */
	struct list_head	drivers;

	/* These two are mainly used for bookkeeping & dynamic unloading of 
	 * kernel modules. inc_use tells the driver that a client is being  
	 * used by another module & that it should increase its ref. counter.
	 * dec_use is the inverse operation.
	 * NB: Make sure you have no circular dependencies, or else you get a 
	 * deadlock when trying to unload the modules.
	* You should use the i2c_{inc,dec}_use_client functions instead of
	* calling this function directly.
	 */
	void (*inc_use)(struct spi_client *client);
	void (*dec_use)(struct spi_client *client);
};

/*flags for the driver struct: */
#define SPI_DF_NOTIFY	0x01		/* notify on bus (de/a)ttaches 	*/
#define SPI_DF_DUMMY	0x02		/* do not connect any clients */


struct spi_algorithm {
	/* textual description */
	char name[32];

	/* perform bus transactions */
	int (*xfer)(struct spi_adapter *, struct spi_msg msgs[], int num);
	/* set bus bit rate */
	int (*set_bitrate)(struct spi_adapter *, unsigned long bitrate);
};

struct spi_cs {
	char name[32];
	int id;
};

struct semaphore;

/*
 * spi_adapter is the structure used to identify a physical SPI bus along
 * with the access algorithms necessary to access it.
 */
struct spi_adapter {
	/*
	 * This name is used to uniquely identify the adapter.
	 * It should be the same as the module name.
	 */
	char			name[32];

	/*
	 * the algorithm to access the bus
	 */
	struct spi_algorithm	*algo;

	/*
	 * Algorithm specific data
	 */
	void			*algo_data;

	/* 
	 * Set ( assret/release ) spi chip select line 
	 * 
	 */	

	int (* set_spi_adap_cs)(struct spi_adapter *,int id,int val);
	/*
	 *
	 * table of assigned CS id:s of this board
	 *
	*/
	struct spi_cs *cs;
	/*
	 * This may be NULL, or should point to the module struct
	 */
	struct module		*owner;

	/*
	 * private data for the adapter
	 */
	void			*data;

	/*
	 * Bus parameters (such as SPI clock polarity, SPI clock phase)
	 */
	unsigned char		bus_params;
#define SPI_CLOCK_ACTIVE_LOW	0x01
#define SPI_CLOCK_DATA_START	0x02

	/*
	 * Our lock.
	 */
	struct semaphore	lock;

	/* No need to set this when you initialize the adapter          */
	int inode;

        /*  unique id */
        int   id;
	/*
	 * List of attached clients.
	 */
	struct list_head	clients;

	/*
	 * List of all adapters.
	 */
	struct list_head	adapters;
};

/*
 * spi_client identifies a single device (i.e. chip) that is connected to an 
 * SPI bus. The behaviour is defined by the routines of the driver. This
 * function is mainly used for lookup & other admin. functions.
 */
struct spi_client {
	char			name[32];
	struct spi_adapter	*adapter;	/* the adapter we sit on	*/
	struct spi_driver	*driver;	/* and our access routines	*/
	void			*driver_data;	/* private driver data		*/
	struct list_head	__adap;
        int                     addr;           /* Target chip chip select number */
};


extern int spi_add_adapter(struct spi_adapter *);
extern int spi_del_adapter(struct spi_adapter *);

extern int spi_add_driver(struct spi_driver *);
extern int spi_del_driver(struct spi_driver *);

extern int spi_attach_client(struct spi_client *, const char *, const char *);
extern void spi_detach_client(struct spi_client *);

extern int spi_transfer(struct spi_adapter *, struct spi_msg msgs[], int);
extern int spi_read_write(struct spi_client *, int, const char *, int);
extern int spi_write(struct spi_client *, int, const char *, int);
extern int spi_read(struct spi_client *, int, char *, int);
/* This call returns a unique low identifier for each registered adapter,
 * or -1 if the adapter was not registered. 
 */
extern int spi_adapter_id(struct spi_adapter *adap);

/**
 * spi_command - send a command to a SPI device driver
 * @client: registered client structure
 * @cmd: device driver command
 * @arg: device driver arguments
 *
 * Ask the SPI device driver to perform some function.  Further information
 * should be sought from the device driver in question.
 *
 * Returns negative error code on failure.
 */
static inline int spi_command(struct spi_client *clnt, int cmd, void *arg)
{
	struct spi_ops *ops = clnt->driver->ops;
	int ret = -EINVAL;

	if (ops && ops->command)
		ret = ops->command(clnt, cmd, arg);

	return ret;
}

static inline int spi_open(struct spi_client *clnt)
{
	struct spi_ops *ops = clnt->driver->ops;
	int ret = 0;

	if (ops && ops->open)
		ret = ops->open(clnt);
	return ret;
}

static inline void spi_close(struct spi_client *clnt)
{
	struct spi_ops *ops = clnt->driver->ops;
	if (ops && ops->close)
		ops->close(clnt);
}
#endif

/* IOCTL codes dor spi-dev.c */

#define SPI_RDWR	0x6207	/** Combined R/W transfer (one stop only)*/
#define SPI_SETADDR	0x6206	/** Combined R/W transfer (one stop only)*/
#define SPI_FUNCS	0x6205	/** Get the adapter functionality */
#define SPI_BITRATE	0x6208  /** Set adapter's bitrate */

/* This is the structure as used in the SPI_RDWR ioctl call */
struct spi_rdwr_ioctl_data {
	struct spi_msg *msgs;	/* pointers to spi_msgs */
	unsigned int   nmsgs;		/* number of spi_msgs */
};


#endif /* SPI_H */
